-- Raven is an addon to monitor auras and cooldowns, providing timer bars, action bar highlights, and helpful notifications.

-- Conditions.lua contains routines to test auras and status info to generate notifications.

-- Exported functions:
-- Raven:InitializeCondition(name, condition) adds a condition (or overwrites a default one) during initialization
-- Raven:CheckCondition(name) evaluates a condition and returns its current value
-- Raven:GetConditionSpell(name) returns the spell associated with a condition
-- Raven:GetConditionText(name) returns a string containing a detailed description of the condition

local MOD = Raven
local L = LibStub("AceLocale-3.0"):GetLocale("Raven")
local LSPELL = MOD.LocalSpellNames
MOD.status = {} -- global status info cached on every update
local fishSpell = GetSpellInfo(7620)
local classificationList = { normal = L["Normal"], worldboss = L["Boss"], elite = L["Elite"], rare = L["Rare"], rareelite = L["Rare Elite"] }
local unitList = { player = "Player", pet = "Pet", target = "Target", focus = "Focus",
	mouseover = "Mouseover", pettarget = "Pet's Target", targettarget = "Target's Target", focustarget = "Focus's Target" }

-- Saved variables don't handle being set to nil properly so need to use alternate value to indicate an option has been turned off
local Off = 0 -- value used to designate an option is turned off
local function IsOff(value) return value == nil or value == Off end -- return true if option is turned off
local function IsOn(value) return value ~= nil and value ~= Off end -- return true if option is turned on

-- Initialization settings for each test type (note that nil values are placeholders for documentation purposes)
MOD.conditionTests = {
	["Player Status"] = { enable = false, inCombat = nil, isResting = nil, hasPet = nil, isStealthed = nil, isMounted = nil,
		isFishing = nil, inGroup = nil, inParty = nil, inRaid = nil, isPvP = nil, inInstance = nil,
		hasMainHand = nil, levelMainHand = 1, hasOffHand = nil, levelOffHand = 1, hasRanged = nil, levelRanged = 1,
		checkTalent = nil, talent = nil, checkLevel = nil, level = 80, checkStance = nil, stance = nil,
		checkHealth = nil, minHealth = 100, checkPower = nil, minPower = 100, checkHolyPower = nil, minHolyPower = 1, checkShards = nil, minShards = 1,
		checkEclipsePower = nil, minEclipsePower = 0, checkEclipse = nil, checkSolar = nil, checkLunar = nil, checkSun = nil, checkMoon = nil,
		checkComboPoints = nil, minComboPoints = 5, checkRunes = nil, checkAnyRune = nil, needBlood = nil, needFrost = nil, needUnholy = nil,
		checkBlood = nil, chargeBlood = nil, checkFrost = nil, chargeFrost = nil, checkUnholy = nil, chargeUnholy = nil },
	["Pet Status"] = { enable = false, checkHealth = nil, minHealth = 100, checkPower = nil, minPower = 100 },
	["Target Status"] = { enable = false, isEnemy = nil, isFriend = nil, inRange = nil, checkHealth = nil, minHealth = 100,
		checkPower = nil, minPower = 100, classification = "normal" },
	["Focus Status"] = { enable = false, isEnemy = nil, isFriend = nil, inRange = nil, checkHealth = nil, minHealth = 100,
		checkPower = nil, minPower = 100, classification = "normal" },
	["All Buffs"] = { enable = false, unit = "player", auras = nil, isMine = nil, toggle = false },
	["Any Buffs"] = { enable = false, unit = "player", auras = nil, isMine = nil, toggle = false },
	["Buff Time Left"] = { enable = false, unit = "player", aura = nil, timeLeft = 10, isMine = nil, toggle = false },
	["Buff Count"] = { enable = false, unit = "player", aura = nil, count = 1, isMine = nil, toggle = false },
	["Buff Type"] = { enable = false, hasBuff = nil, toggle = false },
	["All Debuffs"] = { enable = false, unit = "target", auras = nil, isMine = nil, toggle = false },
	["Any Debuffs"] = { enable = false, unit = "target", auras = nil, isMine = nil, toggle = false },
	["Debuff Time Left"] = { enable = false, unit = "target", aura = nil, timeLeft = 10, isMine = nil, toggle = false },
	["Debuff Count"] = { enable = false, unit = "player", aura = nil, count = 1, isMine = nil, toggle = false },
	["Debuff Type"] = { enable = false, hasDebuff = nil, toggle = false },
	["All Cooldowns"] = { enable = false, spells = nil, timeLeft = 10 },
	["Spell Ready"] = { enable = false, gcd = false, spell = nil, inRange = nil, notUsable = false },
	["Spell Casting"] = { enable = false, spell = nil, unit = "player" },
	["Item Ready"] = { enable = false, item = nil, toggle = false },
}

MOD.testOrder = { "Player Status", "Pet Status", "Target Status", "Focus Status", "All Buffs", "Any Buffs", "Buff Time Left",
	"Buff Count", "Buff Type", "All Debuffs", "Any Debuffs", "Debuff Time Left", "Debuff Count", "Debuff Type",
	"All Cooldowns", "Spell Ready", "Spell Casting", "Item Ready"
}
	
local testNames = {
	["Player Status"]= L["Player Status"],
	["Pet Status"]= L["Pet Status"],
	["Target Status"] = L["Target Status"],
	["Focus Status"] = L["Focus Status"],
	["All Buffs"] = L["All Buffs"],
	["Any Buffs"] = L["Any Buffs"],
	["Buff Time Left"] = L["Buff Time Left"],
	["Buff Count"] = L["Buff Count"],
	["Buff Type"] = L["Buff Type"],
	["All Debuffs"] = L["All Debuffs"],
	["Any Debuffs"] = L["Any Debuffs"],
	["Debuff Time Left"] = L["Debuff Time Left"],
	["Debuff Count"] = L["Debuff Count"],
	["Debuff Type"] = L["Debuff Type"],
	["All Cooldowns"] = L["All Cooldowns"],
	["Spell Ready"] = L["Spell Ready"],
	["Spell Casting"] = L["Spell Casting"],
	["Item Ready"] = L["Item Ready"],
}

-- Initialize data used by this module, and restore any default test settings
function MOD:InitializeConditions()
	local conditions = MOD.db.profile.Conditions[MOD.myClass]
	for _, c in pairs(conditions) do
		if IsOn(c) and c.tests then
			for ttype, test in pairs(c.tests) do
				local ct = MOD.conditionTests[ttype]
				if ct then for k, v in pairs(ct) do if test[k] == nil then test[k] = v end end end
			end
		end
	end
	for _, c in pairs(MOD.db.global.SharedConditions) do
		if IsOn(c) and c.tests then
			for ttype, test in pairs(c.tests) do
				local ct = MOD.conditionTests[ttype]
				if ct then for k, v in pairs(ct) do if test[k] == nil then test[k] = v end end end
			end
		end
	end
end

-- Finalize conditions by stripping out unnecessary values, updating shared conditions, removing default test settings
function MOD:FinalizeConditions()
	for _, c in pairs(MOD.db.profile.Conditions[MOD.myClass]) do
		if IsOn(c) then
			c.result = nil; c.testResult = nil -- these values are left over from evaluations
			if c.tests then
				for ttype, test in pairs(c.tests) do
					local ct = MOD.conditionTests[ttype]
					if ct then for k, v in pairs(ct) do if test[k] == v then test[k] = nil end end end
				end
			end
			if c.shared then
				local t = Raven_CopyTable(c)
				t.shared = false
				MOD.db.global.SharedConditions["[" .. MOD.myClass .. "] " .. c.name] = t
			end
		end
	end
	for _, c in pairs(MOD.db.global.SharedConditions) do
		if IsOn(c) and c.tests then
			for ttype, test in pairs(c.tests) do
				local ct = MOD.conditionTests[ttype]
				if ct then for k, v in pairs(ct) do if test[k] == v then test[k] = nil end end end
			end
		end
	end
end

-- Check if a spell is ready to be cast by the player, if rangeCheck then make sure in range of unit too
local function CheckSpellReady(spell, gcd, unit, rangeCheck, usable)
	if not spell or (spell == "") then return true end
	if usable and not IsUsableSpell(spell) then return false end -- checks player has learned the spell, has mana and/or reagents, and reactive conditions are met
	local dcheck = 1.5 -- global cooldown threshold
	if gcd == true then dcheck = 0 end -- ignore global cooldown
	local start, duration, enabled = GetSpellCooldown(spell) -- checks if spell is on cooldown (includes cases where spell is on global cooldown, distinguished by duration of 1.5 seconds or less)
	if (start == nil) or ((enabled == 0) or (start > 0 and duration > dcheck)) then return false end -- start will be nil for invalid spell names
	if IsOn(rangeCheck) and IsOn(unit) and ((IsSpellInRange(spell, unit) == 1) ~= rangeCheck) then return false end
	return true
end

-- Check if a specified spell is currently being cast or channeled by the unit 
local function CheckSpellCast(spell, unit)
	if IsOff(unit) or not spell or (spell == "") then return true end
	local sp = UnitCastingInfo(unit)
	if (sp ~= nil) and (spell == sp) then return true end
	sp = UnitChannelInfo(unit)
	if (sp ~= nil) and (spell == sp) then return true end
	return false
end

-- Check if an item is ready to be used by the player, can be either itemID (number) or item name (string)
-- The item must be in the player's bag for this to work properly
local function CheckItemReady(item, toggle)
	if not item or (item == "") then return true end
	local id = tonumber(item)
	if id then
		local start, duration = GetItemCooldown(id)
		if (start > 0) and (duration > 0) then return toggle end -- item is not ready
	else
		local cd = MOD:CheckCooldown(item) -- look up in the active cooldowns table
		if cd and (cd[1] ~= nil) then return toggle end
	end
	return not toggle
end

-- Check for either at least one active aura (hasAll is false) or all active auras (hasAll is true)
-- If isMine true then cast by the player, if false then cast by other, if nil cast by anyone
local function CheckAuras(unit, auras, hasAll, isMine, buff, toggle)
	if IsOff(unit) or not auras then return true end -- no test
	if (unit == "target") and MOD.status.noTarget then return not toggle end
	if (unit == "focus") and MOD.status.noFocus then return not toggle end
	local found = false
	for _, aname in pairs(auras) do -- look for each aura and check if buff/debuff and cast by player
		local auraList = MOD:CheckAura(unit, aname, buff)
		if #auraList > 0 then
			for _, aura in pairs(auraList) do
				if IsOff(isMine) or (isMine == (aura[6] == "player")) then -- check cast by setting
					found = true -- found at least one of the auras
				else
					if hasAll then return toggle end -- not all are found
				end
			end
		else
			if hasAll then return toggle end -- not all are found
		end
	end
	if toggle then return not found end -- flip result if necessary
	return found
end

-- Check an aura on the unit to see if active and either expiration time is less than or equal or count is greater
-- If isMine true then cast by the player, if false then cast by other, if nil cast by anyone
local function CheckAuraValue(unit, aname, timeLeft, count, isMine, buff, toggle)
	if IsOff(unit) or not aname or (aname == "") or (IsOff(timeLeft) and IsOff(count)) then return true end -- no test
	if (unit == "target") and MOD.status.noTarget then return not toggle end
	if (unit == "focus") and MOD.status.noFocus then return not toggle end
	local auraList = MOD:CheckAura(unit, aname, buff)
	for _, aura in pairs(auraList) do
		if IsOff(isMine) or (isMine == (aura[6] == "player")) then -- check if found one with right cast by setting
			if timeLeft and (aura[2] >= timeLeft) then return not toggle end -- check expiration time is greater or equal
			if count and (aura[3] >= count) then return not toggle end -- check count is greater or equal
		end
	end
	return toggle
end

-- Return true only if all spells in the list are on cooldown and have at least the specified amount of time left.
-- If toggle is true then check if they all have less than the specified time left instead.
-- If a spell is not ready for some other reason (e.g., too much target HP ala Kill Shot) then consider it on long cooldown.
local function CheckAllCooldowns(spells, timeLeft, toggle)
	for _, spell in pairs(spells) do -- look for each spell and check if on cooldown
		local cdt = 0
		if not IsUsableSpell(spell) then cdt = 3600 end -- checks player has learned the spell, has mana and/or reagents, and reactive conditions are met
		local cd = MOD:CheckCooldown(spell) -- look up in the active cooldowns table
		if cd and (cd[1] ~= nil) then cdt = cd[1] end
		if toggle then
			if cdt >= timeLeft then return false end
		else
			if cdt < timeLeft then return false end
		end
	end
	return true
end

-- Return the current stance or "none" if not in one
local function GetStance()
	local form = "none"
	local index = GetShapeshiftForm(nil)
	if index > 0 then
		 _, form = GetShapeshiftFormInfo(index)
	end
	return form
end

-- Return true only if the specified talent (either spell name or spell id) has points spent in it
function RavenCheckTalent(talent)
	local id = tonumber(talent)
	if id then talent = GetSpellInfo(id) end -- translate spell id
	local t = MOD.talents[talent]
	if t then
		local _, _, _, _, pts = GetTalentInfo(t.tab, t.index, false, false, nil)
		if pts and (pts > 0) then return true end
	end
	return false
end

-- Check if mounted, complicated by having to test for druid flight form
local function CheckMounted()
	local flying = false
	if MOD.myClass == "DRUID" then
		local form = GetShapeshiftForm(true)
		if form ~= 0 then
			local _, fname = GetShapeshiftFormInfo(form)
			flying = (fname == LSPELL["Flight Form"]) or (fname == LSPELL["Swift Flight Form"])
		end
	end
	return (IsMounted() ~= nil) or flying
end

-- Check balance druid eclipse
local function CheckEclipsePower(minPower)
	if (MOD.myClass == "DRUID") and IsSpellKnown(78674) and minPower then -- only Balance if know Starsurge spell
		local power = UnitPower("player", SPELL_POWER_ECLIPSE)
		local maxPower = UnitPowerMax("player", SPELL_POWER_ECLIPSE)
		if (maxPower <= 0) or (power > maxPower) then return false end -- avoid errors from the eclipse API
		local e = 100 * power / maxPower
		return e >= minPower
	end
	return false
end

-- Check balance druid eclipse
local function CheckEclipse(solar, lunar, sun, moon)
	if (MOD.myClass == "DRUID") and IsSpellKnown(78674) then -- only Balance if know Starsurge spell
		if solar then -- check solar eclipse, if required
			local auraList = MOD:CheckAura("player", LSPELL["Eclipse (Solar)"], true)
			if #auraList == 0 then return false end
		end
		if lunar then -- check lunar eclipse, if required
			local auraList = MOD:CheckAura("player", LSPELL["Eclipse (Lunar)"], true)
			if #auraList == 0 then return false end
		end
		local direction = GetEclipseDirection()
		if direction == "none" then direction = (UnitPower("player", SPELL_POWER_ECLIPSE) < 0) and "moon" or "sun" end
		if sun and direction ~= "sun" then return false end
		if moon and direction ~= "moon" then return false end
		return true
	end
	return false
end

-- Check if weapon equipped with at least the minimum specified item level
local function CheckWeapon(slot, level)
	local islot = GetInventorySlotInfo(slot)
	local id = GetInventoryItemLink("player", islot)
	if id then
		local iname, _, _, ilevel = GetItemInfo(id)
		if iname and (ilevel >= level) then return true end
	end
	return false
end

-- Make sure the table contains at least one entry
local function HasTable(t)
	if IsOn(t) and (type(t) == "table") and (next(t) ~= nil) then return true end
	return false
end

-- Check the value of one test (see template at end of file)
local function CheckTest(ttype, t)
	local toggle = (t.toggle == true)
	local stat = MOD.status
	if ttype == "Player Status" then
		if IsOn(t.inCombat) and (t.inCombat ~= stat.inCombat) then return false end
		if IsOn(t.inGroup) and (t.inGroup ~= stat.inGroup) then return false end
		if IsOn(t.inParty) and (t.inParty ~= stat.inParty) then return false end
		if IsOn(t.inRaid) and (t.inRaid ~= stat.inRaid) then return false end
		if IsOn(t.inInstance) and (t.inInstance ~= stat.inInstance) then return false end
		if IsOn(t.isResting) and (t.isResting ~= stat.isResting) then return false end
		if IsOn(t.isStealthed) and (t.isStealthed ~= stat.isStealthed) then return false end
		if IsOn(t.isFishing) and (t.isFishing ~= stat.isFishing) then return false end
		if IsOn(t.isMounted) and (t.isMounted ~= stat.isMounted) then return false end
		if IsOn(t.isPvP) and (t.isPvP ~= stat.isPvP) then return false end
		if IsOn(t.checkLevel) and IsOn(t.level) and (t.checkLevel ~= (stat.level >= t.level)) then return false end
		if IsOn(t.checkHealth) and IsOn(t.minHealth) and (t.checkHealth ~= (stat.health >= t.minHealth)) then return false end
		if IsOn(t.checkPower) and IsOn(t.minPower) and (t.checkPower ~= (stat.power >= t.minPower)) then return false end
		if IsOn(t.checkHolyPower) and IsOn(t.minHolyPower) and (t.checkHolyPower ~= (stat.holyPower >= t.minHolyPower)) then return false end
		if IsOn(t.checkShards) and IsOn(t.minShards) and (t.checkShards ~= (stat.shards >= t.minShards)) then return false end
		if IsOn(t.checkEclipsePower) and (t.checkEclipsePower ~= CheckEclipsePower(t.minEclipsePower)) then return false end
		if IsOn(t.checkEclipse) and not CheckEclipse(t.checkSolar, t.checkLunar, t.checkSun, t.checkMoon) then return false end
		if IsOn(t.checkComboPoints) and IsOn(t.minComboPoints) and (t.checkComboPoints ~= (stat.combo >= t.minComboPoints)) then return false end
		if IsOn(t.hasPet) and (t.hasPet ~= (HasPetUI() ~= nil)) then return false end
		if IsOn(t.checkStance) and IsOn(t.stance) and (t.stance ~= stat.stance) then return false end
		if IsOn(t.checkRunes) then
			if not MOD:IsRuneSpellReady(t.needBlood, t.needFrost, t.needUnholy, t.checkAnyRune) then return false end
			if not MOD:IsRuneRecharging(t.checkBlood, t.chargeBlood, t.checkFrost, t.chargeFrost, t.checkUnholy, t.chargeUnholy) then return false end
		end
		if IsOn(t.checkTalent) and IsOn(t.talent) and not RavenCheckTalent(t.talent) then return false end
		if IsOn(t.hasMainHand) and not CheckWeapon("MainHandSlot", t.levelMainHand) then return false end
		if IsOn(t.hasOffHand) and not CheckWeapon("SecondaryHandSlot", t.levelOffHand) then return false end
		if IsOn(t.hasRanged) and not CheckWeapon("RangedSlot", t.levelRanged) then return false end
	elseif ttype == "Pet Status" then -- pet must exist for these tests to be true
		if IsOn(t.checkHealth) and IsOn(t.minHealth) and (stat.noPet or (t.checkHealth ~= (stat.petHealth >= t.minHealth))) then return false end
		if IsOn(t.checkPower) and IsOn(t.minPower) and (stat.noPet or (t.checkPower ~= (stat.petPower >= t.minPower))) then return false end
	elseif ttype == "Target Status" then -- target must exist for these tests to be true
		if IsOn(t.isEnemy) and (stat.noTarget or (t.isEnemy ~= stat.targetEnemy)) then return false end
		if IsOn(t.isFriend) and (stat.noTarget or (t.isFriend ~= stat.targetFriend)) then return false end
		if IsOn(t.isDead) and (stat.noTarget or (t.isDead ~= stat.targetDead)) then return false end
		if IsOn(t.isSteal) and (stat.noTarget or (t.isSteal ~= MOD:UnitHasBuff("target", "Steal"))) then return false end
		if t.classify and t.classification and (stat.noTarget or (t.classify ~= (stat.targetClassification == t.classification))) then return false end
		if IsOn(t.checkHealth) and IsOn(t.minHealth) and (stat.noTarget or (t.checkHealth ~= (stat.targetHealth >= t.minHealth))) then return false end
		if IsOn(t.checkPower) and IsOn(t.minPower) and (stat.noTarget or (t.checkPower ~= (stat.targetPower >= t.minPower))) then return false end
		if IsOn(t.inRange) and (stat.noTarget or (t.inRange ~= stat.targetInRange)) then return false end
	elseif ttype == "Focus Status" then -- focus must exist for these tests to be true
		if IsOn(t.isEnemy) and (stat.noFocus or (t.isEnemy ~= stat.focusEnemy)) then return false end
		if IsOn(t.isFriend) and (stat.noFocus or (t.isFriend ~= stat.focusFriend)) then return false end
		if IsOn(t.isDead) and (stat.noFocus or (t.isDead ~= stat.focusDead)) then return false end
		if IsOn(t.isSteal) and (stat.noFocus or (t.isSteal ~= MOD:UnitHasBuff("focus", "Steal"))) then return false end
		if IsOn(t.classify) and IsOn(t.classification) and (stat.noFocus or (t.classify ~= (stat.focusClassification == t.classification))) then return false end
		if IsOn(t.checkHealth) and IsOn(t.minHealth) and (stat.noFocus or (t.checkHealth ~= (stat.focusHealth >= t.minHealth))) then return false end
		if IsOn(t.checkPower) and IsOn(t.minPower) and (stat.noFocus or (t.checkPower ~= (stat.focusPower >= t.minPower))) then return false end
		if IsOn(t.inRange) and (stat.noFocus or (t.inRange ~= stat.focusInRange)) then return false end
	elseif ttype == "All Buffs" then
		if HasTable(t.auras) and not CheckAuras(t.unit, t.auras, true, t.isMine, true, toggle) then return false end
	elseif ttype == "Any Buffs" then
		if HasTable(t.auras) and not CheckAuras(t.unit, t.auras, false, t.isMine, true, toggle) then return false end
	elseif ttype == "Buff Time Left" then
		if not CheckAuraValue(t.unit, t.aura, t.timeLeft, nil, t.isMine, true, toggle) then return false end
	elseif ttype == "Buff Count" then
		if not CheckAuraValue(t.unit, t.aura, nil, t.count, t.isMine, true, toggle) then return false end
	elseif ttype == "Buff Type" then
		if IsOn(t.hasBuff) and (MOD:UnitHasBuff("player", t.hasBuff) == toggle) then return false end
	elseif ttype == "All Debuffs" then
		if HasTable(t.auras) and not CheckAuras(t.unit, t.auras, true, t.isMine, false, toggle) then return false end
	elseif ttype == "Any Debuffs" then
		if HasTable(t.auras) and not CheckAuras(t.unit, t.auras, false, t.isMine, false, toggle) then return false end
	elseif ttype == "Debuff Time Left" then
		if not CheckAuraValue(t.unit, t.aura, t.timeLeft, nil, t.isMine, false, toggle) then return false end
	elseif ttype == "Debuff Count" then
		if not CheckAuraValue(t.unit, t.aura, nil, t.count, t.isMine, false, toggle) then return false end
	elseif ttype == "Debuff Type" then
		if IsOn(t.hasDebuff) and (MOD:UnitHasDebuff("player", t.hasDebuff) == toggle) then return false end
	elseif ttype == "All Cooldowns" then
		if HasTable(t.spells) and not CheckAllCooldowns(t.spells, t.timeLeft, toggle) then return false end
	elseif ttype == "Spell Ready" then
		if not CheckSpellReady(t.spell, t.gcd, "target", t.inRange, not t.notUsable) then return false end
	elseif ttype == "Spell Casting" then
		if not CheckSpellCast(t.spell, t.unit) then return false end
	elseif ttype == "Item Ready" then
		if not CheckItemReady(t.item, toggle) then return false end
	else
		return false
	end
	return true
end

-- Check the value of one condition (see template at end of file), all individual tests must evaluate to true (i.e., logical and)
local function CheckCondition(c)
	if not c.enabled then return false end -- if enabled default value is true, if not enabled always return false
	if c.tests then
		for ttype, test in pairs(c.tests) do
			if test.enable and not CheckTest(ttype, test) then return false end
		end
	end
	return true
end

-- Process table of conditions in current profile, checking status and auras.
function MOD:UpdateConditions()
	-- update globally useful conditions
	local stat = MOD.status
	stat.inCombat = (UnitAffectingCombat("player") ~= nil)
	stat.inParty = (GetNumPartyMembers() > 0)
	stat.inRaid = (GetNumRaidMembers() > 1)
	stat.inGroup = stat.inParty or stat.inRaid
	stat.inParty = stat.inParty and not stat.inRaid
	stat.inInstance = (IsInInstance() ~= nil)
	stat.isResting = (IsResting() ~= nil)
	stat.isFishing = (IsUsableSpell(fishSpell) ~= nil)
	stat.isMounted = CheckMounted()
	stat.isPvP = (UnitIsPVP("player") ~= nil)
	stat.isStealthed = (IsStealthed() ~= nil)
	stat.level = UnitLevel("player")
	stat.health = (100 * UnitHealth("player")/UnitHealthMax("player"))
	stat.power = (100 * UnitPower("player")/UnitPowerMax("player"))
	if MOD.myClass == "PALADIN" then stat.holyPower = UnitPower("player", SPELL_POWER_HOLY_POWER) else stat.holyPower = 0 end
	if MOD.myClass == "WARLOCK" then stat.shards = UnitPower("player", SPELL_POWER_SOUL_SHARDS) else stat.shards = 0 end
	stat.combo = GetComboPoints("player")
	stat.stance = GetStance()
	stat.noPet = (UnitExists("pet") == nil)
	if not stat.noPet then
		stat.petHealth = (100 * UnitHealth("pet")/UnitHealthMax("pet"))
		stat.petPower = (100 * UnitPower("pet")/UnitPowerMax("pet"))
	end
	stat.noTarget = (UnitExists("target") == nil)
	if not stat.noTarget then
		stat.targetEnemy = (UnitIsEnemy("player", "target") ~= nil)
		stat.targetFriend = (UnitIsFriend("player", "target") ~= nil)
		stat.targetDead = (UnitIsDead("target") ~= nil)
		stat.targetClassification = UnitClassification("target")
		stat.targetHealth = (100 * UnitHealth("target")/UnitHealthMax("target"))
		stat.targetPower = (100 * UnitPower("target")/UnitPowerMax("target"))
		stat.targetInRange = (CheckInteractDistance("target", 1) ~= nil)
	end
	stat.noFocus = (UnitExists("focus") == nil)
	if not stat.noFocus then
		stat.focusEnemy = (UnitIsEnemy("player", "focus") ~= nil)
		stat.focusFriend = (UnitIsFriend("player", "focus") ~= nil)
		stat.focusDead = (UnitIsDead("focus") ~= nil)
		stat.focusClassification = UnitClassification("focus")
		stat.focusHealth = (100 * UnitHealth("focus")/UnitHealthMax("focus"))
		stat.focusPower = (100 * UnitPower("focus")/UnitPowerMax("focus"))
		stat.focusInRange = (CheckInteractDistance("focus", 1) ~= nil)
	end
	
	-- only check conditions for the player's class
	local ct = MOD.db.profile.Conditions[MOD.myClass]
	if ct then
		-- set all conditions to false first (quicker to clear it than to check IsOn)
		for _, c in pairs(ct) do if IsOn(c) then c.testResult = false; c.result = false end end
		-- don't check conditions if dead or in vehicle or on a taxi
		if UnitIsDeadOrGhost("player") ~= nil then return end
		if UnitUsingVehicle("player") ~= nil then return end
		if UnitOnTaxi("player") ~= nil then return end		
		-- run the tests in each condition to get intermediate testResult
		for _, c in pairs(ct) do
			if IsOn(c) and c.name then c.testResult = CheckCondition(c) end
		end
		-- then check dependencies and overrides to get final result
		for _, c in pairs(ct) do 
			if IsOn(c) and c.name then
				c.result = c.testResult -- start with intermediate result
				if c.result and c.dependencies then -- if starting true then check dependencies
					local ncount, nreqd = 0, false -- look for any non-required that evaluates to true
					for dname, result in pairs(c.dependencies) do -- check each dependency in the table
						local dep = ct[dname] -- dependency may be missing from the conditions table after a copy
						if c.dependencyType and c.dependencyType[dname] then -- check for at least one non-required dependency to be true
							if IsOn(dep) and (dep.testResult == result) then nreqd = true end
							ncount = ncount + 1
						else
							if IsOff(dep) or (dep.testResult ~= result) then c.result = false end -- all required must be true
						end
					end
					if (ncount > 0) and not nreqd then c.result = false end
				end
				if c.toggleResult then c.result = not c.result end
				if IsOn(c.setResult) then c.result = c.setResult end -- force result
			end
		end
	end
end
	
-- Initialize conditions for the player's class from the preset files
function MOD:SetConditionDefaults()
	local nt = MOD.classConditions[MOD.myClass]
	for k, v in pairs(nt) do MOD:InitializeCondition(k, v) end
end

-- Get localized spell name from a field in a condition definition (return nil if called with nil)
local function GetLocalizedSpellName(field)
	if not field then return nil end
	local id, name = tonumber(field), field
	if id then name = GetSpellInfo(id) end
	return name
end
	
-- Add or overwrite a condition from another module. If class is not nil then restrict to designated class.
-- Must be called during OnInitialize to add to default conditions
-- Changes made using the configuration panel will be tracked in Raven's profile.
function MOD:InitializeCondition(name, c, class)
	if class then
		class = string.upper(class)
		if class == "DEATH KNIGHT" then class = "DEATHKNIGHT" end
		if class ~= MOD.myClass then return end
	end
	local ct = MOD.DefaultProfile.profile.Conditions
	if ct and name and c then
		if not ct[MOD.myClass] then ct[MOD.myClass] = {} end
		if c.name == nil then c.name = name end -- set default values for uninitialized fields
		if c.enabled == nil then c.enabled = true end
		if c.notify == nil then c.notify = true end
		if c.tests then
			for ttype, test in pairs(c.tests) do
				test.spell = GetLocalizedSpellName(test.spell)
				test.aura = GetLocalizedSpellName(test.aura)
				test.talent = GetLocalizedSpellName(test.talent)
				test.stance = GetLocalizedSpellName(test.stance)
				if test.spells then for k, n in pairs(test.spells) do test.spells[k] = GetLocalizedSpellName(n) end end
				if test.auras then for k, n in pairs(test.auras) do test.auras[k] = GetLocalizedSpellName(n) end end
				-- don't convert test.item because itemID is much more useful currently
			end
		end
		c.associatedSpell = GetLocalizedSpellName(c.associatedSpell) -- localized associated spell for color and icon
		ct[MOD.myClass][name] = c -- add to the default profile, it is okay to overwrite any with same name
	end
end

-- Return the current value of a condition, nil if not found
function MOD:CheckCondition(name)
	local ct = MOD.db.profile.Conditions[MOD.myClass]
	if ct then
		local c = ct[name]
		if IsOn(c) and c.name then return c.result end
	end
	return nil
end

-- Return the associated spell for a condition, nil if not found
function MOD:GetConditionSpell(name)
	local ct = MOD.db.profile.Conditions[MOD.myClass]
	if ct then
		local c = ct[name]
		if IsOn(c) and c.associatedSpell and (c.associatedSpell ~= "") then return c.associatedSpell end
	end
	return nil
end

-- Return a text description of the current condition
function MOD:GetConditionText(name)
	local description, testString, depString, resultString = "", "", "", ""
	description = L["Condition Name"](name)
	local etext = ""
	local ct, c = MOD.db.profile.Conditions[MOD.myClass], nil
	if ct then c = ct[name] end
	if not c then return description .. " " .. L["(not found)"] end
	if not c.enabled then etext = L["Disable String"] end
	description = description .. etext
	if c.tests then
		for _, tt in pairs(MOD.testOrder) do -- loop through tests in right order
			local t = c.tests[tt]
			if t and t.enable then -- check if a test of this type is defined and enabled
				local a, d = "", " " -- a is string for current test conditions, d gets comma after first test clause added
				local at = string.format("\n|cFF7adbf2%s:|r", testNames[tt]) -- localized test name
				if tt == "Player Status" then
					if IsOn(t.inCombat) then if t.inCombat then a = a .. d .. L["In Combat"] else a = a .. d .. L["Out Of Combat"] end; d = ", " end
					if IsOn(t.isResting) then if t.isResting then a = a .. d .. L["Is Resting"] else a = a .. d .. L["Not Resting"] end; d = ", " end
					if IsOn(t.isStealthed) then if t.isStealthed then a = a .. d .. L["Is Stealthed"] else a = a .. d .. L["Not Stealthed"] end; d = ", " end
					if IsOn(t.isFishing) then if t.isFishing then a = a .. d .. L["Is Fishing"] else a = a .. d .. L["Not Fishing"] end; d = ", " end
					if IsOn(t.isMounted) then if t.isMounted then a = a .. d .. L["Is Mounted"] else a = a .. d .. L["Not Mounted"] end; d = ", " end
					if IsOn(t.isPvP) then if t.isPvP then a = a .. d .. L["Is PvP"] else a = a .. d .. L["Not PvP"] end; d = ", " end
					if IsOn(t.inGroup) then if t.inGroup then a = a .. d .. L["In Group"] else a = a .. d .. L["Not In Group"] end; d = ", " end
					if IsOn(t.inParty) then if t.inParty then a = a .. d .. L["In Party"] else a = a .. d .. L["Not In Party"] end; d = ", " end
					if IsOn(t.inRaid) then if t.inRaid then a = a .. d .. L["In Raid"] else a = a .. d .. L["Not In Raid"] end; d = ", " end
					if IsOn(t.inInstance) then if t.inInstance then a = a .. d .. L["In Instance"] else a = a .. d .. L["Not In Instance"] end; d = ", " end
					if IsOn(t.hasPet) then if t.hasPet then a = a .. d .. L["Has Pet"] else a = a .. d .. L["No Pet"] end; d = ", " end
					if IsOn(t.checkLevel) and t.level then local x = "<"; if t.checkLevel then x = ">=" end;
						a = a .. d .. L["Level String"](x, t.level); d = ", " end
					if IsOn(t.checkHealth) and t.minHealth then local x = "<"; if t.checkHealth then x = ">=" end;
						a = a .. d .. L["Health String"](x, t.minHealth); d = ", " end
					if IsOn(t.checkPower) and t.minPower then local x = "<"; if t.checkPower then x = ">=" end;
						a = a .. d .. L["Power String"](x, t.minPower); d = ", " end
					if IsOn(t.checkHolyPower) and t.minHolyPower then local x = "<"; if t.checkHolyPower then x = ">=" end;
						a = a .. d .. L["Holy Power String"](x, t.minHolyPower); d = ", " end
					if IsOn(t.checkShards) and t.minShards then local x = "<"; if t.checkShards then x = ">=" end;
						a = a .. d .. L["Shards String"](x, t.minShards); d = ", " end
					if IsOn(t.checkEclipsePower) and t.minEclipsePower then local x = "<"; if t.checkEclipsePower then x = ">=" end;
						a = a .. d .. L["Eclipse String"](x, t.minEclipsePower); d = ", " end
					if IsOn(t.checkEclipse) and t.checkSolar then a = a .. d .. L["Solar Eclipse"]; d = ", " end
					if IsOn(t.checkEclipse) and t.checkLunar then a = a .. d .. L["Lunar Eclipse"]; d = ", " end
					if IsOn(t.checkEclipse) and t.checkSun then a = a .. d .. L[">> Sun"]; d = ", " end
					if IsOn(t.checkEclipse) and t.checkMoon then a = a .. d .. L[">> Moon"]; d = ", " end
					if IsOn(t.checkComboPoints) and t.minComboPoints then local x = "<"; if t.checkComboPoints then x = ">=" end;
						a = a .. d .. L["Combo Pts String"](x, t.minComboPoints); d = ", " end
					if IsOn(t.checkStance) and t.stance then a = a .. d .. L["Stance String"](t.stance); d = ", " end
					if IsOn(t.checkTalent) and t.talent then a = a .. d .. L["Talent String"](t.talent); d = ", " end
					if IsOn(t.checkRunes) and (t.checkAnyRune or t.needBlood or t.needFrost or t.needUnholy) then
						a = a .. d
						local p, pc = "", 0
						if t.checkAnyRune then
							a = a .. L["Any"]; pc = 1
						else
							if t.needBlood then a = a .. L["Blood"]; p = "+"; pc = pc + 1 end
							if t.needFrost then a = a .. p .. L["Frost"]; p = "+"; pc = pc + 1 end
							if t.needUnholy then a = a .. p .. L["Unholy"]; pc = pc + 1 end
						end
						if pc > 1 then a = a .. " " .. L["Runes"] else a = a .. " " .. L["Rune"] end
						d = ", "
					end
					if IsOn(t.checkRunes) then
						if t.checkBlood then a = a .. d .. L["Blood"] .. " " .. (t.chargeBlood and L["Recharging"] or L["Not Recharging"]) end
						if t.checkFrost then a = a .. d .. L["Frost"] .. " " .. (t.chargeFrost and L["Recharging"] or L["Not Recharging"]) end
						if t.checkUnholy then a = a .. d .. L["Unholy"] .. " " .. (t.chargeUnholy and L["Recharging"] or L["Not Recharging"]) end
						d = ", "
					end
					if IsOn(t.hasMainHand) and IsOn(t.levelMainHand) then a = a .. d .. L["Mainhand String"](t.levelMainHand); d = ", " end
					if IsOn(t.hasOffHand) and IsOn(t.levelOffHand) then a = a .. d .. L["Offhand String"](t.levelOffHand); d = ", " end
					if IsOn(t.hasRanged) and IsOn(t.levelRanged) then a = a .. d .. L["Ranged String"](t.levelRanged); d = ", " end
				elseif tt == "Pet Status" then
					if IsOn(t.checkHealth) and t.minHealth then local x = "<"; if t.checkHealth then x = ">=" end;
						a = a .. d .. L["Health String"](x, t.minHealth); d = ", " end
					if IsOn(t.checkPower) and t.minPower then local x = "<"; if t.checkPower then x = ">=" end;
						a = a .. d .. L["Power String"](x, t.minPower); d = ", " end
				elseif tt == "Target Status" or tt == "Focus Status" then
					if IsOn(t.isEnemy) then if t.isEnemy then a = a .. d .. L["Is Enemy"] else a = a .. d .. L["Not Enemy"] end; d = ", " end
					if IsOn(t.isFriend) then if t.isFriend then a = a .. d .. L["Is Friendly"] else a = a .. d .. L["Not Friendly"] end; d = ", " end
					if IsOn(t.isDead) then if t.isDead then a = a .. d .. L["Is Dead"] else a = a .. d .. L["Not Dead"] end; d = ", " end
					if t.classify and t.classification then
						local lc = classificationList[t.classification] -- localized classification name
						if not lc then lc = "Unknown" end
						if t.classify then a = a .. d .. L["Is "] .. lc else a = a .. d .. L["Not "] .. lc end
						d = ", "
					end
					if IsOn(t.inRange) then if t.inRange then a = a .. d .. L["In Range"] else a = a .. d .. L["Out Of Range"] end; d = ", " end
					if IsOn(t.isSteal) then if t.isSteal then a = a .. d .. L["Spellsteal"] else a = a .. d .. L["Not Spellsteal"] end; d = ", " end
					if IsOn(t.checkHealth) and t.minHealth then local x = "<"; if t.checkHealth then x = ">=" end;
						a = a .. d .. L["Health String"](x, t.minHealth); d = ", " end
					if IsOn(t.checkPower) and t.minPower then local x = "<"; if t.checkPower then x = ">=" end;
						a = a .. d .. L["Power String"](x, t.minPower); d = ", " end
				elseif tt == "All Buffs" or tt == "All Debuffs" then
					if IsOn(t.unit) and t.auras then
						for _, v in pairs(t.auras) do a = a .. d .. string.format("\"%s\"", v); d = ", " end
						if d == ", " then -- the list of auras contains at least one entry
							if t.toggle == true then a = a .. d .. L["Not All Active"] else a = a .. d .. L["All Active"] end
							a = a .. d .. L["On"] .. " " .. unitList[t.unit]
							if t.isMine == true then a = a .. d .. L["Cast By Player"] else if t.isMine == false then
								a = a .. d .. L["Cast By Other"] else a = a .. d .. L["Cast By Anyone"] end end
						end
					end
				elseif tt == "Any Buffs" or tt == "Any Debuffs" then
					if IsOn(t.unit) and t.auras then
						for _, v in pairs(t.auras) do a = a .. d .. string.format("\"%s\"", v); d = ", " end
						if d == ", " then -- the list of auras contains at least one entry
							if t.toggle == true then a = a .. d .. L["None Active"] else a = a .. d .. L["Any Active"] end
							a = a .. d .. L["On"] .. " " .. unitList[t.unit]
							if t.isMine == true then a = a .. d .. L["Cast By Player"] else if t.isMine == false then
								a = a .. d .. L["Cast By Other"] else a = a .. d .. L["Cast By Anyone"] end end
						end
					end
				elseif tt == "Buff Time Left" or tt == "Debuff Time Left" then
					if IsOn(t.unit) and IsOn(t.timeLeft) and t.aura and t.aura ~= "" then
						a = a .. string.format(" \"%s\"", t.aura); d = ", "
						a = a .. d .. L["On"] .. " " .. unitList[t.unit]
						local ds = string.format("%0d:%02d", math.floor(t.timeLeft / 60), t.timeLeft % 60)
						if t.toggle == true then a = a .. d .. L["Less Than"] .. " " .. ds else a = a .. d .. ds .. " " .. L["Or More"] end
						if t.isMine == true then a = a .. d .. L["Cast By Player"] else if t.isMine == false then
							a = a .. d .. L["Cast By Other"] else a = a .. d .. L["Cast By Anyone"] end end
					end
				elseif tt == "Buff Count" or tt == "Debuff Count" then
					if IsOn(t.unit) and t.aura and t.aura ~= "" then
						a = a .. string.format(" \"%s\"", t.aura); d = ", "
						a = a .. d .. "On" .. " " .. unitList[t.unit]
						local dcount = t.count
						if IsOff(dcount) then dcount = 1 end -- default count is 1
						if t.toggle == true then a = a .. d .. L["Less Than "] .. dcount else a = a .. d .. dcount .. " " .. L["Or More"] end
						if t.isMine == true then a = a .. d .. L["Cast By Player"] else if t.isMine == false then
							a = a .. d .. L["Cast By Other"] else a = a .. d .. L["Cast By Anyone"] end end
					end
				elseif tt == "Buff Type" then
					if IsOn(t.hasBuff) then
						a = a .. string.format(" \"%s\" ", t.hasBuff)
						if t.toggle == true then a = a .. L["Missing"] else a = a .. L["Present"] end
					end
				elseif tt == "Debuff Type" then
					if IsOn(t.hasDebuff) then
						a = a .. string.format(" \"%s\" ", t.hasDebuff)
						if t.toggle == true then a = a .. L["Missing"] else a = a .. L["Present"] end
					end
				elseif tt == "All Cooldowns" then
					if IsOn(t.timeLeft) and t.spells then
						for _, v in pairs(t.spells) do a = a .. d .. string.format("\"%s\"", v); d = ", " end
						if d == ", " then -- the list of spells contains at least one entry
							local ds = string.format("%0.1f ", t.timeLeft)
							if t.toggle == true then a = a .. d .. L["Less Than"] .. " " .. ds .. L["Seconds"] else a = a .. d .. ds .. L["Seconds Or More"] end
						end
					end
				elseif tt == "Spell Ready" then
					if t.spell and t.spell ~= "" then
						a = a .. string.format(" \"%s\"", t.spell); d = ", "
						if t.gcd == true then a = a .. d .. L["Test Global Cooldown"] else a = a .. d .. L["Ignore Global Cooldown"] end
						if t.notUsable == true then a = a .. d .. L["Ignore Spell Usable"] else a = a .. d .. L["Test Spell Usable"] end
						if IsOn(t.inRange) then
							if t.inRange == true then a = a .. d .. L["Target In Range"] else a = a .. d .. L["Target Out Of Range"] end
						end
					end
				elseif tt == "Spell Casting" then
					if t.spell and t.spell ~= "" and IsOn(t.unit) then
						a = a .. string.format(" \"%s\"", t.spell) .. ", " .. L["Cast By"] .. " " .. unitList[t.unit]							
					end
				elseif tt == "Item Ready" then
					if t.item and t.item ~= "" then
						a = a .. string.format(" \"%s\"", t.item); d = ", "
						if t.toggle == true then a = a .. d .. L["Item Is Not Ready"] else a = a .. d .. L["Item Is Ready"] end
					end
				end
				testString = testString .. at .. a -- add caption and clauses for current test
			end
		end
	end
	if IsOn(c.dependencies) then -- have a dependencies list
		local at = string.format(L["Dependencies String"])
		local a, d = "", " " -- a is string for current test conditions, d gets comma after first test clause added
		local i, list = 0, {} -- build a sorted list of dependencies
		for dep in pairs(c.dependencies) do
			i = i + 1
			list[i] = dep
		end
		table.sort(list)
		for _, dep in pairs(list) do
			local value, ctype = c.dependencies[dep], false
			if c.dependencyType then ctype = c.dependencyType[dep] end
			a = a .. d .. string.format("[%s] = ", dep)
			if value then a = a .. L["True Color String"] else a = a .. L["False Color String"] end
			a = a .. " (" .. (ctype and L["Or"] or L["And"]) .. ")"
			d = ", "
		end
		if a ~= "" then depString = depString .. at .. a end -- add caption and dependencies
	end
	if c.toggleResult then resultString = string.format(L["Toggle String"]) end
	if IsOn(c.setResult) then if c.setResult then resultString = L["Set Result True"] else resultString = L["Set Result False"] end end
	return description .. testString .. depString .. resultString
end

--[[ Templates for conditions
conditionTemplate = {			-- indexed by the condition's name (must be unique)
	name = name,				-- condition name
	enabled = true,				-- true if condition is to be evaluated
	shared = true,				-- shared means copy settings to SharedConditions
	tests = {},					-- table of tests, indexed by testType
	testResult = false,			-- result from most recent update (without checking dependencies and overrides)
	dependencies = {},			-- table of condition, result pairs that must match for this one to be true
	result = false,				-- result of most recent update (after checking dependencies and overrides)
}

testTemplate = {
	testType = "type",			-- type of this test
	unit = "unit",				-- unit the test applies to
	toggle = true,				-- toggle result of the test (i.e., logical not)
	inCombat = nil,				-- if not nil, check if in or out of combat
	isResting = nil,			-- if not nil, check if resting
	isStealthed = nil,			-- if not nil, check if stealthed
	isMounted = nil,			-- if not nil, check if mounted
	hasPet = nil,				-- if not nil, check if player has pet
	inRange = nil,				-- if not nil, check if unit is in range
	isEnemy = nil,				-- if not nil, check if unit exists and is enemy
	isFriend = nil,				-- if not nil, check if unit exists and is friend
	hasMainHand = nil,			-- if not nil, check if unit has mainhand weapon equipped
	levelMainHand = nil,		-- if not nil, minimum mainhand weapon level
	hasOffHand = nil,			-- if not nil, check if unit has offhand weapon equipped
	levelOffHand = nil,			-- if not nil, miniumum offhand weapon level
	hasRanged = nil,			-- if not nil, check if unit has ranged weapon equipped
	levelRanged = nil,			-- if not nil, miniumum ranged weapon level
	checkHealth = nil,			-- if not nil, check unit health versus targetHealth
	minHealth = nil,			-- if not nil, % minimum health required
	checkPower = nil,			-- if not nil, check unit power versus minPower
	minPower = nil,				-- if not nil, minimum % energy/rage/runic power required
	checkComboPoints = nil,		-- if not nil, check combo points versus minComboPoints
	minComboPoints = nil,		-- if not nil, minimum combo points required
	checkRunes = nil,			-- if not nil, check runes
	needBlood = nil				-- if true, must have a blood rune (or enough death runes)
	needFrost = nil				-- if true, must have a frost rune (or enough death runes)
	needUnholy = nil			-- if true, must have an unholy rune (or enough death runes)
	checkBlood = nil			-- if true, check if blood rune recharging (compare to chargeBlood)
	checkFrost = nil			-- if true, check if frost rune recharging (compare to chargeFrost)
	checkUnholy = nil			-- if true, check if unholy rune recharging (compare to chargeUnholy)
	checkLevel = nil,			-- if not nil, check player's level
	level = nil,				-- if not nil, minimum level for player
	checkStance = nil,			-- if not nil, check player's stance
	stance = nil,				-- if not nil, stance for player
	checkTalent = nil,			-- if not nil, check for a specific talent
	talent = nil,				-- if not nil, name of talent to check for
	item = nil,					-- if not nil, an item name
	spell = nil,				-- if not nil, a spell name
	gcd = nil,					-- if true then consider global cooldown for spell readiness
	notUsable = nil,			-- for spell ready, check if not usable due to mana, reagents, etc.
	anySpell = false,			-- spell check applies to any spell, not one specified spell
	channel = false,			-- check if spell is being channeled or not
	aura = nil,					-- aura for timeleft and count checks
	auras = {},					-- if not nil, a table of aura names
	isMine = true,				-- if not nil, check if aura cast by player
	timeLeft = nil,				-- if not nil, check time left on aura less than this value
	count = nil,				-- if not nil, check aura count is greater than or equal to this value
	hasDebuff = type,			-- if not nil, check if Poison/Curse/Magic/Disease
}
]]--
