-- Nest is a graphics package that is optimized to display Raven's bar groups.
-- Bar groups share layout and appearance options (dimensions, fonts, textures, configuration, special effects).
-- Each bar has a fixed set of graphical components (icon, iconText, foreground bar, background bar, labelText, timeText, spark).

local L = LibStub("AceLocale-3.0"):GetLocale("Raven")

Nest_SupportedConfigurations = { -- table of configurations can be used in dialogs to select appropriate options
	[1] = { name = L["Right-to-left bars, label left, icon left"], iconOnly = false, bars = "r2l", label = "left", icon = "left" },
	[2] = { name = L["Left-to-right bars, label left, icon left"], iconOnly = false, bars = "l2r", label = "left", icon = "left" },
	[3] = { name = L["Right-to-left bars, label right, icon left"], iconOnly = false, bars = "r2l", label = "right", icon = "left" },
	[4] = { name = L["Left-to-right bars, label right, icon left"], iconOnly = false, bars = "l2r", label = "right", icon = "left" },
	[5] = { name = L["Right-to-left bars, label left, icon right"], iconOnly = false, bars = "r2l", label = "left", icon = "right" },
	[6] = { name = L["Left-to-right bars, label left, icon right"], iconOnly = false, bars = "l2r", label = "left", icon = "right" },
	[7] = { name = L["Right-to-left bars, label right, icon right"], iconOnly = false, bars = "r2l", label = "right", icon = "right" },
	[8] = { name = L["Left-to-right bars, label right, icon right"], iconOnly = false, bars = "l2r", label = "right", icon = "right" },
	[9] = { name = L["Icons in rows, with right-to-left mini-bars"], iconOnly = true, bars = "r2l", orientation = "horizontal" },
	[10] = { name = L["Icons in rows, with left-to-right mini-bars"], iconOnly = true, bars = "l2r", orientation = "horizontal" },
	[11] = { name = L["Icons in columns, right-to-left mini-bars"], iconOnly = true, bars = "r2l", orientation = "vertical" },
	[12] = { name = L["Icons in columns, left-to-right mini-bars"], iconOnly = true, bars = "l2r", orientation = "vertical" },
}
Nest_MaxBarConfiguration = 8
Nest_MaxIconConfiguration = 12

local barGroups = {} -- current barGroups
local usedBarGroups = {} -- cache of recycled barGroups
local usedBars = {} -- cache of recycled bars
local update = false -- set whenever a global change has occured
local buttonName = 0 -- incremented for each button created
local callbacks = {} -- registered callback functions
local displayWidth, displayHeight = UIParent:GetWidth(), UIParent:GetHeight()

local anchorDefaults = { -- backdrop initialization for bar group anchors
	bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
	tile = true, tileSize = 8, edgeSize = 8, insets = { left = 2, right = 2, top = 2, bottom = 2 }
}

local bgTemplate = { -- these fields are cleared when a bar group is deleted
	barWidth  = 0, barHeight = 0, iconSize = 0, scale = 0, spacingX = 0, spacingY = 0, iconOffsetX = 0, iconOffsetY = 0,
	labelOffset = 0, labelInset = 0, labelWrap = 0, labelAlign = 0, timeOffset = 0, timeInset = 0, timeAlign = 0, iconOffset = 0, iconInset = 0,
	configuration = 0, reverse = 0, wrap = 0, wrapDirection = 0, snapCenter = 0, maxBars = 0, width = 0, height = 0,
	labelFont = 0, labelFSize = 0, labelAlpha = 0, labelColor = 0, labelOutline = 0,
	timeFont = 0, timeFSize = 0, timeAlpha = 0, timeColor = 0, timeOutline = 0,
	iconFont = 0, iconFSize = 0, iconAlpha = 0, iconColor = 0, iconOutline = 0,
	fgTexture = 0, bgTexture = 0, fgAlpha = 0, bgAlpha = 0, fgSaturation = 0, fgBrightness = 0, bgSaturation = 0, bgBrightness = 0,
	timeFormat = 0, timeSpaces = 0, timeCase = 0, fgNotTimer = 0,
	showIcon = 0, showCooldown = 0, showBar = 0, showSpark = 0, showLabelText = 0, showTimeText = 0,
	relativeTo = 0, relativeX = 0, relativeY = 0, relativeLastBar = 0, relativeEmpty = 0, lastX = 0, lastY = 0, anchorPoint = 0,
}

local barTemplate = { -- these fields are cleared with a bar is deleted
	startTime = 0, offsetTime = 0, timeLeft = 0, duration = 0, maxTime = 0, alpha = 0, flash = 0, label = 0, iconCount = 0, tooltipAnchor = 0,
	cr = 0, cg = 0, cb = 0, ca = 0, br = 0, bg = 0, bb = 0, ba = 0, ibr = 0, ibg = 0, ibb = 0, iba = 0
}

-- Calculate alpha for flashing bars, period is how long the total flash time should last
function Nest_FlashAlpha(maxAlpha, period)
	local frac = GetTime() / period
	frac = frac - math.floor(frac) -- get fractional part of current period
	if frac >= 0.5 then frac = 1 - frac end -- now goes from 0 to 0.5 then back to 0
	return (maxAlpha / 2) + (maxAlpha * frac) -- drops to half of maxAlpha
end

-- Optional ButtonFacade support
local LBF = LibStub("LibButtonFacade", true)
local LBF_Skin = {} -- local cache for current skin settings
local LBF_Group = LBF and LBF:Group("Raven", "NestIcons") or nil -- ButtonFacade group for skinning bar group icons
local LBF_ButtonData = { AutoCast = false, AutoCastable = false, Border = false, Checked = false, Cooldown= false, Count = false,
	Disabled = false, Flash = false, Highlight = false, HotKey = false, Icon = false, Name = false, Normal = false, Pushed = false }

local function LBF_SkinCallback(arg, SkinID, Gloss, Backdrop, Group, Button, Colors)
	LBF_Skin.SkinID = SkinID; LBF_Skin.Gloss = Gloss; LBF_Skin.Backdrop = Backdrop; LBF_Skin.Colors = Colors
end

do if LBF then LBF:RegisterSkinCallback("Raven", LBF_SkinCallback) end end

function Nest_SetButtonFacadeData(SkinID, Gloss, Backdrop, Colors)
	if LBF and LBF_Group and SkinID then
		LBF_Skin.SkinID = SkinID; LBF_Skin.Gloss = Gloss; LBF_Skin.Backdrop = Backdrop; LBF_Skin.Colors = Colors
		LBF_Group:Skin(LBF_Skin.SkinID, LBF_Skin.Gloss, LBF_Skin.Backdrop, LBF_Skin.Colors)
	end
end

function Nest_GetButtonFacadeData(SkinID, Gloss, Backdrop, Colors)
	return LBF_Skin.SkinID, LBF_Skin.Gloss, LBF_Skin.Backdrop, LBF_Skin.Colors
end

-- Bar sorting functions: alphabetic, time left, duration, bar's start time
-- Values are assumed equal if difference less than 0.05 seconds
local function sortValues(a, b, f, up)
	if math.abs(a[f] - b[f]) >= 0.05 then if up then return a[f] < b[f] else return a[f] > b[f] end end
	return a.name < b.name -- use ascending alphabetic order for equal values
end

local function SortTimeDown(a, b) return sortValues(a, b, "timeLeft", false) end
local function SortTimeUp(a, b) return sortValues(a, b, "timeLeft", true) end
local function SortDurationDown(a, b) return sortValues(a, b, "duration", false) end
local function SortDurationUp(a, b) return sortValues(a, b, "duration", true) end
local function SortStartDown(a, b) return sortValues(a, b, "start", false) end
local function SortStartUp(a, b) return sortValues(a, b, "start", true) end
local function SortAlphaDown(a, b) return a.name > b.name end
local function SortAlphaUp(a, b) return a.name < b.name end

-- Register callbacks that can be used by internal functions to communicate in special cases
function Nest_RegisterCallbacks(cbs) if cbs then for k, v in pairs(cbs) do callbacks[k] = v end end end

-- Event handling functions for bar group anchors, pass both anchor and bar group
local function BarGroup_OnEvent(anchor, callback)
	local bg, bgName = nil, anchor.bgName
	if bgName then bg = barGroups[bgName] end -- locate the bar group associated with the anchor
	if bg then
		local func = bg.callbacks[callback]
		if func then func(anchor, bgName) end
	end
end

local function BarGroup_OnEnter(anchor) BarGroup_OnEvent(anchor, "onEnter") end
local function BarGroup_OnLeave(anchor) BarGroup_OnEvent(anchor, "onLeave") end

-- OnClick does a callback (except for unmodified left click), passing bar group name and button
local function BarGroup_OnClick(anchor, button)
	local bg, bgName = nil, anchor.bgName
	if bgName then bg = barGroups[bgName] end -- locate the bar group associated with the anchor
	if ((button ~= "LeftButton") or IsModifierKeyDown()) and bg and not bg.locked then
		local func = bg.callbacks.onClick -- only pass left clicks if no modifier key is down
		if func then func(anchor, bgName, button) end
	end
end

-- OnMouseDown with no modifier key starts moving if frame unlocked and does callback, passing bar group name
local function BarGroup_OnMouseDown(anchor, button)
	local bg, bgName = nil, anchor.bgName
	if bgName then bg = barGroups[bgName] end -- locate the bar group associated with the anchor
	if (button == "LeftButton") and not IsModifierKeyDown() and bg and not bg.locked then
		bg.startX = bg.frame:GetLeft(); bg.startY = bg.frame:GetTop()
		bg.moving = true
		bg.frame:SetFrameStrata("HIGH")
		bg.frame:StartMoving()
		local func = bg.callbacks.onMove -- called to start movement as long as no modifier key is down
		if func then func(anchor, bgName) end
	end
end

-- OnMouseUp stops moving if frame is in motion and does a callback passing bar group name to indicate movement
local function BarGroup_OnMouseUp(anchor, button)
	local bg, bgName = nil, anchor.bgName
	if bgName then bg = barGroups[bgName] end -- locate the bar group associated with the anchor
	if bg and bg.moving then
		bg.frame:StopMovingOrSizing()
		bg.frame:SetFrameStrata(bg.strata or "MEDIUM")
		local func = bg.callbacks.onMove
		if func then
			local endX = bg.frame:GetLeft(); local endY = bg.frame:GetTop()
			if bg.startX ~= endX or bg.startY ~= endY then func(anchor, bgName) end -- only fires if actually moved
		end
		bg.moving = false
	end
end

-- Initialize and return a new bar group containing either timer bars or enhanced icons
function Nest_CreateBarGroup(name)
	if barGroups[name] then return nil end -- already have one with that name
	local n, bg = next(usedBarGroups) -- get any available recycled bar group
	if n then
		usedBarGroups[n] = nil
	else
		bg = {}
		bg.frame = CreateFrame("Frame", "RavenBarGroup" .. name, UIParent) -- add name for reference from other addons
		bg.frame:SetFrameLevel(bg.frame:GetFrameLevel() + 20) -- higher than other addons
		bg.frame:SetMovable(true); bg.frame:SetClampedToScreen(true)
		bg.frame:ClearAllPoints(); bg.frame:SetPoint("CENTER", UIParent, "CENTER")		
		bg.anchor = CreateFrame("Button", nil, bg.frame)
		bg.anchor:SetBackdrop(anchorDefaults)
		bg.anchor:SetBackdropColor(0.3, 0.3, 0.3, 0.9)
		bg.anchor:SetBackdropBorderColor(0, 0, 0, 0.9)
		bg.anchor:SetNormalFontObject(ChatFontSmall)		
		bg.bars = {}
		bg.sorter = {}
		bg.attributes = {}
		bg.callbacks = {}
		bg.position = {}
		bg.sortFunction = SortAlphaUp
		bg.locked = false; bg.moving = false
		bg.count = 0
	end
	bg.anchor.bgName = name
	bg.anchor:SetScript("OnMouseDown", BarGroup_OnMouseDown)
	bg.anchor:SetScript("OnMouseUp", BarGroup_OnMouseUp)
	bg.anchor:SetScript("OnClick", BarGroup_OnClick)
	bg.anchor:SetScript("OnEnter", BarGroup_OnEnter)
	bg.anchor:SetScript("OnLeave", BarGroup_OnLeave)
	bg.anchor:RegisterForClicks("LeftButtonUp", "RightButtonUp")
	bg.anchor:EnableMouse(true)
	table.wipe(bg.position)
	bg.name = name
	bg.update = true
	barGroups[name] = bg
	update = true
	return bg
end

-- Return the bar group with the specified name
function Nest_GetBarGroup(name) return barGroups[name] end

-- Return the table of bar groups
function Nest_GetBarGroups() return barGroups end

-- Delete a bar group and move it to the recycled bar group table
function Nest_DeleteBarGroup(bg)
	for _, bar in pairs(bg.bars) do Nest_DeleteBar(bg, bar) end -- empty out bars table
	for n in pairs(bg.sorter) do bg.sorter[n] = nil end -- empty the sorting table
	for n in pairs(bg.attributes) do bg.attributes[n] = nil end
	for n in pairs(bg.callbacks) do bg.callbacks[n] = nil end
	for n in pairs(bgTemplate) do bg[n] = nil end -- reset current bar group settings
	bg.frame:ClearAllPoints(); bg.frame:SetPoint("CENTER", UIParent, "CENTER") -- return to neutral position
	bg.anchor:SetScript("OnMouseDown", nil)
	bg.anchor:SetScript("OnMouseUp", nil)
	bg.anchor:SetScript("OnClick", nil)
	bg.anchor:SetScript("OnEnter", nil)
	bg.anchor:SetScript("OnLeave", nil)
	bg.anchor:EnableMouse(false)
	bg.anchor.bgName = nil
	bg.sortFunction = SortAlphaUp
	bg.count = 0
	bg.locked = false; bg.moving = false
	bg.update = false
	bg.anchor:Hide()
	barGroups[bg.name] = nil
	bg.name = nil
	table.insert(usedBarGroups, bg)
	update = true
end

-- Set layout options for a bar group
function Nest_SetBarGroupBarLayout(bg, barWidth, barHeight, iconSize, scale, spacingX, spacingY, iconOffsetX, iconOffsetY,
			labelOffset, labelInset, labelWrap, labelAlign, timeOffset, timeInset, timeAlign, iconOffset, iconInset,
			configuration, reverse, wrap, wrapDirection, snapCenter, maxBars, strata)
	bg.barWidth = barWidth; bg.barHeight = barHeight; bg.iconSize = iconSize; bg.scale = scale or 1; bg.maxBars = maxBars; bg.strata = strata
	bg.spacingX = spacingX or 0; bg.spacingY = spacingY or 0; bg.iconOffsetX = iconOffsetX or 0; bg.iconOffsetY = iconOffsetY or 0
	bg.labelOffset = labelOffset or 0; bg.labelInset = labelInset or 0; bg.labelWrap = labelWrap; bg.labelAlign = labelAlign or "MIDDLE"
	bg.timeOffset = timeOffset or 0; bg.timeInset = timeInset or 0; bg.timeAlign = timeAlign or "normal"
	bg.iconOffset = iconOffset or 0; bg.iconInset = iconInset or 0
	bg.configuration = configuration or 1; bg.reverse = reverse; bg.wrap = wrap or 0; bg.wrapDirection = wrapDirection; bg.snapCenter = snapCenter
	bg.update = true
end
	
-- Set label font options for a bar group
function Nest_SetBarGroupLabelFont(bg, font, fsize, alpha, color, outline)
	if not color then color = { r = 1, g = 1, b = 1, a = 1 } end
	bg.labelFont = font; bg.labelFSize = fsize or 9; bg.labelAlpha = alpha or 1; bg.labelColor = color; bg.labelOutline = outline
	bg.update = true
end

-- Set time text font options for a bar group
function Nest_SetBarGroupTimeFont(bg, font, fsize, alpha, color, outline)
	if not color then color = { r = 1, g = 1, b = 1, a = 1 } end
	bg.timeFont = font; bg.timeFSize = fsize or 9; bg.timeAlpha = alpha or 1; bg.timeColor = color; bg.timeOutline = outline
	bg.update = true
end

-- Set icon text font options for a bar group
function Nest_SetBarGroupIconFont(bg, font, fsize, alpha, color, outline)
	if not color then color = { r = 1, g = 1, b = 1, a = 1 } end
	bg.iconFont = font; bg.iconFSize = fsize or 9; bg.iconAlpha = alpha or 1; bg.iconColor = color; bg.iconOutline = outline
	bg.update = true
end

-- Set texture options for a bar group
function Nest_SetBarGroupTextures(bg, fgTexture, fgAlpha, bgTexture, bgAlpha, fgNotTimer, fgSaturation, fgBrightness, bgSaturation, bgBrightness)
	bg.fgTexture = fgTexture; bg.fgAlpha = fgAlpha; bg.bgTexture = bgTexture; bg.bgAlpha = bgAlpha; bg.fgNotTimer = fgNotTimer
	bg.fgSaturation = fgSaturation or 0; bg.fgBrightness = fgBrightness or 0; bg.bgSaturation = bgSaturation or 0; bg.bgBrightness = bgBrightness or 0
	bg.update = true
end

-- Select visible components for a bar group
function Nest_SetBarGroupVisibles(bg, icon, cooldown, bar, spark, labelText, timeText)
	bg.showIcon = icon; bg.showCooldown = cooldown; bg.showBar = bar; bg.showSpark = spark
	bg.showLabelText = labelText; bg.showTimeText = timeText
	bg.update = true
end

-- Sort the bars in a bar group using the designated sort method and direction (default is sort by name alphabetically)
function Nest_BarGroupSortFunction(bg, sortMethod, sortDirection)
	if sortMethod == "time" then -- sort by time left on the bar
		if sortDirection then bg.sortFunction = SortTimeDown else bg.sortFunction = SortTimeUp end
	elseif sortMethod == "duration" then -- sort by bar duration
		if sortDirection then bg.sortFunction = SortDurationDown else bg.sortFunction = SortDurationUp end
	elseif sortMethod == "start" then -- sort by bar start time
		if sortDirection then bg.sortFunction = SortStartDown else bg.sortFunction = SortStartUp end
	else -- default is sort alphabetically by bar name
		if sortDirection then bg.sortFunction = SortAlphaDown else bg.sortFunction = SortAlphaUp end
	end
	bg.update = true
end

-- Set the time format function for the bar group, if not set will use default
function Nest_SetBarGroupTimeFormat(bg, timeFormat, timeSpaces, timeCase)
	bg.timeFormat = timeFormat; bg.timeSpaces = timeSpaces; bg.timeCase = timeCase
	bg.update = true
end

-- If locked is true then lock the bar group anchor, otherwise unlock it
function Nest_SetBarGroupLock(bg, locked)
	bg.locked = locked
	bg.update = true
end

-- Return a bar group's display position as percentages of actual display size to edges of the anchor frame
-- Return values are descaled to match UIParent and include left, right, bottom and top plus descaled width and height
function Nest_GetAnchorPoint(bg)
	local scale = bg.scale or 1
	local dw, dh = displayWidth, displayHeight
	local w, h = bg.frame:GetWidth() * scale, bg.frame:GetHeight() * scale
	local left, bottom = bg.frame:GetLeft(), bg.frame:GetBottom() -- get scaled coordinates for frame's anchor
	if left and bottom then left = (left * scale); bottom = (bottom * scale) else left = dw / 2; bottom = dh / 2 end -- default to center
	local right, top = dw - (left + w), dh - (bottom + h)
	local p = bg.position; p.left, p.right, p.bottom, p.top, p.width, p.height = left / dw, right / dw, bottom / dh, top / dh, w, h
	return p.left, p.right, p.bottom, p.top, p.width, p.height
end

-- Set a bar group's scaled display position from left, right, bottom, top where left and bottom should always be valid
-- Use right, top, width and height only if valid and closer to that edge to fix position shift when UIParent dimensions change
function Nest_SetAnchorPoint(bg, left, right, bottom, top, scale, width, height)
	if left and bottom and width and height then -- make sure valid settings
		bg.scale = scale -- make sure save scale since may not have been initialized yet
		local p = bg.position; p.left, p.right, p.bottom, p.top, p.width, p.height = left, right, bottom, top, width, height
		local dw, dh = displayWidth, displayHeight
		local xoffset = left * dw
		local yoffset = bottom * dh
		if right and top and width and height then -- optionally set from other edges if closer to them
			if left > 0.5 then xoffset = dw - (right * dw) - width end
			if bottom > 0.5 then yoffset = dh - (top * dh) - height end
		end
		bg.frame:SetScale(scale); bg.frame:SetSize(width, height)
		bg.frame:ClearAllPoints(); bg.frame:SetPoint("BOTTOMLEFT", nil, "BOTTOMLEFT", xoffset / scale, yoffset / scale)
	end
end

-- Set a bar group's display position as relative to another bar group
function Nest_SetRelativeAnchorPoint(bg, rTo, rX, rY, rLB, rEmpty)
	if bg.relativeTo and not rTo then -- removing a relative anchor point
		local left, bottom = bg.frame:GetLeft(), bg.frame:GetBottom()
		bg.frame:ClearAllPoints(); bg.frame:SetPoint("BOTTOMLEFT", nil, "BOTTOMLEFT", left, bottom)
		bg.relativeTo = nil -- remove relative anchor point
	else
		bg.relativeTo = rTo -- if relativeTo is nil then relative anchor point is not set
		bg.relativeX = rX; bg.relativeY = rY; bg.relativeLastBar = rLB; bg.relativeEmpty = rEmpty
	end
end

-- Set callbacks for a bar group
function Nest_SetBarGroupCallbacks(bg, onMove, onClick, onEnter, onLeave)
	bg.callbacks.onMove = onMove; bg.callbacks.onClick = onClick; bg.callbacks.onEnter = onEnter; bg.callbacks.onLeave = onLeave	
end

-- Set a bar group attribute. This is the mechanism to associate application-specific data with bar groups.
function Nest_SetBarGroupAttribute(bg, name, value) bg.attributes[name] = value end

-- Get a bar group attribute. This is the mechanism to associate application-specific data with bar groups.
function Nest_GetBarGroupAttribute(bg, name) return bg.attributes[name] end

-- Set an attribute for all bars in the bar group
function Nest_SetAllAttributes(bg, name, value)
	for _, bar in pairs(bg.bars) do bar.attributes[name] = value end
end

-- Delete all bars in the bar group with the specifed attribute value (useful for mark/sweep garbage collection)
function Nest_DeleteBarsWithAttribute(bg, name, value)
	for barName, bar in pairs(bg.bars) do
		if bar.attributes[name] == value then Nest_DeleteBar(bg, bar) end
	end
end

-- Event handling functions for bars with callback
local function Bar_OnEvent(frame, callback, value)
	local bg, bgName, name = nil, frame.bgName, frame.name
	if bgName then bg = barGroups[bgName] end -- locate the bar group associated with the anchor
	if bg then
		local bar = bg.bars[name]
		if bar then
			if not value then value = bar.tooltipAnchor end
			local func = bar.callbacks[callback]
			if func then func(frame, bgName, name, value) end
		end
	end
end

local function Bar_OnEnter(frame) Bar_OnEvent(frame, "onEnter") end
local function Bar_OnLeave(frame) Bar_OnEvent(frame, "onLeave") end
local function Bar_OnClick(frame, button) Bar_OnEvent(frame, "onClick", button) end

local function GetButtonName() buttonName = buttonName + 1; return "RavenButton" .. tostring(buttonName) end -- unique button name

-- Initialize and return a new bar of the specified type
function Nest_CreateBar(bg, name)
	if bg.bars[name] then return nil end -- already have one with that name
	local n, bar = next(usedBars) -- get any available recycled bar
	if n then
		usedBars[n] = nil
		bar.frame:SetParent(bg.frame)
	else
		bar = {}
		bar.frame = CreateFrame("Frame", nil, bg.frame)
		bar.container = CreateFrame("Frame", nil, bar.frame)
		bar.fgTexture = bar.container:CreateTexture(nil, "BORDER")	
		bar.bgTexture = bar.container:CreateTexture(nil, "BACKGROUND")
		bar.spark = bar.container:CreateTexture(nil, "OVERLAY")
		bar.spark:SetTexture([[Interface\CastingBar\UI-CastingBar-Spark]])
		bar.spark:SetSize(10, 10)
		bar.spark:SetBlendMode("ADD")
		bar.spark:SetTexCoord(0, 1, 0, 1)	
		bar.labelText = bar.container:CreateFontString(nil, "OVERLAY")		
		bar.timeText = bar.container:CreateFontString(nil, "OVERLAY")
		local bname = GetButtonName()
		bar.icon = CreateFrame("Button", bname, bar.frame)
		bar.icon:SetFrameLevel(bar.frame:GetFrameLevel() + 5)
		bar.iconTexture = bar.icon:CreateTexture(nil, "ARTWORK") -- texture for the bar's icon
		bar.cooldown = CreateFrame("Cooldown", bname .. "Cooldown", bar.icon) -- cooldown overlay to animate timer
		bar.cooldown:SetReverse(true)
		bar.iconTextFrame = CreateFrame("Frame", nil, bar.icon)
		bar.iconTextFrame:SetFrameLevel(bar.cooldown:GetFrameLevel() + 5)
		bar.iconText = bar.iconTextFrame:CreateFontString(nil, "OVERLAY")
		bar.iconBorder = bar.icon:CreateTexture() -- only used for LBF
		
		local anim = bar.icon:CreateAnimationGroup()
		anim:SetLooping("NONE")
		local grow = anim:CreateAnimation("Scale")
		grow:SetScale(3, 3); grow:SetOrigin('CENTER', 0, 0); grow:SetDuration(0.25); grow:SetOrder(1)
		local shrink = anim:CreateAnimation("Scale")
		shrink:SetScale(-3, -3); shrink:SetOrigin('CENTER', 0, 0); shrink:SetDuration(0.25); shrink:SetOrder(2)
		bar.icon.anim = anim
		
		if LBF then -- if using ButtonFacade, create and initialize a button data table
			bar.buttonData = {} -- only initialize once so no garbage collection issues
			for k, v in pairs(LBF_ButtonData) do bar.buttonData[k] = v end
		else -- use default button texture
			bar.icon:SetNormalTexture("Interface\\AddOns\\Raven\\Normal")	
		end

		bar.attributes = {}
		bar.callbacks = {}
	end
	bar.frame.name = name
	bar.frame.bgName = bg.name
	bar.frame:SetScript("OnMouseUp", Bar_OnClick)
	bar.frame:SetScript("OnEnter", Bar_OnEnter)
	bar.frame:SetScript("OnLeave", Bar_OnLeave)
	bar.icon.name = name
	bar.icon.bgName = bg.name
	bar.icon:SetScript("OnMouseUp", Bar_OnClick)
	bar.icon:SetScript("OnEnter", Bar_OnEnter)
	bar.icon:SetScript("OnLeave", Bar_OnLeave)
	bar.icon:RegisterForClicks("LeftButtonUp", "RightButtonUp")
	bar.startTime = GetTime()
	bar.name = name
	bar.update = true
	bg.bars[name] = bar
	bg.count = bg.count + 1
	bg.sorter[bg.count] = { name = name }
	bg.update = true
	return bar
end

-- Return the bar with the specified name
function Nest_GetBar(bg, name) return bg.bars[name] end

-- Delete a bar from a bar group, moving it to recycled bar table
function Nest_DeleteBar(bg, bar)
	for n in pairs(bar.attributes) do bar.attributes[n] = nil end
	for n in pairs(bar.callbacks) do bar.callbacks[n] = nil end
	for n in pairs(barTemplate) do bar[n] = nil end -- reset current bar settings
	bar.icon:EnableMouse(false); bar.frame:EnableMouse(false)
	bar.frame:SetScript("OnMouseUp", nil)
	bar.frame:SetScript("OnEnter", nil)
	bar.frame:SetScript("OnLeave", nil)
	bar.frame.name = nil
	bar.frame.bgName = nil
	bar.icon:SetScript("OnMouseUp", nil)
	bar.icon:SetScript("OnEnter", nil)
	bar.icon:SetScript("OnLeave", nil)
	bar.icon.name = nil
	bar.icon.bgName = nil
	bar.icon.anim:Stop()
	bar.cooldown:SetCooldown(0, 0)
	bar.update = false	
	bar.fgTexture:Hide(); bar.bgTexture:Hide(); bar.spark:Hide(); bar.icon:Hide();
	bar.iconText:Hide(); bar.labelText:Hide(); bar.timeText:Hide()
	if callbacks.release then callbacks.release(bar) end
	
	local i = 1
	while i <= bg.count do -- find and remove the corresponding entry in the sorting table
		if bg.sorter[i].name == bar.name then
			if i ~= bg.count then bg.sorter[i] = bg.sorter[bg.count] end -- copy last one to fill the hole
			bg.sorter[bg.count] = nil
			bg.count = bg.count - 1
			break
		end
		i = i + 1
	end
	bg.bars[bar.name] = nil
	bar.name = nil
	table.insert(usedBars, bar)
	bg.update = true
end

-- Delete all bars in a bar group
function Nest_DeleteAllBars(bg)
	for barName, bar in pairs(bg.bars) do Nest_DeleteBar(bg, bar) end
end

-- Start (or restart) a timer bar, note that maxTime is the display maximum which may be less than duration
function Nest_StartTimer(bar, timeLeft, duration, maxTime)
	bar.startTime = GetTime() -- time the timer bar was started (or restarted, which counts as a new timer)
	bar.offsetTime = duration - timeLeft -- save offset since may be sent multiple times
	bar.timeLeft = timeLeft; bar.duration = duration; bar.maxTime = maxTime or duration
	bar.cooldown:SetCooldown(bar.startTime - bar.offsetTime, duration)
	bar.update = true
end

-- Return true if time parameters have been set for a bar
function Nest_IsTimer(bar) return bar.timeLeft ~= nil end

-- Get the time parameters for a bar, including adjusted timeLeft amount
function Nest_GetTimes(bar) return bar.timeLeft, bar.duration, bar.maxTime, bar.startTime, bar.offsetTime end

-- Set bar colors, includes foreground, background and icon border codes
function Nest_SetColors(bar, cr, cg, cb, ca, br, bg, bb, ba, ibr, ibg, ibb, iba)
	bar.cr = cr; bar.cg = cg; bar.cb = cb; bar.ca = ca
	bar.br = br; bar.bg = bg; bar.bb = bb; bar.ba = ba
	bar.ibr = ibr; bar.ibg = ibg; bar.ibb = ibb; bar.iba = iba
end

-- Set the overall alpha for a bar, this is last alpha adjustment made before bar is displayed
function Nest_SetAlpha(bar, alpha) bar.alpha = alpha end

-- Set whether the bar should flash or not
function Nest_SetFlash(bar, flash) bar.flash = flash end

-- Set the label text for a bar
function Nest_SetLabel(bar, label) bar.label = label end

-- Set the icon texture for a bar
function Nest_SetIcon(bar, icon) bar.iconTexture:SetTexture(icon) end

-- Set the numeric value to display on the bar's icon
function Nest_SetCount(bar, iconCount) bar.iconCount = iconCount end

-- Set a bar attribute. This is the mechanism to associate application-specific data with bars.
function Nest_SetAttribute(bar, name, value) bar.attributes[name] = value end

-- Get a bar attribute. This is the mechanism to associate application-specific data with bars.
function Nest_GetAttribute(bar, name) return bar.attributes[name] end

-- Set callbacks for a bar
function Nest_SetCallbacks(bar, onClick, onEnter, onLeave)
	bar.callbacks.onClick = onClick; bar.callbacks.onEnter = onEnter; bar.callbacks.onLeave = onLeave	
end

-- Set saturation and brightness of RGB colors by converting into HSL, adjusting saturation, then converting back to RGB
-- Input color RGB components are values between 0 to 1.0, saturation and brightness are values between -1.0 and 1.0
local function LevelAdjust(v, a) -- apply adjustment in range -1..+1 to either saturation or brightness
	if a ~= 0 then if a >= -1 and a < 0 then return v * (a + 1) elseif a > 0 and a <= 1 then return v + ((1 - v) * a) end end
	return v
end

function Nest_AdjustColor(r, g, b, saturation, brightness)
	if not r or not g or not b then return 0.5, 0.5, 0.5 end -- avoid errors if passed in nil values
	if not saturation then saturation = 0 end; if not brightness then brightness = 0 end -- set to default values
	if (saturation == 0) and (brightness == 0) then return r, g, b end	
	local ch, cs, cl, v, m, sv, h, sextant, fract, mid1, mid2, r2, g2, b2, vsf
	local mincolor, maxcolor = math.min(r, g, b), math.max(r, g, b)
	if mincolor == maxcolor then
		ch, cs, cl = 0, 0, r
	else
		v = maxcolor - mincolor; cl = (mincolor + maxcolor) / 2	
		if cl < 0.5 then cs = v / (maxcolor + mincolor) else cs = v / (2 - maxcolor - mincolor) end	
		r2 = (maxcolor - r) / v; g2 = (maxcolor - g) / v; b2 = (maxcolor - b) / v	
		if r == maxcolor then
			if g == mincolor then ch = b2 + 5 else ch = 1 - g2 end
		elseif g == maxcolor then
			if b == mincolor then ch = r2 + 1 else ch = 3 - b2 end
		else
			if r == mincolor then ch = g2 + 3 else ch = 5 - r2 end
		end		
		ch = ch / 6
	end
	if saturation < -1 then saturation = -1 elseif saturation > 1 then saturation = 1 end
	cs = LevelAdjust(cs, saturation) -- adjust the saturation, using original -1 .. +1 scale
	if brightness < -1 then brightness = -1 elseif brightness > 1 then brightness = 1 end
	cl = LevelAdjust(cl, brightness / 2) -- adjust the brightness, restricting the range to -0.5 to +0.5
	r, g, b = cl, cl, cl
	if cl <= 0.5 then v = cl * (1 + cs) else v = (cl + cs) - (cl * cs) end
	if v > 0 then
		m = cl + cl - v; sv = (v - m) / v; h = ch * 6
		sextant = math.floor(h); fract = h - sextant; vsf = v * sv * fract
		mid1 = m + vsf; mid2 = v - vsf
		if sextant == 0 then
			r, g, b = v, mid1, m
		elseif sextant == 1 then
			r, g, b = mid2, v, m
		elseif sextant == 2 then
			r, g, b = m, v, mid1
		elseif sextant == 3 then
			r, g, b = m, mid2, v
		elseif sextant == 4 then
			r, g, b = mid1, m, v
		elseif sextant == 5 then
			r, g, b = v, m, mid2
		else
			r, g, b = v, mid1, m
		end
	end
	return r, g, b
end

-- Update a bar group's anchor, showing it only if the bar group is unlocked
local function BarGroup_UpdateAnchor(bg, config)
	bg.anchor:SetSize(bg.width, bg.height)
	bg.anchor:SetText(bg.name)
	local align = "BOTTOMLEFT" -- select corner to attach based on configuration
	if config.iconOnly then -- icons can grow in any direction
		if config.orientation == "horizontal" then
			if bg.reverse then align = "BOTTOMRIGHT" end -- align rights for going left (reverse=true), lefts for right (reverse=false)
		else -- must be "vertical"
			if not bg.reverse then align = "TOPLEFT" end -- align bottoms for going up (reverse=true), tops for down (reverse=false)
		end
	else -- bars can grow either up are down
		if not bg.reverse then align = "TOPLEFT" end -- align bottoms for going up (reverse=true), tops for down (reverse=false)
	end
	bg.anchor:ClearAllPoints(); bg.anchor:SetPoint(align, bg.frame, align)
	if not bg.locked then bg.anchor:Show() else bg.anchor:Hide() end
end

-- Update a bar's layout based on the bar group configuration and dimension settings
-- Layout includes relative position of components plus mouse click rectangle and tooltip position
local function Bar_UpdateLayout(bg, bar, config)
	bar.icon:ClearAllPoints(); bar.iconTexture:ClearAllPoints(); bar.spark:ClearAllPoints(); bar.labelText:ClearAllPoints()
	bar.timeText:ClearAllPoints(); bar.fgTexture:ClearAllPoints(); bar.bgTexture:ClearAllPoints()
	bar.icon:SetSize(bg.iconSize, bg.iconSize)
	local w, h = bg.width, bg.height
	if config.iconOnly then -- icon only layouts
		bar.icon:SetPoint("TOPLEFT", bar.frame, "TOPLEFT", 0, 0)
		if (bg.barHeight > 0) and (bg.barWidth > 0) then
			local offset = (w - bg.barWidth) / 2 -- how far bars start from edge of frame
			if config.bars == "r2l" then 
				bar.fgTexture:SetPoint("TOPLEFT", bar.icon, "BOTTOMLEFT", bg.iconOffsetX + offset, -bg.iconOffsetY)
				bar.bgTexture:SetPoint("TOPRIGHT", bar.icon, "BOTTOMRIGHT", bg.iconOffsetX - offset, -bg.iconOffsetY)
			elseif config.bars == "l2r" then
				bar.fgTexture:SetPoint("TOPRIGHT", bar.icon, "BOTTOMRIGHT", bg.iconOffsetX - offset, -bg.iconOffsetY)
				bar.bgTexture:SetPoint("TOPLEFT", bar.icon, "BOTTOMLEFT", bg.iconOffsetX + offset, -bg.iconOffsetY)
			end
			bar.fgTexture:SetHeight(bg.barHeight); bar.bgTexture:SetHeight(bg.barHeight)
		end
		bar.timeText:SetPoint("TOP", bar.icon, "BOTTOM", bg.timeInset, bg.timeOffset)
		bar.timeText:SetPoint("LEFT", bar.icon, "LEFT", bg.timeInset, bg.timeOffset)
		bar.timeText:SetPoint("RIGHT", bar.icon, "RIGHT", bg.timeInset + 2, bg.timeOffset) -- pad right to center time text better
		if bg.timeAlign == "normal" then bar.timeText:SetJustifyH("CENTER") else bar.timeText:SetJustifyH(bg.timeAlign) end
		bar.timeText:SetJustifyV("MIDDLE")
		bar.labelText:SetPoint("LEFT", bar.icon, "LEFT", bg.labelInset, bg.labelOffset)
		bar.labelText:SetPoint("RIGHT", bar.icon, "RIGHT", bg.labelInset, bg.labelOffset)
		bar.labelText:SetJustifyH("CENTER"); bar.labelText:SetJustifyV(bg.labelAlign)
	else -- bar layouts
		local offsetLeft, offsetRight, fudge = 0, 0, 0
		if bg.timeAlign == "normal" then fudge = 4 end
		if bg.showIcon then
			if config.icon == "left" then
				bar.icon:SetPoint("LEFT", bar.frame, "LEFT", bg.iconOffsetX, bg.iconOffsetY)
				offsetLeft = bg.iconSize
			elseif config.icon == "right" then
				bar.icon:SetPoint("RIGHT", bar.frame, "RIGHT", bg.iconOffsetX, bg.iconOffsetY)
				offsetRight = bg.iconSize
			end
		end		
		if config.label == "right" then
			bar.timeText:SetPoint("LEFT", bar.frame, "LEFT", bg.timeInset + offsetLeft + fudge, bg.timeOffset)
			if bg.timeAlign == "normal" then bar.timeText:SetJustifyH("LEFT") else bar.timeText:SetJustifyH(bg.timeAlign) end
			bar.labelText:SetPoint("RIGHT", bar.frame, "RIGHT", -bg.labelInset - offsetRight - fudge, bg.labelOffset)
			if (bg.labelOffset == bg.timeOffset) and (bg.timeAlign == "normal") then
				bar.labelText:SetPoint("LEFT", bar.timeText, "RIGHT", 0, 0)
			else
				bar.labelText:SetPoint("LEFT", bar.frame, "LEFT", 0, bg.labelOffset)
				bar.timeText:SetPoint("RIGHT", bar.frame, "RIGHT", bg.timeInset - offsetRight, bg.timeOffset)
			end
			bar.labelText:SetJustifyH("RIGHT")
		elseif config.label == "left" then
			bar.timeText:SetPoint("RIGHT", bar.frame, "RIGHT", bg.timeInset - offsetRight - fudge, bg.timeOffset)
			if bg.timeAlign == "normal" then bar.timeText:SetJustifyH("RIGHT") else bar.timeText:SetJustifyH(bg.timeAlign) end
			bar.labelText:SetPoint("LEFT", bar.frame, "LEFT", bg.labelInset + offsetLeft + fudge, bg.labelOffset)
			if (bg.labelOffset == bg.timeOffset) and (bg.timeAlign == "normal") then
				bar.labelText:SetPoint("RIGHT", bar.timeText, "LEFT", 0, 0)
			else
				bar.labelText:SetPoint("RIGHT", bar.frame, "RIGHT", 0, bg.labelOffset)
				bar.timeText:SetPoint("LEFT", bar.frame, "LEFT", bg.timeInset + offsetLeft, bg.timeOffset)
			end
			bar.labelText:SetJustifyH("LEFT")
		end
		if config.bars == "r2l" then 
			bar.fgTexture:SetPoint("TOPLEFT", bar.frame, "TOPLEFT", offsetLeft, 0)
			bar.bgTexture:SetPoint("TOPRIGHT", bar.frame, "TOPRIGHT", -offsetRight, 0)
		elseif config.bars == "l2r" then
			bar.fgTexture:SetPoint("TOPRIGHT", bar.frame, "TOPRIGHT", -offsetRight, 0)
			bar.bgTexture:SetPoint("TOPLEFT", bar.frame, "TOPLEFT", offsetLeft, 0)
		end	
		bar.fgTexture:SetHeight(h); bar.bgTexture:SetHeight(h)
		bar.timeText:SetJustifyV("MIDDLE"); bar.labelText:SetJustifyV(bg.labelAlign)
	end

	if config.bars == "r2l" then 
		bar.spark:SetPoint("TOP", bar.fgTexture, "TOPRIGHT", 0, 4)
		bar.spark:SetPoint("BOTTOM", bar.fgTexture, "BOTTOMRIGHT", 0, -4)
	elseif config.bars == "l2r" then
		bar.spark:SetPoint("TOP", bar.fgTexture, "TOPLEFT", 0, 4)
		bar.spark:SetPoint("BOTTOM", bar.fgTexture, "BOTTOMLEFT", 0, -4)
	end
		
	bar.tooltipAnchor = bg.attributes.anchorTips
	bar.labelText:SetWordWrap(bg.labelWrap)
	bar.labelText:SetHeight(h + bg.spacingY) -- limit label height to frame's height plus vertical spacing
	bar.labelText:SetFontObject(ChatFontNormal); bar.timeText:SetFontObject(ChatFontNormal); bar.iconText:SetFontObject(ChatFontNormal)
	local t = nil; if bg.labelOutline then t = "OUTLINE" end; bar.labelText:SetFont(bg.labelFont, bg.labelFSize, t)
	t = nil; if bg.timeOutline then t = "OUTLINE" end; bar.timeText:SetFont(bg.timeFont, bg.timeFSize, t)
	t = nil; if bg.iconOutline then t = "OUTLINE" end; bar.iconText:SetFont(bg.iconFont, bg.iconFSize, t)
	t = bg.labelColor; bar.labelText:SetTextColor(t.r, t.g, t.b, bg.labelAlpha)
	t = bg.timeColor; bar.timeText:SetTextColor(t.r, t.g, t.b, bg.timeAlpha)
	t = bg.iconColor; bar.iconText:SetTextColor(t.r, t.g, t.b, bg.iconAlpha)

	if bg.showIcon then
		bar.cooldown:SetSize(bg.iconSize, bg.iconSize); bar.cooldown:SetPoint("TOPLEFT", bar.icon, "TOPLEFT")
		bar.iconText:SetPoint("CENTER", bar.icon, "CENTER", bg.iconInset, bg.iconOffset)
		bar.iconText:SetJustifyH("CENTER"); bar.iconText:SetJustifyV("MIDDLE")
		if LBF then -- if using ButtonFacade, set custom fields in the button data table and add it to the skinnning group
			bar.iconTexture:SetTexCoord(0, 1, 0, 1)
			bar.iconTexture:SetSize(bg.iconSize, bg.iconSize)
			bar.iconTexture:SetPoint("TOPLEFT", bar.icon, "TOPLEFT")
			LBF_Group:RemoveButton(bar.icon, true) -- needed so size changes work when icon is reused
			local bdata = bar.buttonData
			bdata.Icon = bar.iconTexture
			bdata.Normal = bar.icon:GetNormalTexture()
			bdata.Cooldown = bar.cooldown
			bdata.Border = bar.iconBorder
			LBF_Group:AddButton(bar.icon, bdata)
		else -- if not then use a default button arrangment
			local trim, crop, slice, dice = 0.05, 0.95, 0.9 * bg.iconSize, 0.05 * bg.iconSize
			bar.iconTexture:SetTexCoord(trim, crop, trim, crop)
			bar.iconTexture:SetSize(slice, slice)
			bar.iconTexture:SetPoint("TOPLEFT", bar.icon, "TOPLEFT", dice, -dice)
		end
	end	
	bar.frame:SetSize(w, h)
	bar.container:SetSize(w, h); bar.container:SetAllPoints()
	if config.iconOnly then
		bar.container:SetFrameLevel(bar.frame:GetFrameLevel() + 5)
		bar.icon:SetFrameLevel(bar.frame:GetFrameLevel() + 1)
		bar.iconTextFrame:SetFrameLevel(bar.container:GetFrameLevel() + 6)
	else
		bar.container:SetFrameLevel(bar.frame:GetFrameLevel() + 1)
		bar.icon:SetFrameLevel(bar.frame:GetFrameLevel() + 5)
		bar.iconTextFrame:SetFrameLevel(bar.cooldown:GetFrameLevel() + 5)
	end
end

-- Convert a time value into a compact text string
Nest_TimeFormatOptions = { { 1, 1, 1, 1, 1 }, { 1, 1, 1, 3, 5 }, { 1, 1, 1, 3, 4 },
						{ 2, 3, 1, 2, 3 }, { 2, 3, 1, 2, 2 }, { 2, 3, 1, 3, 4 }, { 2, 3, 1, 3, 5 },
						{ 2, 2, 2, 2, 3 }, { 2, 2, 2, 2, 2 }, { 2, 2, 2, 2, 4 }, { 2, 2, 2, 3, 4 }, { 2, 2, 2, 3, 5 },
						{ 2, 3, 2, 2, 3 }, { 2, 3, 2, 2, 2 }, { 2, 3, 2, 2, 4 }, { 2, 3, 2, 3, 4 }, { 2, 3, 2, 3, 5 },
						{ 2, 3, 3, 2, 3 }, { 2, 3, 3, 2, 2 }, { 2, 3, 3, 2, 4 }, { 2, 3, 3, 3, 4 }, { 2, 3, 3, 3, 5 },
						{ 3, 3, 3, 2, 3 }, { 3, 3, 3, 3, 5 } }
function Nest_FormatTime(t, timeFormat, timeSpaces, timeCase)
	if not timeFormat or (timeFormat > #Nest_TimeFormatOptions) then timeFormat = #Nest_TimeFormatOptions end -- last is most compact
	local opt = Nest_TimeFormatOptions[timeFormat]
	local func = opt.custom
	local h, m, mplus, s, ts, f
	if func then -- check for custom time formatting options
		f = func(t)
	else
		local o1, o2, o3, o4, o5 = opt[1], opt[2], opt[3], opt[4], opt[5]
		h = math.floor(t / 3600); m = math.floor((t - (h * 3600)) / 60); s = math.floor(t - (h * 3600) - (m * 60))
		mplus = math.floor((t - (h * 3600) + 59.99) / 60) -- provides compatibility with tooltips
		ts = math.floor(t * 10) / 10 -- truncated to a tenth second
		if t >= 3600 then
			if o1 == 1 then f = string.format("%.0f:%02.0f:%02.0f", h, m, s) elseif o1 == 2 then f = string.format("%.0fh %.0fm", h, m)
				else f = string.format("%.0fh", h) end
		elseif t >= 120 then
			if o2 == 1 then f = string.format("%.0f:%02.0f", m, s) elseif o2 == 2 then f = string.format("%.0fm %.0fs", m, s)
				else f = string.format("%.0fm", mplus) end
		elseif t >= 60 then
			if o3 == 1 then f = string.format("%.0f:%02.0f", m, s) elseif o3 == 2 then f = string.format("%.0fm %.0fs", m, s)
				else f = string.format("%.0fm", mplus) end
		elseif t >= 10 then
			if o4 == 1 then f = string.format(":%02.0f", s) elseif o4 == 2 then f = string.format("%.0fs", s)
				else f = string.format("%.0f", s) end
		else
			if o5 == 1 then f = string.format(":%02.0f", s) elseif o5 == 2 then f = string.format("%.1fs", ts)
				elseif o5 == 3 then f = string.format("%.0fs", s) elseif o5 == 4 then f = string.format("%.1f", ts)
				else f = string.format("%.0f", s) end
		end
	end
	if not timeSpaces then f = string.gsub(f, " ", "") end
	if timeCase then f = string.upper(f) end
	return f
end

-- Add a formatting function to the table of time format options.
function Nest_RegisterTimeFormat(func)
	local index = #Nest_TimeFormatOptions
	index = index + 1
	Nest_TimeFormatOptions[index] = { custom = func }
	return index
end

-- Update labels and colors plus for timer bars adjust bar length and formatted time text
-- This function is called on every update and the settings in it do not need to invoke other updates
local function Bar_UpdateSettings(bg, bar, config)
	local fill, sparky, offsetX = 1, false, 0 -- fill is fraction of the bar to display, default to full bar
	local timeText, bt, bl, bi, bf, bb, ba = "", bar.timeText, bar.labelText, bar.iconText, bar.fgTexture, bar.bgTexture, bar.icon.anim
	if bar.timeLeft and bar.duration and bar.maxTime and bar.offsetTime then -- only update if key parameters are set
		local remaining = bar.duration - (GetTime() - bar.startTime + bar.offsetTime) -- remaining time in seconds
		if remaining < 0 then remaining = 0 end -- make sure no rounding funnies
		if remaining > bar.duration then remaining = bar.duration end -- and no inaccurate durations!
		bar.timeLeft = remaining -- update saved value
		if remaining < bar.maxTime then fill = remaining / bar.maxTime end -- calculate fraction of time remaining
		timeText = Nest_FormatTime(remaining, bg.timeFormat, bg.timeSpaces, bg.timeCase) -- set timer text
	end
	if bg.showIcon then
		offsetX = bg.iconSize
		bar.icon:Show()
		local pulseStart, pulseEnd = (bg.attributes.pulseStart or bar.attributes.pulseStart), (bg.attributes.pulseEnd or bar.attributes.pulseEnd)
		if pulseStart and bar.timeLeft and ((bar.duration - bar.timeLeft) < 0.25) and not ba:IsPlaying() then ba:Play() end
		if pulseEnd and bar.timeLeft and (bar.timeLeft < 0.45) and not ba:IsPlaying() then ba:Play() end
		if LBF then -- icon border coloring
			local btex = bar.buttonData.Border
			if Raven.db.global.ButtonFacadeBorder and btex and btex.SetVertexColor then
				btex:SetVertexColor(bar.ibr, bar.ibg, bar.ibb, bar.iba)
				bar.iconBorder:Show()
			else
				bar.iconBorder:Hide()
			end
			local ntex = LBF:GetNormalTexture(bar.icon)
			if Raven.db.global.ButtonFacadeNormal and ntex and ntex.SetVertexColor then ntex:SetVertexColor(bar.ibr, bar.ibg, bar.ibb, bar.iba) end
		else
			local ntex = bar.icon:GetNormalTexture()
			if ntex then ntex:SetVertexColor(bar.ibr, bar.ibg, bar.ibb, bar.iba) end
			bar.iconBorder:Hide()
		end
	else
		bar.icon:Hide()
	end
	if bg.showIcon and bar.iconCount then bi:SetText(tostring(bar.iconCount)); bi:Show() else bi:Hide() end
	if bg.showIcon and bg.showCooldown and bar.timeLeft and (bar.timeLeft > 0) and not ba:IsPlaying() then bar.cooldown:Show() else bar.cooldown:Hide() end
	if bg.showTimeText then bt:SetText(timeText); bt:Show() else bt:Hide() end
	if bg.showLabelText and bar.label then bl:SetText(bar.label); bl:Show() else bl:Hide() end
	local w, h = bg.width - offsetX, bg.height; if config.iconOnly then w = bg.barWidth; h = bg.barHeight end
	if bg.showBar and (w > 0) and (h > 0) then -- non-zero dimensions to fix the zombie bar bug
		local ar, ag, ab = Nest_AdjustColor(bar.br, bar.bg, bar.bb, bg.bgSaturation or 0, bg.bgBrightness or 0)
		bb:SetVertexColor(ar, ag, ab, 1); bb:SetTexture(bg.bgTexture); bb:SetAlpha(bg.bgAlpha)
		bb:SetWidth(w); bb:SetTexCoord(0, 1, 0, 1); bb:Show()
		if (fill > 0) and (bg.fgNotTimer or bar.timeLeft) then
			if bg.showSpark and (fill < 1) then sparky = true end
			local ar, ag, ab = Nest_AdjustColor(bar.cr, bar.cg, bar.cb, bg.fgSaturation or 0, bg.fgBrightness or 0)
			bf:SetVertexColor(ar, ag, ab, 1); bf:SetTexture(bg.fgTexture); bf:SetAlpha(bg.fgAlpha)
			bf:SetWidth(w * fill); bf:SetTexCoord(0, 1, 0, 1); bf:Show()
		else bf:Hide() end
	else bf:Hide(); bb:Hide() end
	if sparky then bar.spark:Show() else bar.spark:Hide() end
	local alpha = bar.alpha or 1 -- adjust by bar alpha
	if bar.flash then alpha = Nest_FlashAlpha(alpha, 1) end -- adjust alpha if flashing
	bar.frame:SetAlpha(alpha) -- final alpha adjustment
	if bg.attributes.noMouse or (bg.attributes.iconMouse and not bg.showIcon) then -- non-interactive or "only icon" but icon disabled
		bar.icon:EnableMouse(false); bar.frame:EnableMouse(false); if callbacks.deactivate then callbacks.deactivate(bar.overlay) end
	elseif bg.attributes.iconMouse then -- only mouse is interactive
		bar.icon:EnableMouse(true); bar.frame:EnableMouse(false); if callbacks.activate then callbacks.activate(bar, bar.icon) end
	else -- entire bar is interactive
		bar.icon:EnableMouse(false); bar.frame:EnableMouse(true); if callbacks.activate then callbacks.activate(bar, bar.frame) end
	end
end

-- Update the lengths of timer bars, spark positions, and alphas of flashing bars
local function Bar_RefreshAnimations(bg, bar, config)
	local fill, sparky, offsetX = 1, false, 0
	if bar.timeLeft and bar.duration and bar.maxTime and bar.offsetTime then -- only update if key parameters are set
		local remaining = bar.duration - (GetTime() - bar.startTime + bar.offsetTime) -- remaining time in seconds
		if remaining < 0 then remaining = 0 end -- make sure no rounding funnies
		if remaining > bar.duration then remaining = bar.duration end -- and no inaccurate durations!
		bar.timeLeft = remaining -- update saved value
		if remaining < bar.maxTime then fill = remaining / bar.maxTime end -- calculate fraction of time remaining
	end
	if bg.showBar and bg.iconSize and (fill > 0) and (bg.fgNotTimer or bar.timeLeft) then
		if bg.showIcon then offsetX = bg.iconSize end
		local w, h = bg.width - offsetX, bg.height; if config.iconOnly then w = bg.barWidth; h = bg.barHeight end
		if (w > 0) and (h > 0) then
			if bg.showSpark and (fill < 1) then sparky = true end
			bar.fgTexture:SetWidth(w * fill); bar.fgTexture:Show()
		else bar.fgTexture:Hide() end
	end
	if sparky then bar.spark:Show() else bar.spark:Hide() end
	local alpha = bar.alpha or 1 -- adjust by bar alpha
	if bar.flash then alpha = Nest_FlashAlpha(alpha, 1) end -- adjust alpha if flashing
	bar.frame:SetAlpha(alpha) -- final alpha adjustment
end

-- Update bar order and calculate offsets within the bar stack plus overall width and height of the frame
local function BarGroup_SortBars(bg, config)
	for i = 1, bg.count do -- fill data into the sorting table
		local s = bg.sorter[i]
		local bar = bg.bars[s.name]
		if not bar.startTime or not bar.offsetTime then s.start = 0 else s.start = bar.startTime - bar.offsetTime end
		if not bar.timeLeft or not bar.duration then
			s.timeLeft = 100000; s.duration = 100000 -- really big number so sorts like infinite time
		else
			s.timeLeft = bar.timeLeft; s.duration = bar.duration
		end
	end
	table.sort(bg.sorter, bg.sortFunction)
	local wrap = 0 -- indicates default of not wrapping
	local dir = bg.reverse and 1 or -1 -- plus or minus depending on direction
	local x0, y0 = 0, 0 -- starting position, must be offset by dimensions of anchor if unlocked
	local dx, dy = dir * (bg.width + bg.spacingX), dir * (bg.height + bg.spacingY) -- offsets for each new bar
	local wx, wy = 0, 0 -- offsets from starting point when need to wrap
	local count, maxBars, cdir = bg.count, bg.maxBars, 0
	if not maxBars or (maxBars == 0) then maxBars = count end
	if count > maxBars then count = maxBars end
	local anchorPoint = "BOTTOMLEFT"
	if config.iconOnly then -- icons can go any direction from anchor
		if bg.wrap then
			wrap = bg.wrap
			if (wrap > 0) and (count > wrap) then count = wrap end
		end
		if config.orientation == "vertical" then
			wx = -dx; dx = 0
			if not bg.locked then y0 = dy end
			if not bg.reverse then anchorPoint = "TOPLEFT"; wx = -wx; cdir = -1 end
			if bg.snapCenter and bg.locked then y0 = y0 - (dy * (((count - dir) / 2) + cdir)) end
		else -- horizontal
			wy = dy; dy = 0
			if not bg.locked then x0 = dx end
			if bg.reverse then anchorPoint = "BOTTOMRIGHT"; wy = -wy; cdir = -1 end
			if bg.snapCenter and bg.locked then x0 = x0 - (dx * (((count + dir) / 2) + cdir)) end
		end
		if bg.wrapDirection then wx = -wx; wy = -wy end
	else -- bars just go up or down with anchor set to top or bottom
		dx = 0 -- only going straight up, at least until need to wrap
		if not bg.locked then y0 = y0 + dy end
		if not bg.reverse then anchorPoint = "TOPLEFT" end
	end
	for i = 1, bg.count do
		local bar = bg.bars[bg.sorter[i].name]
		bar.frame:ClearAllPoints()
		if i <= maxBars then
			local w, skip = i - 1, 0
			if wrap > 0 then skip = floor(w / wrap); w = w % wrap end
			local x1 = x0 + (dx * w) + (wx * skip)
			local y1 = y0 + (dy * w) + (wy * skip)	
			bar.frame:SetPoint(anchorPoint, bg.frame, anchorPoint, x1, y1)
			bar.frame:Show()
		else
			bar.frame:Hide()
		end
	end
	bg.lastX = x0 + (dx * count); bg.lastY = y0 + (dy * count); bg.anchorPoint = anchorPoint -- offset from frame to last bar
	bg.frame:SetSize(bg.width, bg.height)
	local scale = bg.frame:GetScale()
	if math.abs(bg.scale - scale) > 0.001 then -- only adjust scale if it has changed by a detectable amount
		if bg.relativeTo then -- if anchored to another bar group then just change the scale
			bg.frame:SetScale(bg.scale)
		else -- if not anchored make sure the position doesn't get changed
			scale = scale / bg.scale -- compute scaling factor
			x0 = bg.frame:GetLeft() * scale; y0 = bg.frame:GetBottom() * scale -- normalize by scale factor
			bg.frame:SetScale(bg.scale)
			bg.frame:ClearAllPoints(); bg.frame:SetPoint("BOTTOMLEFT", nil, "BOTTOMLEFT", x0, y0)
		end
	end
end

-- Update relative positions between bar groups, has to be called on every update for the lastbar feature to work right
local function UpdateRelativePositions()
	for _, bg in pairs(barGroups) do
		if bg.configuration and bg.relativeTo then
			local rbg = barGroups[bg.relativeTo]
			if rbg then
				if rbg.count == 0 then
					local i = 0
					while (rbg.count == 0) and rbg.relativeEmpty and rbg.relativeTo do
						rbg = barGroups[rbg.relativeTo]
						i = i + 1; if i > 20 then break end -- safety check, never loop more than 20 deep
					end
				end
				local align, offsetX, offsetY = "BOTTOMLEFT", 0, 0
				if (rbg.count > 0) or not bg.relativeEmpty then
					offsetX, offsetY = bg.relativeX / bg.scale, bg.relativeY / bg.scale
					if bg.relativeLastBar then -- alternative is to attach to the last bar rendered back in the other bar group
						offsetX = offsetX + rbg.lastX; offsetY = offsetY + rbg.lastY; align = rbg.anchorPoint
					end
				end
				bg.frame:ClearAllPoints(); bg.frame:SetPoint(align, rbg.frame, align, offsetX, offsetY)
			end
		end
	end
end

-- Check configuration and minimum values to determine frame width and height
local function SetBarGroupEffectiveDimensions(bg, config)
	local w, h, minimumWidth, minimumHeight = bg.barWidth or 10, bg.barHeight or 10, 5, 5
	if config.iconOnly then
		w = bg.iconSize; h = bg.iconSize -- icon configs start with icon size and add room for bar and time text, if they are displayed
		h = h + (bg.showBar and (bg.barHeight + math.max(0, bg.iconOffsetY)) or 0)
	else
		if bg.showIcon then w = w + bg.iconSize end -- bar config start with bar size and add room for icon if it is displayed
	end
	if h < minimumHeight then h = minimumHeight end -- enforce minimums for dimensions
	if w < minimumWidth then w = minimumWidth end
	bg.width = w; bg.height = h
	if not bg.scale then bg.scale = 1 end
end

-- Check if display dimensions have changed and update bar group locations
function Nest_CheckDisplayDimensions()
	local dw, dh = UIParent:GetWidth(), UIParent:GetHeight()
	if (displayWidth ~= dw) or (displayHeight ~= dh) then
		displayWidth = dw; displayHeight = dh
		for _, bg in pairs(barGroups) do
			if bg.configuration then -- make sure configuration is valid
				local p = bg.position
				Nest_SetAnchorPoint(bg, p.left, p.right, p.bottom, p.top, bg.scale, p.width, p.height) -- restore cached position
			end
		end
	end
end

-- Update routine does all the actual work of setting up and displaying bar groups.
function Nest_Update()
	for _, bg in pairs(barGroups) do
		if bg.configuration then -- make sure configuration is valid
			local config = Nest_SupportedConfigurations[bg.configuration]
			if not bg.moving then bg.frame:SetFrameStrata(bg.strata or "MEDIUM") end
			SetBarGroupEffectiveDimensions(bg, config) -- stored in bg.width and bg.height
			if update or bg.update then BarGroup_UpdateAnchor(bg, config) end
			for _, bar in pairs(bg.bars) do
				if update or bg.update or bar.update then -- see if any bar configurations need to be updated
					Bar_UpdateLayout(bg, bar, config) -- configure internal bar layout
				end
				Bar_UpdateSettings(bg, bar, config) -- update bar color, times, and texts plus activate buff buttons
				bar.update = false
			end
			BarGroup_SortBars(bg, config) -- update bar order and relative positions plus overall frame dimensions
			bg.update = false
		end
	end
	UpdateRelativePositions() -- has to be done every time to support relative positioning to last bar
	update = false
end

-- Just refresh timers and flashing bars without checking settings.
function Nest_Refresh()
	for _, bg in pairs(barGroups) do
		if bg.configuration then -- make sure configuration is valid
			local config = Nest_SupportedConfigurations[bg.configuration]
			SetBarGroupEffectiveDimensions(bg, config) -- stored in bg.width and bg.height
			for _, bar in pairs(bg.bars) do if not bar.update then Bar_RefreshAnimations(bg, bar, config) end end
		end
	end
end