-- Raven is an addon to monitor auras and cooldowns, providing timer bars, action bar highlights, and helpful notifications.

-- Profile.lua contains the default profile for initializing the player's selected profile.
-- It includes routines to process preset aura and cooldown info for all classes and races.
-- Profile settings are accessed by reference to MOD.db.global or MOD.db.profile.
-- It also maintains a persistent database that caches icons, colors and labels for auras and spells.

-- Exported functions for looking up aura and spell-related info:
-- Raven:SetIcon(name) save icon in cache
-- Raven:GetIcon(name) returns cached icon for spell with the specified name, nil if not found
-- Raven:SetColor(name) save color in cache
-- Raven:GetColor(name) returns cached color for spell with the specified name, nil if not found
-- Raven:SetLabel(name) save label in cache
-- Raven:GetLabel(name) returns cached label for spell with the specified name, nil if not found
-- Raven:GetHyperlink(name) return a hyperlink for the spell with the specified name, nil if not found
-- Raven:FormatTime(time, index, spaces, upperCase) returns time in seconds converted into a text string
-- Raven:RegisterTimeFormat(func) adds a custom time format and returns its assigned index

local MOD = Raven
local L = LibStub("AceLocale-3.0"):GetLocale("Raven")
MOD.BuffTable = {}
MOD.DebuffTable = {}
MOD.CooldownTable = {}
local LSPELL = MOD.LocalSpellNames

local dispelTypes = {} -- table of debuff types that the character can dispel
local cooldownSpells = {} -- table of spells that have cooldowns

-- Saved variables don't handle being set to nil properly so need to use alternate value to indicate an option has been turned off
local Off = 0 -- value used to designate an option is turned off
local function IsOff(value) return value == nil or value == Off end -- return true if option is turned off
local function IsOn(value) return value ~= nil and value ~= Off end -- return true if option is turned on

-- Convert color codes from hex number to array with r, g, b, a fields (alpha set to 1.0)
function Raven_HexColor(hex)
	local n = tonumber(hex, 16)
	local red = math.floor(n / (256 * 256))
	local green = math.floor(n / 256) % 256
	local blue = n % 256

	return { r = red/255, g = green/255, b = blue/255, a = 1.0 }
end

-- Return a copy of a color, if c is nil then return white
function Raven_CopyColor(c)
	if not c then return nil end
	return { r = c.r, g = c.g, b = c.b, a = c.a }
end

-- Copy a table, including its metatable
function Raven_CopyTable(object)
    local lookup_table = {}
    local function _copy(object)
        if type(object) ~= "table" then
            return object
        elseif lookup_table[object] then
            return lookup_table[object]
        end
        local new_table = {}
        lookup_table[object] = new_table
        for index, value in pairs(object) do
            new_table[_copy(index)] = _copy(value)
        end
        return setmetatable(new_table, getmetatable(object))
    end
    return _copy(object)
end

-- Global color palette containing the standard colors for this addon
MOD.ColorPalette = {
	Yellow1=Raven_HexColor("fce94f"), Yellow2=Raven_HexColor("edd400"), Yellow3=Raven_HexColor("c4a000"),
	Orange1=Raven_HexColor("fcaf3e"), Orange2=Raven_HexColor("f57900"), Orange3=Raven_HexColor("ce5c00"),
	Brown1=Raven_HexColor("e9b96e"), Brown2=Raven_HexColor("c17d11"), Brown3=Raven_HexColor("8f5902"),
	Green1=Raven_HexColor("8ae234"), Green2=Raven_HexColor("73d216"), Green3=Raven_HexColor("4e9a06"),
	Blue1=Raven_HexColor("729fcf"), Blue2=Raven_HexColor("3465a4"), Blue3=Raven_HexColor("204a87"),
	Purple1=Raven_HexColor("ad7fa8"), Purple2=Raven_HexColor("75507b"), Purple3=Raven_HexColor("5c3566"),
	Red1=Raven_HexColor("ef2929"), Red2=Raven_HexColor("cc0000"), Red3=Raven_HexColor("a40000"),
	Pink=Raven_HexColor("ff6eb4"), Cyan=Raven_HexColor("7adbf2"), Gray=Raven_HexColor("888a85"),
}

-- Remove unneeded variables from the profile before logout
local function OnProfileShutDown()
	for n, k in pairs(MOD.db.global.SpellIDs) do if k == 0 then MOD.db.global.SpellIDs[n] = nil end end
	MOD:FinalizeBars() -- strip out all default values to significantly reduce profile file size
	MOD:FinalizeConditions() -- strip out temporary values from conditions
	MOD:FinalizeSettings() -- strip default values from layouts
	MOD:FinalizeInCombatBar() -- save linked settings
end
	
-- Initialize profile used to customize the addon
function MOD:InitializeProfile()
	MOD:SetSpellDefaults()
	MOD:SetCooldownDefaults()
	MOD:SetConditionDefaults()
	MOD:SetHighlightDefaults()
	MOD:SetIconDefaults()
	MOD:SetSpellNameDefaults()
	MOD:SetDimensionDefaults(MOD.DefaultProfile.global.Defaults)
	MOD:SetFontTextureDefaults(MOD.DefaultProfile.global.Defaults)
	MOD:SetInCombatBarDefaults()
	
	-- Get profile from database, providing default profile for initialization
	MOD.db = LibStub("AceDB-3.0"):New("RavenDB", MOD.DefaultProfile)
	MOD.db.RegisterCallback(MOD, "OnDatabaseShutdown", OnProfileShutDown)
	
	MOD:CleanupProfile() -- Cleanup from previous versions, if necessary
	MOD:InitializeSettings() -- Initialize bar group settings with default values
end

-- Cleanup any old profile information, if any
function MOD:CleanupProfile()
end

-- Process an aura description and add info to tables for buffs, debuffs, cooldowns
-- p[1] = name, p[2] = isBuff (nil if not actually an aura), p[3] = color
-- isBuff=true for BUFFs, false for DEBUFFs, nil for other spells
-- color = array of r, g, b, a to be used for associated bars
-- cooldown = true for spells that have a cooldown
-- school = school of magic that the spell belongs to
-- lockout = true for spells that can be used to test if a school is locked out
local function ProcessSpellInfo(p, pet, class, race, special)
	local c = p[3]
	local name = nil
	if p.id then name = GetSpellInfo(p.id) elseif p.ids then name = GetSpellInfo(p.ids[1]) end -- get localized name
	if not name then name = p[1] end -- if no id specified then use the non-localized name
	MOD.DefaultProfile.global.SpellColors[name] = c -- sets default color in the shared color table
	if p.cooldown then -- add to the cooldown table
		local cd = { pet = pet, class = class, race = race, school = p.school, id = p.id, ids = p.ids }
		MOD.CooldownTable[name] = cd -- add to table, indexed by name
	end
	if not pet and ((class == MOD.myClass) or (race == MOD.myRace) or special) then
		if p.lockout and p.school then MOD.lockSpells[name] = { school = p.school, id = p.id } end
		if p.cooldown and (not cooldownSpells[name] or (class == MOD.myClass)) then -- prefer current class
			cooldownSpells[name] = { school = p.school, id = p.id, ids = p.ids, refer = p.refer, profession = p.profession }
		end
	end
	if p[2] ~= nil then -- add to the buff or debuff table, prefering own class if already present
		local aura = { pet = pet, class = class, race = race, school = p.school, id = p.id, ids = p.ids }
		local bt = p[2] and MOD.BuffTable or MOD.DebuffTable
		if not bt[name] or (class == MOD.myClass) then bt[name] = aura end -- prefer current class
	end	
end
	
-- Initialize spells for class auras and cooldowns, also scan other classes for group buffs and cooldowns
-- Buffs, debuffs and cooldowns are tracked in tables containing name, color, class, race
function MOD:SetSpellDefaults()
	for class, at in pairs(MOD.petSpells) do -- look at all pet spells
		for _, p in pairs(at) do ProcessSpellInfo(p, true, class, nil, false) end
	end	
	for class, at in pairs(MOD.classSpells) do -- look at all class spells
		for _, p in pairs(at) do ProcessSpellInfo(p, false, class, nil, false) end
	end	
	for race, at in pairs(MOD.racialSpells) do -- look at racial spells
		for _, p in pairs(at) do ProcessSpellInfo(p, false, nil, race, false) end
	end	
	for _, p in pairs(MOD.generalSpells) do ProcessSpellInfo(p, false, nil, nil, true) end -- look at general purpose spells
	
	MOD.classSpells = nil -- not used again after initialization so okay to delete
	MOD.petSpells = nil
	MOD.racialSpells = nil
	MOD.generalSpells = nil
	
	if MOD.myClass == "DEATHKNIGHT" then -- localize rune spell names
		local t = {}
		for k, p in pairs(MOD.runeSpells) do if p.id then name = GetSpellInfo(p.id); t[name] = p end end
		MOD.runeSpells = t
	end
end

-- Initialize cooldown info from spellbook, should be called whenever spell book changes
function MOD:SetCooldownDefaults()
	table.wipe(MOD.cdSpells)
	local k, numTabs, numSpells = 1, GetNumSpellTabs(), 0
	local name, icon
	for i = 1, numTabs do local _, _, _, n = GetSpellTabInfo(i); numSpells = numSpells + n end
	for spell, p in pairs(cooldownSpells) do
		if p.id and IsSpellKnown(p.refer or p.id) then -- only include known spells in the cooldown list
			name, _, icon = GetSpellInfo(p.id)
			MOD.cdSpells[k] = { name = name, id = p.id, school = p.school, icon = icon, link = GetSpellLink(p.id) }
			k = k + 1
		end
		if p.ids then
			for _, id in pairs(p.ids) do
				if IsSpellKnown(p.refer or id) then
					name, _, icon = GetSpellInfo(id)
					MOD.cdSpells[k] = { name = name, id = id, school = p.school, icon = icon, link = GetSpellLink(id) }
					k = k + 1
				end
			end
		end
	end
	
	if numSpells then
		for i = 1, numSpells do
			name, _, icon = GetSpellInfo(i, "spell")
			local cd = cooldownSpells[name]
			if cd and not cd.id and not cd.ids then
				MOD.cdSpells[k] = { name = name, index = i, school = cd.school, icon = icon, link = GetSpellLink(i, "spell") }
				k = k + 1
			end
			local ls = MOD.lockSpells[name]
			if ls then
				ls.index = i
				if ls.school == "Frost" then ls.label = L["Frost School"]; ls.text = L["Locked out of Frost school of magic."]
				elseif ls.school == "Fire" then ls.label = L["Fire School"]; ls.text = L["Locked out of Fire school of magic."]
				elseif ls.school == "Nature" then ls.label = L["Nature School"]; ls.text = L["Locked out of Nature school of magic."]
				elseif ls.school == "Shadow" then ls.label = L["Shadow School"]; ls.text = L["Locked out of Shadow school of magic."]
				elseif ls.school == "Arcane" then ls.label = L["Arcane School"]; ls.text = L["Locked out of Arcane school of magic."]
				elseif ls.school == "Holy" then ls.label = L["Holy School"]; ls.text = L["Locked out of Holy school of magic."]
				elseif ls.school == "Physical" then ls.label = L["Physical School"]; ls.text = L["Locked out of Physical school of magic."]
				end
			end
		end
	end
end

-- Reset a particular spell color to its default value
function MOD:ResetColorDefault(name)
	if name then
		local dct = MOD.DefaultProfile.global.SpellColors
		local sct = MOD.db.global.SpellColors
		local c = dct[name]
		if not c then
			sct[name] = nil -- if not default value then just clear the spell color
		else
			local t = sct[name]
			if t then
				t.r, t.g, t.b, t.a = c.r, c.g, c.b, c.a
			else
				sct[name] = Raven_CopyColor(c)
			end
		end	
	end
end

-- Reset all colors to default values
function MOD:ResetColorDefaults()
	local dct = MOD.DefaultProfile.global.SpellColors
	local sct = MOD.db.global.SpellColors
	for n, c in pairs(dct) do -- copy all original values from the default color table
		local t = sct[n]
		if t then
			t.r, t.g, t.b, t.a = c.r, c.g, c.b, c.a
		else
			sct[n] = Raven_CopyColor(c)
		end
	end 
	for n in pairs(sct) do if not dct[n] then sct[n] = nil end end -- remove any extras
end

-- Initialize cache of icons
local iconCache = {}
function MOD:SetIconDefaults()
	local tcount = GetNumSpellTabs()
	for i = 1, tcount do
		local tname, _, toffset, tspells = GetSpellTabInfo(i)	
		
		for k = toffset + 1, toffset + tspells do
			local name, icon
			name = GetSpellBookItemName(k, BOOKTYPE_SPELL)
			icon = GetSpellBookItemTexture(k, BOOKTYPE_SPELL)
			iconCache[name] = icon
		end
	end
	local _, _, iconGCD = GetSpellInfo(28730) -- cached for global cooldown (using same icon as Arcane Torrent)
	iconCache[L["GCD"]] = iconGCD
end

-- Initialize dimension defaults
function MOD:SetDimensionDefaults(p)
	p.barWidth = 150; p.barHeight = 15; p.iconSize = 15; p.scale = 1; p.spacingX = 0; p.spacingY = 0; p.iconOffsetX = 0; p.iconOffsetY = 0
	p.hideIcon = false; p.hideClock = false; p.hideBar = false; p.hideSpark = false
	p.hideLabel = false; p.hideCount = true; p.hideValue = false; p.showTooltips = true
	p.i_barWidth = 20; p.i_barHeight = 5; p.i_iconSize = 25; p.i_scale = 1; p.i_spacingX = 2; p.i_spacingY = 15; p.i_iconOffsetX = 0; p.i_iconOffsetY = 0
	p.i_hideIcon = false; p.i_hideClock = false; p.i_hideBar = true; p.i_hideSpark = false
	p.i_hideLabel = true; p.i_hideCount = true; p.i_hideValue = false; p.i_showTooltips = true
end

-- Copy dimensions, destination is always a bar group, check which configuration type and copy either bar or icon defaults
function MOD:CopyDimensions(s, d)
	local iconOnly = d.configuration and Nest_SupportedConfigurations[d.configuration].iconOnly or false
	if iconOnly then
		d.barWidth = s.i_barWidth; d.barHeight = s.i_barHeight; d.iconSize = s.i_iconSize; d.scale = s.i_scale				
		d.spacingX = s.i_spacingX; d.spacingY = s.i_spacingY; d.iconOffsetX = s.i_iconOffsetX; d.iconOffsetY = s.i_iconOffsetY
		d.hideIcon = s.i_hideIcon; d.hideClock = s.i_hideClock; d.hideBar = s.i_hideBar; d.hideSpark = s.i_hideSpark
		d.hideLabel = s.i_hideLabel; d.hideCount = s.i_hideCount; d.hideValue = s.i_hideValue; d.showTooltips = s.i_showTooltips
	else
		d.barWidth = s.barWidth; d.barHeight = s.barHeight; d.iconSize = s.iconSize; d.scale = s.scale				
		d.spacingX = s.spacingX; d.spacingY = s.spacingY; d.iconOffsetX = s.iconOffsetX; d.iconOffsetY = s.iconOffsetY
		d.hideIcon = s.hideIcon; d.hideClock = s.hideClock; d.hideBar = s.hideBar; d.hideSpark = s.hideSpark
		d.hideLabel = s.hideLabel; d.hideCount = s.hideCount; d.hideValue = s.hideValue; d.showTooltips = s.showTooltips
	end
end

-- Initialize default fonts and textures
function MOD:SetFontTextureDefaults(p)
	p.labelFont = "Arial Narrow"; p.labelFSize = 10; p.labelAlpha = 1; p.labelColor = { r = 1, g = 1, b = 1, a = 1 }; p.labelOutline = false
	p.timeFont = "Arial Narrow"; p.timeFSize = 10; p.timeAlpha = 1; p.timeColor = { r = 1, g = 1, b = 1, a = 1 }; p.timeOutline = false
	p.iconFont = "Arial Narrow"; p.iconFSize = 10; p.iconAlpha = 1; p.iconColor = { r = 1, g = 1, b = 1, a = 1 }; p.iconOutline = true
	p.texture = "Blizzard"; p.bgtexture = "Blizzard"; p.alpha = 1; p.combatAlpha = 1; p.fgAlpha = 1; p.bgAlpha = 0.65
	p.fgSaturation = 0; p.fgBrightness = 0; p.bgSaturation = 0; p.bgBrightness = 0; p.borderSaturation = 0; p.borderBrightness = 0
end

-- Copy fonts and textures between tables
function MOD:CopyFontsAndTextures(s, d)
	if s and d and (s ~= d) then
		d.labelFont = s.labelFont; d.labelFSize = s.labelFSize; d.labelAlpha = s.labelAlpha; d.labelColor = Raven_CopyColor(s.labelColor); d.labelOutline = s.labelOutline
		d.timeFont = s.timeFont; d.timeFSize = s.timeFSize; d.timeAlpha = s.timeAlpha; d.timeColor = Raven_CopyColor(s.timeColor); d.timeOutline = s.timeOutline
		d.iconFont = s.iconFont; d.iconFSize = s.iconFSize; d.iconAlpha = s.iconAlpha; d.iconColor = Raven_CopyColor(s.iconColor); d.iconOutline = s.iconOutline
		d.texture = s.texture; d.bgtexture = s.bgtexture; d.alpha = s.alpha; d.combatAlpha = s.combatAlpha; d.fgAlpha = s.fgAlpha; d.bgAlpha = s.bgAlpha
		d.fgSaturation = s.fgSaturation; d.fgBrightness = s.fgBrightness; 
		d.borderSaturation = s.borderSaturation; d.borderBrightness = s.borderBrightness
	end
end

-- Copy standard colors between tables
function MOD:CopyStandardColors(s, d)
	if s and d and (s ~= d) then
		d.buffColor = Raven_CopyColor(s.buffColor); d.debuffColor = Raven_CopyColor(s.debuffColor)
		d.cooldownColor = Raven_CopyColor(s.cooldownColor); d.notificationColor = Raven_CopyColor(s.notificationColor)
		d.poisonColor = Raven_CopyColor(s.poisonColor); d.curseColor = Raven_CopyColor(s.curseColor)
		d.magicColor = Raven_CopyColor(s.magicColor); d.diseaseColor = Raven_CopyColor(s.diseaseColor)
	end
end

-- Find and cache spell ids (this should be used rarely, primarily when entering spell names manually
local function GetSpellID(name)
	local id = nil
	id = MOD.db.global.SpellIDs[name]
	if id == 0 then return nil end -- only scan invalid ones once in a session
	if id and (name ~= GetSpellInfo(id)) then id = nil end -- verify it is still valid

	if not id then
		id = 0
		while id < 120000 do
			id = id + 1
			local n = GetSpellInfo(id)
			if n == name then
				MOD.db.global.SpellIDs[n] = id
				return id
			end
		end
		MOD.db.global.SpellIDs[name] = 0 -- this marks an invalid spell name
		id = nil
	end
	return id
end

-- Add a texture to the icons cache
function MOD:SetIcon(name, texture)
	if name and texture then iconCache[name] = texture end
end

-- Get a texture from the icons cache, if not there try to get by spell name and cache if found.
-- If not found then look up spell identifier and use it to locate a texture.
function MOD:GetIcon(name)
	if not name or (name == "none") or (name == "") then return nil end
	local tex = iconCache[name] -- this is initialized from the player's spell book
	if not tex then
		local n
		n, _, tex = GetSpellInfo(name) -- first try to find it based on the name
		if n and tex then
			iconCache[name] = tex -- only cache textures found by looking up the name
		else
			local id = GetSpellID(name)
			if id then _, _, tex = GetSpellInfo(id) end -- then try based on id
		end
	end
	return tex
end

-- Get a hyperlink for a spell, looking up as a spell identifier if name is unknown
function MOD:GetHyperlink(name)
	if not name or (name == "") then return nil end
	local link = GetSpellLink(name) -- first try to find it based on the name
	if not link then
		local id = GetSpellID(name)
		if id then link = GetSpellLink(id) end -- then try based on id
	end
	return link
end

-- Add a color to the cache, update values in case they have changed
function MOD:SetColor(name, c)
	if name and c then
		local t = MOD.db.global.SpellColors[name]
		if t then
			t.r, t.g, t.b, t.a = c.r, c.g, c.b, c.a
		else
			MOD.db.global.SpellColors[name] = Raven_CopyColor(c)
		end
	end
end

-- Get a color from the cache of given name, but if not in cache then return nil
function MOD:GetColor(name)
	local c = MOD.db.global.SpellColors[name]
	return c
end

-- Add a label to the cache but only if different from name
function MOD:SetLabel(name, label)
	if name and label then
		if name == label then MOD.db.global.Labels[name] = nil else MOD.db.global.Labels[name] = label end
	end
end

-- Get a label from the cache, but if not in the cache then return the name
function MOD:GetLabel(name)
	local label = MOD.db.global.Labels[name]
	if not label then label = name end
	return label
end

-- Get localized names for all spells used internally or in built-in conditions
function MOD:SetSpellNameDefaults()
	LSPELL["Freezing Trap"] = GetSpellInfo(1499)
	LSPELL["Ice Trap"] = GetSpellInfo(13809)
	LSPELL["Immolation Trap"] = GetSpellInfo(13795)
	LSPELL["Explosive Trap"] = GetSpellInfo(13813)
	LSPELL["Black Arrow"] = GetSpellInfo(3674)
	LSPELL["Frost Shock"] = GetSpellInfo(8056)
	LSPELL["Flame Shock"] = GetSpellInfo(8050)
	LSPELL["Earth Shock"] = GetSpellInfo(8042)
	LSPELL["Flight Form"] = GetSpellInfo(33943)
	LSPELL["Swift Flight Form"] = GetSpellInfo(40120)
	LSPELL["Earthliving Weapon"] = GetSpellInfo(51730)
	LSPELL["Flametongue Weapon"] = GetSpellInfo(8024)
	LSPELL["Frostbrand Weapon"] = GetSpellInfo(8033)
	LSPELL["Rockbiter Weapon"] = GetSpellInfo(8017)
	LSPELL["Windfury Weapon"] = GetSpellInfo(8232)
	LSPELL["Eclipse (Solar)"] = GetSpellInfo(48517)
	LSPELL["Eclipse (Lunar)"] = GetSpellInfo(48518)
	LSPELL["Crusader Strike"] = GetSpellInfo(35395)
	LSPELL["Hammer of the Righteous"] = GetSpellInfo(53595)
end

-- Check if a spell id is known and usable by the player
function RavenCheckSpellKnown(spellID)
	local name = GetSpellInfo(spellID)
	return IsUsableSpell(name)
end

-- Initialize the dispel table which lists what types of debuffs the player can dispel
-- This needs to be updated when the player changes talent specs or learns new spells
function MOD:SetDispelDefaults()
	dispelTypes.Poison = false; dispelTypes.Curse = false; dispelTypes.Magic = false; dispelTypes.Disease = false
	if MOD.myClass == "DRUID" then
		if RavenCheckSpellKnown(2782) then -- Remove Corruption
			dispelTypes.Poison = true; dispelTypes.Curse = true
			if RavenCheckTalent(88423) then dispelTypes.Magic = true end -- Nature's Cure
		end
	elseif MOD.myClass == "PRIEST" then
		if RavenCheckSpellKnown(527) then dispelTypes.Magic = true end -- Dispel Magic
		if RavenCheckSpellKnown(528) then -- Cure Disease
			dispelTypes.Disease = true
			if RavenCheckTalent(64129) then dispelTypes.Poison = "player" end -- Body and Soul is self only
		end
	elseif MOD.myClass == "MAGE" then
		if RavenCheckSpellKnown(475) then dispelTypes.Curse = true end -- Remove Curse
	elseif MOD.myClass == "PALADIN" then
		if RavenCheckSpellKnown(4987) then -- Cleanse
			dispelTypes.Poison = true; dispelTypes.Disease = true
			if RavenCheckTalent(53551) then dispelTypes.Magic = true end -- Sacred Cleansing
		end
	elseif MOD.myClass == "SHAMAN" then
		if RavenCheckSpellKnown(51886) then -- Cleanse Spirit
			dispelTypes.Curse = true
			if RavenCheckTalent(77130) then dispelTypes.Magic = true end -- Improved Cleanse Spirit
		end
	end
end

-- Return true if the player can dispel the type of debuff on the unit
function MOD:IsDebuffDispellable(n, unit, debuffType)
	if not debuffType then return false end
	local t = dispelTypes[debuffType]
	if not t then return false end
	if (t == "player") and (unit ~= "player") then return false end -- special case for self-only dispels
	if unit == "player" then return true end -- always can dispel debuffs on self
	if UnitIsFriend("player", unit) ~= nil then return true end -- only can dispel on friendly units
	return false
end

-- Format a time value in seconds, return converted string or nil if invalid index
function MOD:FormatTime(t, index, spaces, upperCase)
	if (index > 0) and (index <= #Nest_TimeFormatOptions) then
		return Nest_FormatTime(t, index, spaces, upperCase)
	end
	return nil
end

-- Register a new time format option and return its assigned index
function MOD:RegisterTimeFormat(func) return Nest_RegisterTimeFormat(func) end

-- Default profile description used to initialize the SavedVariables persistent database
MOD.DefaultProfile = {
	global = {
		Labels = {},					-- cache of labels for actions and spells
		SpellColors = {},				-- cache of colors for actions and spells
		SpellIDs = {},					-- cache of spell ids that had to be looked up
		Settings = {},					-- settings table indexed by bar group names
		Defaults = {},					-- default settings for bar group layout, fonts and textures
		FilterBuff = {},				-- shared bar group filter for buffs
		FilterDebuff = {},				-- shared bar group filter for debuffs
		FilterCooldown = {},			-- shared bar group filter for cooldowns
		SharedConditions = {},			-- shared condition settings
		BuffDurations = {},				-- cache of buff durations used for weapon buffs
		DefaultBuffColor = Raven_HexColor("8ae234"), -- Green1
		DefaultDebuffColor = Raven_HexColor("fcaf3e"), -- Orange1
		DefaultCooldownColor = Raven_HexColor("fce94f"), -- Yellow1
		DefaultNotificationColor = Raven_HexColor("729fcf"), -- Blue1
		DefaultPoisonColor = Raven_CopyColor(DebuffTypeColor["Poison"]),
		DefaultCurseColor = Raven_CopyColor(DebuffTypeColor["Curse"]),
		DefaultMagicColor = Raven_CopyColor(DebuffTypeColor["Magic"]),
		DefaultDiseaseColor = Raven_CopyColor(DebuffTypeColor["Disease"]),
		HighlightsEnabled = true,		-- enable highlight support
		PlayerBuffHighlights = true, 	-- enable highlights for player buffs
		PlayerDebuffHighlights = true,	-- enable highlights for player debuffs
		TargetBuffHighlights = true,	-- enable highlights for target buffs
		TargetDebuffHighlights = true,	-- enable highlights for target debuffs
		FocusBuffHighlights = true,		-- enable highlights for target buffs
		FocusDebuffHighlights = true,	-- enable highlights for target debuffs
		PlayerBuffColor = Raven_HexColor("8ae234"), -- Green1
		PlayerDebuffColor = Raven_HexColor("fcaf3e"), -- Orange1
		TargetBuffColor = Raven_HexColor("3465a4"), -- Blue2
		TargetDebuffColor = Raven_HexColor("cc0000"), -- Red2
		FocusBuffColor = Raven_HexColor("ad7fa8"), -- Purple1
		FocusDebuffColor = Raven_HexColor("fce94f"), -- Yellow1
		FlashExpiring = true,			-- enable flashing of expiring buffs and debuffs
		FlashTime = 5,					-- how many seconds to flash before expiration
		CooldownText = true,			-- enable cooldown text overlays
		CooldownFont = "Arial Narrow",	-- default font for cooldown text
		CooldownFsize = 10,				-- default font size for cooldown text
		CooldownTimeFormat = 24,		-- default time format for cooldown text
		CooldownTimeSpaces = false,		-- default is no spaces in cooldown text
		CooldownTimeCase = false,		-- default is lowercase for cooldown text
		ButtonFacade = false,			-- enable highlights support for ButtonFacade if it is loaded
		ButtonFacadeNormal = true,		-- enable color of normal texture in ButtonFacade
		ButtonFacadeBorder = true,		-- enable color of border texture in ButtonFacade
		Minimap = { hide = false, minimapPos = 180, radius = 80, }, -- saved DBIcon minimap settings
		InCombatBar = {},				-- shared settings for the in-combat bar
	},
	profile = {
		enabled = true,					-- enable Raven
		hideBlizz = true,				-- enable hiding the Blizzard buff and temp enchant frames
		BarGroups = {},					-- bar group options to be filled in and saved between sessions
		Conditions = {}, 				-- conditions for the player's class
		ButtonFacadeSkin = {},			-- skin settings from ButtonFacade
		InCombatBar = {},				-- settings for the in-combat bar used to cancel buffs in combat
		InCombatBuffs = {},				-- list of buffs that can be cancelled in-combat
	},
}
