-- Raven is an addon to monitor auras and cooldowns, providing timer bars, action bar highlights, and helpful notifications.
-- Raven_Options is a load-on-demand module that only gets loaded and initialized when the user requests the GUI

-- Options.lua contains the tables used by the options panel as well as all the supporting functions for displaying
-- current settings, bar groups, conditions, etc.

local MOD = Raven
local acereg = LibStub("AceConfigRegistry-3.0")
local acedia = LibStub("AceConfigDialog-3.0")
local LBF = LibStub("LibButtonFacade", true)
local L = LibStub("AceLocale-3.0"):GetLocale("Raven")
-- local L = LibStub("AceLocale-3.0"):GetLocale("Raven_Options")
local initialized = false -- set when options are first accessed

local standardPlayerBuffs = false -- standard bar groups for "getting started"
local standardPlayerDebuffs = false
local standardCooldowns = false
local standardTarget = false
local standardFocus = false
local standardRunes = false
local standardNotifications = false
local standardShortBuffs = false
local standardLongBuffs = false

local enterNewBarGroup = false -- trigger showing bar group name entry
local toggleNewBarGroup = false -- used for automatic cancel
local typeNewBarGroup = false -- if true then auto-added bars, otherwise manually added
local enterNewBar = false -- used to trigger "enter bar mode"
local newbar = nil -- new bar template for editing
local oldbar = nil -- cache previous new bar settings
local configlist = nil -- cache pointer to generated configuration list

local enterNewCondition= false -- trigger showing notification name entry
local toggleNewCondition = false -- used for automatic cancel
local selectedDependency = nil -- used for selecting dependency condition
local selectedCondition = nil -- used for finding out when switching conditions
local tableSelection = {} -- stores current selection for a profile table
local selectedSpellName = nil -- stores current selection for a spell name in defaults panel
local selectedBuffName = nil -- stores current selection for a buff name in defaults panel
local weaponBuffs = { [L["Mainhand Weapon"]] = true, [L["Offhand Weapon"]] = true, [L["Ranged Weapon"]] = true }
local anchorPoints = { BOTTOMLEFT = "BOTTOMLEFT", CURSOR = "CURSOR", DEFAULT = "DEFAULT", LEFT = "LEFT", RIGHT = "RIGHT", TOPLEFT = "TOPLEFT", TOPRIGHT = "TOPRIGHT" }
local stratas = { BACKGROUND = "BACKGROUND", LOW = "LOW", MEDIUM = "MEDIUM", HIGH = "HIGH" }

-- Saved variables don't handle being set to nil properly so need to use alternate value to indicate an option has been turned off
local Off = 0 -- value used to designate an option is turned off
local function IsOff(value) return value == nil or value == Off end -- return true if option is turned off
local function IsOn(value) return value ~= nil and value ~= Off end -- return true if option is turned on

local function InMode(t) -- return true if locked into entry mode of specified type
	if t == "Bar" then return enterNewBar end
	if t == "BG" then return enterNewBar or enterNewBarGroup end
	if t == "Not" then return enterNewCondition end
	return enterNewBar or enterNewBarGroup or enterNewCondition
end 

-- Check if spell name is valid and display a warning if it is not, return validation result
local function ValidateSpellName(name)
	if not name or (name == "") then return false end
	if not MOD:GetIcon(name) then
		print(L["Not valid string"](name))
		return false
	end
	return true
end

-- Update the addon when the profile changes
local function OnProfileChanged()
	tableSelection = {} -- necessary since this is not cached in the profile itself
	MOD:InitializeBars() -- required because of the linkage between the profile and LibBars
end

-- Initialize options for the configuration GUI
-- Register the options table and link to the Blizz addons interface
local function InitializeOptions()
	initialized = true -- only do this once	
	local options = MOD.OptionsTable	
	options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(MOD.db) -- fill in the profile section
	options.args.profile.disabled = function(info) return InMode() end,
	
	acereg:RegisterOptionsTable("Raven", options)
	acereg:RegisterOptionsTable("Raven: "..options.args.FrontPage.name, options.args.FrontPage)
	acereg:RegisterOptionsTable("Raven: "..options.args.BarGroups.name, options.args.BarGroups)
	acereg:RegisterOptionsTable("Raven: "..options.args.Conditions.name, options.args.Conditions)
	acereg:RegisterOptionsTable("Raven: "..options.args.Highlights.name, options.args.Highlights)
	acereg:RegisterOptionsTable("Raven: "..options.args.profile.name, options.args.profile)
	acereg:RegisterOptionsTable("Raven Options", MOD.TopOptionsTable)
	acedia:AddToBlizOptions("Raven Options", "Raven")

	local w, h = 860, 680 -- somewhat arbitrary numbers that seem to work for the configuration dialog layout
	acedia:SetDefaultSize("Raven", w, h)
	
	MOD.db.RegisterCallback(MOD, "OnProfileChanged", OnProfileChanged)
	MOD.db.RegisterCallback(MOD, "OnProfileCopied", OnProfileChanged)
	MOD.db.RegisterCallback(MOD, "OnProfileReset", OnProfileChanged)
end

-- Update options in case anything changes
function MOD:UpdateOptions()
	if initialized and acedia.OpenFrames["Raven"] then
		acereg:NotifyChange("Raven")
	end
end

-- Toggle display of the options panel
function MOD:ToggleOptions()
	if not initialized then InitializeOptions() end
	if acedia.OpenFrames["Raven"] then
		acedia:Close("Raven")
	else
		acedia:Open("Raven")
	end
	collectgarbage("collect")
end

-- Generic function to build and return a sorted list of names stored in a profile table
-- Assumes table entries are arrays with the names indexed by "name"
local function GetNameList(id, t)
	local i, list = 0, {} -- build the name list from the profile table
	
	if IsOff(t) then
		tableSelection[id] = nil -- no selection for empty list
		return list -- empty list if t is not defined
	end

	local selection = tableSelection[id] -- make sure still a valid selection
	local found = false
	for _, n in pairs(t) do
		if IsOn(n) and n.name then
			i = i + 1
			list[i] = n.name
			if n.name == selection then found = true end
		end
	end
	table.sort(list)
	
	if not found and (i > 0) then -- check that current selection, if any, was found in the table
		tableSelection[id] = list[1] -- default is first entry in the list
	end
	return list
end

-- Generic function to add a new entry to a profile table, assumes "name" is field in the entry
local function AddNameEntry(id, t, newEntry)
	t[newEntry.name] = newEntry -- add the new entry indexed by its name to facilitate reverse lookup
	GetNameList(id, t) -- initializes the cached settings, if necessary
	tableSelection[id] = newEntry.name
end

-- Generic function to delete a name entry from a profile table
local function DeleteNameEntry(id, t)
	-- Get the cached list of names and the selected name
	local list = GetNameList(id, t)
	local selection = tableSelection[id]
	if not selection then return end -- empty table
	
	-- Look in the profile table to find the matching entry
	for k, n in pairs(t) do
		if IsOn(n) and (n.name == selection) then
			t[k] = Off -- remove the table entry by setting it to Off so that it will not come back when reload profile
			tableSelection[id] = nil -- will be set on next call to GetNameList, if table not empty
			return
		end
	end
end

-- Generic function to return index into the sorted name list of current selection
local function GetNameSelection(id, t)
	-- Get the cached list of names and the selected name
	local list = GetNameList(id, t)
	local selection = tableSelection[id]
	if not selection then return nil end -- empty table
	
	-- Look in the cached list to find the index of the selection
	for pos, n in pairs(list) do
		if n == selection then return pos end
	end
	return nil -- should never get here, just in case return nil
end

-- Generic function to set the selection in a name list
local function SetNameSelection(id, t, value)
	local list = GetNameList(id, t)
	if value > 0 and value <= table.maxn(list) then
		tableSelection[id] = list[value]
	else
		tableSelection[id] = nil
	end
end

-- Generic function to return table position (nil if empty table) and pointer to the currently selection in the profile table
local function GetNameEntry(id, t)
	-- Get the cached list of names and the selected name
	local list = GetNameList(id, t)
	local selection = tableSelection[id]
	if not selection then return nil end -- empty table
	
	-- Look in the profile table to find the matching entry
	for _, n in pairs(t) do
		if IsOn(n) then
			if n.name == selection then return n end
		end
	end
	return nil -- should never get here, just in case return nil
end

-- Functions for accessing bar group lists
local function GetBarGroupList() return GetNameList("BarGroups", MOD.db.profile.BarGroups) end
local function GetBarGroupEntry() return GetNameEntry("BarGroups", MOD.db.profile.BarGroups) end
local function GetSelectedBarGroup() return GetNameSelection("BarGroups", MOD.db.profile.BarGroups) end
local function SetSelectedBarGroup(value) SetNameSelection("BarGroups", MOD.db.profile.BarGroups, value) end
local function NoBarGroup() return enterNewBarGroup or (GetSelectedBarGroup() == nil) end

-- Create new bar group in either bar or icon configuration style
local function CreateBarGroup(name, auto, link, style, offsetX, offsetY)
	local bp = {}
	for n, k in pairs(MOD.db.global.Defaults) do bp[n] = k end -- current default settings for layouts fonts textures
	for n, k in pairs(MOD.BarGroupTemplate) do bp[n] = k end -- add settings in the template
	bp.name = name
	bp.configuration = style and 1 or (Nest_MaxBarConfiguration + 1)
	bp.growDirection = style
	bp.auto = auto
	bp.linkSettings = link
	bp.locked = false
	bp.bars = {}

	AddNameEntry("BarGroups", MOD.db.profile.BarGroups, bp)
	MOD:InitializeBarGroup(bp, offsetX, offsetY)
	MOD:UpdateAllBarGroups()
	return bp
end

local function DeleteBarGroup()
	local bg = GetBarGroupEntry()
	local oldName = bg.name
	MOD:ReleaseBarGroup(bg)
	DeleteNameEntry("BarGroups", MOD.db.profile.BarGroups)
	for _, bp in pairs(MOD.db.profile.BarGroups) do if IsOn(bp) and (bp.anchor == oldName) then bp.anchor = false end end
	MOD:UpdateAllBarGroups()
end

-- Rename involves deleting current bar group and then adding it back renamed
local function RenameBarGroup(newName)
	local bg = GetBarGroupEntry()
	local oldName = bg.name
	MOD:ReleaseBarGroup(bg)
	DeleteNameEntry("BarGroups", MOD.db.profile.BarGroups)
	bg.name = newName
	AddNameEntry("BarGroups", MOD.db.profile.BarGroups, bg)
	MOD:InitializeBarGroup(bg, 0, 0)
	for _, bp in pairs(MOD.db.profile.BarGroups) do if IsOn(bp) and (bp.anchor == oldName) then bp.anchor = newName end end
	MOD:UpdateAllBarGroups()
end

-- Check if any bar groups currently exist...
local function CheckBarGroupsExist()
	local bgs = MOD.db.profile.BarGroups
	for n, bg in pairs(bgs) do
		if IsOn(bg) then
			if standardPlayerBuffs and (n == L["Buffs"]) then return true end
			if standardShortBuffs and (n == L["Short Buffs"]) then return true end
			if standardLongBuffs and (n == L["Long Buffs"]) then return true end
			if standardPlayerDebuffs and (n == L["Debuffs"]) then return true end
			if standardCooldowns and (n == L["Cooldowns"]) then return true end
			if standardTarget and (n == L["Target"]) then return true end
			if standardFocus and (n == L["Focus"]) then return true end
			if standardRunes and (n == L["Runes"]) then return true end
			if standardNotifications and (n == L["Notifications"]) then return true end
		end
	end
	return false
end

-- Return the bar group specific suffix for the description
local function BarGroupString(name)
	local suffix = ""
	local bg = MOD.db.profile.BarGroups[name]
	if IsOn(bg) then
		local f, style
		if bg.configuration and (bg.configuration > Nest_MaxBarConfiguration) then style = "Icons" else style = "Bars" end
		if bg.linkSettings then f = "\n|cFF7adbf2Exists (%s), Linked|r" else f = "\n|cFF7adbf2Exists (%s)|r" end
		suffix = string.format(f, style)
	end
	return suffix
end

-- Set link settings for existing and selected standard bar groups
local function LinkStandardBarGroups()
	local bgs = MOD.db.profile.BarGroups
	for n, bg in pairs(bgs) do
		if IsOn(bg) then
			if (standardPlayerBuffs and (n == L["Buffs"])) or
			(standardShortBuffs and (n == L["Short Buffs"])) or (standardLongBuffs and (n == L["Long Buffs"])) or
			(standardPlayerDebuffs and (n == L["Debuffs"])) or (standardCooldowns and (n == L["Cooldowns"])) or
			(standardTarget and (n == L["Target"])) or (standardFocus and (n == L["Focus"])) or
			(standardRunes and (n == L["Runes"])) or (standardNotifications and (n == L["Notifications"])) then
				bg.linkSettings = not bg.linkSettings
				MOD:InitializeBarGroupSettings(bg)
			end
		end
	end
	MOD:UpdateAllBarGroups()
end

-- Delete existing and selected standard bar groups
local function DeleteStandardBarGroups()
	local bgs = MOD.db.profile.BarGroups
	for n, bg in pairs(bgs) do
		if IsOn(bg) then
			if (standardPlayerBuffs and (n == L["Buffs"])) or
			(standardShortBuffs and (n == L["Short Buffs"])) or (standardLongBuffs and (n == L["Long Buffs"])) or
			(standardPlayerDebuffs and (n == L["Debuffs"])) or (standardCooldowns and (n == L["Cooldowns"])) or
			(standardTarget and (n == L["Target"])) or (standardFocus and (n == L["Focus"])) or
			(standardRunes and (n == L["Runes"])) or (standardNotifications and (n == L["Notifications"])) then
				MOD:ReleaseBarGroup(bg)
				MOD.db.profile.BarGroups[n] = Off
				for _, bp in pairs(MOD.db.profile.BarGroups) do if IsOn(bp) and (bp.anchor == n) then bp.anchor = false end end
			end
		end
	end
	MOD:UpdateAllBarGroups()
end

-- Check if any standard bar groups are selected
local function AnySelectedStandardBarGroups(exists)
	if not exists then return standardPlayerBuffs or standardShortBuffs or standardLongBuffs or standardPlayerDebuffs or standardCooldowns or
		standardTarget or standardFocus or standardRunes or standardNotifications end
	return CheckBarGroupsExist()
end

-- Reset all standard bar group selections
local function ResetSelectedStandardBarGroups()
	standardPlayerBuffs = false; standardShortBuffs = false; standardLongBuffs = false; standardPlayerDebuffs = false; standardCooldowns = false
	standardTarget = false; standardFocus = false; standardRunes = false; standardNotifications = false
end

-- Check if a bar group with the name already exists, return a confirmation string if it does, otherwise return false
local function ConfirmNewBarGroup(name)
	local bg = MOD.db.profile.BarGroups[name]
	if IsOn(bg) then return L["Dup bar group string"](name) end -- name already exists
	return false
end

-- Copy layout configuration settings from another bar group
local function CopyBarGroupConfiguration(name)
	local bg = GetBarGroupEntry()
	local abg = MOD.db.profile.BarGroups[name]
	if IsOn(abg) and (bg.name ~= name) then -- make sure not same bar groups
		for n in pairs(MOD.BarGroupLayoutTemplate) do bg[n] = abg[n] end
		MOD:UpdateAllBarGroups()
	end
end

-- Copy fonts and textures from another bar group
local function CopyBarGroupFontsAndTextures(name)
	local bg = GetBarGroupEntry()
	local abg = MOD.db.profile.BarGroups[name]
	if IsOn(abg) and (bg.name ~= name) then -- make sure not same bar groups
		MOD:CopyFontsAndTextures(abg, bg)
		MOD:UpdateAllBarGroups()
	end
end

-- Copy standard colors from another bar group
local function CopyBarGroupStandardColors(name)
	local bg = GetBarGroupEntry()
	local abg = MOD.db.profile.BarGroups[name]
	if IsOn(abg) and (bg.name ~= name) then -- make sure not same bar groups
		MOD:CopyStandardColors(abg, bg)
		MOD:UpdateAllBarGroups()
	end
end

-- Functions for accessing bars, special case since the group tree function to list bars
local function GetBarEntryWithLabel(bname)
	local bg = GetBarGroupEntry()
	if not bg then return nil end

	for k, n in pairs(bg.bars) do if n.barLabel == bname then return n, k end end
	return nil
end

local function GetBarEntry(info)
	local barlist = MOD.OptionsTable.args.BarGroups.args.BarTab.args
	local bname = barlist[info[#info-2]].name
	return GetBarEntryWithLabel(bname)
end

-- Return true if there is no bar (or we are in the middle of creating a new bar)
local function NoBar()
	local bg = GetBarGroupEntry()
	if not bg or InMode("Bar") then return true end
	return next(bg.bars) == nil
end

-- Functions for sorting bar lists
local function SortAlphaUp(a, b) return string.upper(a) < string.upper(b) end
local function SortCustomOrder(a, b) return GetBarEntryWithLabel(a).sorder < GetBarEntryWithLabel(b).sorder end

-- Populate the options table for all the bars in the currently selected bar group
local function UpdateBarList()
	local bg = GetBarGroupEntry()
	if not bg then return end

	local barlist = MOD.OptionsTable.args.BarGroups.args.BarTab.args
	for j in pairs(barlist) do -- first remove any old entries
		if string.find(j, "XXX") then barlist[j] = nil end
	end
	if bg.auto then return end -- no bars if auto bar group
	
	local i, bp = 0, {} -- build a list of bar names from the profile table
	for _, n in pairs(bg.bars) do
		i = i + 1
		bp[i] = n.barLabel
	end

	if bg.sor ~= "X" then table.sort(bp, SortAlphaUp) else table.sort(bp, SortCustomOrder) end -- sort names by appropriate sorting algorithm
	
	for k, bname in pairs(bp) do -- now populate the groups in the tree with bar options
		barlist["XXX"..k] = {
			type = "group", order = k + 10, name = bname,
			disabled = function(info) return InMode("Bar") end,
			args = MOD.barOptions,
		}
		local bar = GetBarEntryWithLabel(bname)
		bar.sorder = k -- normalizes the sorder values to ascending integers
	end
	
	local key = "EnterNewBar"
	if enterNewBar then -- support entering new bar into the current bar group
		barlist[key] = { type = "group", order = 10, name = "-- New Bar --", args = MOD.barOptions }
		acedia:SelectGroup("Raven", "BarGroups", "BarTab", key)
	else
		barlist[key] = nil
	end
end

-- Select a particular bar
local function SelectBar(label)
	local barlist = MOD.OptionsTable.args.BarGroups.args.BarTab.args
	for k, b in pairs(barlist) do
		if string.find(k, "XXX") then -- only check special keys for bars
			if b.name == label then
				acedia:SelectGroup("Raven", "BarGroups", "BarTab", k)
				return
			end
		end
	end
end

-- Delete the current bar
local function DeleteBar()
	local bg = GetBarGroupEntry()
	if not bg then return end
	
	local statustable = acedia:GetStatusTable("Raven", { "BarGroups", "BarTab" })
	local selected = statustable.groups.selected

	if string.find(selected, "XXX") then -- only check special keys for bars
		local barlist = MOD.OptionsTable.args.BarGroups.args.BarTab.args
		local bname = barlist[selected].name
		local bar, key = GetBarEntryWithLabel(bname)
		if bar then -- remove the table entry, essential since recycling would require reset of special values
			bg.bars[key] = nil
			UpdateBarList()
		end
	end
	MOD:UpdateAllBarGroups()
end

-- Default bar information tables
local barSources = { "Death Knight", "Druid", "Hunter", "Mage", "Paladin", "Priest", "Rogue", "Shaman", "Warlock", "Warrior", "Racial", "Spell" }
local classIndex = { DEATHKNIGHT = 1, DRUID = 2, HUNTER = 3, MAGE = 4, PALADIN = 5, PRIEST = 6, ROGUE = 7, SHAMAN = 8, WARLOCK = 9, WARRIOR = 10 }
local classList = { "DEATHKNIGHT", "DRUID", "HUNTER", "MAGE", "PALADIN", "PRIEST", "ROGUE", "SHAMAN", "WARLOCK", "WARRIOR" }
local petClassList = { true, false, true, false, false, false, false, false, true, false }
local unitList = { player = "Player", pet = "Pet", target = "Target", focus = "Focus",
	mouseover = "Mouseover", pettarget = "Pet's Target", targettarget = "Target's Target", focustarget = "Focus's Target" }
local castList = { player = "Player", pet = "Pet", other = "Other", anyone = "Anyone" }
local lastdatecheck = ""
local lastdatecount = 0

-- Add a new bar to a bar group, initializing the sort order value and adding a unique identifier
-- Verify that the bar is not a duplicate of one already in the bar group before adding
local function AddBarToGroup(bg, bar)
	for k, n in pairs(bg.bars) do -- scan current bars and see if this bar is a duplicate, only tricky part is the castBy setting
		local ok = true
		if (bar.action == n.action) and (bar.barType == n.barType) and (bar.monitor == n.monitor) then
			if bar.castBy == n.castBy then return false end -- everything matches so is a duplicate for sure
			if n.castBy == "anyone" then return false end -- if already have everything then duplicate
			if bar.castBy == "anyone" then bg.bars[k] = nil; ok = false end -- more inclusive that the original so need to delete the original
		end
		if ok and (bar.barLabel == n.barLabel) then -- found duplicate label, add a suffix 
			bar.barLabel = bar.barLabel .. "*"
			bar.labelLink = true -- make sure not linked (true means not linked)
		end
	end
	
	bar.sorder = 0
	local d = date("%m%d%y-%H%M%S-")
	if d == lastdatecheck then
		lastdatecount = lastdatecount + 1
	else
		lastdatecheck = d
		lastdatecount = 0
	end
	
	bar.uniqueID = d .. lastdatecount
	table.insert(bg.bars, bar)
	MOD:UpdateAllBarGroups()
	return true
end

-- Process entering a new bar, state = "start", "ok", "cancel" to move between stages of bar entry
local function EnterNewBar(state)
	local bg = GetBarGroupEntry()
	if not bg then return end

	if state == "start" then
		enterNewBar = true
		if oldbar then -- cache of previous settings
			newbar = { -- only preserve selected fields from saved bar description
				barType = oldbar.barType, barSource = oldbar.barSource, barClass = oldbar.barClass, barLabel = "",
				monitor = oldbar.monitor, castBy = oldbar.castBy
			}
			oldbar = nil
		else
			newbar = { -- set defaults
				barType = "Notification", barSource = "Class", barClass = classIndex[MOD.myClass],
				barLabel = "", monitor = "player", castBy = "player"
			}
		end
		if not newbar.monitor then newbar.monitor = "player" else newbar.monitor = string.lower(newbar.monitor) end
		if not newbar.castBy then newbar.castBy = "player" else newbar.castBy = string.lower(newbar.castBy) end
	elseif state == "ok" then
		enterNewBar = false		
		local source = barSources[newbar.barClass]
		if newbar.barSource == "Pet" then source = "Pet" end
		if (newbar.barType ~= "Notification") and (newbar.barSource == "Other") and ValidateSpellName(selectedSpellName) then
			if not newbar.actionList then newbar.actionList = {} end
			if not newbar.selectList then newbar.selectList = {} end
			newbar.actionList[1] = selectedSpellName
			newbar.selectList[1] = true
		end
		local sel = nil
		if (newbar.barType == "Buff") or (newbar.barType == "Debuff") then -- buff or debuff
			if newbar.actionList then
				for k, n in pairs(newbar.selectList) do
					if n then
						local bname = newbar.actionList[k]
						local label = MOD:GetLabel(bname)
						local bar = {
							action = bname, enableBar = true, barLabel = label, barType = newbar.barType,
							barSource = source, monitor = newbar.monitor, castBy = newbar.castBy
						}
						if AddBarToGroup(bg, bar) then if not sel then sel = bar.barLabel end end
					end
				end
			end
		elseif newbar.barType == "Cooldown" then -- cooldown
			if newbar.actionList then
				for k, n in pairs(newbar.selectList) do
					if n then
						local bname = newbar.actionList[k]
						local label = MOD:GetLabel(bname)
						local bar = {
							action = bname, enableBar = true, barLabel = label, barType = newbar.barType,
							barSource = source, monitor = nil, castBy = nil, enableReady = nil, readyNotUsable = nil
						}
						if AddBarToGroup(bg, bar) then if not sel then sel = bar.barLabel end end
					end
				end
			end
		elseif newbar.barType == "Notification" then -- notification
			if newbar.conditionList and newbar.selectCondition then
				for k, n in pairs(newbar.selectCondition) do
					if n then
						local bname = newbar.conditionList[k]
						local bar = {
							action = bname, enableBar = true, barLabel = bname, barType = newbar.barType,
							barSource = nil, monitor = nil, castBy = nil
						}
						if AddBarToGroup(bg, bar) then if not sel then sel = bar.barLabel end end
					end
				end
			end
		end
		if sel then UpdateBarList(); SelectBar(sel) end
		oldbar = newbar
		newbar = nil
	elseif state == "cancel" then
		enterNewBar = false
		oldbar = newbar
		newbar = nil
	end
end

-- Get a formatted text description of the current bar
local function GetBarDescription(info)
	local a = ""
	local n = GetBarEntry(info)
	if n then
		if n.barType == "Notification" then
			a = L["Type and condition string"](n.barType, n.action)
			local spell = MOD:GetAssociatedSpellForBar(n)
			if spell then a = a .. L["Associated spell string"](spell) end
		elseif n.barType == "Buff" or n.barType == "Debuff" or n.barType == "Cooldown" then
			if n.action then
				a = L["Action type source string"](n.action, n.barType, n.barSource)
			else
				a = L["Type source string"](n.barType, n.barSource)
			end
			if n.barType ~= "Cooldown" then
				if not n.monitor then n.monitor = "player" else n.monitor = string.lower(n.monitor) end
				if not n.castBy then n.castBy = "player" else n.castBy = string.lower(n.castBy) end
				local unit, cb = unitList[n.monitor], castList[n.castBy]
				if unit and cb then a = a .. L["Action on cast by string"](unit, cb) end
			end
		end
	end
	return a
end

-- Move a bar up or down in the list
local function MoveBarInList(info, direction)
	local bar = GetBarEntry(info)
	if bar then
		if direction == "up" then bar.sorder = bar.sorder - 1.5 else bar.sorder = bar.sorder + 1.5 end
		UpdateBarList()
		SelectBar(bar.barLabel)
	end
end

-- Return a list for selection of configuration of either bar or icon oriented bar groups
local function GetOrientationList(iconOnly)
	if iconOnly == nil then iconOnly = false end
	configlist = {}
	for i, config in pairs(Nest_SupportedConfigurations) do if config.iconOnly == iconOnly then configlist[i] = config.name end end
	return configlist
end

-- Set a field in the current bar
local function SetBarField(info, fname, value)
	local n = GetBarEntry(info)
	if n then
		n[fname] = value
	end
end

-- Get a field in the current bar
local function GetBarField(info, fname)
	local n = GetBarEntry(info)
	if n then return n[fname] end
	return nil
end

-- Get the icon associated with the current bar
local function GetBarIcon(info)
	local b = GetBarEntry(info)
	return MOD:GetIconForBar(b)
end

-- Check for duplicate labels, including "other labels"
local function IsDuplicateLabel(bg, newLabel)
	for _, n in pairs(bg.bars) do if n.barLabel == newLabel then return true end end
	return false
end

-- Change a bar's label, have to redo the list afterwards, also do some validation and return false if not valid
local function SetBarLabel(info, newLabel, isLinked)
	local bg = GetBarGroupEntry()
	if not bg then return false end
	
	local bar = GetBarEntry(info)
	if not bar then return false end
	
	if bar.barLabel == newLabel then return true end  -- make sure new label is actually a change
	if newLabel == "" then return false end -- make sure not an empty string
	
	if IsDuplicateLabel(bg, newLabel) then -- make sure not a duplicate within this bar group
		print('"' .. newLabel .. '" ' .. L["is a duplicate of an existing bar label"])
		return false
	end

	local a, b = string.find(newLabel, "^%a[%a%d%b()%[%] %-%+%!%*_':]*") -- make sure new label has correct format and doesn't contain bad characters
	if a ~= 1 and b ~= string.len(newLabel) then
		print('"' .. newLabel .. '" ' .. L["does not begin with a letter or contains restricted characters"])
		return false
	end 

	DeleteBar()
	if isLinked and ((bar.barType == "Buff") or (bar.barType == "Debuff") or (bar.barType == "Cooldown")) then MOD:SetLabel(bar.action, newLabel) end
	bar.barLabel = newLabel
	table.insert(bg.bars, bar)
	UpdateBarList()
	SelectBar(newLabel)
	return true
end

-- Generate a list of auras that meet the criteria of type and source specified in the newbar table
local function GenerateBarActionList()
	local i = 0
	newbar.actionList = {}
	newbar.selectList = {}
	selectedSpellName = nil -- for entered spell names

	local btype = newbar.barType
	local c = classList[newbar.barClass]
	if (btype == "Buff") or (btype == "Debuff") or (btype == "Cooldown") then
		local btable = MOD[btype .. "Table"]
		if newbar.barSource == "Class" then
			for name, action in pairs(btable) do
				if (action.class == c) and not action.pet then i = i + 1; newbar.actionList[i] = name; end
			end
		elseif newbar.barSource == "Pet" then -- racial buffs
			for name, action in pairs(btable) do
				if (action.class == c) and action.pet then i = i + 1; newbar.actionList[i] = name; end
			end
		elseif newbar.barSource == "Racial" then -- racial buffs
			for name, action in pairs(btable) do
				if action.race then i = i + 1; newbar.actionList[i] = name; end
			end
		end
	end

	table.sort(newbar.actionList) -- alphabetical order to improve readability
	for k in pairs(newbar.actionList) do newbar.selectList[k] = false end -- default to all not selected
	newbar.actionListCount = i
end

-- Functions for accessing bar type lists
local function SetSelectedBarType(value)
	if newbar.barType ~= value then
		newbar.barType = value
		GenerateBarActionList()
	end
end

-- Functions for accessing bar source lists
local function SetSelectedBarSource(value)
	if newbar.barSource ~= value then 
		newbar.barSource = value
		if (value == "Pet") and not petClassList[newbar.barClass] then newbar.barClass = 3 end
		GenerateBarActionList()
	end
end

-- Functions for accessing bar class lists
local function SetSelectedBarClass(value)
	if newbar.barClass ~= value then 
		newbar.barClass = value
		GenerateBarActionList()
	end
end

-- Generate a list of conditions to select from
local function GetSelectConditionList()
	local i, t = 0, {}
	for _, n in pairs(MOD.db.profile.Conditions[MOD.myClass]) do
		if IsOn(n) and n.name then
			i = i + 1
			t[i] = n.name
		end
	end
	table.sort(t)
	return t, i
end

-- Get the, potentially cached, list of conditions to select from
local function GetBarConditionList()
	if not newbar.conditionList then
		newbar.conditionList, newbar.conditionListCount = GetSelectConditionList()
		newbar.selectCondition = {}
		for k in pairs(newbar.conditionList) do newbar.selectCondition[k] = false end -- default to all not selected
	end
	return newbar.conditionList
end

local function GetBarConditionListCount() GetBarConditionList(); return newbar.conditionListCount end
local function GetSelectedBarCondition(key) return newbar.selectCondition[key] end
local function SetSelectedBarCondition(key, value) newbar.selectCondition[key] = value end
local function SetAllBarConditions(value) for k in pairs(newbar.conditionList) do newbar.selectCondition[k] = value end end

-- Generate a list of auras that meet the criteria of type and source specified in the newbar table
local function GetBarActionList() if not newbar.actionList then GenerateBarActionList() end; return newbar.actionList end
local function GetBarActionListCount() if not newbar.actionList then GenerateBarActionList() end; return newbar.actionListCount end
local function GetSelectedBarAction(key) return newbar.selectList[key] end
local function SetSelectedBarAction(key, value) newbar.selectList[key] = value end
local function SetAllBarActions(value) for k in pairs(newbar.actionList) do newbar.selectList[k] = value end end

-- Functions for accessing condition lists
local function GetConditionList() return GetNameList("Conditions", MOD.db.profile.Conditions[MOD.myClass]) end
local function GetCondition() return GetNameEntry("Conditions", MOD.db.profile.Conditions[MOD.myClass]) end
local function GetSelectedCondition() return GetNameSelection("Conditions", MOD.db.profile.Conditions[MOD.myClass]) end
local function SetSelectedCondition(value) SetNameSelection("Conditions", MOD.db.profile.Conditions[MOD.myClass], value) end
local function DeleteCondition() DeleteNameEntry("Conditions", MOD.db.profile.Conditions[MOD.myClass]) end
local function NoCondition() return enterNewCondition or (GetSelectedCondition() == nil) end

local function CreateCondition(name)
	local n = { enabled = true, name = name, notify = true, tooltip = true, tests = {}, testResult = false, dependencies = {}, result = false }
	AddNameEntry("Conditions", MOD.db.profile.Conditions[MOD.myClass], n)
end

-- Check if a condition with the name already exists, return a confirmation string if it does, otherwise return false
local function ConfirmNewCondition(name)
	local cs = MOD.db.profile.Conditions[MOD.myClass]
	local c = cs[name]
	if IsOn(c) and c.name then return L["Dup condition string"](name) end -- name already exists
	return false
end

-- Purge an existing condition, have to delete from bar groups, notifications, dependencies
local function PurgeCondition()
	local con = GetCondition()
	local oldName = con.name
	for _, bp in pairs(MOD.db.profile.BarGroups) do -- check bar group conditions
		if IsOn(bp) then
			if bp.checkCondition and (bp.condition == oldName) then bp.condition = nil; bp.checkCondition = false end
			for _, bar in pairs(bp.bars) do -- check for notifications
				if (bar.barType == "Notification") and (bar.action == oldName) then bar.enableBar = false end
			end
		end
	end
	for _, c in pairs(MOD.db.profile.Conditions[MOD.myClass]) do -- check dependencies
		if IsOn(c) and c.name and c.dependencies then
			local result = c.dependencies[oldName]
			if result ~= nil then
				c.dependencies[oldName] = nil
			end
		end
	end
end

-- Rename an existing condition, have to change for bar groups, bars, dependencies
local function RenameCondition(newName)
	local con = GetCondition()
	local oldName = con.name
	DeleteNameEntry("Conditions", MOD.db.profile.Conditions[MOD.myClass])
	con.name = newName
	AddNameEntry("Conditions", MOD.db.profile.Conditions[MOD.myClass], con)
	for _, bp in pairs(MOD.db.profile.BarGroups) do -- check bar group and bar conditions
		if IsOn(bp) then
			if bp.checkCondition and (bp.condition == oldName) then bp.condition = newName end
			if not bp.auto then -- have to scan bars in custom bar groups
				for _, bar in pairs(bp.bars) do
					if (bar.barType == "Notification") and (bar.action == oldName) then bar.action = newName end
					if bar.hideCondition == oldName then bar.hideCondition = newName end
					if bar.flashCondition == oldName then bar.flashCondition = newName end
					if bar.fadeCondition == oldName then bar.fadeCondition = newName end
				end
			end
		end
	end
	for _, c in pairs(MOD.db.profile.Conditions[MOD.myClass]) do -- check dependencies
		if IsOn(c) and c.name and c.dependencies then
			local result = c.dependencies[oldName]
			if result ~= nil then
				c.dependencies[oldName] = nil
				c.dependencies[newName] = result
			end
		end
	end
end

-- Return the selected condition associated with a bar group
local function GetBarGroupSelectedCondition(list)
	local bp = GetBarGroupEntry()
	if bp and bp.checkCondition and bp.condition then
		for j, c in pairs(list) do
			if c == bp.condition then return j end
		end
	end
	return nil
end

-- Return the index of the condition in the list
local function GetBarSelectedCondition(list, condition)
	for j, c in pairs(list) do
		if c == condition then return j end
	end
	return nil
end

-- Get a list of dependencies to select from
local function GetDependenciesList()
	local con = GetCondition()
	local i, t = 0, {}
	local nt = MOD.db.profile.Conditions[MOD.myClass]
	for _, n in pairs(nt) do
		if IsOn(n) and n.name and (n.name ~= con.name) then
			i = i + 1
			t[i] = n.name
		end
	end
	table.sort(t)
	return t
end

-- Set a dependency with the specified value in the current condition's dependencies list
local function SetDependency(dep, value)
	local con = GetCondition()
	if not con.dependencies then con.dependencies = {} end
	con.dependencies[dep] = value
end

-- Get a dependency type
local function GetDependencyType(dep)
	local con = GetCondition()
	if con.dependencyType then return con.dependencyType[dep] end
	return nil
end

-- Set a dependency type with the specified value
local function SetDependencyType(dep, value)
	local con = GetCondition()
	if not con.dependencyType then con.dependencyType = {} end
	con.dependencyType[dep] = value
end

-- Check if the dependency is included in the current condition's dependencies list
local function CheckDependency(dep)
	local con = GetCondition()
	if con.dependencies then
		for d in pairs(con.dependencies) do
			if dep == d then return true end
		end
	end
	return false
end

-- Return the selected dependency, clear it when switching conditions
local function GetSelectedDependency()
	local con = GetCondition()
	if con ~= selectedCondition then
		selectedCondition = con
		selectedDependency = nil
	end
	return selectedDependency
end

-- Set a field in the current bar group
local function SetBarGroupField(fname, value)
	local n = GetBarGroupEntry()
	if n then
		n[fname] = value
		if fname == "linkSettings" then MOD:InitializeBarGroupSettings(n) end -- special case for linked layouts
		if (fname == "pointX") or (fname == "pointY") then MOD:SetBarGroupPosition(n) end -- special case for display position
		if fname == "filterBuffLink" then MOD:InitializeFilterList(n, "Buff") end -- special case for linked buff filters
		if fname == "filterDebuffLink" then MOD:InitializeFilterList(n, "Debuff") end -- special case for linked debuff filters
		if fname == "filterCooldownLink" then MOD:InitializeFilterList(n, "Cooldown") end -- special case for linked cooldown filters
		MOD:UpdateAllBarGroups() -- any field changes need to be reflected in the actual bars on the display
	end
end

-- Get a field in the current bar group
local function GetBarGroupField(fname)
	local n = GetBarGroupEntry()
	if n then return n[fname] end
	return nil
end

-- Return the current merge bar group for the selected bar group
local function GetMergeBarGroup()
	local bg = GetBarGroupEntry()
	if bg then
		local bglist = GetBarGroupList()
		for i, bgname in pairs(bglist) do if bgname == bg.mergeInto then return i end end
	end
	return nil
end

-- Set the merge bar group for the selected bar group
local function SetMergeBarGroup(value)
	local bg = GetBarGroupEntry()
	if bg then
		local bglist = GetBarGroupList()
		local mergeInto = bglist[value]
		local mbg = MOD.db.profile.BarGroups[mergeInto]
		if IsOn(mbg) and mbg.enabled and not mbg.merged then
			bg.mergeInto = mergeInto		
			MOD:UpdateAllBarGroups() -- any changes need to be reflected in the actual bars on the display
		end
	end
end

-- Return the current anchor bar group for the selected bar group
local function GetBarGroupAnchor()
	local bg = GetBarGroupEntry()
	if bg then
		local anchor = bg.name
		if bg.anchor and (bg.anchor ~= bg.name) then anchor = bg.anchor end
		local bglist = GetBarGroupList()
		for i, bgname in pairs(bglist) do if bgname == anchor then return i end end
	end
	return nil
end

-- Set the current anchor bar group for the selected bar group
local function SetBarGroupAnchor(value)
	local bg = GetBarGroupEntry()
	if bg then
		if not value then
			bg.anchor = false
			bg.anchorX = 0
			bg.anchorY = 0
			bg.anchorLastBar = false
			bg.anchorEmpty = false
			bg.locked = false
		else
			local bglist = GetBarGroupList()
			local anchor = bglist[value]
			if anchor ~= bg.name then -- need to make sure no loops being set up
				local i, a = 0, anchor
				while i < 1000 do -- safety net, should never get this high
					i = i + 1
					local bp = MOD.db.profile.BarGroups[a]
					if not bp.anchor then break end -- found end of chain without looping back
					a = bp.anchor
					if a == bg.name then -- found a loop so print a warning and return
						print(L["Circular string"](a, anchor))
						return
					end
				end
				bg.anchor = anchor
			else bg.anchor = false end
		end
		MOD:UpdateAllBarGroups() -- any changes need to be reflected in the actual bars on the display
	end
end

-- Return the filter list of a given type for a bar group
local function GetBarGroupFilter(actionType)
	local n = GetBarGroupEntry()
	if n then
		local listName, selectName = "filter" .. actionType .. "List", "filter" .. actionType .. "Selection"
		local filterList = n[listName]
		if filterList then 
			table.sort(filterList)
			if not n[selectName] then n[selectName] = next(filterList, nil) end
			return filterList
		end
		n[selectName] = nil
	end
	return {}
end

-- Return selection within the filter list of a given type for a bar group
local function GetBarGroupFilterSelection(actionType)
	local n = GetBarGroupEntry()
	if n then
		local listName, selectName = "filter" .. actionType .. "List", "filter" .. actionType .. "Selection"
		local filterList = n[listName]
		local filterSelect = n[selectName]
		if filterList then 
			table.sort(filterList)
			if not filterSelect or not filterList[filterSelect] then n[selectName] = next(filterList, nil) end
			return n[selectName]
		end
	end
	return nil
end

-- Reset the filter list of a given type for a bar group
local function ResetBarGroupFilter(actionType)
	local n = GetBarGroupEntry()
	if n then
		local listName, selectName = "filter" .. actionType .. "List", "filter" .. actionType .. "Selection"
		n[listName] = nil
		n[selectName] = nil
	end
end

-- Add an action to be filtered by a bar group
local function AddBarGroupFilter(actionType, actionName)
	local n = GetBarGroupEntry()
	if n then
		local listName, selectName = "filter" .. actionType .. "List", "filter" .. actionType .. "Selection"
		local filterList = n[listName]
		if not filterList then filterList = {}; n[listName] = filterList end
		filterList[actionName] = actionName
		n[selectName] = actionName
		MOD:UpdateAllBarGroups() -- any field changes need to be reflected in the actual bars on the display
	end
end

-- Delete an action from a bar group filter list
local function DeleteBarGroupFilter(actionType, actionName)
	local n = GetBarGroupEntry()
	if n then
		local listName, selectName = "filter" .. actionType .. "List", "filter" .. actionType .. "Selection"
		local filterList = n[listName]
		if filterList and actionName then
			filterList[actionName] = nil
			n[selectName] = nil
		end
		MOD:UpdateAllBarGroups() -- any field changes need to be reflected in the actual bars on the display
	end
end

-- Set a field in the current notification
local function SetConditionField(fname, value)
	local n = GetCondition()
	if n then n[fname] = value end
	MOD:UpdateAllBarGroups() -- any field changes need to be reflected in the actual bars on the display
end

-- Get a field in the current notification
local function GetConditionField(fname)
	local n = GetCondition()
	if n then return n[fname] end
	return nil
end

-- Return a text description of the current condition
local function GetConditionDescription()
	local c = GetCondition()
	if c then return MOD:GetConditionText(c.name) end
	return ""
end

-- Copy test settings from a shared condition
local function CopyConditionSettings(name)
	local s = MOD.db.global.SharedConditions[name]
	local d = GetCondition()
	d.tests = Raven_CopyTable(s.tests)
	d.dependencies = Raven_CopyTable(s.dependencies)
	d.associatedSpell = s.associatedSpell
end

-- Return a sorted list of shared conditions
local function GetSharedConditionList()
	local i, t = 0, {}
	for k in pairs(MOD.db.global.SharedConditions) do
		i = i + 1
		t[i] = k
	end
	table.sort(t)
	return t
end
							
-- Get current setting for a test field in the currently selected condition
local function GetTestField(ttype, fname)
	local c = GetCondition()
	if c and c.tests then
		local t = c.tests[ttype] -- find the right test, if it exists
		if t then return t[fname] end -- found the field
	end
	return nil
end

-- Set value of a test field in the current condition, initializing fields as necessary
local function SetTestField(ttype, fname, value)
	local c = GetCondition()
	if c then
		if not c.tests then
			c.tests = {}
		else
			local t = c.tests[ttype] -- find the right test, if it has already been defined
			if t then -- found the field
				t[fname] = value
				MOD:UpdateAllBarGroups()
				return
			end
		end
		local newTest = {}
		for f, v in pairs(MOD.conditionTests[ttype]) do newTest[f] = v end -- initialize test parameters
		newTest[fname] = value -- after initializing the test can set the field
		c.tests[ttype] = newTest
		MOD:UpdateAllBarGroups()
	end
end

-- Check if an option field in the current test is turned on
local function IsTestFieldOn(ttype, fname)
	local value = GetTestField(ttype, fname)
	return (value ~= nil) and (value ~= Off)
end

-- Check if an option field in the current test is turned off
local function IsTestFieldOff(ttype, fname)
	local value = GetTestField(ttype, fname)
	return value == nil or value == Off
end

-- Set a field associated with a test, strip off white space at start and end of the string
local function SetTestFieldString(ttype, fname, s)
	local whiteStart, whiteEnd = string.find(s, "%s*", 1) -- skip white space
	if whiteStart == 1 then s = string.sub(s, whiteEnd + 1) end	
	s = string.reverse(s) -- reverse so can check for end white space
	whiteStart, whiteEnd = string.find(s, "%s*", 1) -- skip white space
	if whiteStart == 1 then s = string.sub(s, whiteEnd + 1) end
	local str = string.reverse(s) -- get the string after white space, reverse again to restore original order
	if str ~= "" then -- make sure not empty string
		if fname ~= "item" then ValidateSpellName(str) end
		SetTestField(ttype, fname, str)
	else
		SetTestField(ttype, fname, nil)
	end
end

-- Get a comma-separated list of the comma-seperated strings associated with a test
local function GetTestFieldSpellList(ttype, fname)
	local a, d = "", ""
	local strs = GetTestField(ttype, fname)
	if strs then
		for _, str in pairs(strs) do
			a = a .. d .. str -- append each to the comma-seperated string being built
			d = ", "
		end
	end
	return a
end

-- Set the auras associated with a test from a string with comma-separated names
local function SetTestFieldSpellList(ttype, fname, st)
	local start, strlist, s = 1, {}, st .. ',' -- initialize array, add ending comma
	repeat
		local whiteStart, whiteEnd = string.find(s, "%s*", start) -- skip white space
		if whiteStart == start then start = whiteEnd + 1 end
		if start > string.len(s) then break end
		local comma, nexts = string.find(s, '%s*,', start) -- find next comma, skipping white space preceding it
		if (comma - start) > 1 then -- make sure there is a string left after stripping white space
			table.insert(strlist, string.sub(s, start, comma - 1)) -- store the string in the table of aura names
		end
		start = nexts + 1
	until start > string.len(s)
	for _, k in pairs(strlist) do ValidateSpellName(k) end -- validates spell names
	SetTestField(ttype, fname, strlist)
end

-- Set a standard color
local function SetStandardColor(i, r, g, b, a)
	local c = MOD.ColorPalette[i]
	if c then c.r = r; c.g = g; c.b = b; c.a = a end
end

-- Get a standard color
local function GetStandardColor(i)
	local c = MOD.ColorPalette[i]
	if c then return c.r, c.g, c.b, c.a end
	return 0, 0, 0, 0
end

-- Generate text array using the standard colors
local function GetStandardColorList()
	local t = { }
	for name, c in pairs(MOD.ColorPalette) do
		t[name] = string.format("|cFF%02x%02x%02x%s", c.r*255, c.g*255, c.b*255, name)
	end
	return t
end

-- Add class, group and racial buffs or debuffs from the known auras table to a bar group
local function AddAurasToBarGroup(bg, buff, monitor, castBy)
	if buff then
		bg.detectBuffs = true
		bg.detectBuffsMonitor = monitor
		bg.detectBuffsCastBy = castBy
	else
		bg.detectDebuffs = true
		bg.detectDebuffsMonitor = monitor
		bg.detectDebuffsCastBy = castBy
	end
end

-- Add cooldowns from the known cooldowns table to a bar group
local function AddCooldownsToBarGroup(bg)
	bg.detectCooldowns = true
end

-- Add notifications for conditions to a bar group
local function AddNotificationsToBarGroup(bg)
	-- only check conditions for the player's class
	local ct = MOD.db.profile.Conditions[MOD.myClass]
	if ct then
		for _, c in pairs(ct) do
			if IsOn(c) and c.name and c.notify then
				local bname = c.name
				local bar = { action = bname, enableBar = true, barLabel = bname, barType = "Notification", barSource = nil, monitor = nil, castBy = nil, }
				AddBarToGroup(bg, bar)
			end
		end
	end
end

-- Configure Raven with the selected standard bar groups
local function ConfigureBarGroups(style)
	local offsetY, delta = 0, style and 20 or 30
	if standardNotifications then
		local bg = CreateBarGroup(L["Notifications"], false, false, style, 0, offsetY)
		AddNotificationsToBarGroup(bg)
		offsetY = offsetY + delta
	end
	if standardRunes then
		local bg = CreateBarGroup(L["Runes"], true, false, style, 0, offsetY)
		bg.showNoDuration = true
		bg.detectRuneCooldowns = true
		offsetY = offsetY + delta
	end
	if standardFocus then
		local bg = CreateBarGroup(L["Focus"], true, false, style, 0, offsetY)
		AddAurasToBarGroup(bg, true, "focus", "player")
		AddAurasToBarGroup(bg, false, "focus", "player")
		bg.showNoDuration = false
		offsetY = offsetY + delta
	end
	if standardTarget then
		local bg = CreateBarGroup(L["Target"], true, false, style, 0, offsetY)
		AddAurasToBarGroup(bg, true, "target", "player")
		AddAurasToBarGroup(bg, false, "target", "player")
		bg.showNoDuration = false
		offsetY = offsetY + delta
	end
	if standardCooldowns then
		local bg = CreateBarGroup(L["Cooldowns"], true, false, style, 0, offsetY)
		bg.checkDuration = true
		bg.minimumDuration = true
		bg.filterDuration = 2
		AddCooldownsToBarGroup(bg)
		offsetY = offsetY + delta
	end
	if standardPlayerDebuffs then
		local bg = CreateBarGroup(L["Debuffs"], true, false, style, 0, offsetY)
		AddAurasToBarGroup(bg, false, "player", "anyone")
		bg.showNoDuration = true
		bg.iconColors = "Debuffs"
		offsetY = offsetY + delta
	end
	if standardPlayerBuffs then
		local bg = CreateBarGroup(L["Buffs"], true, false, style, 0, offsetY)
		bg.showNoDuration = true
		bg.sor = "T"
		AddAurasToBarGroup(bg, true, "player", "anyone")
		offsetY = offsetY + delta
	end
	if standardShortBuffs then
		local bg = CreateBarGroup(L["Short Buffs"], true, false, style, 0, offsetY)
		bg.checkDuration = true
		bg.minimumDuration = false
		bg.filterDuration = 120
		bg.showNoDuration = false
		AddAurasToBarGroup(bg, true, "player", "anyone")
		offsetY = offsetY + delta
	end
	if standardLongBuffs then -- minimum duration of two minutes
		local bg = CreateBarGroup(L["Long Buffs"], true, false, style, 0, offsetY)
		bg.checkDuration = true
		bg.minimumDuration = true
		bg.filterDuration = 120
		-- bg.detectTracking = true
		bg.showNoDuration = true
		bg.barColors = "Class"
		bg.sor = "C"
		AddAurasToBarGroup(bg, true, "player", "anyone")
		offsetY = offsetY + delta
	end
	MOD:UpdateAllBarGroups()
end

-- Prepare a table of time format options showing examples to choose from with select dropdown menu
local function GetTimeFormatList(s, c)
	local i, menu = 1, {}
	while i <= #Nest_TimeFormatOptions do
		local f = Nest_FormatTime
		local t1, t2, t3, t4, t5 = f(8125.8, i, s, c), f(343.8, i, s, c), f(75.3, i, s, c), f(42.7, i, s, c), f(3.6, i, s, c)
		menu[i] = t1 .. ", " .. t2 .. ", " .. t3 .. ", " .. t4 .. ", " .. t5
		i = i + 1
	end
	return menu
end

-- Create a mini-options table to be inserted at top level in the Bliz interface
-- L["Top Options"] = "This addon lets you monitor buffs and debuffs for player, target and focus. Highlights and cooldown counts can be added to buttons to show status. Monitored buffs and debuffs can trigger helpful notifications."
-- For some reason cannot localize the strings in this table!
MOD.TopOptionsTable = {
	type = "group", order = 1,
	args = {
		Configure = {
			type = "execute", order = 90, name = L["Configure"],
			desc = L["Open Raven's standalone options panel."],
			func = function(info) MOD:OptionsPanel() end,
		},
	},
}

-- Create the options table to be used by the configuration GUI
MOD.OptionsTable = {
	type = "group", childGroups = "tab",
	args = {
		FrontPage = {
			type = "group", order = 10, name = L["Setup"],
			disabled = function(info) return InMode() end,
			args = {
				EnableGroup = {
					type = "group", order = 1, name = L["Enable"], inline = true,
					args = {
						EnableGroup = {
							type = "toggle", order = 10, name = L["Enable Raven"],
							desc = L["If checked, Raven is enabled, otherwise all features are disabled."],
							get = function(info) return MOD.db.profile.enabled end,
							set = function(info, value) MOD.db.profile.enabled = value end,
						},
						HideBlizzBuffsGroup = {
							type = "toggle", order = 20, name = L["Hide Blizzard Buffs"],
							desc = L["If checked, Raven will hide the default user interface for buffs."],
							get = function(info) return MOD.db.profile.hideBlizz end,
							set = function(info, value) MOD.db.profile.hideBlizz = value end,
						},
						MinimapGroup = {
							type = "toggle", order = 30, name = L["Minimap Icon"],
							desc = L["If checked, Raven will add an icon to the minimap."],
							hidden = function(info) return MOD.ldbi == nil end,
							get = function(info) return not MOD.db.global.Minimap.hide end,
							set = function(info, value)
								MOD.db.global.Minimap.hide = not value
								if value then MOD.ldbi:Show("Raven") else MOD.ldbi:Hide("Raven") end
							end,
						},
					},
				},
				StandardGroups = {
					type = "group", order = 10, name = L["Standard Bar Groups"], inline = true,
					args = {
						Anchors = {
							type = "description", order = 1,
							name = L["Anchor description"]
						},
						Buffs = {
							type = "toggle", order = 5, name = L["Buffs"],
							get = function(info) return standardPlayerBuffs end,
							set = function(info, value) standardPlayerBuffs = value end,
						},
						spacer = { type = "description", order = 6, width = "double",
							name = function() return L["All buffs on you, sort by time left."] .. BarGroupString(L["Buffs"]) end,
						},
						spacerA = { type = "description", name = "", order = 7 },
						ShortBuffs = {
							type = "toggle", order = 10, name = L["Short Buffs"],
							get = function(info) return standardShortBuffs end,
							set = function(info, value) standardShortBuffs = value end,
						},
						spacer0 = { type = "description", order = 11, width = "double",
							name = function() return L["Buffs on you lasting less than 2 minutes."] .. BarGroupString(L["Short Buffs"]) end,
						},
						spacer0A = { type = "description", name = "", order = 12 },
						LongBuffs = {
							type = "toggle", order = 20, name = L["Long Buffs"],
							get = function(info) return standardLongBuffs end,
							set = function(info, value) standardLongBuffs = value end,
						},
						spacer1 = { type = "description", order = 21, width = "double",
							name = function() return L["Buffs on you lasting at least 2 minutes, sort and color by class."] .. BarGroupString(L["Long Buffs"]) end, 
						},
						spacer1A = { type = "description", name = "", order = 22 },
						Debuffs = {
							type = "toggle", order = 25, name = L["Debuffs"],
							get = function(info) return standardPlayerDebuffs end,
							set = function(info, value) standardPlayerDebuffs = value end,
						},
						spacer2 = { type = "description", order = 26, width = "double",
							name = function() return L["Debuffs on you."] .. BarGroupString(L["Debuffs"]) end,
						},
						spacer2A = { type = "description", name = "", order = 27 },
						Cooldowns = {
							type = "toggle", order = 30, name = L["Cooldowns"],
							get = function(info) return standardCooldowns end,
							set = function(info, value) standardCooldowns = value end,
						},
						spacer3 = { type = "description", order = 31, width = "double",
							name = function() return L["Cooldowns on your spells and trinkets lasting at least 2 seconds."] .. BarGroupString(L["Cooldowns"]) end, 
						},
						spacer3A = { type = "description", name = "", order = 32 },
						Target = {
							type = "toggle", order = 35, name = L["Target"],
							get = function(info) return standardTarget end,
							set = function(info, value) standardTarget = value end,
						},
						spacer4 = { type = "description", order = 36, width = "double",
							name = function() return L["Buffs and debuffs you cast on the target."] .. BarGroupString(L["Target"]) end, 
						},
						spacer4A = { type = "description", name = "", order = 37 },
						Focus = {
							type = "toggle", order = 40, name = L["Focus"],
							get = function(info) return standardFocus end,
							set = function(info, value) standardFocus = value end,
						},
						spacer5 = { type = "description", order = 41, width = "double",
							name = function() return L["Buffs and debuffs you cast on the focus."] .. BarGroupString(L["Focus"]) end, 
						},
						spacer5A = { type = "description", name = "", order = 42,
							hidden = function(info) return MOD.myClass ~= "DEATHKNIGHT" end,
						},
						Runes = {
							type = "toggle", order = 45, name = L["Runes"],
							hidden = function(info) return MOD.myClass ~= "DEATHKNIGHT" end,
							get = function(info) return standardRunes end,
							set = function(info, value) standardRunes = value end,
						},
						spacer6 = { type = "description", order = 46, width = "double",
							name = function() return L["Rune cooldown bars for Death Knights."] .. BarGroupString(L["Runes"]) end, 
							hidden = function(info) return MOD.myClass ~= "DEATHKNIGHT" end,
						},
						spacer6A = { type = "description", name = "", order = 47 },
						Notifications = {
							type = "toggle", order = 50, name = L["Notifications"],
							get = function(info) return standardNotifications end,
							set = function(info, value) standardNotifications = value end,
						},
						spacer7 = { type = "description", order = 51, width = "double",
							name = function() return L["Common class-specific notifications."] .. BarGroupString(L["Notifications"]) end, 
						},
						spacer7A = { type = "description", name = "", order = 52 },
						ConfigureBars = {
							type = "execute", order = 90, name = L["Create As Bars"],
							desc = L["Create bars string"],
							disabled = function(info) return not AnySelectedStandardBarGroups(false) end,
							func = function(info) ConfigureBarGroups(true) end,
							confirm = function(info) if CheckBarGroupsExist() then return L["Configure string"] end end,
						},
						ConfigureIcons = {
							type = "execute", order = 91, name = L["Create As Icons"],
							desc = L["Create icons string"],
							disabled = function(info) return not AnySelectedStandardBarGroups(false) end,
							func = function(info) ConfigureBarGroups(false) end,
							confirm = function(info) if CheckBarGroupsExist() then return L["Configure string"] end end,
						},
						spacer10 = { type = "description", name = "", order = 95, width = "half" },
						Reset = {
							type = "execute", order = 96, name = L["Reset"], width = "half",
							desc = L["Reset selections."],
							disabled = function(info) return not AnySelectedStandardBarGroups(false) end,
							func = function(info) ResetSelectedStandardBarGroups() end,
						},
						LinkSettings = {
							type = "execute", order = 97, name = L["Link"], width = "half",
							desc = L["Link settings string"],
							disabled = function(info) return not AnySelectedStandardBarGroups(true) end,
							func = function(info) LinkStandardBarGroups() end,
						},
						DeleteBars = {
							type = "execute", order = 98, name = L["Delete"], width = "half",
							desc = L["Delete the selected bar groups."],
							disabled = function(info) return not AnySelectedStandardBarGroups(true) end,
							func = function(info) DeleteStandardBarGroups() end,
							confirm = function(info) return L["Delete standard string"] end,
						},
					},
				},
				AnchorGroup = {
					type = "group", order = 50, name = L["Bar Group Anchors"], inline = true,
					args = {
						Anchors = {
							type = "description", order = 10,
							name = L["Bar group anchor string"],
						},
						LockBars = {
							type = "execute", order = 11, name = L["Lock All Anchors"], 
							desc = L["Lock and hide the anchors for all bar groups."],
							func = function(info) MOD:LockBarGroups(true) end,
						},
						UnlockBars = {
							type = "execute", order = 12, name = L["Unlock All Anchors"],
							desc = L["Unlock and show the anchors for all bar groups."],
							func = function(info) MOD:LockBarGroups(false) end,
						},
					},
				},
			},
		},
		DefaultsPage = {
			type = "group", order = 15, name = L["Defaults"],
			disabled = function(info) return InMode() end,
			args = {
				Defaults = {
					type = "description", order = 10,
					name = L["Defaults string"],
				},
				DimensionGroup = {
					type = "group", order = 10, name = L["Format"], inline = true,
					args = {
						BarsGroup = {
							type = "group", order = 10, name = L["Bar Configurations"], inline = true,
							args = {
								BarWidth = {
									type = "range", order = 10, name = L["Bar Width"], min = 5, max = 500, step = 1,
									desc = L["Set width of bars."],
									get = function(info) return MOD.db.global.Defaults.barWidth end,
									set = function(info, value) MOD.db.global.Defaults.barWidth = value; MOD:UpdateAllBarGroups() end,
								},
								BarHeight = {
									type = "range", order = 15, name = L["Bar Height"], min = 5, max = 100, step = 1,
									desc = L["Set height of bars."],
									get = function(info) return MOD.db.global.Defaults.barHeight end,
									set = function(info, value) MOD.db.global.Defaults.barHeight = value; MOD:UpdateAllBarGroups() end,
								},
								IconSize = {
									type = "range", order = 20, name = L["Icon Size"], min = 5, max = 100, step = 1,
									desc = L["Set width/height for icons."],
									get = function(info) return MOD.db.global.Defaults.iconSize end,
									set = function(info, value) MOD.db.global.Defaults.iconSize = value; MOD:UpdateAllBarGroups() end,
								},
								BarScale = {
									type = "range", order = 25, name = L["Scale"], min = 0.1, max = 2, step = 0.05,
									desc = L["Set scale factor for bars and icons."],
									get = function(info) return MOD.db.global.Defaults.scale end,
									set = function(info, value) MOD.db.global.Defaults.scale = value; MOD:UpdateAllBarGroups() end,
								},
								Space1 = { type = "description", name = "", order = 40 },
								HorizontalSpacing = {
									type = "range", order = 60, name = L["Horizontal Spacing"], min = -100, max = 100, step = 1,
									desc = L["Adjust horizontal spacing between bars."],
									get = function(info) return MOD.db.global.Defaults.spacingX end,
									set = function(info, value) MOD.db.global.Defaults.spacingX = value; MOD:UpdateAllBarGroups() end,
								},
								VerticalSpacing = {
									type = "range", order = 65, name = L["Vertical Spacing"], min = -100, max = 100, step = 1,
									desc = L["Adjust vertical spacing between bars."],
									get = function(info) return MOD.db.global.Defaults.spacingY end,
									set = function(info, value) MOD.db.global.Defaults.spacingY = value; MOD:UpdateAllBarGroups() end,
								},
								IconOffsetX = {
									type = "range", order = 70, name = L["Icon Inset"], min = -200, max = 200, step = 1,
									desc = L["Set icon's horizontal inset from bar."],
									get = function(info) return MOD.db.global.Defaults.iconOffsetX end,
									set = function(info, value) MOD.db.global.Defaults.iconOffsetX = value; MOD:UpdateAllBarGroups() end,
								},
								IconOffsetY = {
									type = "range", order = 75, name = L["Icon Offset"], min = -200, max = 200, step = 1,
									desc = L["Set vertical offset between icon and bar."],
									get = function(info) return MOD.db.global.Defaults.iconOffsetY end,
									set = function(info, value) MOD.db.global.Defaults.iconOffsetY = value; MOD:UpdateAllBarGroups() end,
								},
								Space2 = { type = "description", name = "", order = 80 },
								BarFormatGroup = {
									type = "group", order = 90, name = "", inline = true,
									args = {
										HideIconGroup = {
											type = "toggle", order = 30, name = L["Icon"], width = "half",
											desc = L["Show icons with bars (required for tooltips)."],
											get = function(info) return not MOD.db.global.Defaults.hideIcon end,
											set = function(info, value) MOD.db.global.Defaults.hideIcon = not value; MOD:UpdateAllBarGroups() end,
										},
										HideClockGroup = {
											type = "toggle", order = 31, name = L["Clock"], width = "half",
											desc = L["Show clock animation on icons for timer bars."],
											get = function(info) return not MOD.db.global.Defaults.hideClock end,
											set = function(info, value) MOD.db.global.Defaults.hideClock = not value; MOD:UpdateAllBarGroups() end,
										},
										HideBarGroup = {
											type = "toggle", order = 32, name = L["Bar"], width = "half",
											desc = L["Show colored bar and background."],
											get = function(info) return not MOD.db.global.Defaults.hideBar end,
											set = function(info, value) MOD.db.global.Defaults.hideBar = not value; MOD:UpdateAllBarGroups() end,
										},
										HideSparkGroup = {
											type = "toggle", order = 33, name = L["Spark"], width = "half",
											desc = L["Show spark that moves across bars to indicate remaining time."],
											disabled = function(info) return MOD.db.global.Defaults.hideBar end,
											get = function(info) return not MOD.db.global.Defaults.hideSpark end,
											set = function(info, value) MOD.db.global.Defaults.hideSpark = not value; MOD:UpdateAllBarGroups() end,
										},
										HideLabelGroup = {
											type = "toggle", order = 34, name = L["Label"], width = "half",
											desc = L["Show label text on bars."],
											get = function(info) return not MOD.db.global.Defaults.hideLabel end,
											set = function(info, value) MOD.db.global.Defaults.hideLabel = not value; MOD:UpdateAllBarGroups() end,
										},
										HideCountGroup = {
											type = "toggle", order = 35, name = L["Count"], width = "half",
											desc = L["Show stack count in parentheses after label (it is also displayed as overlay on icon)."],
											get = function(info) return not MOD.db.global.Defaults.hideCount end,
											set = function(info, value) MOD.db.global.Defaults.hideCount = not value; MOD:UpdateAllBarGroups() end,
										},
										HideTimerGroup = {
											type = "toggle", order = 36, name = L["Time"], width = "half",
											desc = L["Show time left on bars that have a duration."],
											get = function(info) return not MOD.db.global.Defaults.hideValue end,
											set = function(info, value) MOD.db.global.Defaults.hideValue = not value; MOD:UpdateAllBarGroups() end,
										},
										TooltipsGroup = {
											type = "toggle", order = 37, name = L["Tooltips"], width = "half",
											desc = L["Show tooltips when the cursor is over a bar's icon."],
											get = function(info) return MOD.db.global.Defaults.showTooltips end,
											set = function(info, value) MOD.db.global.Defaults.showTooltips = value; MOD:UpdateAllBarGroups() end,
										},
									},
								},
							},
						},
						IconsGroup = {
							type = "group", order = 10, name = L["Icon Configurations"], inline = true,
							args = {
								BarWidth = {
									type = "range", order = 10, name = L["Bar Width"], min = 5, max = 500, step = 1,
									desc = L["Set width of bars."],
									get = function(info) return MOD.db.global.Defaults.i_barWidth end,
									set = function(info, value) MOD.db.global.Defaults.i_barWidth = value; MOD:UpdateAllBarGroups() end,
								},
								BarHeight = {
									type = "range", order = 15, name = L["Bar Height"], min = 5, max = 100, step = 1,
									desc = L["Set height of bars."],
									get = function(info) return MOD.db.global.Defaults.i_barHeight end,
									set = function(info, value) MOD.db.global.Defaults.i_barHeight = value; MOD:UpdateAllBarGroups() end,
								},
								IconSize = {
									type = "range", order = 20, name = L["Icon Size"], min = 5, max = 100, step = 1,
									desc = L["Set width/height for icons."],
									get = function(info) return MOD.db.global.Defaults.i_iconSize end,
									set = function(info, value) MOD.db.global.Defaults.i_iconSize = value; MOD:UpdateAllBarGroups() end,
								},
								BarScale = {
									type = "range", order = 25, name = L["Scale"], min = 0.1, max = 2, step = 0.05,
									desc = L["Set scale factor for bars and icons."],
									get = function(info) return MOD.db.global.Defaults.i_scale end,
									set = function(info, value) MOD.db.global.Defaults.i_scale = value; MOD:UpdateAllBarGroups() end,
								},
								Space1 = { type = "description", name = "", order = 40 },
								HorizontalSpacing = {
									type = "range", order = 60, name = L["Horizontal Spacing"], min = -100, max = 100, step = 1,
									desc = L["Adjust horizontal spacing between bars."],
									get = function(info) return MOD.db.global.Defaults.i_spacingX end,
									set = function(info, value) MOD.db.global.Defaults.i_spacingX = value; MOD:UpdateAllBarGroups() end,
								},
								VerticalSpacing = {
									type = "range", order = 65, name = L["Vertical Spacing"], min = -100, max = 100, step = 1,
									desc = L["Adjust vertical spacing between bars."],
									get = function(info) return MOD.db.global.Defaults.i_spacingY end,
									set = function(info, value) MOD.db.global.Defaults.i_spacingY = value; MOD:UpdateAllBarGroups() end,
								},
								IconOffsetX = {
									type = "range", order = 70, name = L["Icon Inset"], min = -200, max = 200, step = 1,
									desc = L["Set icon's horizontal inset from bar."],
									get = function(info) return MOD.db.global.Defaults.i_iconOffsetX end,
									set = function(info, value) MOD.db.global.Defaults.i_iconOffsetX = value; MOD:UpdateAllBarGroups() end,
								},
								IconOffsetY = {
									type = "range", order = 75, name = L["Icon Offset"], min = -200, max = 200, step = 1,
									desc = L["Set vertical offset between icon and bar."],
									get = function(info) return MOD.db.global.Defaults.i_iconOffsetY end,
									set = function(info, value) MOD.db.global.Defaults.i_iconOffsetY = value; MOD:UpdateAllBarGroups() end,
								},
								Space2 = { type = "description", name = "", order = 80 },
								BarFormatGroup = {
									type = "group", order = 90, name = "", inline = true,
									args = {
										HideIconGroup = {
											type = "toggle", order = 30, name = L["Icon"], width = "half",
											desc = L["Show icons with bars (required for tooltips)."],
											get = function(info) return not MOD.db.global.Defaults.i_hideIcon end,
											set = function(info, value) MOD.db.global.Defaults.i_hideIcon = not value; MOD:UpdateAllBarGroups() end,
										},
										HideClockGroup = {
											type = "toggle", order = 31, name = L["Clock"], width = "half",
											desc = L["Show clock animation on icons for timer bars."],
											get = function(info) return not MOD.db.global.Defaults.i_hideClock end,
											set = function(info, value) MOD.db.global.Defaults.i_hideClock = not value; MOD:UpdateAllBarGroups() end,
										},
										HideBarGroup = {
											type = "toggle", order = 32, name = L["Bar"], width = "half",
											desc = L["Show colored bar and background."],
											get = function(info) return not MOD.db.global.Defaults.i_hideBar end,
											set = function(info, value) MOD.db.global.Defaults.i_hideBar = not value; MOD:UpdateAllBarGroups() end,
										},
										HideSparkGroup = {
											type = "toggle", order = 33, name = L["Spark"], width = "half",
											desc = L["Show spark that moves across bars to indicate remaining time."],
											disabled = function(info) return MOD.db.global.Defaults.i_hideBar end,
											get = function(info) return not MOD.db.global.Defaults.i_hideSpark end,
											set = function(info, value) MOD.db.global.Defaults.i_hideSpark = not value; MOD:UpdateAllBarGroups() end,
										},
										HideLabelGroup = {
											type = "toggle", order = 34, name = L["Label"], width = "half",
											desc = L["Show label text on bars."],
											get = function(info) return not MOD.db.global.Defaults.i_hideLabel end,
											set = function(info, value) MOD.db.global.Defaults.i_hideLabel = not value; MOD:UpdateAllBarGroups() end,
										},
										HideCountGroup = {
											type = "toggle", order = 35, name = L["Count"], width = "half",
											desc = L["Show stack count in parentheses after label (it is also displayed as overlay on icon)."],
											get = function(info) return not MOD.db.global.Defaults.i_hideCount end,
											set = function(info, value) MOD.db.global.Defaults.i_hideCount = not value; MOD:UpdateAllBarGroups() end,
										},
										HideTimerGroup = {
											type = "toggle", order = 36, name = L["Time"], width = "half",
											desc = L["Show time left on bars that have a duration."],
											get = function(info) return not MOD.db.global.Defaults.i_hideValue end,
											set = function(info, value) MOD.db.global.Defaults.i_hideValue = not value; MOD:UpdateAllBarGroups() end,
										},
										TooltipsGroup = {
											type = "toggle", order = 37, name = L["Tooltips"], width = "half",
											desc = L["Show tooltips when the cursor is over a bar's icon."],
											get = function(info) return MOD.db.global.Defaults.i_showTooltips end,
											set = function(info, value) MOD.db.global.Defaults.i_showTooltips = value; MOD:UpdateAllBarGroups() end,
										},
									},
								},
							},
						},
						ResetDimensions = {
							type = "execute", order = 90, name = L["Reset Format"],
							desc = L["Reset format to default settings."],
							confirm = function(info) return L["RESET FORMAT\nAre you sure you want to reset the format options?"] end,
							func = function(info) MOD:SetDimensionDefaults(MOD.db.global.Defaults); MOD:UpdateAllBarGroups() end,
						},
					},
				},
				FontsGroup = {
					type = "group", order = 20, name = L["Fonts and Textures"], inline = true,
					args = {
						LabelFont = {
							type = "select", order = 21, name = L["Label Font"], 
							desc = L["Select font for label text."],
							dialogControl = 'LSM30_Font',
							values = AceGUIWidgetLSMlists.font,
							get = function(info) return MOD.db.global.Defaults.labelFont end,
							set = function(info, value) MOD.db.global.Defaults.labelFont = value; MOD:UpdateAllBarGroups() end,
						},
						LabelFontSize = {
							type = "range", order = 23, name = L["Font Size"], min = 5, max = 30, step = 1,
							desc = L["Set font size for labels."],
							get = function(info) return MOD.db.global.Defaults.labelFSize end,
							set = function(info, value) MOD.db.global.Defaults.labelFSize = value; MOD:UpdateAllBarGroups() end,
						},
						LabelAlpha = {
							type = "range", order = 25, name = L["Opacity"], min = 0, max = 1, step = 0.05,
							desc = L["Set opacity for labels."],
							get = function(info) return MOD.db.global.Defaults.labelAlpha end,
							set = function(info, value) MOD.db.global.Defaults.labelAlpha = value; MOD:UpdateAllBarGroups() end,
						},
						LabelColor = {
							type = "color", order = 27, name = L["Color"], hasAlpha = false, width = "half",
							get = function(info)
								local t = MOD.db.global.Defaults.labelColor
								if t then return t.r, t.g, t.b, t.a else return 1, 1, 1, 1 end
							end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.Defaults.labelColor
								if not t then MOD.db.global.Defaults.labelColor = { r = r, g = g, b = b, a = a }
									else t.r = r; t.g = g; t.b = b; t.a = a end
								MOD:UpdateAllBarGroups()
							end,
						},
						LabelOutline = {
							type = "toggle", order = 29, name = L["Outline"], width = "half",
							desc = L["Add black outline to label."],
							get = function(info) return MOD.db.global.Defaults.labelOutline end,
							set = function(info, value) MOD.db.global.Defaults.labelOutline = value; MOD:UpdateAllBarGroups() end,
						},
						Space2 = { type = "description", name = "", order = 30 },
						TimeFont = {
							type = "select", order = 31, name = L["Time Font"], 
							desc = L["Select font for time text."],
							dialogControl = 'LSM30_Font',
							values = AceGUIWidgetLSMlists.font,
							get = function(info) return MOD.db.global.Defaults.timeFont end,
							set = function(info, value) MOD.db.global.Defaults.timeFont = value; MOD:UpdateAllBarGroups() end,
						},
						TimeFontSize = {
							type = "range", order = 33, name = L["Font Size"], min = 5, max = 30, step = 1,
							desc = L["Set font size for time text."],
							get = function(info) return MOD.db.global.Defaults.timeFSize end,
							set = function(info, value) MOD.db.global.Defaults.timeFSize = value; MOD:UpdateAllBarGroups() end,
						},
						TimeAlpha = {
							type = "range", order = 35, name = L["Opacity"], min = 0, max = 1, step = 0.05,
							desc = L["Set opacity for time text."],
							get = function(info) return MOD.db.global.Defaults.timeAlpha end,
							set = function(info, value) MOD.db.global.Defaults.timeAlpha = value; MOD:UpdateAllBarGroups() end,
						},
						TimeColor = {
							type = "color", order = 37, name = L["Color"], hasAlpha = false, width = "half",
							get = function(info)
								local t = MOD.db.global.Defaults.timeColor
								if t then return t.r, t.g, t.b, t.a else return 1, 1, 1, 1 end
							end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.Defaults.timeColor
								if not t then MOD.db.global.Defaults.timeColor = { r = r, g = g, b = b, a = a }
									else t.r = r; t.g = g; t.b = b; t.a = a end
								MOD:UpdateAllBarGroups()
							end,
						},
						TimeOutline = {
							type = "toggle", order = 39, name = L["Outline"], width = "half",
							desc = L["Add black outline to time text."],
							get = function(info) return MOD.db.global.Defaults.timeOutline end,
							set = function(info, value) MOD.db.global.Defaults.timeOutline = value; MOD:UpdateAllBarGroups() end,
						},
						Space3 = { type = "description", name = "", order = 40 },
						IconFont = {
							type = "select", order = 41, name = L["Icon Font"], 
							desc = L["Select font for icon text (i.e., spell stack count)."],
							dialogControl = 'LSM30_Font',
							values = AceGUIWidgetLSMlists.font,
							get = function(info) return MOD.db.global.Defaults.iconFont end,
							set = function(info, value) MOD.db.global.Defaults.iconFont = value; MOD:UpdateAllBarGroups() end,
						},
						IconFontSize = {
							type = "range", order = 43, name = L["Font Size"], min = 5, max = 30, step = 1,
							desc = L["Set font size for icon text."],
							get = function(info) return MOD.db.global.Defaults.iconFSize end,
							set = function(info, value) MOD.db.global.Defaults.iconFSize = value; MOD:UpdateAllBarGroups() end,
						},
						IconAlpha = {
							type = "range", order = 45, name = L["Opacity"], min = 0, max = 1, step = 0.05,
							desc = L["Set opacity for icon text."],
							get = function(info) return MOD.db.global.Defaults.iconAlpha end,
							set = function(info, value) MOD.db.global.Defaults.iconAlpha = value; MOD:UpdateAllBarGroups() end,
						},
						IconColor = {
							type = "color", order = 47, name = L["Color"], hasAlpha = false, width = "half",
							get = function(info)
								local t = MOD.db.global.Defaults.iconColor
								if t then return t.r, t.g, t.b, t.a else return 1, 1, 1, 1 end
							end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.Defaults.iconColor
								if not t then MOD.db.global.Defaults.iconColor = { r = r, g = g, b = b, a = a }
									else t.r = r; t.g = g; t.b = b; t.a = a end
								MOD:UpdateAllBarGroups()
							end,
						},
						IconOutline = {
							type = "toggle", order = 49, name = L["Outline"], width = "half",
							desc = L["Add black outline to icon text."],
							get = function(info) return MOD.db.global.Defaults.iconOutline end,
							set = function(info, value) MOD.db.global.Defaults.iconOutline = value; MOD:UpdateAllBarGroups() end,
						},
						Space7 = { type = "description", name = "", order = 70 },
						ForegroundTexture = {
							type = "select", order = 72, name = L["Foreground Texture"], 
							desc = L["Select foreground texture for bars."],
							dialogControl = 'LSM30_Statusbar',
							values = AceGUIWidgetLSMlists.statusbar,
							get = function(info) return MOD.db.global.Defaults.texture end,
							set = function(info, value) MOD.db.global.Defaults.texture = value; MOD:UpdateAllBarGroups() end,
						},
						ForegroundAlpha = {
							type = "range", order = 74, name = L["Foreground Opacity"], min = 0, max = 1, step = 0.05,
							desc = L["Set foreground opacity for bars."],
							get = function(info) return MOD.db.global.Defaults.fgAlpha end,
							set = function(info, value) MOD.db.global.Defaults.fgAlpha = value; MOD:UpdateAllBarGroups() end,
						},
						ForegroundSaturation = {
							type = "range", order = 76, name = L["Foreground Saturation"], min = -1, max = 1, step = 0.05,
							desc = L["Set saturation for foreground colors."],
							get = function(info) return MOD.db.global.Defaults.fgSaturation end,
							set = function(info, value) MOD.db.global.Defaults.fgSaturation = value; MOD:UpdateAllBarGroups() end,
						},
						ForegroundBrightness = {
							type = "range", order = 78, name = L["Foreground Brightness"], min = -1, max = 1, step = 0.05,
							desc = L["Set brightness for foreground colors."],
							get = function(info) return MOD.db.global.Defaults.fgBrightness end,
							set = function(info, value) MOD.db.global.Defaults.fgBrightness = value; MOD:UpdateAllBarGroups() end,
						},
						Space8 = { type = "description", name = "", order = 80 },
						BackgroundTexture = {
							type = "select", order = 82, name = L["Background Texture"], 
							desc = L["Select background texture for bars."],
							dialogControl = 'LSM30_Statusbar',
							values = AceGUIWidgetLSMlists.statusbar,
							get = function(info) return MOD.db.global.Defaults.bgtexture end,
							set = function(info, value) MOD.db.global.Defaults.bgtexture = value; MOD:UpdateAllBarGroups() end,
						},
						BackgroundAlpha = {
							type = "range", order = 84, name = L["Background Opacity"], min = 0, max = 1, step = 0.05,
							desc = L["Set background opacity for bars."],
							get = function(info) return MOD.db.global.Defaults.bgAlpha end,
							set = function(info, value) MOD.db.global.Defaults.bgAlpha = value; MOD:UpdateAllBarGroups() end,
						},
						BackgroundSaturation = {
							type = "range", order = 86, name = L["Background Saturation"], min = -1, max = 1, step = 0.05,
							desc = L["Set saturation for background colors."],
							get = function(info) return MOD.db.global.Defaults.bgSaturation end,
							set = function(info, value) MOD.db.global.Defaults.bgSaturation = value; MOD:UpdateAllBarGroups() end,
						},
						BackgroundBrightness = {
							type = "range", order = 88, name = L["Background Brightness"], min = -1, max = 1, step = 0.05,
							desc = L["Set brightness for background colors."],
							get = function(info) return MOD.db.global.Defaults.bgBrightness end,
							set = function(info, value) MOD.db.global.Defaults.bgBrightness = value; MOD:UpdateAllBarGroups() end,
						},
						Space9 = { type = "description", name = "", order = 90 },
						NormalAlpha = {
							type = "range", order = 92, name = L["Opacity (Not Combat)"], min = 0, max = 1, step = 0.05,
							desc = L["Set opacity for bars when not in combat."],
							get = function(info) return MOD.db.global.Defaults.alpha end,
							set = function(info, value) MOD.db.global.Defaults.alpha = value; MOD:UpdateAllBarGroups() end,
						},
						CombatAlpha = {
							type = "range", order = 94, name = L["Opacity (In Combat)"], min = 0, max = 1, step = 0.05,
							desc = L["Set opacity for bars when in combat."],
							get = function(info) return MOD.db.global.Defaults.combatAlpha end,
							set = function(info, value) MOD.db.global.Defaults.combatAlpha = value; MOD:UpdateAllBarGroups() end,
						},
						IconBorderSaturation = {
							type = "range", order = 95, name = L["Icon Border Saturation"], min = -1, max = 1, step = 0.05,
							desc = L["Set saturation for icon border colors."],
							get = function(info) return MOD.db.global.Defaults.borderSaturation end,
							set = function(info, value) MOD.db.global.Defaults.borderSaturation = value; MOD:UpdateAllBarGroups() end,
						},
						IconBorderBrightness = {
							type = "range", order = 96, name = L["Icon Border Brightness"], min = -1, max = 1, step = 0.05,
							desc = L["Set brightness for icon border colors."],
							get = function(info) return MOD.db.global.Defaults.borderBrightness end,
							set = function(info, value) MOD.db.global.Defaults.borderBrightness = value; MOD:UpdateAllBarGroups() end,
						},
						Space10 = { type = "description", name = "", order = 98 },
						ResetFonts = {
							type = "execute", order = 99, name = L["Reset Fonts/Textures"],
							desc = L["Reset fonts and textures to default settings."],
							confirm = function(info) return L["RESET FONTS/TEXTURES\nAre you sure you want to reset the font and texture options?"] end,
							func = function(info) MOD:SetFontTextureDefaults(MOD.db.global.Defaults); MOD:UpdateAllBarGroups() end,
						},
					},
				},
				ColorsGroup = {
					type = "group", order = 30, name = L["Standard and Spell Colors"], inline = true,
					args = {
						ColorText = { type = "description", name = L["Bar Colors:"], order = 1, width = "half" },
						NotificationColor = {
							type = "color", order = 10, name = L["Notify"], hasAlpha = false, width = "half",
							get = function(info) local t = MOD.db.global.DefaultNotificationColor; return t.r, t.g, t.b, t.a end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.DefaultNotificationColor
								t.r = r; t.g = g; t.b = b; t.a = a
								MOD:UpdateAllBarGroups()
							end,
						},
						BuffColor = {
							type = "color", order = 11, name = L["Buff"], hasAlpha = false, width = "half",
							get = function(info) local t = MOD.db.global.DefaultBuffColor; return t.r, t.g, t.b, t.a end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.DefaultBuffColor
								t.r = r; t.g = g; t.b = b; t.a = a
								MOD:UpdateAllBarGroups()
							end,
						},
						DebuffColor = {
							type = "color", order = 12, name = L["Debuff"], hasAlpha = false, width = "half",
							get = function(info) local t = MOD.db.global.DefaultDebuffColor; return t.r, t.g, t.b, t.a end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.DefaultDebuffColor
								t.r = r; t.g = g; t.b = b; t.a = a
								MOD:UpdateAllBarGroups()
							end,
						},
						CooldownColor = {
							type = "color", order = 13, name = L["Cooldown"], hasAlpha = false,
							get = function(info) local t = MOD.db.global.DefaultCooldownColor; return t.r, t.g, t.b, t.a end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.DefaultCooldownColor
								t.r = r; t.g = g; t.b = b; t.a = a
								MOD:UpdateAllBarGroups()
							end,
						},
						Space0 = { type = "description", name = "", order = 14 },
						DebuffText = { type = "description", name = L["Debuff Colors:"], order = 15, width = "half" },
						PoisonColor = {
							type = "color", order = 15, name = L["Poison"], hasAlpha = false, width = "half",
							get = function(info) local t = MOD.db.global.DefaultPoisonColor; return t.r, t.g, t.b, t.a end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.DefaultPoisonColor
								t.r = r; t.g = g; t.b = b; t.a = a
								MOD:UpdateAllBarGroups()
							end,
						},
						CurseColor = {
							type = "color", order = 16, name = L["Curse"], hasAlpha = false, width = "half",
							get = function(info) local t = MOD.db.global.DefaultCurseColor; return t.r, t.g, t.b, t.a end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.DefaultCurseColor
								t.r = r; t.g = g; t.b = b; t.a = a
								MOD:UpdateAllBarGroups()
							end,
						},
						MagicColor = {
							type = "color", order = 17, name = L["Magic"], hasAlpha = false, width = "half",
							get = function(info) local t = MOD.db.global.DefaultMagicColor; return t.r, t.g, t.b, t.a end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.DefaultMagicColor
								t.r = r; t.g = g; t.b = b; t.a = a
								MOD:UpdateAllBarGroups()
							end,
						},
						DiseaseColor = {
							type = "color", order = 18, name = L["Disease"], hasAlpha = false, width = "half",
							get = function(info) local t = MOD.db.global.DefaultDiseaseColor; return t.r, t.g, t.b, t.a end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.DefaultDiseaseColor
								t.r = r; t.g = g; t.b = b; t.a = a
								MOD:UpdateAllBarGroups()
							end,
						},
						Space1 = { type = "description", name = "", order = 20 },
						SpellName = {
							type = "input", order = 30, name = L["Spell Name"],
							desc = L["Enter spell name (or numeric identifier) whose color you want to change."],
							get = function(info) return selectedSpellName end,
							set = function(info, value)
								local n, t = value, tonumber(value)
								if t then n = GetSpellInfo(t) end
								ValidateSpellName(n)
								selectedSpellName = n
							end,
						},
						StandardColors = {
							type = "select", order = 40, name = L["Color"], width = "half",
							desc = L["Select a standard color or click to set a custom color."],
							disabled = function(info) return not selectedSpellName or (selectedSpellName == "") end,
							get = function(info) return nil end,
							set = function(info, value)
								if MOD:GetIcon(selectedSpellName) then -- make sure this is a valid spell before setting a color in save table
									local t = {}
									t.r, t.g, t.b, t.a = GetStandardColor(value)
									MOD:SetColor(selectedSpellName, t)
									MOD:UpdateAllBarGroups()
								end
							end,
							values = function(info) return GetStandardColorList() end,
							style = "dropdown",
						},
						SpellColor = {
							type = "color", order = 50, name = "", hasAlpha = false, width = "half",
							disabled = function(info) return not selectedSpellName or (selectedSpellName == "") end,
							get = function(info)
								local t = MOD:GetColor(selectedSpellName)
								if not t then return nil end
								return t.r, t.g, t.b, t.a
							end,
							set = function(info, r, g, b, a)
								if MOD:GetIcon(selectedSpellName) then -- make sure this is a valid spell before setting a color in save table
									local t = {}
									t.r = r; t.g = g; t.b = b; t.a = a
									MOD:SetColor(selectedSpellName, t)
									MOD:UpdateAllBarGroups()
								end
							end,
						},
						SpellLabel = {
							type = "input", order = 55, name = L["Spell Label"],
							desc = L["Enter a label to be used by default with this spell."],
							get = function(info) return MOD:GetLabel(selectedSpellName) end,
							set = function(info, value) MOD:SetLabel(selectedSpellName, value) end,
						},
						SpellIcon = {
							type = "description", order = 60, name = "", width = "half", 
							disabled = function(info) return not selectedSpellName or (selectedSpellName == "") end,
							image = function(info) return MOD:GetIcon(selectedSpellName) end,
							imageWidth = 24, imageHeight = 24,
						},
						ResetSpellColor = {
							type = "execute", order = 70, name = L["Reset"], width = "half",
							desc = function(info) return L["Reset color and label string"](selectedSpellName) end,
							disabled = function(info) return not selectedSpellName or (selectedSpellName == "") end,
							confirm = function(info) return L["Reset color and label confirm"](selectedSpellName) end,
							func = function(info)
								if MOD:GetIcon(selectedSpellName) then -- make sure this is a valid spell name before changing the color
									MOD:ResetColorDefault(selectedSpellName)
									MOD:SetLabel(selectedSpellName, selectedSpellName)
									MOD:UpdateAllBarGroups()
								end
							end,
						},
						Space2 = { type = "description", name = "", order = 80 },
						ResetBarColors = {
							type = "execute", order = 85, name = L["Reset Bar Colors"],
							desc = L["Reset bar colors back to default."],
							confirm = function(info) return (L["RESET BAR COLORS\nAre you sure you want to reset bar colors back to default?"]) end,
							func = function(info)
								MOD.db.global.DefaultBuffColor = Raven_HexColor("8ae234") -- Green1
								MOD.db.global.DefaultDebuffColor = Raven_HexColor("fcaf3e") -- Orange1
								MOD.db.global.DefaultCooldownColor = Raven_HexColor("fce94f") -- Yellow1
								MOD.db.global.DefaultNotificationColor = Raven_HexColor("729fcf") -- Blue1								
							end,
						},
						ResetDebuffColors = {
							type = "execute", order = 86, name = L["Reset Debuff Colors"],
							desc = L["Reset debuff colors back to default."],
							confirm = function(info) return (L["RESET DEBUFF COLORS\nAre you sure you want to reset debuff colors back to default?"]) end,
							func = function(info)
								MOD.db.global.DefaultPoisonColor = Raven_CopyColor(DebuffTypeColor["Poison"])
								MOD.db.global.DefaultCurseColor = Raven_CopyColor(DebuffTypeColor["Curse"])
								MOD.db.global.DefaultMagicColor = Raven_CopyColor(DebuffTypeColor["Magic"])
								MOD.db.global.DefaultDiseaseColor = Raven_CopyColor(DebuffTypeColor["Disease"])						
							end,
						},
						ResetSpellColors = {
							type = "execute", order = 87, name = L["Reset Spell Colors"],
							desc = L["Reset spell colors back to defaults."],
							confirm = function(info) return L["RESET SPELL COLORS\nAre you sure you want to reset all spell colors back to defaults?"] end,
							func = function(info) MOD:ResetColorDefaults(); MOD:UpdateAllBarGroups() end,
						},
					},
				},
				ButtonFacadeGroup = {
					type = "group", order = 40, name = L["Button Facade"], inline = true,
					hidden = function(info) return not LBF end,
					args = {
						NormalTexture = {
							type = "toggle", order = 10, name = L["Color Normal Texture"],
							desc = L["If checked, icon border color is applied to the normal texture."],
							get = function(info) return MOD.db.global.ButtonFacadeNormal end,
							set = function(info, value) MOD.db.global.ButtonFacadeNormal = value; MOD:UpdateAllBarGroups() end,
						},
						BorderTexture = {
							type = "toggle", order = 20, name = L["Color Border Texture"],
							desc = L["If checked, icon border color is applied to the border texture."],
							get = function(info) return MOD.db.global.ButtonFacadeBorder end,
							set = function(info, value) MOD.db.global.ButtonFacadeBorder = value; MOD:UpdateAllBarGroups() end,
						},
					},
				},
			},
		},
		BarGroups = {
			type = "group", order = 20, name = L["Bar Groups"], childGroups = "tab",
			disabled = function(info) return InMode("Not") end,
			args = {
				SelectBarGroup = {
					type = "select", order = 1, name = L["Bar Group"],
					get = function(info) UpdateBarList(); return GetSelectedBarGroup() end,
					set = function(info, value) SetSelectedBarGroup(value) end,
					disabled = function(info) return NoBarGroup() or InMode("Bar") end,
					values = function(info) return GetBarGroupList() end,
					style = "dropdown",
				},
				Space1 = { type = "description", name = "", order = 2, width = "half" },
				NewBarGroupButton = {
					type = "execute", order = 3, name = L["New Custom Group"],
					desc = L["Create a new bar group with manually added bars."],
					disabled = function(info) return InMode("Bar") end,
					hidden = function(info) return enterNewBarGroup end,
					func = function(info) enterNewBarGroup, toggleNewBarGroup, typeNewBarGroup = true, true, false end,
				},
				NewAutoBarGroupButton = {
					type = "execute", order = 4, name = L["New Auto Group"],
					desc = L["Create a new bar group with automatically displayed bars."],
					disabled = function(info) return InMode("Bar") end,
					hidden = function(info) return enterNewBarGroup end,
					func = function(info) enterNewBarGroup, toggleNewBarGroup, typeNewBarGroup = true, true, true end,
				},
				NewCustomBarGroupName = {
					type = "input", order = 5, name = L["Enter Custom Group Name"],
					desc = L["Enter name of new custom bar group."],
					hidden = function(info) return not enterNewBarGroup or typeNewBarGroup end,
					validate = function(info, n) if not n or (n == "") then return L["Invalid name."] else return true end end,
					confirm = function(info, value) return ConfirmNewBarGroup(value) end,
					get = function(info)
						enterNewBarGroup = toggleNewBarGroup
						enterNewBarGroupType = false
						if toggleNewBarGroup then toggleNewBarGroup = false end
						if not enterNewBarGroup then MOD:UpdateOptions() end
						return false
					end,
					set = function(info, value) enterNewBarGroup = false; CreateBarGroup(value, false, false, true, 0, 0) end,
				},
				NewAutoBarGroupName = {
					type = "input", order = 6, name = L["Enter Auto Group Name"],
					desc = L["Enter name of new auto bar group."],
					hidden = function(info) return not enterNewBarGroup or not typeNewBarGroup end,
					validate = function(info, n) if not n or (n == "") then return L["Invalid name."] else return true end end,
					confirm = function(info, value) return ConfirmNewBarGroup(value) end,
					get = function(info)
						enterNewBarGroup = toggleNewBarGroup
						enterNewBarGroupType = false
						if toggleNewBarGroup then toggleNewBarGroup = false end
						if not enterNewBarGroup then MOD:UpdateOptions() end
						return false
					end,
					set = function(info, value) enterNewBarGroup = false; CreateBarGroup(value, true, false, true, 0, 0) end,
				},
				CancelNewBarGroup = {
					type = "execute", order = 7, name = L["Cancel"], width = "half",
					desc = L["Cancel creating a new bar group."],
					hidden = function(info) return not enterNewBarGroup end,
					func = function(info) enterNewBarGroup, toggleNewBarGroup = false, false end,
				},
				DeleteBarGroup = {
					type = "execute", order = 8, name = L["Delete Bar Group"],
					desc = L["Delete the selected bar group."],
					disabled = function(info) return NoBarGroup() or InMode("Bar") end,
					hidden = function(info) return enterNewBarGroup end,
					func = function(info) DeleteBarGroup() end,
					confirm = function(info) return L["Delete bar group string"](GetBarGroupField("name")) end,
				},
				GeneralTab = {
					type = "group", order = 10, name = L["General"],
					disabled = function(info) return InMode("Bar") end,
					hidden = function(info) return NoBarGroup() end,
					args = {
						SettingsGroup = {
							type = "group", order = 1, name = L["Settings"], inline = true,
							args = {
								EnableBarGroup = {
									type = "toggle", order = 10, name = L["Enable Bar Group"],
									desc = L["Enable bar group string"],
									get = function(info) return GetBarGroupField("enabled") end,
									set = function(info, value) SetBarGroupField("enabled", value) end,
								},
								EnableMerge = {
									type = "toggle", order = 20, name = L["Merge Bar Group"],
									desc = L["Merge bar group string"],
									get = function(info) return GetBarGroupField("merged") end,
									set = function(info, value) SetBarGroupField("merged", value) end,
								},
								LinkSettings = {
									type = "toggle", order = 30, name = L["Link Settings"],
									desc = L["Link settings string"],
									get = function(info) return GetBarGroupField("linkSettings") end,
									set = function(info, value) SetBarGroupField("linkSettings", value) end,
								},
								SaveSettings = {
									type = "execute", order = 40, name = L["Save Settings"],
									hidden = function(info) return GetBarGroupField("linkSettings") end,
									desc = L["Click to save current settings to linked settings."],
									func = function(info) MOD:SaveBarGroupSettings(GetBarGroupEntry()) end,
								},
								spacer = { type = "description", name = "", order = 50, hidden = function(info) return not GetBarGroupField("merged") end },
								MergeBarGroup = {
									type = "select", order = 60, name = L["Bar Group To Merge Into"],
									desc = L["Select a bar group to merge into."],
									hidden = function(info) return not GetBarGroupField("merged") end,
									get = function(info) return GetMergeBarGroup() end,
									set = function(info, value) SetMergeBarGroup(value) end,
									values = function(info) return GetBarGroupList() end,
									style = "dropdown",
								},
							},
						},
						SortingGroup = {
							type = "group", order = 10, name = L["Sort Order"], inline = true,
							hidden = function(info) return GetBarGroupField("merged") end,
							args = {
								AtoZOrder = {
									type = "toggle", order = 10, name = L["A to Z"], width = "half", 
									desc = L["If checked, sort in ascending alphabetical order starting at bar closest to the anchor."],
									get = function(info) return GetBarGroupField("sor") == "A" end,
									set = function(info, value) SetBarGroupField("sor", "A") end,
								},
								TimeLeftOrder = {
									type = "toggle", order = 20, name = L["Time Left"], width = "half", 
									desc = L["If checked, sort by time left in ascending order starting at bar closest to the anchor."],
									get = function(info) return GetBarGroupField("sor") == "T" end,
									set = function(info, value) SetBarGroupField("sor", "T") end,
								},
								DurationOrder = {
									type = "toggle", order = 30, name = L["Duration"], width = "half", 
									desc = L["If checked, sort by overall duration in ascending order starting at bar closest to the anchor."],
									get = function(info) return GetBarGroupField("sor") == "D" end,
									set = function(info, value) SetBarGroupField("sor", "D") end,
								},
								StartOrder = {
									type = "toggle", order = 35, name = L["Creation"], width = "half", 
									desc = L["If checked, show bars in order created with oldest bar closest to the anchor."],
									get = function(info) return GetBarGroupField("sor") == "S" end,
									set = function(info, value) SetBarGroupField("sor", "S") end,
								},
								ClassOrder = {
									type = "toggle", order = 40, name = L["Class"], width = "half", 
									desc = L["If checked, sort by associated class name, if any, in ascending alphabetical order starting at bar closest to the anchor."],
									get = function(info) return GetBarGroupField("sor") == "C" end,
									set = function(info, value) SetBarGroupField("sor", "C") end,
								},
								CustomOrder = {
									type = "toggle", order = 50, name = L["Custom"], width = "half",
									desc = L["If checked, allow manually setting the order of bars."],
									hidden = function(info) return GetBarGroupField("auto") end,
									get = function(info) return GetBarGroupField("sor") == "X" end,
									set = function(info, value) SetBarGroupField("sor", "X") end,
								},
								ReverseSortOrder = {
									type = "toggle", order = 70, name = L["Reverse Order"], 
									desc = L['If checked, reverse the sort order (e.g., "A to Z" becomes "Z to A").'],
									get = function(info) return GetBarGroupField("reverseSort") end,
									set = function(info, value) SetBarGroupField("reverseSort", value) end,
								},
							},
						},
						ShowWhenGroup = {
							type = "group", order = 20, name = L["Show When"], inline = true,
							args = {
								InCombatGroup = {
									type = "toggle", order = 20, name = L["In Combat"],
									desc = L["If checked, bar group is shown when the player is in combat."],
									get = function(info) return GetBarGroupField("showCombat") end,
									set = function(info, value) SetBarGroupField("showCombat", value) end,
								},
								OutOfCombatGroup = {
									type = "toggle", order = 21, name = L["Out Of Combat"],
									desc = L["If checked, bar group is shown when the player is out of combat."],
									get = function(info) return GetBarGroupField("showOOC") end,
									set = function(info, value) SetBarGroupField("showOOC", value) end,
								},
								RestingGroup = {
									type = "toggle", order = 22, name = L["Resting"],
									desc = L["If checked, bar group is shown when the player is resting."],
									get = function(info) return GetBarGroupField("showResting") end,
									set = function(info, value) SetBarGroupField("showResting", value) end,
								},
								MountedGroup = {
									type = "toggle", order = 23, name = L["Mounted"],
									desc = L["If checked, bar group is shown when the player is mounted."],
									get = function(info) return GetBarGroupField("showMounted") end,
									set = function(info, value) SetBarGroupField("showMounted", value) end,
								},
								SoloGroup = {
									type = "toggle", order = 30, name = L["Solo"],
									desc = L["If checked, bar group is shown when the player is not in a party or raid."],
									get = function(info) return GetBarGroupField("showSolo") end,
									set = function(info, value) SetBarGroupField("showSolo", value) end,
								},
								PartyGroup = {
									type = "toggle", order = 31, name = L["In Party"],
									desc = L["If checked, bar group is shown when the player is in a party."],
									get = function(info) return GetBarGroupField("showParty") end,
									set = function(info, value) SetBarGroupField("showParty", value) end,
								},
								RaidGroup = {
									type = "toggle", order = 32, name = L["In Raid"],
									desc = L["If checked, bar group is shown when the player is in a raid."],
									get = function(info) return GetBarGroupField("showRaid") end,
									set = function(info, value) SetBarGroupField("showRaid", value) end,
								},
								FishingGroup = {
									type = "toggle", order = 33, name = L["Fishing"],
									desc = L["If checked, bar group is shown when the player is ready to fish (i.e., trained with fishing pole equipped)."],
									get = function(info) return GetBarGroupField("showFishing") end,
									set = function(info, value) SetBarGroupField("showFishing", value) end,
								},
								spacer = { type = "description", name = "", order = 40, },
								ShowIfBlizzard = {
									type = "toggle", order = 45, name = L["Blizzard Buffs Enabled"],
									desc = L["If checked, the bar group is shown if the default user interface for buffs is enabled."],
									get = function(info) return GetBarGroupField("showBlizz") end,
									set = function(info, value) SetBarGroupField("showBlizz", value) end,
								},
								ShowNotBlizzard = {
									type = "toggle", order = 46, name = L["Blizzard Buffs Disabled"],
									desc = L["If checked, the bar group is shown if the default user interface for buffs is disabled."],
									get = function(info) return GetBarGroupField("showNotBlizz") end,
									set = function(info, value) SetBarGroupField("showNotBlizz", value) end,
								},
								CheckCondition = {
									type = "toggle", order = 60, name = L["Condition Is True"],
									desc = L["If checked, bar group is shown only when the selected condition is true."],
									get = function(info) return GetBarGroupField("checkCondition") end,
									set = function(info, value) if not value then SetBarGroupField("condition", nil) end; SetBarGroupField("checkCondition", value) end,
								},
								SelectCondition = {
									type = "select", order = 65, name = L["Condition"],
									disabled = function(info) return not GetBarGroupField("checkCondition") end,
									get = function(info) return GetBarGroupSelectedCondition(GetSelectConditionList()) end,
									set = function(info, value) SetBarGroupField("condition", GetSelectConditionList()[value]) end,
									values = function(info) return GetSelectConditionList() end,
									style = "dropdown",
								},
								spacer = { type = "description", name = "", order = 70, },
								SelectClass = {
									type = "group", order = 75, name = L["Player Class"], inline = true,
									args = {
										Druid = {
											type = "toggle", order = 10, name = L["Druid"], width = "half",
											get = function(info) local t = GetBarGroupField("showClasses"); return not t or not t.DRUID end,
											set = function(info, value)
												local t = GetBarGroupField("showClasses")
												if not t then SetBarGroupField("showClasses", { DRUID = not value } ) else t.DRUID = not value end
											end
										},
										Hunter = {
											type = "toggle", order = 15, name = L["Hunter"], width = "half",
											get = function(info) local t = GetBarGroupField("showClasses"); return not t or not t.HUNTER end,
											set = function(info, value)
												local t = GetBarGroupField("showClasses")
												if not t then SetBarGroupField("showClasses", { HUNTER = not value } ) else t.HUNTER = not value end
											end
										},
										Mage = {
											type = "toggle", order = 20, name = L["Mage"], width = "half",
											get = function(info) local t = GetBarGroupField("showClasses"); return not t or not t.MAGE end,
											set = function(info, value)
												local t = GetBarGroupField("showClasses")
												if not t then SetBarGroupField("showClasses", { MAGE = not value } ) else t.MAGE = not value end
											end
										},
										Paladin = {
											type = "toggle", order = 25, name = L["Paladin"], width = "half",
											get = function(info) local t = GetBarGroupField("showClasses"); return not t or not t.PALADIN end,
											set = function(info, value)
												local t = GetBarGroupField("showClasses")
												if not t then SetBarGroupField("showClasses", { PALADIN = not value } ) else t.PALADIN = not value end
											end
										},
										Priest = {
											type = "toggle", order = 30, name = L["Priest"], width = "half",
											get = function(info) local t = GetBarGroupField("showClasses"); return not t or not t.PRIEST end,
											set = function(info, value)
												local t = GetBarGroupField("showClasses")
												if not t then SetBarGroupField("showClasses", { PRIEST = not value } ) else t.PRIEST = not value end
											end
										},
										Rogue = {
											type = "toggle", order = 35, name = L["Rogue"], width = "half",
											get = function(info) local t = GetBarGroupField("showClasses"); return not t or not t.ROGUE end,
											set = function(info, value)
												local t = GetBarGroupField("showClasses")
												if not t then SetBarGroupField("showClasses", { ROGUE = not value } ) else t.ROGUE = not value end
											end
										},
										Shaman = {
											type = "toggle", order = 40, name = L["Shaman"], width = "half",
											get = function(info) local t = GetBarGroupField("showClasses"); return not t or not t.SHAMAN end,
											set = function(info, value)
												local t = GetBarGroupField("showClasses")
												if not t then SetBarGroupField("showClasses", { SHAMAN = not value } ) else t.SHAMAN = not value end
											end
										},
										Warlock = {
											type = "toggle", order = 45, name = L["Warlock"], width = "half",
											get = function(info) local t = GetBarGroupField("showClasses"); return not t or not t.WARLOCK end,
											set = function(info, value)
												local t = GetBarGroupField("showClasses")
												if not t then SetBarGroupField("showClasses", { WARLOCK = not value } ) else t.WARLOCK = not value end
											end
										},
										Warrior = {
											type = "toggle", order = 50, name = L["Warrior"], width = "half",
											get = function(info) local t = GetBarGroupField("showClasses"); return not t or not t.WARRIOR end,
											set = function(info, value)
												local t = GetBarGroupField("showClasses")
												if not t then SetBarGroupField("showClasses", { WARRIOR = not value } ) else t.WARRIOR = not value end
											end
										},
										DeathKnight = {
											type = "toggle", order = 55, name = L["Death Knight"],
											get = function(info) local t = GetBarGroupField("showClasses"); return not t or not t.DEATHKNIGHT end,
											set = function(info, value)
												local t = GetBarGroupField("showClasses")
												if not t then SetBarGroupField("showClasses", { DEATHKNIGHT = not value } ) else t.DEATHKNIGHT = not value end
											end
										},
									},
								},
							},
						},
						EffectsGroup = {
							type = "group", order = 30, name = L["Special Effects"],  inline = true,
							args = {
								PulseStart = {
									type = "toggle", order = 10, name = L["Pulse At Start"],
									desc = L["Enable icon pulse when bar is started."],
									get = function(info) return GetBarGroupField("pulseStart") end,
									set = function(info, value) SetBarGroupField("pulseStart", value) end,
								},
								PulseEnd = {
									type = "toggle", order = 15, name = L["Pulse When Expiring"],
									desc = L["Enable icon pulse when bar is expiring."],
									get = function(info) return GetBarGroupField("pulseEnd") end,
									set = function(info, value) SetBarGroupField("pulseEnd", value) end,
								},
								FlashExpiring = {
									type = "toggle", order = 20, name = L["Flash When Expiring"],
									desc = L["Enable flashing of expiring bars."],
									get = function(info) return GetBarGroupField("flashExpiring") end,
									set = function(info, value) SetBarGroupField("flashExpiring", value) end,
								},
								FlashTime = {
									type = "range", order = 25, name = L["Flash Time"], min = 1, max = 300, step = 1,
									desc = L["Set number of seconds before expiration that bar should start flashing."],
									disabled = function(info) return not GetBarGroupField("flashExpiring") end,
									get = function(info) return GetBarGroupField("flashTime") end,
									set = function(info, value) SetBarGroupField("flashTime", value) end,
								},
								space = { type = "description", name = "", order = 30 },
								HideEnable = {
									type = "toggle", order = 35, name = L["Hide After Start"],
									desc = L["Enable hide after start for timer bars (does not hide bars with unlimited duration)."],
									get = function(info) return GetBarGroupField("hide") end,
									set = function(info, value) SetBarGroupField("hide", value) end,
								},
								FadeEnable = {
									type = "toggle", order = 40, name = L["Fade After Start"],
									desc = L["Enable fade after start (i.e., switch to fade opacity after a delay)."],
									disabled = function(info) return GetBarGroupField("hide") end,
									get = function(info) return GetBarGroupField("fade") end,
									set = function(info, value) SetBarGroupField("fade", value) end,
								},
								FadeDelay = {
									type = "range", order = 45, name = L["Delay"], min = 1, max = 300, step = 1,
									desc = L["Set number of seconds before bar will hide or fade."],
									disabled = function(info) return not GetBarGroupField("fade") and not GetBarGroupField("hide") end,
									get = function(info) return GetBarGroupField("delayTime") or 5 end,
									set = function(info, value) SetBarGroupField("delayTime", value) end,
								},
								FadeAlpha = {
									type = "range", order = 50, name = L["Fade Opacity"], min = 0, max = 1, step = 0.05,
									desc = L["Set opacity for faded bars."],
									disabled = function(info) return GetBarGroupField("hide") or not GetBarGroupField("fade") end,
									get = function(info) return GetBarGroupField("fadeAlpha") end,
									set = function(info, value) SetBarGroupField("fadeAlpha", value) end,
								},
							},
						},
						OptionsGroup = {
							type = "group", order = 40, name = L["Miscellaneous Options"], inline = true,
							hidden = function(info) return GetBarGroupField("merged") end,
							args = {
								Rename = {
									type = "input", order = 10, name = L["Rename Bar Group"],
									validate = function(info, n) if not n or (n == "") then return L["Invalid name."] else return true end end,
									confirm = function(info, value) return ConfirmNewBarGroup(value) end,
									desc = L["Enter new name for the bar group."],
									get = function(info) return GetBarGroupField("name") end,
									set = function(info, value) RenameBarGroup(value) end,
								},
								Space0 = { type = "description", name = "", order = 11, width = "half" },
								FrameStrata = {
									type = "select", order = 12, name = L["Frame Strata"],
									desc = L["Frame strata string"],
									get = function(info) return GetBarGroupField("strata") end,
									set = function(info, value) SetBarGroupField("strata", value) end,
									values = function(info) return stratas end,
									style = "dropdown",
								},
								Space1 = { type = "description", name = "", order = 15, width = "half" },
								LockBars = {
									type = "execute", order = 20, name = L["Lock"], width = "half",
									desc = L["Lock and hide the anchor for the bar group."],
									func = function(info) SetBarGroupField("locked", true) end,
								},
								UnlockBars = {
									type = "execute", order = 25, name = L["Unlock"], width = "half",
									desc = L["Unlock and show the anchor for the bar group."],
									func = function(info) SetBarGroupField("locked", false) end,
								},
								Space2 = { type = "description", name = "", order = 30 },
								NoMouse = {
									type = "toggle", order = 35, name = L["Non-Interactive"],
									desc = L["If checked, the bar group is non-interactive and doesn't show tooltips or respond to clicks."],
									get = function(info) return GetBarGroupField("noMouse") end,
									set = function(info, value) SetBarGroupField("noMouse", value) end,
								},
								BarOrIcon = {
									type = "toggle", order = 40, name = L["Only Icons Interact"],
									desc = L["If checked, only icons show tooltips and respond to clicks, otherwise entire bar does."],
									disabled = function(info) return GetBarGroupField("noMouse") end,
									get = function(info) return GetBarGroupField("iconMouse") end,
									set = function(info, value) SetBarGroupField("iconMouse", value) end,
								},
								CombatTooltips = {
									type = "toggle", order = 45, name = L["Combat Tooltips"],
									desc = L["If checked, tooltips are shown during combat."],
									disabled = function(info) return GetBarGroupField("noMouse") end,
									get = function(info) return GetBarGroupField("combatTips") end,
									set = function(info, value) SetBarGroupField("combatTips", value) end,
								},
								TooltipAnchor = {
									type = "select", order = 50, name = L["Tooltip Anchor"],
									desc = L["Tooltip anchor string"],
									disabled = function(info) return GetBarGroupField("noMouse") or GetBarGroupField("autoTips") end,
									get = function(info) return GetBarGroupField("anchorTips") end,
									set = function(info, value) SetBarGroupField("anchorTips", value) end,
									values = function(info) return anchorPoints end,
									style = "dropdown",
								},
							},
						},
					},
				},
				BarTab = {
					type = "group", order = 15, name = L["Custom Bars"],
					hidden = function(info) return NoBarGroup() or GetBarGroupField("auto") end,
					args = {
						NewBarButton = {
							type = "execute", order = 1, name = L["New"], width = "half",
							desc = L["Create a new bar."],
							disabled = function(info) return InMode("Bar") end,
							func = function(info) EnterNewBar("start") end,
						},
						DeleteBar = {
							type = "execute", order = 2, name = L["Delete"], width = "half",
							desc = L["Delete the selected bar."],
							disabled = function(info) return NoBar() end,
							func = function(info) DeleteBar() end,
							confirm = function(info) return L['DELETE BAR\nAre you sure you want to delete the selected bar?'] end,
						},
						-- Bars get plugged in here, with order starting at 10
					},
				},
				DetectBuffsTab = {
					type = "group", order = 20, name = L["Buffs"],
					disabled = function(info) return InMode("Bar") end,
					hidden = function(info) return NoBarGroup() or not GetBarGroupField("auto") end,
					args = {
						EnableGroup = {
							type = "group", order = 1, name = L["Enable"], inline = true,
							args = {
								DetectEnable = {
									type = "toggle", order = 1, name = L["Auto Buffs"],
									desc = L['Enable automatically displaying bars for buffs that match these settings.'],
									get = function(info) return GetBarGroupField("detectBuffs") end,
									set = function(info, value) SetBarGroupField("detectBuffs", value) end,
								},
								Tracking = {
									type = "toggle", order = 10, name = L["Include Tracking"],
									disabled = function(info) return not GetBarGroupField("detectBuffs") end,
									desc = L['Include the active tracking types as buffs.'],
									get = function(info) return GetBarGroupField("detectTracking") end,
									set = function(info, value) SetBarGroupField("detectTracking", value) end,
								},
								OnlyTracking = {
									type = "toggle", order = 20, name = L["Show Only Tracking"],
									disabled = function(info) return not (GetBarGroupField("detectBuffs") and GetBarGroupField("detectTracking")) end,
									desc = L['Include only the active tracking types.'],
									get = function(info) return GetBarGroupField("detectOnlyTracking") end,
									set = function(info, value) SetBarGroupField("detectOnlyTracking", value) end,
								},
							},
						},
						MonitorUnitGroup = {
							type = "group", order = 10, name = L["Action On"], inline = true,
							disabled = function(info) return not GetBarGroupField("detectBuffs") end,
							args = {
								PlayerBuff = {
									type = "toggle", order = 10, name = L["Player"],
									desc = L["If checked, only add bars for buffs if they are on the player."],
									get = function(info) return GetBarGroupField("detectBuffsMonitor") == "player" end,
									set = function(info, value) SetBarGroupField("detectBuffsMonitor", "player") end,
								},
								PetBuff = {
									type = "toggle", order = 15, name = L["Pet"],
									desc = L["If checked, only add bars for buffs if they are on the player's pet."],
									get = function(info) return GetBarGroupField("detectBuffsMonitor") == "pet" end,
									set = function(info, value) SetBarGroupField("detectBuffsMonitor", "pet") end,
								},
								TargetBuff = {
									type = "toggle", order = 20, name = L["Target"],
									desc = L["If checked, only add bars for buffs if they are on the target."],
									get = function(info) return GetBarGroupField("detectBuffsMonitor") == "target" end,
									set = function(info, value) SetBarGroupField("detectBuffsMonitor", "target") end,
								},
								FocusBuff = {
									type = "toggle", order = 30, name = L["Focus"],
									desc = L["If checked, only add bars for buffs if they are on the focus."],
									get = function(info) return GetBarGroupField("detectBuffsMonitor") == "focus" end,
									set = function(info, value) SetBarGroupField("detectBuffsMonitor", "focus") end,
								},
								Space1 = { type = "description", name = "", order = 35 },
								MouseoverBuff = {
									type = "toggle", order = 40, name = L["Mouseover"],
									desc = L["If checked, only add bars for buffs if they are on the mouseover unit."],
									get = function(info) return GetBarGroupField("detectBuffsMonitor") == "mouseover" end,
									set = function(info, value) SetBarGroupField("detectBuffsMonitor", "mouseover") end,
								},
								PetTargetBuff = {
									type = "toggle", order = 45, name = L["Pet's Target"],
									desc = L["If checked, only add bars for buffs if they are on the pet's target."],
									get = function(info) return GetBarGroupField("detectBuffsMonitor") == "pettarget" end,
									set = function(info, value) SetBarGroupField("detectBuffsMonitor", "pettarget") end,
								},
								TargetTargetBuff = {
									type = "toggle", order = 50, name = L["Target's Target"],
									desc = L["If checked, only add bars for buffs if they are on the target's target."],
									get = function(info) return GetBarGroupField("detectBuffsMonitor") == "targettarget" end,
									set = function(info, value) SetBarGroupField("detectBuffsMonitor", "targettarget") end,
								},
								FocusTargetBuff = {
									type = "toggle", order = 60, name = L["Focus's Target"],
									desc = L["If checked, only add bars for buffs if they are on the focus's target."],
									get = function(info) return GetBarGroupField("detectBuffsMonitor") == "focustarget" end,
									set = function(info, value) SetBarGroupField("detectBuffsMonitor", "focustarget") end,
								},
							},
						},
						CastUnitGroup = {
							type = "group", order = 20, name = L["Cast By"], inline = true, width = "full",
							disabled = function(info) return not GetBarGroupField("detectBuffs") end,
							args = {
								MyBuff = {
									type = "toggle", order = 10, name = L["Player"],
									desc = L["If checked, only add bars for buffs if cast by the player."],
									get = function(info) return GetBarGroupField("detectBuffsCastBy") == "player" end,
									set = function(info, value) SetBarGroupField("detectBuffsCastBy", "player") end,
								},
								PetBuff = {
									type = "toggle", order = 15, name = L["Pet"],
									desc = L["If checked, only add bars for buffs if cast by the player's pet."],
									get = function(info) return GetBarGroupField("detectBuffsCastBy") == "pet" end,
									set = function(info, value) SetBarGroupField("detectBuffsCastBy", "pet") end,
								},
								YourBuff = {
									type = "toggle", order = 20, name = L["Other"],
									desc = L["If checked, only add bars for buffs if cast by anyone other than the player or pet."],
									get = function(info) return GetBarGroupField("detectBuffsCastBy") == "other" end,
									set = function(info, value) SetBarGroupField("detectBuffsCastBy", "other") end,
								},
								OurBuff = {
									type = "toggle", order = 30, name = L["Anyone"],
									desc = L["If checked, add bars for buffs if cast by anyone, including player."],
									get = function(info) return GetBarGroupField("detectBuffsCastBy") == "anyone" end,
									set = function(info, value) SetBarGroupField("detectBuffsCastBy", "anyone") end,
								},
							},
						},
						FilterByType = {
							type = "group", order = 30, name = L["Filter By Type"], inline = true, width = "full",
							disabled = function(info) return not GetBarGroupField("detectBuffs") or
											(GetBarGroupField("detectTracking") and GetBarGroupField("detectOnlyTracking")) end,
							args = {
								Enable = {
									type = "toggle", order = 10, name = L["Enable"],
									desc = L["Filter buff types string"],
									get = function(info) return GetBarGroupField("filterBuffTypes") end,
									set = function(info, v) SetBarGroupField("filterBuffTypes", v) end,
								},
								Stealable = {
									type = "toggle", order = 20, name = L["Stealable"],
									disabled = function(info) return not GetBarGroupField("filterBuffTypes") or not GetBarGroupField("detectBuffs") or
											(GetBarGroupField("detectTracking") and GetBarGroupField("detectOnlyTracking")) end,
									desc = L['Include buffs that mages can spellsteal.'],
									get = function(info) return GetBarGroupField("detectStealable") end,
									set = function(info, value) SetBarGroupField("detectStealable", value) end,
								},
								Magic = {
									type = "toggle", order = 30, name = L["Magic"],
									disabled = function(info) return not GetBarGroupField("filterBuffTypes") or not GetBarGroupField("detectBuffs") or
											(GetBarGroupField("detectTracking") and GetBarGroupField("detectOnlyTracking")) end,
									desc = L['Include magic buffs (these can usually be removed with abilities like Purge).'],
									get = function(info) return GetBarGroupField("detectMagicBuffs") end,
									set = function(info, value) SetBarGroupField("detectMagicBuffs", value) end,
								},
								Other = {
									type = "toggle", order = 40, name = L["Other"],
									disabled = function(info) return not GetBarGroupField("filterBuffTypes") or not GetBarGroupField("detectBuffs") or
											(GetBarGroupField("detectTracking") and GetBarGroupField("detectOnlyTracking")) end,
									desc = L['Include other buffs not selected with filter types.'],
									get = function(info) return GetBarGroupField("detectOtherBuffs") end,
									set = function(info, value) SetBarGroupField("detectOtherBuffs", value) end,
								},
							},
						},
						FilterGroup = {
							type = "group", order = 40, name = L["Filter List"], inline = true, width = "full",
							disabled = function(info) return not GetBarGroupField("detectBuffs") end,
							args = {
								BlackList = {
									type = "toggle", order = 10, name = L["Black List"],
									desc = L["If checked, don't display any buffs that are in the filter list."],
									get = function(info) return GetBarGroupField("filterBuff") end,
									set = function(info, v) SetBarGroupField("filterBuff", v); if v then SetBarGroupField("showBuff", false) end end,
								},
								WhiteList = {
									type = "toggle", order = 11, name = L["White List"],
									desc = L["If checked, only display buffs that are in the filter list."],
									get = function(info) return GetBarGroupField("showBuff") end,
									set = function(info, v) SetBarGroupField("showBuff", v); if v then SetBarGroupField("filterBuff", false) end  end,
								},
								LinkFilters = {
									type = "toggle", order = 15, name = L["Link Filter"],
									desc = L["If checked, the buff filter list is shared with bar groups in other profiles with the same name."],
									disabled = function(info) return not GetBarGroupField("detectBuffs") or not (GetBarGroupField("filterBuff") or GetBarGroupField("showBuff")) end,
									get = function(info) return GetBarGroupField("filterBuffLink") end,
									set = function(info, value) SetBarGroupField("filterBuffLink", value) end,
								},
								Space1 = { type = "description", name = "", order = 25 },
								AddFilter = {
									type = "input", order = 30, name = L["Enter Buff"],
									desc = L["Enter a buff to be added to the filter list."],
									disabled = function(info) return not GetBarGroupField("detectBuffs") or not (GetBarGroupField("filterBuff") or GetBarGroupField("showBuff")) end,
									get = function(info) return nil end,
									set = function(info, value) ValidateSpellName(value); AddBarGroupFilter("Buff", value) end,
								},
								SelectFilter = {
									type = "select", order = 40, name = L["Filter List"],
									disabled = function(info) return not GetBarGroupField("detectBuffs") or not (GetBarGroupField("filterBuff") or GetBarGroupField("showBuff")) end,
									get = function(info) return GetBarGroupFilterSelection("Buff") end,
									set = function(info, value) SetBarGroupField("filterBuffSelection", value) end,
									values = function(info) return GetBarGroupFilter("Buff") end,
									style = "dropdown",
								},
								DeleteFilter = {
									type = "execute", order = 50, name = L["Delete"], width = "half",
									disabled = function(info) return not GetBarGroupField("detectBuffs") or not (GetBarGroupField("filterBuff") or GetBarGroupField("showBuff")) end,
									desc = L["Delete the selected buff from the filter list."],
									func = function(info) DeleteBarGroupFilter("Buff", GetBarGroupField("filterBuffSelection")) end,
								},
								ResetFilter = {
									type = "execute", order = 60, name = L["Reset"], width = "half",
									desc = L["Reset the buff filter list."],
									disabled = function(info) return not GetBarGroupField("detectBuffs") or not (GetBarGroupField("filterBuff") or GetBarGroupField("showBuff")) end,
									confirm = function(info) return L['RESET\nAre you sure you want to reset the buff filter list?'] end,
									func = function(info) ResetBarGroupFilter("Buff") end,
								},
							},
						},
						FilterBarGroup = {
							type = "group", order = 50, name = L["Filter Bar Group"], inline = true, width = "full",
							disabled = function(info) return not GetBarGroupField("detectBuffs") end,
							args = {
								Enable = {
									type = "toggle", order = 10, name = L["Enable"],
									desc = L["Filter buff bar group string"],
									get = function(info) return GetBarGroupField("filterBuffBars") end,
									set = function(info, v) SetBarGroupField("filterBuffBars", v) end,
								},
								SelectBarGroup = {
									type = "select", order = 20, name = L["Bar Group"],
									desc = L["Select filter bar group."],
									disabled = function(info) return not GetBarGroupField("detectBuffs") or not GetBarGroupField("filterBuffBars") end,
									get = function(info) local t = GetBarGroupList(); for k, v in pairs(t) do if v == GetBarGroupField("filterBuffBarGroup") then return k end end end,
									set = function(info, value) SetBarGroupField("filterBuffBarGroup", GetBarGroupList()[value]) end,
									values = function(info) return GetBarGroupList() end,
									style = "dropdown",
								},
							},
						},
					},
				},
				DetectDebuffsTab = {
					type = "group", order = 25, name = L["Debuffs"],
					disabled = function(info) return InMode("Bar") end,
					hidden = function(info) return NoBarGroup() or not GetBarGroupField("auto") end,
					args = {
						EnableGroup = {
							type = "group", order = 1, name = L["Enable"], inline = true,
							args = {
								DetectEnable = {
									type = "toggle", order = 1, name = L["Auto Debuffs"],
									desc = L['Enable automatically displaying bars for debuffs that match these settings.'],
									get = function(info) return GetBarGroupField("detectDebuffs") end,
									set = function(info, value) SetBarGroupField("detectDebuffs", value) end,
								},
							},
						},
						MonitorUnitGroup = {
							type = "group", order = 10, name = L["Action On"], inline = true,
							disabled = function(info) return not GetBarGroupField("detectDebuffs") end,
							args = {
								PlayerBuff = {
									type = "toggle", order = 10, name = L["Player"],
									desc = L["If checked, only add bars for debuffs if they are on the player."],
									get = function(info) return GetBarGroupField("detectDebuffsMonitor") == "player" end,
									set = function(info, value) SetBarGroupField("detectDebuffsMonitor", "player") end,
								},
								PetBuff = {
									type = "toggle", order = 15, name = L["Pet"],
									desc = L["If checked, only add bars for debuffs if they are on the player's pet."],
									get = function(info) return GetBarGroupField("detectDebuffsMonitor") == "pet" end,
									set = function(info, value) SetBarGroupField("detectDebuffsMonitor", "pet") end,
								},
								TargetBuff = {
									type = "toggle", order = 20, name = L["Target"],
									desc = L["If checked, only add bars for debuffs if they are on the target."],
									get = function(info) return GetBarGroupField("detectDebuffsMonitor") == "target" end,
									set = function(info, value) SetBarGroupField("detectDebuffsMonitor", "target") end,
								},
								FocusBuff = {
									type = "toggle", order = 30, name = L["Focus"],
									desc = L["If checked, only add bars for debuffs if they are on the focus."],
									get = function(info) return GetBarGroupField("detectDebuffsMonitor") == "focus" end,
									set = function(info, value) SetBarGroupField("detectDebuffsMonitor", "focus") end,
								},
								Space1 = { type = "description", name = "", order = 35 },
								MouseoverDebuff = {
									type = "toggle", order = 40, name = L["Mouseover"],
									desc = L["If checked, only add bars for debuffs if they are on the mouseover unit."],
									get = function(info) return GetBarGroupField("detectDebuffsMonitor") == "mouseover" end,
									set = function(info, value) SetBarGroupField("detectDebuffsMonitor", "mouseover") end,
								},
								PetTargetDebuff = {
									type = "toggle", order = 45, name = L["Pet's Target"],
									desc = L["If checked, only add bars for debuffs if they are on the pet's target."],
									get = function(info) return GetBarGroupField("detectDebuffsMonitor") == "pettarget" end,
									set = function(info, value) SetBarGroupField("detectDebuffsMonitor", "pettarget") end,
								},
								TargetTargetDebuff = {
									type = "toggle", order = 50, name = L["Target's Target"],
									desc = L["If checked, only add bars for debuffs if they are on the target's target."],
									get = function(info) return GetBarGroupField("detectDebuffsMonitor") == "targettarget" end,
									set = function(info, value) SetBarGroupField("detectDebuffsMonitor", "targettarget") end,
								},
								FocusTargetDebuff = {
									type = "toggle", order = 60, name = L["Focus's Target"],
									desc = L["If checked, only add bars for debuffs if they are on the focus's target."],
									get = function(info) return GetBarGroupField("detectDebuffsMonitor") == "focustarget" end,
									set = function(info, value) SetBarGroupField("detectDebuffsMonitor", "focustarget") end,
								},
							},
						},
						CastUnitGroup = {
							type = "group", order = 20, name = L["Cast By"], inline = true,
							disabled = function(info) return not GetBarGroupField("detectDebuffs") end,
							args = {
								MyBuff = {
									type = "toggle", order = 10, name = L["Player"],
									desc = L["If checked, only add bars for debuffs if cast by the player."],
									get = function(info) return GetBarGroupField("detectDebuffsCastBy") == "player" end,
									set = function(info, value) SetBarGroupField("detectDebuffsCastBy", "player") end,
								},
								PetBuff = {
									type = "toggle", order = 15, name = L["Pet"],
									desc = L["If checked, only add bars for debuffs if cast by the player's pet."],
									get = function(info) return GetBarGroupField("detectDebuffsCastBy") == "pet" end,
									set = function(info, value) SetBarGroupField("detectDebuffsCastBy", "pet") end,
								},
								YourBuff = {
									type = "toggle", order = 20, name = L["Other"],
									desc = L["If checked, only add bars for debuffs if cast by anyone other than the player or pet."],
									get = function(info) return GetBarGroupField("detectDebuffsCastBy") == "other" end,
									set = function(info, value) SetBarGroupField("detectDebuffsCastBy", "other") end,
								},
								OurBuff = {
									type = "toggle", order = 30, name = L["Anyone"],
									desc = L["If checked, add bars for debuffs if cast by anyone, including player."],
									get = function(info) return GetBarGroupField("detectDebuffsCastBy") == "anyone" end,
									set = function(info, value) SetBarGroupField("detectDebuffsCastBy", "anyone") end,
								},
							},
						},
						FilterByType = {
							type = "group", order = 30, name = L["Filter By Type"], inline = true, width = "full",
							disabled = function(info) return not GetBarGroupField("detectDebuffs") end,
							args = {
								Enable = {
									type = "toggle", order = 10, name = L["Enable"],
									desc = L["Filter debuff types string"],
									get = function(info) return GetBarGroupField("filterDebuffTypes") end,
									set = function(info, v) SetBarGroupField("filterDebuffTypes", v) end,
								},
								Dispellable = {
									type = "toggle", order = 20, name = L["Dispellable"],
									disabled = function(info) return not GetBarGroupField("filterDebuffTypes") end,
									desc = L['Include debuffs that the player can dispel.'],
									get = function(info) return GetBarGroupField("detectDispellable") end,
									set = function(info, value) SetBarGroupField("detectDispellable", value) end,
								},
								Boss = {
									type = "toggle", order = 30, name = L["Boss"],
									disabled = function(info) return not GetBarGroupField("filterDebuffTypes") end,
									desc = L['Include debuffs cast by boss.'],
									get = function(info) return GetBarGroupField("detectBoss") end,
									set = function(info, value) SetBarGroupField("detectBoss", value) end,
								},
								Other = {
									type = "toggle", order = 40, name = L["Other"],
									disabled = function(info) return not GetBarGroupField("filterDebuffTypes") end,
									desc = L['Include other debuffs not selected with filter types.'],
									get = function(info) return GetBarGroupField("detectOtherDebuffs") end,
									set = function(info, value) SetBarGroupField("detectOtherDebuffs", value) end,
								},
								Space1 = { type = "description", name = "", order = 50 },
								Poison = {
									type = "toggle", order = 60, name = L["Poison"],
									disabled = function(info) return not GetBarGroupField("filterDebuffTypes") end,
									desc = L['Include poison debuffs.'],
									get = function(info) return GetBarGroupField("detectPoison") end,
									set = function(info, value) SetBarGroupField("detectPoison", value) end,
								},
								Curse = {
									type = "toggle", order = 70, name = L["Curse"],
									disabled = function(info) return not GetBarGroupField("filterDebuffTypes") end,
									desc = L['Include curse debuffs.'],
									get = function(info) return GetBarGroupField("detectCurse") end,
									set = function(info, value) SetBarGroupField("detectCurse", value) end,
								},
								Magic = {
									type = "toggle", order = 80, name = L["Magic"],
									disabled = function(info) return not GetBarGroupField("filterDebuffTypes") end,
									desc = L['Include magic debuffs.'],
									get = function(info) return GetBarGroupField("detectMagic") end,
									set = function(info, value) SetBarGroupField("detectMagic", value) end,
								},
								Disease = {
									type = "toggle", order = 90, name = L["Disease"],
									disabled = function(info) return not GetBarGroupField("filterDebuffTypes") end,
									desc = L['Include disease debuffs.'],
									get = function(info) return GetBarGroupField("detectDisease") end,
									set = function(info, value) SetBarGroupField("detectDisease", value) end,
								},
							},
						},
						FilterGroup = {
							type = "group", order = 40, name = L["Filter List"], inline = true, width = "full",
							disabled = function(info) return not GetBarGroupField("detectDebuffs") end,
							args = {
								BlackList = {
									type = "toggle", order = 10, name = L["Black List"],
									desc = L["If checked, don't display any debuffs that are in the filter list."],
									get = function(info) return GetBarGroupField("filterDebuff") end,
									set = function(info, v) SetBarGroupField("filterDebuff", v); if v then SetBarGroupField("showDebuff", false) end end,
								},
								WhiteList = {
									type = "toggle", order = 11, name = L["White List"],
									desc = L["If checked, only display debuffs that are in the filter list."],
									get = function(info) return GetBarGroupField("showDebuff") end,
									set = function(info, v) SetBarGroupField("showDebuff", v); if v then SetBarGroupField("filterDebuff", false) end  end,
								},
								LinkFilters = {
									type = "toggle", order = 15, name = L["Link Filter"],
									desc = L["If checked, the debuff filter list is shared with bar groups in other profiles with the same name."],
									disabled = function(info) return not GetBarGroupField("detectDebuffs") or not (GetBarGroupField("filterDebuff") or GetBarGroupField("showDebuff")) end,
									get = function(info) return GetBarGroupField("filterDebuffLink") end,
									set = function(info, value) SetBarGroupField("filterDebuffLink", value) end,
								},
								Space1 = { type = "description", name = "", order = 25 },
								AddFilter = {
									type = "input", order = 30, name = L["Enter Debuff"],
									desc = L["Enter a debuff to be added to the filter list."],
									disabled = function(info) return not GetBarGroupField("detectDebuffs") or not (GetBarGroupField("filterDebuff") or GetBarGroupField("showDebuff")) end,
									get = function(info) return nil end,
									set = function(info, value) ValidateSpellName(value); AddBarGroupFilter("Debuff", value) end,
								},
								SelectFilter = {
									type = "select", order = 40, name = L["Filter List"],
									disabled = function(info) return not GetBarGroupField("detectDebuffs") or not (GetBarGroupField("filterDebuff") or GetBarGroupField("showDebuff")) end,
									get = function(info) return GetBarGroupField("filterDebuffSelection") end,
									set = function(info, value) SetBarGroupField("filterDebuffSelection", value) end,
									values = function(info) return GetBarGroupFilter("Debuff") end,
									style = "dropdown",
								},
								DeleteFilter = {
									type = "execute", order = 50, name = L["Delete"], width = "half",
									disabled = function(info) return not GetBarGroupField("detectDebuffs") or not (GetBarGroupField("filterDebuff") or GetBarGroupField("showDebuff")) end,
									desc = L["Delete the selected debuff from the filter list."],
									func = function(info) DeleteBarGroupFilter("Debuff", GetBarGroupField("filterDebuffSelection")) end,
								},
								ResetFilter = {
									type = "execute", order = 60, name = L["Reset"], width = "half",
									desc = L["Reset the debuff filter list."],
									disabled = function(info) return not GetBarGroupField("detectDebuffs") or not (GetBarGroupField("filterDebuff") or GetBarGroupField("showDebuff")) end,
									confirm = function(info) return L['RESET\nAre you sure you want to reset the debuff filter list?'] end,
									func = function(info) ResetBarGroupFilter("Debuff") end,
								},
							},
						},
						FilterBarGroup = {
							type = "group", order = 50, name = L["Filter Bar Group"], inline = true, width = "full",
							disabled = function(info) return not GetBarGroupField("detectDebuffs") end,
							args = {
								Enable = {
									type = "toggle", order = 10, name = L["Enable"],
									desc = L["Filter debuff bar group string"],
									get = function(info) return GetBarGroupField("filterDebuffBars") end,
									set = function(info, v) SetBarGroupField("filterDebuffBars", v) end,
								},
								SelectBarGroup = {
									type = "select", order = 20, name = L["Bar Group"],
									desc = L["Select filter bar group."],
									disabled = function(info) return not GetBarGroupField("detectDebuffs") or not GetBarGroupField("filterDebuffBars") end,
									get = function(info) local t = GetBarGroupList(); for k, v in pairs(t) do if v == GetBarGroupField("filterDebuffBarGroup") then return k end end end,
									set = function(info, value) SetBarGroupField("filterDebuffBarGroup", GetBarGroupList()[value]) end,
									values = function(info) return GetBarGroupList() end,
									style = "dropdown",
								},
							},
						},
					},
				},
				DetectCooldownsTab = {
					type = "group", order = 30, name = L["Cooldowns"],
					disabled = function(info) return InMode("Bar") end,
					hidden = function(info) return NoBarGroup() or not GetBarGroupField("auto") end,
					args = {
						EnableGroup = {
							type = "group", order = 1, name = L["Enable"], inline = true,
							args = {
								DetectEnable = {
									type = "toggle", order = 1, name = L["Auto Cooldowns"],
									desc = L['Enable automatically displaying bars for cooldowns that match these settings.'],
									get = function(info) return GetBarGroupField("detectCooldowns") end,
									set = function(info, value) SetBarGroupField("detectCooldowns", value) end,
								},
							},
						},
						ActionUnitGroup = {
							type = "group", order = 20, name = L["Action By"], inline = true,
							disabled = function(info) return not GetBarGroupField("detectCooldowns") end,
							args = {
								MyBuff = {
									type = "toggle", order = 10, name = L["Player"],
									desc = L["If checked, only add bars for cooldowns associated with the player."],
									get = function(info) return GetBarGroupField("detectCooldownsBy") == "player" end,
									set = function(info, value) SetBarGroupField("detectCooldownsBy", "player") end,
								},
								PetBuff = {
									type = "toggle", order = 20, name = L["Pet"],
									desc = L["If checked, only add bars for cooldowns associated with the player's pet."],
									get = function(info) return GetBarGroupField("detectCooldownsBy") == "pet" end,
									set = function(info, value) SetBarGroupField("detectCooldownsBy", "pet") end,
								},
								AnyBuff = {
									type = "toggle", order = 30, name = L["Anyone"],
									desc = L["If checked, add bars for cooldowns cast by either player or player's pet."],
									get = function(info) return GetBarGroupField("detectCooldownsBy") == "anyone" end,
									set = function(info, value) SetBarGroupField("detectCooldownsBy", "anyone") end,
								},
							},
						},
						CooldownTypeGroup = {
							type = "group", order = 30, name = L["Include"], inline = true,
							disabled = function(info) return not GetBarGroupField("detectCooldowns") end,
							args = {
								SpellCooldowns = {
									type = "toggle", order = 10, name = L["Spells"],
									desc = L["If checked, include bars for spell cooldowns."],
									get = function(info) return GetBarGroupField("detectSpellCooldowns") end,
									set = function(info, value) SetBarGroupField("detectSpellCooldowns", value) end,
								},
								TrinketCooldowns = {
									type = "toggle", order = 20, name = L["Trinkets"],
									desc = L["If checked, include bars for cooldowns on equipped trinkets."],
									get = function(info) return GetBarGroupField("detectTrinketCooldowns") end,
									set = function(info, value) SetBarGroupField("detectTrinketCooldowns", value) end,
								},
								PotionCooldowns = {
									type = "toggle", order = 30, name = L["Potions/Elixirs"],
									desc = L["If checked, include bars for shared potion/elixir cooldowns (an item subject to the shared cooldown must be in your bags in order for the cooldown to be detected)."],
									get = function(info) return GetBarGroupField("detectPotionCooldowns") end,
									set = function(info, value) SetBarGroupField("detectPotionCooldowns", value) end,
								},
								OtherCooldowns = {
									type = "toggle", order = 40, name = L["Other"],
									desc = L["If checked, include bars for other cooldowns (for item cooldowns, an item subject to the cooldown must be in your bags in order for the cooldown to be detected)."],
									get = function(info) return GetBarGroupField("detectOtherCooldowns") end,
									set = function(info, value) SetBarGroupField("detectOtherCooldowns", value) end,
								},
							},
						},
						SharedCooldownGroup = {
							type = "group", order = 30, name = L["Shared Cooldowns"], inline = true,
							disabled = function(info) return not GetBarGroupField("detectCooldowns") end,
							args = {
								FrostTrapsCooldowns = {
									type = "toggle", order = 10, name = L["Frost Traps"],
									desc = L["If checked, only show one cooldown for hunter Frost Traps."],
									get = function(info) return GetBarGroupField("detectSharedFrostTraps") end,
									set = function(info, value) SetBarGroupField("detectSharedFrostTraps", value) end,
								},
								FireTrapsCooldowns = {
									type = "toggle", order = 15, name = L["Fire Traps"],
									desc = L["If checked, only show one cooldown for hunter Fire Traps."],
									get = function(info) return GetBarGroupField("detectSharedFireTraps") end,
									set = function(info, value) SetBarGroupField("detectSharedFireTraps", value) end,
								},
								ShockCooldowns = {
									type = "toggle", order = 20, name = L["Shocks"],
									desc = L["If checked, only show one cooldown for shaman Shocks."],
									get = function(info) return GetBarGroupField("detectSharedShocks") end,
									set = function(info, value) SetBarGroupField("detectSharedShocks", value) end,
								},
								PallyCooldowns = {
									type = "toggle", order = 25, name = L["Crusader/Hammer"],
									desc = L["If checked, only show one cooldown for paladin Crusader Strike and Hammer of the Righteous."],
									get = function(info) return GetBarGroupField("detectSharedCrusader") end,
									set = function(info, value) SetBarGroupField("detectSharedCrusader", value) end,
								},
							},
						},
						FilterGroup = {
							type = "group", order = 40, name = L["Filter List"], inline = true, width = "full",
							disabled = function(info) return not GetBarGroupField("detectCooldowns") end,
							args = {
								BlackList = {
									type = "toggle", order = 10, name = L["Black List"],
									desc = L["If checked, don't display any cooldowns that are in the filter list."],
									get = function(info) return GetBarGroupField("filterCooldown") end,
									set = function(info, v) SetBarGroupField("filterCooldown", v); if v then SetBarGroupField("showCooldown", false) end end,
								},
								WhiteList = {
									type = "toggle", order = 11, name = L["White List"],
									desc = L["If checked, only display cooldowns that are in the filter list."],
									get = function(info) return GetBarGroupField("showCooldown") end,
									set = function(info, v) SetBarGroupField("showCooldown", v); if v then SetBarGroupField("filterCooldown", false) end  end,
								},
								LinkFilters = {
									type = "toggle", order = 15, name = L["Link Filter"],
									desc = L["If checked, the cooldown filter list is shared with bar groups in other profiles with the same name."],
									disabled = function(info) return not GetBarGroupField("detectCooldowns") or not (GetBarGroupField("filterCooldown") or GetBarGroupField("showCooldown"))end,
									get = function(info) return GetBarGroupField("filterCooldownLink") end,
									set = function(info, value) SetBarGroupField("filterCooldownLink", value) end,
								},
								Space1 = { type = "description", name = "", order = 25 },
								AddFilter = {
									type = "input", order = 30, name = L["Enter Cooldown"],
									desc = L["Enter a cooldown to be added to the filter list."],
									disabled = function(info) return not GetBarGroupField("detectCooldowns") or not (GetBarGroupField("filterCooldown") or GetBarGroupField("showCooldown"))end,
									get = function(info) return nil end,
									set = function(info, value) ValidateSpellName(value); AddBarGroupFilter("Cooldown", value) end,
								},
								SelectFilter = {
									type = "select", order = 40, name = L["Filter List"],
									disabled = function(info) return not GetBarGroupField("detectCooldowns") or not (GetBarGroupField("filterCooldown") or GetBarGroupField("showCooldown"))end,
									get = function(info) return GetBarGroupField("filterCooldownSelection") end,
									set = function(info, value) SetBarGroupField("filterCooldownSelection", value) end,
									values = function(info) return GetBarGroupFilter("Cooldown") end,
									style = "dropdown",
								},
								DeleteFilter = {
									type = "execute", order = 50, name = L["Delete"], width = "half",
									disabled = function(info) return not GetBarGroupField("detectCooldowns") or not (GetBarGroupField("filterCooldown") or GetBarGroupField("showCooldown"))end,
									desc = L["Delete the selected cooldown from the filter list."],
									func = function(info) DeleteBarGroupFilter("Cooldown", GetBarGroupField("filterCooldownSelection")) end,
								},
								ResetFilter = {
									type = "execute", order = 60, name = L["Reset"], width = "half",
									desc = L["Reset the cooldown filter list."],
									disabled = function(info) return not GetBarGroupField("detectCooldowns") or not (GetBarGroupField("filterCooldown") or GetBarGroupField("showCooldown"))end,
									confirm = function(info) return 'RESET\nAre you sure you want to reset the cooldown filter list?' end,
									func = function(info) ResetBarGroupFilter("Cooldown") end,
								},
							},
						},
						FilterBarGroup = {
							type = "group", order = 50, name = L["Filter Bar Group"], inline = true, width = "full",
							disabled = function(info) return not GetBarGroupField("detectCooldowns") end,
							args = {
								Enable = {
									type = "toggle", order = 10, name = L["Enable"],
									desc = L["Filter cooldown bar group string"],
									get = function(info) return GetBarGroupField("filterCooldownBars") end,
									set = function(info, v) SetBarGroupField("filterCooldownBars", v) end,
								},
								SelectBarGroup = {
									type = "select", order = 20, name = L["Bar Group"],
									desc = L["Select filter bar group."],
									disabled = function(info) return not GetBarGroupField("detectCooldowns") or not GetBarGroupField("filterCooldownBars") end,
									get = function(info) local t = GetBarGroupList(); for k, v in pairs(t) do if v == GetBarGroupField("filterCooldownBarGroup") then return k end end end,
									set = function(info, value) SetBarGroupField("filterCooldownBarGroup", GetBarGroupList()[value]) end,
									values = function(info) return GetBarGroupList() end,
									style = "dropdown",
								},
							},
						},
					},
				},
				LayoutTab = {
					type = "group", order = 40, name = L["Layout"],
					disabled = function(info) return InMode("Bar") end,
					hidden = function(info) return NoBarGroup() or GetBarGroupField("merged") end,
					args = {
						ConfigurationGroup = {
							type = "group", order = 10, name = L["Configuration"], inline = true,
							args = {
								BarConfiguration = {
									type = "toggle", order = 10, name = L["Bars"],
									desc = L["If checked, use a bar-oriented configuration."],
									get = function(info)
										local config = Nest_SupportedConfigurations[GetBarGroupField("configuration")]
										return not config.iconOnly
									end,
									set = function(info, value)
										local config = Nest_SupportedConfigurations[GetBarGroupField("configuration")]
										if config.iconOnly then SetBarGroupField("configuration", 1) end
									end,
								},
								IconConfiguration = {
									type = "toggle", order = 20, name = L["Icons"],
									desc = L["If checked, use an icon-oriented configuration."],
									get = function(info)
										local config = Nest_SupportedConfigurations[GetBarGroupField("configuration")]
										return config.iconOnly
									end,
									set = function(info, value)
										local config = Nest_SupportedConfigurations[GetBarGroupField("configuration")]
										if not config.iconOnly then SetBarGroupField("configuration", 9) end
									end,
								},
								Space1 = { type = "description", name = "", order = 40 },
								Configuration = {
									type = "select", order = 50, name = L["Options"], width = "double",
									desc = L["Select a configuration option for bars or icons."],
									get = function(info) return GetBarGroupField("configuration") end,
									set = function(info, value) SetBarGroupField("configuration", value) end,
									values = function(info)
										local config = Nest_SupportedConfigurations[GetBarGroupField("configuration")]
										return GetOrientationList(config.iconOnly)
									end,
									style = "dropdown",
								},
								ReverseGrowthGroup = {
									type = "toggle", order = 60, name = L["Direction"],
									desc = L["If checked, grow up or to the right, otherwise grow down or to the left."],
									get = function(info) return GetBarGroupField("growDirection") end,
									set = function(info, value) SetBarGroupField("growDirection", value) end,
								},
								CopyLayoutGroup = {
									type = "select", order = 70, name = L["Copy Layout From"],
									desc = L["Select bar group to copy all layout settings from."],
									get = function(info) return nil end,
									set = function(info, value) CopyBarGroupConfiguration(GetBarGroupList()[value]) end,
									values = function(info) return GetBarGroupList() end,
									style = "dropdown",
								},
							},
						},
						DimensionGroup = {
							type = "group", order = 20, name = L["Format"], inline = true,
							args = {
								UseDefaultsGroup = {
									type = "toggle", order = 1, name = L["Use Defaults"],
									desc = L["If checked, format options are set to default values."],
									get = function(info) return GetBarGroupField("useDefaultDimensions") end,
									set = function(info, value) SetBarGroupField("useDefaultDimensions", value) end,
								},
								RestoreDefaults = {
									type = "execute", order = 5, name = L["Restore Defaults"],
									desc = L["Reset format for this bar group back to the current defaults."],
									disabled = function(info) return GetBarGroupField("useDefaultDimensions") end,
									func = function(info) MOD:CopyDimensions(MOD.db.global.Defaults, GetBarGroupEntry()); MOD:UpdateAllBarGroups() end,
								},
								Space1 = { type = "description", name = "", order = 10 },
								BarWidth = {
									type = "range", order = 20, name = L["Bar Width"], min = 5, max = 500, step = 1,
									desc = L["Set width of bars."],
									disabled = function(info) return GetBarGroupField("useDefaultDimensions") end,
									get = function(info) return GetBarGroupField("barWidth") end,
									set = function(info, value) SetBarGroupField("barWidth", value) end,
								},
								BarHeight = {
									type = "range", order = 25, name = L["Bar Height"], min = 5, max = 100, step = 1,
									desc = L["Set height of bars."],
									disabled = function(info) return GetBarGroupField("useDefaultDimensions") end,
									get = function(info) return GetBarGroupField("barHeight") end,
									set = function(info, value) SetBarGroupField("barHeight", value) end,
								},
								IconSize = {
									type = "range", order = 30, name = L["Icon Size"], min = 5, max = 100, step = 1,
									desc = L["Set width/height for icons."],
									disabled = function(info) return GetBarGroupField("useDefaultDimensions") end,
									get = function(info) return GetBarGroupField("iconSize") end,
									set = function(info, value) SetBarGroupField("iconSize", value) end,
								},
								Scale = {
									type = "range", order = 35, name = L["Scale"], min = 0.1, max = 2, step = 0.05,
									desc = L["Set scale factor for bars and icons."],
									disabled = function(info) return GetBarGroupField("useDefaultDimensions") end,
									get = function(info) return GetBarGroupField("scale") end,
									set = function(info, value) SetBarGroupField("scale", value) end,
								},
								Space2 = { type = "description", name = "", order = 40 },
								HorizontalSpacing = {
									type = "range", order = 60, name = L["Horizontal Spacing"], min = -100, max = 100, step = 1,
									desc = L["Adjust horizontal spacing between bars."],
									disabled = function(info) return GetBarGroupField("useDefaultDimensions") end,
									get = function(info) return GetBarGroupField("spacingX") end,
									set = function(info, value) SetBarGroupField("spacingX", value) end,
								},
								VerticalSpacing = {
									type = "range", order = 65, name = L["Vertical Spacing"], min = -100, max = 100, step = 1,
									desc = L["Adjust vertical spacing between bars."],
									disabled = function(info) return GetBarGroupField("useDefaultDimensions") end,
									get = function(info) return GetBarGroupField("spacingY") end,
									set = function(info, value) SetBarGroupField("spacingY", value) end,
								},
								IconOffsetX = {
									type = "range", order = 70, name = L["Icon Inset"], min = -200, max = 200, step = 1,
									desc = L["Set icon's horizontal inset from bar."],
									disabled = function(info) return GetBarGroupField("useDefaultDimensions") end,
									get = function(info) return GetBarGroupField("iconOffsetX") end,
									set = function(info, value) SetBarGroupField("iconOffsetX", value) end,
								},
								IconOffsetY = {
									type = "range", order = 75, name = L["Icon Offset"], min = -200, max = 200, step = 1,
									desc = L["Set vertical offset between icon and bar."],
									disabled = function(info) return GetBarGroupField("useDefaultDimensions") end,
									get = function(info) return GetBarGroupField("iconOffsetY") end,
									set = function(info, value) SetBarGroupField("iconOffsetY", value) end,
								},
								Space2 = { type = "description", name = "", order = 80 },
								BarFormatGroup = {
									type = "group", order = 90, name = "", inline = true,
									disabled = function(info) return GetBarGroupField("useDefaultDimensions") end,
									args = {
										HideIconGroup = {
											type = "toggle", order = 30, name = L["Icon"], width = "half",
											desc = L["Show icons with bars (required for tooltips)."],
											get = function(info) return not GetBarGroupField("hideIcon") end,
											set = function(info, value) SetBarGroupField("hideIcon", not value) end,
										},
										HideClockGroup = {
											type = "toggle", order = 31, name = L["Clock"], width = "half",
											desc = L["Show clock animation on icons for timer bars."],
											get = function(info) return not GetBarGroupField("hideClock") end,
											set = function(info, value) SetBarGroupField("hideClock", not value) end,
										},
										HideBarGroup = {
											type = "toggle", order = 32, name = L["Bar"], width = "half",
											desc = L["Show colored bar and background."],
											get = function(info) return not GetBarGroupField("hideBar") end,
											set = function(info, value) SetBarGroupField("hideBar", not value) end,
										},
										HideSparkGroup = {
											type = "toggle", order = 33, name = L["Spark"], width = "half",
											desc = L["Show spark that moves across bars to indicate remaining time."],
											disabled = function(info) return GetBarGroupField("useDefaultDimensions") or GetBarGroupField("hideBar") end,
											get = function(info) return not GetBarGroupField("hideSpark") end,
											set = function(info, value) SetBarGroupField("hideSpark", not value) end,
										},
										HideLabelGroup = {
											type = "toggle", order = 34, name = L["Label"], width = "half",
											desc = L["Show label text on bars."],
											get = function(info) return not GetBarGroupField("hideLabel") end,
											set = function(info, value) SetBarGroupField("hideLabel", not value) end,
										},
										HideCountGroup = {
											type = "toggle", order = 35, name = L["Count"], width = "half",
											desc = L["Show stack count in parentheses after label (it is also displayed as overlay on icon)."],
											get = function(info) return not GetBarGroupField("hideCount") end,
											set = function(info, value) SetBarGroupField("hideCount", not value) end,
										},
										HideTimerGroup = {
											type = "toggle", order = 36, name = L["Time"], width = "half",
											desc = L["Show time left on bars that have a duration."],
											get = function(info) return not GetBarGroupField("hideValue") end,
											set = function(info, value) SetBarGroupField("hideValue", not value) end,
										},
										TooltipsGroup = {
											type = "toggle", order = 37, name = L["Tooltips"], width = "half",
											desc = L["Show tooltips when the cursor is over a bar's icon."],
											get = function(info) return GetBarGroupField("showTooltips") end,
											set = function(info, value) SetBarGroupField("showTooltips", value) end,
										},
									},
								},
							},
						},
						TextSettings = {
							type = "group", order = 30, name = L["Text Settings"], inline = true,
							args = {
								LabelInset = {
									type = "range", order = 10, name = L["Label Text Inset"], min = -200, max = 200, step = 1,
									desc = L["Set horizontal inset for label from edge of bar."],
									get = function(info) return GetBarGroupField("labelInset") end,
									set = function(info, value) SetBarGroupField("labelInset", value) end,
								},
								LabelOffset = {
									type = "range", order = 15, name = L["Label Text Offset"], min = -200, max = 200, step = 1,
									desc = L["Set vertical offset for label text from center of bar."],
									get = function(info) return GetBarGroupField("labelOffset") end,
									set = function(info, value) SetBarGroupField("labelOffset", value) end,
								},
								LabelWrapGroup = {
									type = "toggle", order = 20, name = L["Wrap"], width = "half",
									desc = L["If checked, wrap label text when it doesn't fit in the bar's width."],
									get = function(info) return GetBarGroupField("labelWrap") end,
									set = function(info, value) SetBarGroupField("labelWrap", value) end,
								},
								LabelTopGroup = {
									type = "toggle", order = 21, name = L["Top"], width = "half",
									desc = L["If checked, set \"Top\" alignment for label text."],
									get = function(info) return GetBarGroupField("labelAlign") == "TOP" end,
									set = function(info, value) SetBarGroupField("labelAlign", "TOP") end,
								},
								LabelCenterGroup = {
									type = "toggle", order = 22, name = L["Middle"], width = "half",
									desc = L["If checked, set \"Middle\" alignment for label text."],
									get = function(info) return GetBarGroupField("labelAlign") == "MIDDLE" end,
									set = function(info, value) SetBarGroupField("labelAlign", "MIDDLE") end,
								},
								LabelBottomGroup = {
									type = "toggle", order = 23, name = L["Bottom"], width = "half",
									desc = L["If checked, set \"Bottom\" alignment for label text."],
									get = function(info) return GetBarGroupField("labelAlign") == "BOTTOM" end,
									set = function(info, value) SetBarGroupField("labelAlign", "BOTTOM") end,
								},
								Space1 = { type = "description", name = "", order = 30 },
								TimeTextInset = {
									type = "range", order = 40, name = L["Time Text Inset"], min = -200, max = 200, step = 1,
									desc = L["Set horizontal inset for time text from edge of bar."],
									get = function(info) return GetBarGroupField("timeInset") end,
									set = function(info, value) SetBarGroupField("timeInset", value) end,
								},
								TimeTextOffset = {
									type = "range", order = 45, name = L["Time Text Offset"], min = -200, max = 200, step = 1,
									desc = L["Set vertical offset for time text from center of bar."],
									get = function(info) return GetBarGroupField("timeOffset") end,
									set = function(info, value) SetBarGroupField("timeOffset", value) end,
								},
								TimeNormalGroup = {
									type = "toggle", order = 46, name = L["Normal"], width = "half",
									desc = L["If checked, use normal alignment with small inset for time and label texts."],
									get = function(info) return GetBarGroupField("timeAlign") == "normal" end,
									set = function(info, value) SetBarGroupField("timeAlign", "normal") end,
								},
								TimeLeftGroup = {
									type = "toggle", order = 47, name = L["Left"], width = "half",
									desc = L["If checked, set \"Left\" alignment for time text and normal alignment for label."],
									get = function(info) return GetBarGroupField("timeAlign") == "LEFT" end,
									set = function(info, value) SetBarGroupField("timeAlign", "LEFT") end,
								},
								TimeCenterGroup = {
									type = "toggle", order = 48, name = L["Center"], width = "half",
									desc = L["If checked, set \"Center\" alignment for time text and normal alignment for label."],
									get = function(info) return GetBarGroupField("timeAlign") == "CENTER" end,
									set = function(info, value) SetBarGroupField("timeAlign", "CENTER") end,
								},
								TimeRightGroup = {
									type = "toggle", order = 49, name = L["Right"], width = "half",
									desc = L["If checked, set \"Right\" alignment for time text and normal alignment for label."],
									get = function(info) return GetBarGroupField("timeAlign") == "RIGHT" end,
									set = function(info, value) SetBarGroupField("timeAlign", "RIGHT") end,
								},
								Space2 = { type = "description", name = "", order = 50 },
								IconTextInset = {
									type = "range", order = 60, name = L["Icon Text Inset"], min = -200, max = 200, step = 1,
									desc = L["Set horizontal inset for icon text from middle of icon."],
									get = function(info) return GetBarGroupField("iconInset") end,
									set = function(info, value) SetBarGroupField("iconInset", value) end,
								},
								IconTextOffset = {
									type = "range", order = 65, name = L["Icon Text Offset"], min = -200, max = 200, step = 1,
									desc = L["Set vertical offset for icon text from center of icon."],
									get = function(info) return GetBarGroupField("iconOffset") end,
									set = function(info, value) SetBarGroupField("iconOffset", value) end,
								},
							},
						},
						IconOnlyGroup = {
							type = "group", order = 35, name = L["Icon Configuration Options"], inline = true,
							hidden = function(info)
								local config = GetBarGroupField("configuration")
								if config then return not Nest_SupportedConfigurations[config].iconOnly else return true end
							end,
							args = {
								Wrap = {
									type = "range", order = 20, name = L["Wrap"], min = 0, max = 50, step = 1,
									desc = L["Set how many icons to line up before wrapping to next row or column."],
									get = function(info) return GetBarGroupField("wrap") end,
									set = function(info, value) SetBarGroupField("wrap", value) end,
								},
								WrapDirection = {
									type = "toggle", order = 25, name = L["Wrap Direction"],
									desc = L["If checked, wrap icons up when arranged in rows or to the right when arranged in columns, otherwise wrap down or to the left."],
									get = function(info) return GetBarGroupField("wrapDirection") end,
									set = function(info, value) SetBarGroupField("wrapDirection", value) end,
								},
								SnapCenter = {
									type = "toggle", order = 30, name = L["Snap To Center"],
									desc = L["If checked and the bar group is locked, icons snap to center at the anchor position."],
									get = function(info) return GetBarGroupField("snapCenter") end,
									set = function(info, value) SetBarGroupField("snapCenter", value) end,
								},
							},
						},
						PositionGroup = {
							type = "group", order = 40, name = L["Display Position/Bar Limit"], inline = true,
							args = {
								Horizontal = {
									type = "range", order = 10, name = L["Horizontal"], min = 0, max = 100, step = 0.01,
									desc = L["Set horizontal position as percentage of overall width (cannot move beyond edge of display)."],
									get = function(info) return GetBarGroupField("pointX") * 100 end,
									set = function(info, value) SetBarGroupField("pointX", value / 100); SetBarGroupField("pointXR", nil) end,
								},
								Vertical = {
									type = "range", order = 20, name = L["Vertical"], min = 0, max = 100, step = 0.01,
									desc = L["Set vertical position as percentage of overall height (cannot move beyond edge of display)."],
									get = function(info) return GetBarGroupField("pointY") * 100 end,
									set = function(info, value) SetBarGroupField("pointY", value / 100); SetBarGroupField("pointYT", nil) end,
								},
								Space1 = { type = "description", name = "", order = 30, width = "half" },
								MaxBars = {
									type = "range", order = 40, name = L["Bar Limit"], min = 0, max = 100, step = 1,
									desc = L["Set the maximum number of bars to display (bars that sort closest to the anchor have priority). If this is set to 0 then the number of bars is not limited."],
									get = function(info) return GetBarGroupField("maxBars") end,
									set = function(info, value) SetBarGroupField("maxBars", value) end,
								},
							},
						},
						AnchorGroup = {
							type = "group", order = 50, name = L["Attachment"], inline = true,
							args = {
								Anchor = {
									type = "select", order = 10, name = L["Bar Group"],
									desc = L["Select a bar group to attach to (for independent position, attach to self)."],
									get = function(info) return GetBarGroupAnchor() end,
									set = function(info, value) SetBarGroupAnchor(value) end,
									values = function(info) return GetBarGroupList() end,
									style = "dropdown",
								},
								OffsetX = {
									type = "range", order = 20, name = L["Offset X"], min = -500, max = 500, step = 1,
									desc = L["Set horizontal offset from the selected bar group."],
									disabled = function(info) return not GetBarGroupField("anchor") end,
									get = function(info) return GetBarGroupField("anchorX") end,
									set = function(info, value) SetBarGroupField("anchorX", value) end,
								},
								OffsetY = {
									type = "range", order = 30, name = L["Offset Y"], min = -500, max = 500, step = 1,
									desc = L["Set vertical offset from the selected bar group."],
									disabled = function(info) return not GetBarGroupField("anchor") end,
									get = function(info) return GetBarGroupField("anchorY") end,
									set = function(info, value) SetBarGroupField("anchorY", value) end,
								},
								Relative = {
									type = "toggle", order = 40, name = L["Last Bar"], width = "half",
									desc = L["If checked, position is relative to last bar in the selected bar group."],
									disabled = function(info) return not GetBarGroupField("anchor") end,
									get = function(info) return GetBarGroupField("anchorLastBar") end,
									set = function(info, value) SetBarGroupField("anchorLastBar", value) end,
								},
								Empty = {
									type = "toggle", order = 50, name = L["Empty"], width = "half",
									desc = L["If checked, offsets are not applied if the selected bar group is empty."],
									disabled = function(info) return not GetBarGroupField("anchor") end,
									get = function(info) return GetBarGroupField("anchorEmpty") end,
									set = function(info, value) SetBarGroupField("anchorEmpty", value) end,
								},
								ResetAnchor = {
									type = "execute", order = 60, name = L["Reset"], width = "half",
									desc = L["Reset position to screen center, unlock anchor, and reset attachment options."],
									func = function(info) SetBarGroupAnchor(nil) end,
								},
							},
						},
					},
				},
				AppearanceTab = {
					type = "group", order = 45, name = L["Appearance"],
					disabled = function(info) return InMode("Bar") end,
					hidden = function(info) return NoBarGroup() or GetBarGroupField("merged") end,
					args = {
						FontsGroup = {
							type = "group", order = 20, name = L["Fonts and Textures"], inline = true,
							args = {
								UseDefaultsGroup = {
									type = "toggle", order = 1, name = L["Use Defaults"],
									desc = L["If checked, fonts and textures use the default values."],
									get = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									set = function(info, value) SetBarGroupField("useDefaultFontsAndTextures", value) end,
								},
								RestoreDefaults = {
									type = "execute", order = 3, name = L["Restore Defaults"],
									desc = L["Reset fonts and textures for this bar group back to the current defaults."],
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									func = function(info) MOD:CopyFontsAndTextures(MOD.db.global.Defaults, GetBarGroupEntry()); MOD:UpdateAllBarGroups() end,
								},
								CopyFromGroup = {
									type = "select", order = 4, name = L["Copy From"],
									desc = L["Select bar group to copy font and texture settings from."],
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return nil end,
									set = function(info, value) CopyBarGroupFontsAndTextures(GetBarGroupList()[value]) end,
									values = function(info) return GetBarGroupList() end,
									style = "dropdown",
								},
								Space1 = { type = "description", name = "", order = 5 },
								LabelFont = {
									type = "select", order = 21, name = L["Label Font"], 
									desc = L["Select font for label text."],
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("labelFont") end,
									set = function(info, value) SetBarGroupField("labelFont", value) end,
								},
								LabelFontSize = {
									type = "range", order = 23, name = L["Font Size"], min = 5, max = 30, step = 1,
									desc = L["Set font size for labels."],
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("labelFSize") end,
									set = function(info, value) SetBarGroupField("labelFSize", value) end,
								},
								LabelAlpha = {
									type = "range", order = 25, name = L["Opacity"], min = 0, max = 1, step = 0.05,
									desc = L["Set opacity for labels."],
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("labelAlpha") end,
									set = function(info, value) SetBarGroupField("labelAlpha", value) end,
								},
								LabelColor = {
									type = "color", order = 27, name = L["Color"], hasAlpha = false, width = "half",
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info)
										local t = GetBarGroupField("labelColor")
										if t then return t.r, t.g, t.b, t.a else return 1, 1, 1, 1 end
									end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("labelColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("labelColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
								LabelOutline = {
									type = "toggle", order = 29, name = L["Outline"], width = "half",
									desc = L["Add black outline to label."],
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("labelOutline") end,
									set = function(info, value) SetBarGroupField("labelOutline", value) end,
								},
								Space2 = { type = "description", name = "", order = 30 },
								TimeFont = {
									type = "select", order = 31, name = L["Time Font"], 
									desc = L["Select font for time text."],
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("timeFont") end,
									set = function(info, value) SetBarGroupField("timeFont", value) end,
								},
								TimeFontSize = {
									type = "range", order = 33, name = L["Font Size"], min = 5, max = 30, step = 1,
									desc = L["Set font size for time text."],
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("timeFSize") end,
									set = function(info, value) SetBarGroupField("timeFSize", value) end,
								},
								TimeAlpha = {
									type = "range", order = 35, name = L["Opacity"], min = 0, max = 1, step = 0.05,
									desc = L["Set opacity for time text."],
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("timeAlpha") end,
									set = function(info, value) SetBarGroupField("timeAlpha", value) end,
								},
								TimeColor = {
									type = "color", order = 37, name = L["Color"], hasAlpha = false, width = "half",
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info)
										local t = GetBarGroupField("timeColor")
										if t then return t.r, t.g, t.b, t.a else return 1, 1, 1, 1 end
									end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("timeColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("timeColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
								TimeOutline = {
									type = "toggle", order = 39, name = L["Outline"], width = "half",
									desc = L["Add black outline to time text."],
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("timeOutline") end,
									set = function(info, value) SetBarGroupField("timeOutline", value) end,
								},
								Space3 = { type = "description", name = "", order = 40 },
								IconFont = {
									type = "select", order = 41, name = L["Icon Font"], 
									desc = L["Select font for icon text (i.e., spell stack count)."],
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("iconFont") end,
									set = function(info, value) SetBarGroupField("iconFont", value) end,
								},
								IconFontSize = {
									type = "range", order = 43, name = L["Font Size"], min = 5, max = 30, step = 1,
									desc = L["Set font size for icon text."],
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("iconFSize") end,
									set = function(info, value) SetBarGroupField("iconFSize", value) end,
								},
								IconAlpha = {
									type = "range", order = 45, name = L["Opacity"], min = 0, max = 1, step = 0.05,
									desc = L["Set opacity for icon text."],
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("iconAlpha") end,
									set = function(info, value) SetBarGroupField("iconAlpha", value) end,
								},
								IconColor = {
									type = "color", order = 47, name = L["Color"], hasAlpha = false, width = "half",
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info)
										local t = GetBarGroupField("iconColor")
										if t then return t.r, t.g, t.b, t.a else return 1, 1, 1, 1 end
									end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("iconColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("iconColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
								IconOutline = {
									type = "toggle", order = 49, name = L["Outline"], width = "half",
									desc = L["Add black outline to icon text."],
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("iconOutline") end,
									set = function(info, value) SetBarGroupField("iconOutline", value) end,
								},
								Space7 = { type = "description", name = "", order = 70 },
								ForegroundTexture = {
									type = "select", order = 72, name = L["Foreground Texture"], 
									desc = L["Select foreground texture for bars."],
									dialogControl = 'LSM30_Statusbar',
									values = AceGUIWidgetLSMlists.statusbar,
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("texture") end,
									set = function(info, value) SetBarGroupField("texture", value) end,
								},
								ForegroundAlpha = {
									type = "range", order = 74, name = L["Foreground Opacity"], min = 0, max = 1, step = 0.05,
									desc = L["Set foreground opacity for bars."],
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("fgAlpha") end,
									set = function(info, value) SetBarGroupField("fgAlpha", value) end,
								},
								ForegroundSaturation = {
									type = "range", order = 76, name = L["Foreground Saturation"], min = -1, max = 1, step = 0.05,
									desc = L["Set saturation for foreground colors."],
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("fgSaturation") end,
									set = function(info, value) SetBarGroupField("fgSaturation", value) end,
								},
								ForegroundBrightness = {
									type = "range", order = 78, name = L["Foreground Brightness"], min = -1, max = 1, step = 0.05,
									desc = L["Set brightness for foreground colors."],
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("fgBrightness") end,
									set = function(info, value) SetBarGroupField("fgBrightness", value) end,
								},
								Space8 = { type = "description", name = "", order = 80 },
								BackgroundTexture = {
									type = "select", order = 82, name = L["Background Texture"], 
									desc = L["Select background texture for bars."],
									dialogControl = 'LSM30_Statusbar',
									values = AceGUIWidgetLSMlists.statusbar,
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("bgtexture") end,
									set = function(info, value) SetBarGroupField("bgtexture", value) end,
								},
								Background = {
									type = "range", order = 84, name = L["Background Opacity"], min = 0, max = 1, step = 0.05,
									desc = L["Set background opacity for bars."],
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("bgAlpha") end,
									set = function(info, value) SetBarGroupField("bgAlpha", value) end,
								},
								BackgroundSaturation = {
									type = "range", order = 86, name = L["Background Saturation"], min = -1, max = 1, step = 0.05,
									desc = L["Set saturation for foreground colors."],
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("bgSaturation") end,
									set = function(info, value) SetBarGroupField("bgSaturation", value) end,
								},
								BackgroundBrightness = {
									type = "range", order = 88, name = L["Background Brightness"], min = -1, max = 1, step = 0.05,
									desc = L["Set brightness for foreground colors."],
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("bgBrightness") end,
									set = function(info, value) SetBarGroupField("bgBrightness", value) end,
								},
								Space9 = { type = "description", name = "", order = 90 },
								NormalAlpha = {
									type = "range", order = 92, name = L["Opacity (Not Combat)"], min = 0, max = 1, step = 0.05,
									desc = L["Set opacity for bars when not in combat."],
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("alpha") end,
									set = function(info, value) SetBarGroupField("alpha", value) end,
								},
								CombatAlpha = {
									type = "range", order = 94, name = L["Opacity (In Combat)"], min = 0, max = 1, step = 0.05,
									desc = L["Set opacity for bars when in combat."],
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("combatAlpha") end,
									set = function(info, value) SetBarGroupField("combatAlpha", value) end,
								},
								IconBorderSaturation = {
									type = "range", order = 95, name = L["Icon Border Saturation"], min = -1, max = 1, step = 0.05,
									desc = L["Set saturation for icon border colors."],
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("borderSaturation") end,
									set = function(info, value) SetBarGroupField("borderSaturation", value) end,
								},
								IconBorderBrightness = {
									type = "range", order = 96, name = L["Icon Border Brightness"], min = -1, max = 1, step = 0.05,
									desc = L["Set brightness for icon border colors."],
									disabled = function(info) return GetBarGroupField("useDefaultFontsAndTextures") end,
									get = function(info) return GetBarGroupField("borderBrightness") end,
									set = function(info, value) SetBarGroupField("borderBrightness", value) end,
								},
							},
						},
						ColorsGroup = {
							type = "group", order = 30, name = L["Standard and Spell Colors"], inline = true,
							args = {
								UseDefaultsGroup = {
									type = "toggle", order = 1, name = L["Use Defaults"],
									desc = L["If checked, colors use the default values."],
									get = function(info) return GetBarGroupField("useDefaultColors") end,
									set = function(info, value) SetBarGroupField("useDefaultColors", value) end,
								},
								RestoreDefaults = {
									type = "execute", order = 3, name = L["Restore Defaults"],
									desc = L["Reset standard colors for this bar group back to the current defaults."],
									disabled = function(info) return GetBarGroupField("useDefaultColors") end,
									func = function(info) local bg = GetBarGroupEntry()
										bg.buffColor = nil; bg.debuffColor = nil; bg.cooldownColor = nil; bg.notificationColor = nil
										bg.poisonColor = nil; bg.curseColor = nil; bg.magicColor = nil; bg.diseaseColor = nil
										MOD:UpdateAllBarGroups()
									end,
								},
								CopyFromGroup = {
									type = "select", order = 4, name = L["Copy From"],
									desc = L["Select bar group to copy standard colors from."],
									disabled = function(info) return GetBarGroupField("useDefaultColors") end,
									get = function(info) return nil end,
									set = function(info, value) CopyBarGroupStandardColors(GetBarGroupList()[value]) end,
									values = function(info) return GetBarGroupList() end,
									style = "dropdown",
								},
								Space0 = { type = "description", name = "", order = 5 },
								ColorText = { type = "description", name = L["Bar Colors:"], order = 7, width = "half" },
								NotificationColor = {
									type = "color", order = 10, name = L["Notify"], hasAlpha = false, width = "half",
									disabled = function(info) return GetBarGroupField("useDefaultColors") end,
									get = function(info) local t = GetBarGroupField("notificationColor") or MOD.db.global.DefaultNotificationColor
										return t.r, t.g, t.b, t.a end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("notificationColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("notificationColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
								BuffColor = {
									type = "color", order = 11, name = L["Buff"], hasAlpha = false, width = "half",
									disabled = function(info) return GetBarGroupField("useDefaultColors") end,
									get = function(info) local t = GetBarGroupField("buffColor") or MOD.db.global.DefaultBuffColor
										return t.r, t.g, t.b, t.a end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("buffColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("buffColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
								DebuffColor = {
									type = "color", order = 12, name = L["Debuff"], hasAlpha = false, width = "half",
									disabled = function(info) return GetBarGroupField("useDefaultColors") end,
									get = function(info) local t = GetBarGroupField("debuffColor") or MOD.db.global.DefaultDebuffColor
										return t.r, t.g, t.b, t.a end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("debuffColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("debuffColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
								CooldownColor = {
									type = "color", order = 13, name = L["Cooldown"], hasAlpha = false,
									disabled = function(info) return GetBarGroupField("useDefaultColors") end,
									get = function(info) local t = GetBarGroupField("cooldownColor") or MOD.db.global.DefaultCooldownColor
										return t.r, t.g, t.b, t.a end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("cooldownColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("cooldownColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
								Space1 = { type = "description", name = "", order = 14 },
								DebuffText = { type = "description", name = L["Debuff Colors:"], order = 15, width = "half" },
								PoisonColor = {
									type = "color", order = 15, name = L["Poison"], hasAlpha = false, width = "half",
									disabled = function(info) return GetBarGroupField("useDefaultColors") end,
									get = function(info) local t = GetBarGroupField("poisonColor") or MOD.db.global.DefaultPoisonColor
										return t.r, t.g, t.b, t.a end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("poisonColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("poisonColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
								CurseColor = {
									type = "color", order = 16, name = L["Curse"], hasAlpha = false, width = "half",
									disabled = function(info) return GetBarGroupField("useDefaultColors") end,
									get = function(info) local t = GetBarGroupField("curseColor") or MOD.db.global.DefaultCurseColor
										return t.r, t.g, t.b, t.a end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("curseColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("curseColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
								MagicColor = {
									type = "color", order = 17, name = L["Magic"], hasAlpha = false, width = "half",
									disabled = function(info) return GetBarGroupField("useDefaultColors") end,
									get = function(info) local t = GetBarGroupField("magicColor") or MOD.db.global.DefaultMagicColor
										return t.r, t.g, t.b, t.a end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("magicColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("magicColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
								DiseaseColor = {
									type = "color", order = 18, name = L["Disease"], hasAlpha = false, width = "half",
									disabled = function(info) return GetBarGroupField("useDefaultColors") end,
									get = function(info) local t = GetBarGroupField("diseaseColor") or MOD.db.global.DefaultDiseaseColor
										return t.r, t.g, t.b, t.a end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("diseaseColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("diseaseColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
							},
						},
						BarColorGroup = {
							type = "group", order = 40, name = L["Bar Color Scheme"], inline = true,
							args = {
								ForegroundText = { type = "description", name = L["Foreground:"], order = 1, width = "half" },
								StandardColors = {
									type = "toggle", order = 10, name = "Standard Colors",
									desc = L["Show bars in default colors for their type, including special debuff colors when applicable."],
									get = function(info) return GetBarGroupField("barColors") == "Standard" end,
									set = function(info, value) SetBarGroupField("barColors", "Standard") end,
								},
								CustomForeground = {
									type = "toggle", order = 15, name = L["Custom"], width = "half",
									desc = L["Color the bars with a custom color."],
									get = function(info) return GetBarGroupField("barColors") == "Custom" end,
									set = function(info, value) SetBarGroupField("barColors", "Custom") end,
								},
								ForegroundColor = {
									type = "color", order = 16, name = L["Color"], hasAlpha = false, width = "half",
									disabled = function(info) return GetBarGroupField("barColors") ~= "Custom" end,
									get = function(info)
										local t = GetBarGroupField("fgColor")
										if t then return t.r, t.g, t.b, t.a else return 1, 1, 1, 1 end
									end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("fgColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("fgColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
								ClassColors = {
									type = "toggle", order = 20, name = L["Class"], width = "half",
									desc = L["Show bars in class colors of associated spells when possible, otherwise use default bar colors."],
									get = function(info) return GetBarGroupField("barColors") == "Class" end,
									set = function(info, value) SetBarGroupField("barColors", "Class") end,
								},
								SpellColors = {
									type = "toggle", order = 30, name = L["Spell"], width = "half",
									desc = L["Show bars using spell colors when possible, otherwise use default bar colors."],
									get = function(info) return GetBarGroupField("barColors") == "Spell" end,
									set = function(info, value) SetBarGroupField("barColors", "Spell") end,
								},
								spacer1 = { type = "description", name = "", order = 40 },
								BackgroundText = { type = "description", name = L["Background:"], order = 41, width = "half" },
								NormalBackground = {
									type = "toggle", order = 50, name = L["Same as Foreground"],
									desc = L["Color the background the same as the foreground."],
									get = function(info) return GetBarGroupField("bgColors") == "Normal" end,
									set = function(info, value) SetBarGroupField("bgColors", "Normal") end,
								},
								CustomBackground = {
									type = "toggle", order = 60, name = L["Custom"], width = "half",
									desc = L["Color the background with a custom color."],
									get = function(info) return GetBarGroupField("bgColors") == "Custom" end,
									set = function(info, value) SetBarGroupField("bgColors", "Custom") end,
								},
								BackgroundColor = {
									type = "color", order = 70, name = L["Color"], hasAlpha = false, width = "half",
									disabled = function(info) return GetBarGroupField("bgColors") ~= "Custom" end,
									get = function(info)
										local t = GetBarGroupField("bgColor")
										if t then return t.r, t.g, t.b, t.a else return 1, 1, 1, 1 end
									end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("bgColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("bgColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
								spacer2 = { type = "description", name = "", order = 80 },
								IconBorderText = { type = "description", name = L["Icon Border:"], order = 81, width = "half" },
								NormalIcon = {
									type = "toggle", order = 85, name = L["Same as Foreground"],
									desc = L["Color the icon border the same as the bar foreground."],
									get = function(info) return GetBarGroupField("iconColors") == "Normal" end,
									set = function(info, value) SetBarGroupField("iconColors", "Normal") end,
								},
								CustomIcon = {
									type = "toggle", order = 86, name = L["Custom"], width = "half",
									desc = L["Color the icon border with a custom color."],
									get = function(info) return GetBarGroupField("iconColors") == "Custom" end,
									set = function(info, value) SetBarGroupField("iconColors", "Custom") end,
								},
								IconBorderColor = {
									type = "color", order = 87, name = L["Color"], hasAlpha = true, width = "half",
									disabled = function(info) return GetBarGroupField("iconColors") ~= "Custom" end,
									get = function(info)
										local t = GetBarGroupField("iconBorderColor")
										if t then return t.r, t.g, t.b, t.a else return 1, 1, 1, 1 end
									end,
									set = function(info, r, g, b, a)
										local t = GetBarGroupField("iconBorderColor")
										if t then t.r = r; t.g = g; t.b = b; t.a = a else
											t = { r = r, g = g, b = b, a = a }; SetBarGroupField("iconBorderColor", t) end
										MOD:UpdateAllBarGroups()
									end,
								},
								DebuffIcon = {
									type = "toggle", order = 90, name = L["Debuffs"], width = "half",
									desc = L["Color the icon border string"],
									get = function(info) return GetBarGroupField("iconColors") == "Debuffs" end,
									set = function(info, value) SetBarGroupField("iconColors", "Debuffs") end,
								},
								NoneIcon = {
									type = "toggle", order = 95, name = L["None"], width = "half",
									desc = L["Do not color the icon border."],
									get = function(info) return GetBarGroupField("iconColors") == "None" end,
									set = function(info, value) SetBarGroupField("iconColors", "None") end,
								},
							},
						},
					},
				},
				TimerOptionsTab = {
					type = "group", order = 50, name = L["Timer Options"],
					disabled = function(info) return InMode("Bar") end,
					hidden = function(info) return NoBarGroup() end,
					args = {
						DurationMaxGroup = {
							type = "group", order = 10, name = L["Show With Uniform Duration"], inline = true,
							args = {
								DurationCheck = {
									type = "toggle", order = 3, name = L["Enable"], width = "half",
									desc = L["Show timer bars scaled with a uniform duration (text still shows actual time left)."],
									get = function(info) return GetBarGroupField("setDuration") end,
									set = function(info, value) SetBarGroupField("setDuration", value) end,
								},
								DurationMinutes = {
									type = "range", order = 5, name = L["Minutes"], min = 0, max = 120, step = 1,
									desc = L["Enter minutes in the uniform duration."],
									disabled = function(info) return NoBarGroup() or not GetBarGroupField("setDuration") end,
									get = function(info) local d = GetBarGroupField("uniformDuration"); return math.floor(d / 60) end,
									set = function(info, value) local d = GetBarGroupField("uniformDuration"); SetBarGroupField("uniformDuration", (value * 60) + (d % 60)) end,
								},
								DurationSeconds = {
									type = "range", order = 7, name = L["Seconds"], min = 0, max = 59, step = 1,
									desc = L["Enter seconds in the uniform duration."],
									disabled = function(info) return NoBarGroup() or not GetBarGroupField("setDuration") end,
									get = function(info) local d = GetBarGroupField("uniformDuration"); return d % 60 end,
									set = function(info, value) local d = GetBarGroupField("uniformDuration"); SetBarGroupField("uniformDuration", value + (60 * math.floor(d / 60))) end,
								},
								DurationRange = {
									type = "description", order = 9,
									disabled = function(info) return not GetBarGroupField("setDuration") end, width = "half",
									name = function(info) local d = GetBarGroupField("uniformDuration"); return string.format("      %0d:%02d", math.floor(d / 60), d % 60) end,
								},
							},
						},
						DurationLimitGroup = {
							type = "group", order = 20, name = L["Show If Unlimited Duration"], inline = true,
							args = {
								LongDurationCheck = {
									type = "toggle", order = 1, name = L["Enable"], width = "half",
									desc = L["Show bars for actions with unlimited duration (e.g., paladin auras)."],
									get = function(info) return GetBarGroupField("showNoDuration") end,
									set = function(info, value) SetBarGroupField("showNoDuration", value) end,
								},
								LongDurationLimit = {
									type = "toggle", order = 10, name = L["Only Show Unlimited"],
									desc = L["Show bars for actions only if they have unlimited duration."],
									disabled = function(info) return not GetBarGroupField("showNoDuration") end,
									get = function(info) return GetBarGroupField("showOnlyNoDuration") end,
									set = function(info, value) SetBarGroupField("showOnlyNoDuration", value) end,
								},
								ForegroundBackground = {
									type = "toggle", order = 20, name = L["Full Bars"],
									desc = L["If checked, bars with unlimited duration show as full bars, otherwise they show as empty bars."],
									get = function(info) return not GetBarGroupField("showNoDurationBackground") end,
									set = function(info, value) SetBarGroupField("showNoDurationBackground", not value) end,
								},
							},
						},
						DurationGroup = {
							type = "group", order = 30, name = L["Check Overall Duration"], inline = true,
							args = {
								DurationCheck = {
									type = "toggle", order = 3, name = L["Enable"], width = "half",
									desc = L["Only include timer bars with a specified minimum (or maximum) duration."],
									get = function(info) return GetBarGroupField("checkDuration") end,
									set = function(info, value) SetBarGroupField("checkDuration", value) end,
								},
								DurationMinutes = {
									type = "range", order = 4, name = L["Minutes"], min = 0, max = 120, step = 1,
									desc = L["Enter minutes for overall duration check."],
									disabled = function(info) return NoBarGroup() or not GetBarGroupField("checkDuration") end,
									get = function(info) local d = GetBarGroupField("filterDuration"); return math.floor(d / 60) end,
									set = function(info, value) local d = GetBarGroupField("filterDuration"); SetBarGroupField("filterDuration", (value * 60) + (d % 60)) end,
								},
								DurationSeconds = {
									type = "range", order = 5, name = L["Seconds"], min = 0, max = 59, step = 1,
									desc = L["Enter seconds for overall duration check."],
									disabled = function(info) return NoBarGroup() or not GetBarGroupField("checkDuration") end,
									get = function(info) local d = GetBarGroupField("filterDuration"); return d % 60 end,
									set = function(info, value) local d = GetBarGroupField("filterDuration"); SetBarGroupField("filterDuration", value + (60 * math.floor(d / 60))) end,
								},
								DurationMinMax = {
									type = "select", order = 6, name = L["Duration"],
									disabled = function(info) return NoBarGroup() or not GetBarGroupField("checkDuration") end,
									get = function(info) if GetBarGroupField("minimumDuration") then return 1 else return 2 end end,
									set = function(info, value) if value == 1 then SetBarGroupField("minimumDuration", true) else SetBarGroupField("minimumDuration", false) end end,
									values = function(info)
										local d = GetBarGroupField("filterDuration")
										local ds = string.format("%0d:%02d", math.floor(d / 60), d % 60)
										return { "Show if " .. ds .. " or more", "Show if less than " .. ds }
									end,
									style = "dropdown",
								},
							},
						},
						TimeLeftGroup = {
							type = "group", order = 40, name = L["Check Time Left"], inline = true,
							disabled = function(info) return GetBarGroupField("showOnlyNoDuration") end,
							args = {
								TimeLeftCheck = {
									type = "toggle", order = 3, name = L["Enable"], width = "half",
									desc = L["Only show timer bars with a specified minimum (or maximum) time left."],
									get = function(info) return GetBarGroupField("checkTimeLeft") end,
									set = function(info, value) SetBarGroupField("checkTimeLeft", value) end,
								},
								TimeLeftMinutes= {
									type = "range", order = 4, name = L["Minutes"], min = 0, max = 120, step = 1,
									desc = L["Enter minutes for time left check."],
									disabled = function(info) return NoBarGroup() or not GetBarGroupField("checkTimeLeft") end,
									get = function(info) local d = GetBarGroupField("filterTimeLeft"); return math.floor(d / 60) end,
									set = function(info, value) local d = GetBarGroupField("filterTimeLeft"); SetBarGroupField("filterTimeLeft", (value * 60) + (d % 60)) end,
								},
								TimeLeftSeconds = {
									type = "range", order = 5, name = L["Seconds"], min = 0, max = 59, step = 1,
									desc = L["Enter seconds for time left check."],
									disabled = function(info) return NoBarGroup() or not GetBarGroupField("checkTimeLeft") end,
									get = function(info) local d = GetBarGroupField("filterTimeLeft"); return d % 60 end,
									set = function(info, value) local d = GetBarGroupField("filterTimeLeft"); SetBarGroupField("filterTimeLeft", value + (60 * math.floor(d / 60))) end,
								},
								TimeLeftMinMax = {
									type = "select", order = 6, name = L["Time Left"],
									disabled = function(info) return NoBarGroup() or not GetBarGroupField("checkTimeLeft") end,
									get = function(info) if GetBarGroupField("minimumTimeLeft") then return 1 else return 2 end end,
									set = function(info, value) if value == 1 then SetBarGroupField("minimumTimeLeft", true) else SetBarGroupField("minimumTimeLeft", false) end end,
									values = function(info)
										local d = GetBarGroupField("filterTimeLeft")
										local ds = string.format("%0d:%02d", math.floor(d / 60), d % 60)
										return { "Show if " .. ds .. " or more", "Show if less than " .. ds }
									end,
									style = "dropdown",
								},
							},
						},
						TimeFormatGroup = {
							type = "group", order = 50, name = L["Time Format"],  inline = true,
							args = {
								TimeFormat = {
									type = "select", order = 10, name = L["Options"], width = "double",
									desc = L["Time format string"],
									get = function(info) return GetBarGroupField("timeFormat") end,
									set = function(info, value) SetBarGroupField("timeFormat", value) end,
									values = function(info)
										local bg = GetBarGroupEntry()
										local s, c = bg.timeSpaces, bg.timeCase
										return GetTimeFormatList(s, c)
									end,
									style = "dropdown",
								},
								Space1 = { type = "description", name = "", order = 15, width = "half" },
								Spaces = {
									type = "toggle", order = 20, name = L["Spaces"], width = "half",
									desc = L["Include spaces between values in time format."],
									get = function(info) return GetBarGroupField("timeSpaces") end,
									set = function(info, value) SetBarGroupField("timeSpaces", value) end,
								},
								Capitals = {
									type = "toggle", order = 30, name = L["Uppercase"],
									desc = L["If checked, use uppercase H, M and S in time format, otherwise use lowercase."],
									get = function(info) return GetBarGroupField("timeCase") end,
									set = function(info, value) SetBarGroupField("timeCase", value) end,
								},
							},
						},
					},
				},
			},
		},
		Conditions = {
			type = "group", order = 30, name = L["Conditions"], childGroups = "tab",
			disabled = function(info) return InMode("BG") end,
			args = {
				SelectCondition = {
					type = "select", order = 1, name = L["Condition"],
					get = function(info) return GetSelectedCondition() end,
					set = function(info, value) SetSelectedCondition(value) end,
					disabled = function(info) return NoCondition() end,
					values = function(info) return GetConditionList() end,
					style = "dropdown",
				},
				Space1 = { type = "description", name = "", order = 2, width = "half" },
				NewConditionButton = {
					type = "execute", order = 3, name = L["New Condition"],
					desc = L["Create a new condition."],
					hidden = function(info) return enterNewCondition end,
					func = function(info) enterNewCondition, toggleNewCondition = true, true end,
				},
				NewConditionName = {
					type = "input", order = 4, name = L["Enter Condition Name"],
					desc = L["Enter name of new condition."],
					hidden = function(info) return not enterNewCondition end,
					validate = function(info, n) if not n or (n == "") then return "Invalid name." else return true end end,
					confirm = function(info, value) return ConfirmNewCondition(value) end,
					get = function(info)
						enterNewCondition = toggleNewCondition
						if toggleNewCondition then toggleNewCondition = false end
						if not enterNewCondition then MOD:UpdateOptions() end
						return false
					end,
					set = function(info, value) enterNewCondition = false; CreateCondition(value) end,
				},
				CancelNewCondition = {
					type = "execute", order = 5, name = L["Cancel"], width = "half",
					desc = L["Cancel creating a new condition."],
					hidden = function(info) return not enterNewCondition end,
					func = function(info) enterNewCondition, toggleNewCondition = false, false end,
				},
				DeleteCondition = {
					type = "execute", order = 8, name = L["Delete Condition"],
					desc = L["Delete the selected condition."],
					hidden = function(info) return enterNewCondition end,
					func = function(info) PurgeCondition(); DeleteCondition() end,
					confirm = function(info) return L["Delete condition string"](GetConditionField("name")) end,
				},
				GeneralTab = {
					type = "group", order = 20, name = L["General"],
					hidden = function(info) return NoCondition() end,
					args = {
						EnableGroup = {
							type = "group", order = 1, name = L["Enable"], inline = true,
							args = {
								EnableCondition = {
									type = "toggle", order = 1, name = L["Enable Condition"],
									desc = L["If checked, the condition is enabled and its value is determined by evaluating the associated tests, otherwise the condition is disabled and its value is always false."],
									get = function(info) return GetConditionField("enabled") end,
									set = function(info, value) SetConditionField("enabled", value) end,
								},
								EnableNotification = {
									type = "toggle", order = 10, name = L["Notification"],
									desc = L["If checked, this condition is suitable for a notification and will show up as an option for new notify bars."],
									get = function(info) return GetConditionField("notify") end,
									set = function(info, value) SetConditionField("notify", value) end,
								},
								EnableTooltips = {
									type = "toggle", order = 20, name = L["Detailed Tooltip"],
									desc = L["If checked, tooltip for notifications based on this condition will include a detailed description."],
									get = function(info) return GetConditionField("tooltip") end,
									set = function(info, value) SetConditionField("tooltip", value) end,
								},
							},
						},
						SharingGroup = {
							type = "group", order = 5, name = L["Share Condition"], inline = true,
							args = {
								EnableSharing = {
									type = "toggle", order = 5, name = L["Enable"],
									desc = L["If checked, this condition's settings can be copied between characters."],
									get = function(info) return GetConditionField("shared") end,
									set = function(info, value) SetConditionField("shared", value) end,
								},
								CopyConditionGroup = {
									type = "select", order = 10, name = L["Copy Condition Settings From"], width = "double",
									desc = L["Select a shared condition to copy settings from, including tests, associated spells, and dependencies."],
									get = function(info) return nil end,
									set = function(info, value) CopyConditionSettings(GetSharedConditionList()[value]) end,
									values = function(info) return GetSharedConditionList() end,
									style = "dropdown",
								},
								DeleteSharedConditions = {
									type = "execute", order = 20, name = L["Reset"], width = "half",
									desc = L["Reset and delete all current shared condition settings."],
									func = function(info) MOD.db.global.SharedConditions = {} end, -- clear the table recklessly
									confirm = function(info) return 'RESET\nAre you sure you want to delete all shared condition settings?' end,
								},
							},
						},
						RenameGroup = {
							type = "group", order = 20, name = L["Rename Condition"], inline = true,
							args = {
								Rename = {
									type = "input", order = 10, name = L["New Name"],
									validate = function(info, n) if not n or (n == "") then return L["Invalid name."] else return true end end,
									confirm = function(info, value) return ConfirmNewCondition(value) end,
									desc = L["Enter new name for the condition."],
									get = function(info) return GetConditionField("name") end,
									set = function(info, value) RenameCondition(value) end,
								},
							},
						},
						SpellGroup = {
							type = "group", order = 30, name = L["Associated Spell"], inline = true,
							args = {
								Rename = {
									type = "input", order = 10, name = L["Spell Name"],
									desc = L["Enter spell name (or numeric identifier) whose color and icon can be used by notification bars based on this condition."],
									get = function(info) return GetConditionField("associatedSpell") end,
									set = function(info, value)
										local n, t = value, tonumber(value)
										if t then n = GetSpellInfo(t) end
										ValidateSpellName(n)
										SetConditionField("associatedSpell", n)
									end,
								},
							},
						},
					},
				},
				TestsTab = {
					type = "group", order = 40, name = L["Tests"],
					hidden = function(info) return NoCondition() end,
					args = {
						SummaryGroup = {
							type = "group", order = 10, name = L["Summary"], inline = true,
							args = {
								ConditionDescription = {
									type = "description", order = 1, name = function(info) return GetConditionDescription() end,
								},
							},
						},
						PlayerStatusGroup = {
							type = "group", order = 20, name = L["Player Status"],
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = L["Enable Test"], inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, enable this test."],
											get = function(info) return GetTestField("Player Status", "enable") end,
											set = function(info, value) SetTestField("Player Status", "enable", value) end,
										},
									},
								},
								CombatGroup = {
									type = "group", order = 10, name = L["Combat"], inline = true,
									args = {
										CheckCombat = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, test the player's combat status."],
											get = function(info) return IsTestFieldOn("Player Status", "inCombat") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "inCombat", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = L["In Combat"],
											desc = L["If checked, player must be in combat."],
											disabled = function(info) return IsTestFieldOff("Player Status", "inCombat") end,
											get = function(info) return GetTestField("Player Status", "inCombat") == true end,
											set = function(info, value) SetTestField("Player Status", "inCombat", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["Out Of Combat"],
											desc = L["If checked, player must be out of combat."],
											disabled = function(info) return IsTestFieldOff("Player Status", "inCombat") end,
											get = function(info) return GetTestField("Player Status", "inCombat") == false end,
											set = function(info, value) SetTestField("Player Status", "inCombat", false) end,
										},
									},
								},
								RestingGroup = {
									type = "group", order = 20, name = L["Resting"], inline = true,
									args = {
										CheckResting = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, test if the player is resting."],
											get = function(info) return IsTestFieldOn("Player Status", "isResting") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "isResting", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = L["Is Resting"],
											desc = L["If checked, player must be resting (e.g., in an inn)."],
											disabled = function(info) return IsTestFieldOff("Player Status", "isResting") end,
											get = function(info) return GetTestField("Player Status", "isResting") == true end,
											set = function(info, value) SetTestField("Player Status", "isResting", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["Not Resting"],
											desc = L["If checked, player must not be resting."],
											disabled = function(info) return IsTestFieldOff("Player Status", "isResting") end,
											get = function(info) return GetTestField("Player Status", "isResting") == false end,
											set = function(info, value) SetTestField("Player Status", "isResting", false) end,
										},
									},
								},
								MountedGroup = {
									type = "group", order = 25, name = L["Mounted"], inline = true,
									args = {
										CheckMounted = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, test if the player is mounted."],
											get = function(info) return IsTestFieldOn("Player Status", "isMounted") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "isMounted", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = L["Is Mounted"],
											desc = L["If checked, player must be mounted."],
											disabled = function(info) return IsTestFieldOff("Player Status", "isMounted") end,
											get = function(info) return GetTestField("Player Status", "isMounted") == true end,
											set = function(info, value) SetTestField("Player Status", "isMounted", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["Not Mounted"],
											desc = L["If checked, player must not be mounted."],
											disabled = function(info) return IsTestFieldOff("Player Status", "isMounted") end,
											get = function(info) return GetTestField("Player Status", "isMounted") == false end,
											set = function(info, value) SetTestField("Player Status", "isMounted", false) end,
										},
									},
								},
								StealthedGroup = {
									type = "group", order = 30, name = L["Stealthed"], inline = true,
									args = {
										CheckStealthed = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, test if the player is stealthed."],
											get = function(info) return IsTestFieldOn("Player Status", "isStealthed") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "isStealthed", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = L["Is Stealthed"],
											desc = L["If checked, player must be stealthed."],
											disabled = function(info) return IsTestFieldOff("Player Status", "isStealthed") end,
											get = function(info) return GetTestField("Player Status", "isStealthed") == true end,
											set = function(info, value) SetTestField("Player Status", "isStealthed", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["Not Stealthed"],
											desc = L["If checked, player must not be stealthed."],
											disabled = function(info) return IsTestFieldOff("Player Status", "isStealthed") end,
											get = function(info) return GetTestField("Player Status", "isStealthed") == false end,
											set = function(info, value) SetTestField("Player Status", "isStealthed", false) end,
										},
									},
								},
								FishingGroup = {
									type = "group", order = 32, name = L["Fishing"], inline = true,
									args = {
										CheckFishing = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, test if the player is fishing."],
											get = function(info) return IsTestFieldOn("Player Status", "isFishing") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "isFishing", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = L["Is Fishing"],
											desc = L["If checked, player must be ready to fish (i.e., trained with fishing pole equipped)."],
											disabled = function(info) return IsTestFieldOff("Player Status", "isFishing") end,
											get = function(info) return GetTestField("Player Status", "isFishing") == true end,
											set = function(info, value) SetTestField("Player Status", "isFishing", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["Not Fishing"],
											desc = L["If checked, player must not be ready to fish."],
											disabled = function(info) return IsTestFieldOff("Player Status", "isFishing") end,
											get = function(info) return GetTestField("Player Status", "isFishing") == false end,
											set = function(info, value) SetTestField("Player Status", "isFishing", false) end,
										},
									},
								},
								PvPGroup = {
									type = "group", order = 33, name = L["PvP"], inline = true,
									args = {
										CheckFishing = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, test if the player has PvP enabled."],
											get = function(info) return IsTestFieldOn("Player Status", "isPvP") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "isPvP", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = L["Is PvP"],
											desc = L["If checked, player must have PvP enabled."],
											disabled = function(info) return IsTestFieldOff("Player Status", "isPvP") end,
											get = function(info) return GetTestField("Player Status", "isPvP") == true end,
											set = function(info, value) SetTestField("Player Status", "isPvP", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["Not PvP"],
											desc = L["If checked, player must not have PvP enabled."],
											disabled = function(info) return IsTestFieldOff("Player Status", "isPvP") end,
											get = function(info) return GetTestField("Player Status", "isPvP") == false end,
											set = function(info, value) SetTestField("Player Status", "isPvP", false) end,
										},
									},
								},
								InParty = {
									type = "group", order = 35, name = L["In Party"], inline = true,
									args = {
										CheckGroup = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, test if player is in a party."],
											get = function(info) return IsTestFieldOn("Player Status", "inParty") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "inParty", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = L["In Party"],
											desc = L["If checked, player must be in a party."],
											disabled = function(info) return IsTestFieldOff("Player Status", "inParty") end,
											get = function(info) return GetTestField("Player Status", "inParty") == true end,
											set = function(info, value) SetTestField("Player Status", "inParty", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["Not In Party"],
											desc = L["If checked, player must not be in a party."],
											disabled = function(info) return IsTestFieldOff("Player Status", "inParty") end,
											get = function(info) return GetTestField("Player Status", "inParty") == false end,
											set = function(info, value) SetTestField("Player Status", "inParty", false) end,
										},
									},
								},
								InRaid = {
									type = "group", order = 36, name = L["In Raid"], inline = true,
									args = {
										CheckGroup = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, test if player is in raid group."],
											get = function(info) return IsTestFieldOn("Player Status", "inRaid") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "inRaid", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = L["In Raid"],
											desc = L["If checked, player must be in a raid group."],
											disabled = function(info) return IsTestFieldOff("Player Status", "inRaid") end,
											get = function(info) return GetTestField("Player Status", "inRaid") == true end,
											set = function(info, value) SetTestField("Player Status", "inRaid", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["Not In Raid"],
											desc = L["If checked, player must not be in a raid group."],
											disabled = function(info) return IsTestFieldOff("Player Status", "inRaid") end,
											get = function(info) return GetTestField("Player Status", "inRaid") == false end,
											set = function(info, value) SetTestField("Player Status", "inRaid", false) end,
										},
									},
								},
								InGroup = {
									type = "group", order = 37, name = L["In Group"], inline = true,
									args = {
										CheckGroup = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, test if player is in either a party or raid with other players."],
											get = function(info) return IsTestFieldOn("Player Status", "inGroup") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "inGroup", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = L["In Group"],
											desc = L["If checked, player must be in either a party or raid with other players."],
											disabled = function(info) return IsTestFieldOff("Player Status", "inGroup") end,
											get = function(info) return GetTestField("Player Status", "inGroup") == true end,
											set = function(info, value) SetTestField("Player Status", "inGroup", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["Not In Group"],
											desc = L["If checked, player must not be in either a party or raid with other players."],
											disabled = function(info) return IsTestFieldOff("Player Status", "inGroup") end,
											get = function(info) return GetTestField("Player Status", "inGroup") == false end,
											set = function(info, value) SetTestField("Player Status", "inGroup", false) end,
										},
									},
								},
								InInstance = {
									type = "group", order = 40, name = L["In Instance"], inline = true,
									args = {
										CheckGroup = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, test if player is in an instance."],
											get = function(info) return IsTestFieldOn("Player Status", "inInstance") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "inInstance", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = L["In Instance"],
											desc = L["If checked, player must be in an instance."],
											disabled = function(info) return IsTestFieldOff("Player Status", "inInstance") end,
											get = function(info) return GetTestField("Player Status", "inInstance") == true end,
											set = function(info, value) SetTestField("Player Status", "inInstance", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["Not In Instance"],
											desc = L["If checked, player must not be in an instance."],
											disabled = function(info) return IsTestFieldOff("Player Status", "inInstance") end,
											get = function(info) return GetTestField("Player Status", "inInstance") == false end,
											set = function(info, value) SetTestField("Player Status", "inInstance", false) end,
										},
									},
								},
								PetGroup = {
									type = "group", order = 44, name = L["Pet"], inline = true,
									args = {
										CheckStealthed = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, test if the player has a pet."],
											get = function(info) return IsTestFieldOn("Player Status", "hasPet") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "hasPet", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = L["Has Pet"],
											desc = L["If checked, player must have a pet."],
											disabled = function(info) return IsTestFieldOff("Player Status", "hasPet") end,
											get = function(info) return GetTestField("Player Status", "hasPet") == true end,
											set = function(info, value) SetTestField("Player Status", "hasPet", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["No Pet"],
											desc = L["If checked, player must not have a pet"],
											disabled = function(info) return IsTestFieldOff("Player Status", "hasPet") end,
											get = function(info) return GetTestField("Player Status", "hasPet") == false end,
											set = function(info, value) SetTestField("Player Status", "hasPet", false) end,
										},
									},
								},
								CheckLevelGroup = {
									type = "group", order = 45, name = L["Level"], inline = true,
									args = {
										CheckLevelEnable = {
											type = "toggle", order = 1, name = L["Enable"], width = "half",
											desc = L["If checked, test the player's level."],
											get = function(info) return IsTestFieldOn("Player Status", "checkLevel") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "checkLevel", v) end,
										},
										CheckLevel = {
											type = "toggle", order = 2, name = L["Minimum"],
											desc = L["If checked, player must be at least at this level, otherwise must be lower."],
											disabled = function(info) return IsTestFieldOff("Player Status", "checkLevel") end,
											get = function(info) return GetTestField("Player Status", "checkLevel") == true end,
											set = function(info, value) SetTestField("Player Status", "checkLevel", value) end,
										},
										LevelRange = {
											type = "range", order = 3, name = "", min = 1, max = 80, step = 1,
											disabled = function(info) return IsTestFieldOff("Player Status", "checkLevel") end,
											get = function(info) return GetTestField("Player Status", "level") end,
											set = function(info, value) SetTestField("Player Status", "level", value) end,
										},
									},
								},
								CheckHealthGroup = {
									type = "group", order = 50, name = L["Health"], inline = true, 
									args = {
										CheckHealthEnable = {
											type = "toggle", order = 1, name = L["Enable"], width = "half",
											desc = L["If checked, test the player's health."],
											get = function(info) return IsTestFieldOn("Player Status", "checkHealth") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "checkHealth", v) end,
										},
										CheckHealth = {
											type = "toggle", order = 2, name = L["Minimum"],
											desc = L["If checked, player's health must be at least this percentage, otherwise must be less."],
											disabled = function(info) return IsTestFieldOff("Player Status", "checkHealth") end,
											get = function(info) return GetTestField("Player Status", "checkHealth") == true end,
											set = function(info, value) SetTestField("Player Status", "checkHealth", value) end,
										},
										HealthRange = {
											type = "range", order = 3, name = "", min = 1, max = 100, step = 1,
											disabled = function(info) return IsTestFieldOff("Player Status", "checkHealth") end,
											get = function(info) return GetTestField("Player Status", "minHealth") end,
											set = function(info, value) SetTestField("Player Status", "minHealth", value) end,
										},
									},
								},
								CheckPowerGroup = {
									type = "group", order = 70, name = L["Power"], inline = true, 
									args = {
										CheckPowerEnable = {
											type = "toggle", order = 1, name = L["Enable"], width = "half",
											desc = L["If checked, test the player's power (i.e., mana, rage, energy, focus, runic power)."],
											get = function(info) return IsTestFieldOn("Player Status", "checkPower") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "checkPower", v) end,
										},
										CheckPower = {
											type = "toggle", order = 2, name = L["Minimum"],
											desc = L["If checked, player's power must be at least this percentage, otherwise must be less."],
											disabled = function(info) return IsTestFieldOff("Player Status", "checkPower") end,
											get = function(info) return GetTestField("Player Status", "checkPower") == true end,
											set = function(info, value) SetTestField("Player Status", "checkPower", value) end,
										},
										PowerRange = {
											type = "range", order = 3, name = "", min = 1, max = 100, step = 1,
											disabled = function(info) return IsTestFieldOff("Player Status", "checkPower") end,
											get = function(info) return GetTestField("Player Status", "minPower") end,
											set = function(info, value) SetTestField("Player Status", "minPower", value) end,
										},
									},
								},
								CheckHolyPowerGroup = {
									type = "group", order = 71, name = L["Holy Power"], inline = true, 
									args = {
										CheckPowerEnable = {
											type = "toggle", order = 1, name = L["Enable"], width = "half",
											desc = L["If checked, test the player's holy power."],
											get = function(info) return IsTestFieldOn("Player Status", "checkHolyPower") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "checkHolyPower", v) end,
										},
										CheckPower = {
											type = "toggle", order = 2, name = L["Minimum"],
											desc = L["If checked, player's holy power must be at least this many charges, otherwise must be less."],
											disabled = function(info) return IsTestFieldOff("Player Status", "checkHolyPower") end,
											get = function(info) return GetTestField("Player Status", "checkHolyPower") == true end,
											set = function(info, value) SetTestField("Player Status", "checkHolyPower", value) end,
										},
										PowerRange = {
											type = "range", order = 3, name = "", min = 1, max = 3, step = 1,
											disabled = function(info) return IsTestFieldOff("Player Status", "checkHolyPower") end,
											get = function(info) return GetTestField("Player Status", "minHolyPower") end,
											set = function(info, value) SetTestField("Player Status", "minHolyPower", value) end,
										},
									},
								},
								CheckShardsGroup = {
									type = "group", order = 72, name = L["Soul Shards"], inline = true, 
									args = {
										CheckShardsEnable = {
											type = "toggle", order = 1, name = L["Enable"], width = "half",
											desc = L["If checked, test the player's number of soul shards."],
											get = function(info) return IsTestFieldOn("Player Status", "checkShards") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "checkShards", v) end,
										},
										CheckShards = {
											type = "toggle", order = 2, name = L["Minimum"],
											desc = L["If checked, player's holy power must be at least this many charges, otherwise must be less."],
											disabled = function(info) return IsTestFieldOff("Player Status", "checkShards") end,
											get = function(info) return GetTestField("Player Status", "checkShards") == true end,
											set = function(info, value) SetTestField("Player Status", "checkShards", value) end,
										},
										ShardsRange = {
											type = "range", order = 3, name = "", min = 1, max = 3, step = 1,
											disabled = function(info) return IsTestFieldOff("Player Status", "checkShards") end,
											get = function(info) return GetTestField("Player Status", "minShards") end,
											set = function(info, value) SetTestField("Player Status", "minShards", value) end,
										},
									},
								},
								CheckEclipsePower = {
									type = "group", order = 74, name = L["Eclipse Power"], inline = true, 
									args = {
										CheckEclipseEnable = {
											type = "toggle", order = 1, name = L["Enable"], width = "half",
											desc = L["If checked, test the player's eclipse power. You can set power level (-100 to -1 for lunar power, 1 to 100 for solar power) and the comparison to use (either less or greater than the power level)."],
											get = function(info) return IsTestFieldOn("Player Status", "checkEclipsePower") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "checkEclipsePower", v) end,
										},
										CheckEclipse = {
											type = "toggle", order = 10, name = L["Minimum"],
											desc = L["If checked, player's eclipse power must be at least this level, otherwise must be less."],
											disabled = function(info) return IsTestFieldOff("Player Status", "checkEclipsePower") end,
											get = function(info) return GetTestField("Player Status", "checkEclipsePower") == true end,
											set = function(info, value) SetTestField("Player Status", "checkEclipsePower", value) end,
										},
										EclipseRange = {
											type = "range", order = 20, name = "", min = -100, max = 100, step = 1,
											disabled = function(info) return IsTestFieldOff("Player Status", "checkEclipsePower") end,
											get = function(info) return GetTestField("Player Status", "minEclipsePower") end,
											set = function(info, value) SetTestField("Player Status", "minEclipsePower", value) end,
										},
									},
								},
								CheckEclipseGroup = {
									type = "group", order = 74, name = L["Eclipse State"], inline = true, 
									args = {
										CheckEclipseEnable = {
											type = "toggle", order = 1, name = L["Enable"], width = "half",
											desc = L["If checked, test the player's eclipse status. You can test current eclipse direction (toward sun or moon) and whether lunar or solar eclipse is active."],
											get = function(info) return IsTestFieldOn("Player Status", "checkEclipse") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "checkEclipse", v) end,
										},
										SolarEclipse = {
											type = "toggle", order = 40, name = L["Solar"], width = "half",
											desc = L["If checked, player must be in Solar Eclipse."],
											disabled = function(info) return IsTestFieldOff("Player Status", "checkEclipse") end,
											get = function(info) return GetTestField("Player Status", "checkSolar") end,
											set = function(info, value) SetTestField("Player Status", "checkSolar", value) end,
										},
										LunarEclipse = {
											type = "toggle", order = 50, name = L["Lunar"], width = "half",
											desc = L["If checked, player must be in Lunar Eclipse."],
											disabled = function(info) return IsTestFieldOff("Player Status", "checkEclipse") end,
											get = function(info) return GetTestField("Player Status", "checkLunar") end,
											set = function(info, value) SetTestField("Player Status", "checkLunar", value) end,
										},
										EclipseSun = {
											type = "toggle", order = 60, name = L[">> Sun"], width = "half",
											desc = L["If checked, eclipse direction must be going toward the sun."],
											disabled = function(info) return IsTestFieldOff("Player Status", "checkEclipse") end,
											get = function(info) return GetTestField("Player Status", "checkSun") end,
											set = function(info, value) SetTestField("Player Status", "checkSun", value) end,
										},
										EclipseMoon = {
											type = "toggle", order = 70, name = L[">> Moon"], width = "half",
											desc = L["If checked, eclipse direction must be going toward the moon."],
											disabled = function(info) return IsTestFieldOff("Player Status", "checkEclipse") end,
											get = function(info) return GetTestField("Player Status", "checkMoon") end,
											set = function(info, value) SetTestField("Player Status", "checkMoon", value) end,
										},
									},
								},
								CheckRunesGroup = {
									type = "group", order = 75, name = L["Runes"], inline = true,
									args = {
										CheckRunes = {
											type = "toggle", order = 10, name = L["Enable"], width = "half",
											desc = L["If checked, test the player's rune status."],
											get = function(info) return IsTestFieldOn("Player Status", "checkRunes") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "checkRunes", v) end,
										},
										CheckAvailableGroup = {
											type = "group", order = 30, name = L["Available"], inline = true,
											args = {
												DoBlood = {
													type = "toggle", order = 20, name = L["Blood"], width = "half",
													desc = L["If checked, must have at least one blood rune available (or enough death runes)."],
													disabled = function(info) return IsTestFieldOff("Player Status", "checkRunes") end,
													get = function(info) return GetTestField("Player Status", "needBlood") == true end,
													set = function(info, value) SetTestField("Player Status", "needBlood", value) end,
												},
												DoFrost = {
													type = "toggle", order = 30, name = L["Frost"], width = "half",
													desc = L["If checked, must have at least one frost rune available (or enough death runes)."],
													disabled = function(info) return IsTestFieldOff("Player Status", "checkRunes") end,
													get = function(info) return GetTestField("Player Status", "needFrost") == true end,
													set = function(info, value) SetTestField("Player Status", "needFrost", value) end,
												},
												DoUnholy = {
													type = "toggle", order = 40, name = L["Unholy"], width = "half",
													desc = L["If checked, must have at least one unholy rune available (or enough death runes)."],
													disabled = function(info) return IsTestFieldOff("Player Status", "checkRunes") end,
													get = function(info) return GetTestField("Player Status", "needUnholy") == true end,
													set = function(info, value) SetTestField("Player Status", "needUnholy", value) end,
												},
												DoAny = {
													type = "toggle", order = 50, name = L["Any"], width = "half",
													desc = L["If checked, at least one rune of any type must be available."],
													disabled = function(info) return IsTestFieldOff("Player Status", "checkRunes") end,
													get = function(info) return GetTestField("Player Status", "checkAnyRune") == true end,
													set = function(info, value) SetTestField("Player Status", "checkAnyRune", value) end,
												},
											},
										},
										CheckRechargeGroup = {
											type = "group", order = 40, name = L["Recharging"], inline = true,
											disabled = function(info) return IsTestFieldOff("Player Status", "checkRunes") end,
											args = {
												CheckBlood = {
													type = "toggle", order = 10, name = L["Blood"], width = "half",
													desc = L["If checked, test blood runes."],
													get = function(info) return GetTestField("Player Status", "checkBlood") end,
													set = function(info, v) SetTestField("Player Status", "checkBlood", v) end,
												},
												BloodRecharge = {
													type = "toggle", order = 15, name = L[">= 1"], width = "half",
													desc = L["If checked, at least one blood rune must be recharging, otherwise none."],
													disabled = function(info) return IsTestFieldOff("Player Status", "checkRunes") or not GetTestField("Player Status", "checkBlood") end,
													get = function(info) return GetTestField("Player Status", "chargeBlood") == true end,
													set = function(info, value) SetTestField("Player Status", "chargeBlood", value) end,
												},
												CheckFrost = {
													type = "toggle", order = 20, name = L["Frost"], width = "half",
													desc = L["If checked, test frost runes."],
													get = function(info) return GetTestField("Player Status", "checkFrost") end,
													set = function(info, v) SetTestField("Player Status", "checkFrost", v) end,
												},
												FrostRecharge = {
													type = "toggle", order = 25, name = L[">= 1"], width = "half",
													desc = L["If checked, at least one frost rune must be recharging, otherwise none."],
													disabled = function(info) return IsTestFieldOff("Player Status", "checkRunes") or not GetTestField("Player Status", "checkFrost") end,
													get = function(info) return GetTestField("Player Status", "chargeFrost") == true end,
													set = function(info, value) SetTestField("Player Status", "chargeFrost", value) end,
												},
												CheckUnholy = {
													type = "toggle", order = 30, name = L["Unholy"], width = "half",
													desc = L["If checked, test unholy runes."],
													get = function(info) return GetTestField("Player Status", "checkUnholy") end,
													set = function(info, v) SetTestField("Player Status", "checkUnholy", v) end,
												},
												UnholyRecharge = {
													type = "toggle", order = 35, name = L[">= 1"], width = "half",
													desc = L["If checked, at least one unholy rune must be recharging, otherwise none."],
													disabled = function(info) return IsTestFieldOff("Player Status", "checkRunes") or not GetTestField("Player Status", "checkUnholy") end,
													get = function(info) return GetTestField("Player Status", "chargeUnholy") == true end,
													set = function(info, value) SetTestField("Player Status", "chargeUnholy", value) end,
												},
											},
										},
									},
								},
								CheckComboPointsGroup = {
									type = "group", order = 80, name = L["Combo Points"], inline = true, 
									args = {
										CheckComboPointsEnable = {
											type = "toggle", order = 1, name = L["Enable"], width = "half",
											desc = L["If checked, test how many combo points the player has on the target."],
											get = function(info) return IsTestFieldOn("Player Status", "checkComboPoints") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "checkComboPoints", v) end,
										},
										CheckComboPoints = {
											type = "toggle", order = 2, name = L["Minimum"],
											desc = L["If checked, player must have at least this many combo points, otherwise must be fewer."],
											disabled = function(info) return IsTestFieldOff("Player Status", "checkComboPoints") end,
											get = function(info) return GetTestField("Player Status", "checkComboPoints") == true end,
											set = function(info, value) SetTestField("Player Status", "checkComboPoints", value) end,
										},
										ComboPointsRange = {
											type = "range", order = 3, name = "", min = 1, max = 5, step = 1,
											disabled = function(info) return IsTestFieldOff("Player Status", "checkComboPoints") end,
											get = function(info) return GetTestField("Player Status", "minComboPoints") end,
											set = function(info, value) SetTestField("Player Status", "minComboPoints", value) end,
										},
									},
								},
								CheckWeaponGroup = {
									type = "group", order = 85, name = L["Weapons"], inline = true, 
									args = {
										CheckMainHandEnable = {
											type = "toggle", order = 1, name = L["Mainhand"], width = "half",
											desc = L["If checked, test if the player has a mainhand weapon equipped with at least the specified item level."],
											get = function(info) return IsTestFieldOn("Player Status", "hasMainHand") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "hasMainHand", v) end,
										},
										MainHandLevel = {
											type = "range", order = 2, name = "", min = 1, max = 500, step = 1,
											disabled = function(info) return IsTestFieldOff("Player Status", "hasMainHand") end,
											get = function(info)
												local level = GetTestField("Player Status", "levelMainHand")
												if not level then level = 1; SetTestField("Player Status", "levelMainHand", level) end
												return level
											end,
											set = function(info, value) SetTestField("Player Status", "levelMainHand", value) end,
										},
										CheckOffHandEnable = {
											type = "toggle", order = 3, name = L["Offhand"], width = "half",
											desc = L["If checked, test if the player has an offhand weapon equipped with at least the specified item level."],
											get = function(info) return IsTestFieldOn("Player Status", "hasOffHand") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "hasOffHand", v) end,
										},
										OffHandLevel = {
											type = "range", order = 4, name = "", min = 1, max = 500, step = 1,
											disabled = function(info) return IsTestFieldOff("Player Status", "hasOffHand") end,
											get = function(info)
												local level = GetTestField("Player Status", "levelOffHand")
												if not level then level = 1; SetTestField("Player Status", "levelOffHand", level) end
												return level
											end,
											set = function(info, value) SetTestField("Player Status", "levelOffHand", value) end,
										},
										CheckRangedEnable = {
											type = "toggle", order = 5, name = L["Ranged"], width = "half",
											desc = L["If checked, test if the player has an ranged weapon equipped with at least the specified item level."],
											get = function(info) return IsTestFieldOn("Player Status", "hasRanged") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "hasRanged", v) end,
										},
										RangedLevel = {
											type = "range", order = 6, name = "", min = 1, max = 500, step = 1,
											disabled = function(info) return IsTestFieldOff("Player Status", "hasRanged") end,
											get = function(info)
												local level = GetTestField("Player Status", "levelRanged")
												if not level then level = 1; SetTestField("Player Status", "levelRanged", level) end
												return level
											end,
											set = function(info, value) SetTestField("Player Status", "levelRanged", value) end,
										},
									},
								},
								CheckStanceGroup = {
									type = "group", order = 90, name = L["Stance"], inline = true, 
									args = {
										CheckStanceEnable = {
											type = "toggle", order = 10, name = L["Enable"], width = "half",
											desc = L["If checked, test if the player is in a stance."],
											get = function(info) return IsTestFieldOn("Player Status", "checkStance") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "checkStance", v) end,
										},
										StanceEntry = {
											type = "input", order = 20, name = L["Stance"],
											desc = L['Enter the stance to check for (enter "none" to check for no stance).'],
											get = function(info) return GetTestField("Player Status", "stance") end,
											set = function(info, value) SetTestFieldString("Player Status", "stance", value) end,
										},
									},
								},
								CheckTalentGroup = {
									type = "group", order = 95, name = L["Talent"], inline = true, 
									args = {
										CheckTalentsEnable = {
											type = "toggle", order = 10, name = L["Enable"], width = "half",
											desc = L["If checked, test if points have been spent in one of the player's talents."],
											get = function(info) return IsTestFieldOn("Player Status", "checkTalent") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Player Status", "checkTalent", v) end,
										},
										SelectTalent = {
											type = "select", order = 20, name = L["Talent"],
											get = function(info) local v = GetTestField("Player Status", "talent"); if v then local t = MOD.talents[v]; if t then return t.select end end return nil end,
											set = function(info, value) SetTestField("Player Status", "talent", MOD.talentList[value]) end,
											values = function(info) return MOD.talentList end,
											style = "dropdown",
										},
									},
								},
							},
						},
						PetStatusGroup = {
							type = "group", order = 35, name = L["Pet Status"],
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = L["Enable Test"], inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, enable this test."],
											get = function(info) return GetTestField("Pet Status", "enable") end,
											set = function(info, value) SetTestField("Pet Status", "enable", value) end,
										},
									},
								},
								CheckHealthGroup = {
									type = "group", order = 50, name = L["Health"], inline = true, 
									args = {
										CheckHealthEnable = {
											type = "toggle", order = 1, name = L["Enable"], width = "half",
											desc = L["If checked, test the pet's health."],
											get = function(info) return IsTestFieldOn("Pet Status", "checkHealth") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Pet Status", "checkHealth", v) end,
										},
										CheckHealth = {
											type = "toggle", order = 2, name = L["Minimum"],
											desc = L["If checked, pet's health must be at least this percentage, otherwise must be less."],
											disabled = function(info) return IsTestFieldOff("Pet Status", "checkHealth") end,
											get = function(info) return GetTestField("Pet Status", "checkHealth") == true end,
											set = function(info, value) SetTestField("Pet Status", "checkHealth", value) end,
										},
										HealthRange = {
											type = "range", order = 3, name = "", min = 1, max = 100, step = 1,
											disabled = function(info) return IsTestFieldOff("Pet Status", "checkHealth") end,
											get = function(info) return GetTestField("Pet Status", "minHealth") end,
											set = function(info, value) SetTestField("Pet Status", "minHealth", value) end,
										},
									},
								},
								CheckPowerGroup = {
									type = "group", order = 70, name = L["Power"], inline = true, 
									args = {
										CheckPowerEnable = {
											type = "toggle", order = 1, name = L["Enable"], width = "half",
											desc = L["If checked, test the pet's power (i.e., mana, rage, energy, focus)."],
											get = function(info) return IsTestFieldOn("Pet Status", "checkPower") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Pet Status", "checkPower", v) end,
										},
										CheckPower = {
											type = "toggle", order = 2, name = L["Minimum"],
											desc = L["If checked, pet's power must be at least this percentage, otherwise must be less."],
											disabled = function(info) return IsTestFieldOff("Pet Status", "checkPower") end,
											get = function(info) return GetTestField("Pet Status", "checkPower") == true end,
											set = function(info, value) SetTestField("Pet Status", "checkPower", value) end,
										},
										PowerRange = {
											type = "range", order = 3, name = "", min = 1, max = 100, step = 1,
											disabled = function(info) return IsTestFieldOff("Pet Status", "checkPower") end,
											get = function(info) return GetTestField("Pet Status", "minPower") end,
											set = function(info, value) SetTestField("Pet Status", "minPower", value) end,
										},
									},
								},
							},
						},
						TargetStatusGroup = {
							type = "group", order = 35, name = L["Target Status"],
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = L["Enable Test"], inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, enable this test."],
											get = function(info) return GetTestField("Target Status", "enable") end,
											set = function(info, value) SetTestField("Target Status", "enable", value) end,
										},
									},
								},
								EnemyGroup = {
									type = "group", order = 10, name = L["Enemy"], inline = true,
									args = {
										CheckEnemy = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, test if the target is an enemy."],
											get = function(info) return IsTestFieldOn("Target Status", "isEnemy") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Target Status", "isEnemy", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = L["Is Enemy"],
											desc = L["If checked, target must be an enemy."],
											disabled = function(info) return IsTestFieldOff("Target Status", "isEnemy") end,
											get = function(info) return GetTestField("Target Status", "isEnemy") == true end,
											set = function(info, value) SetTestField("Target Status", "isEnemy", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["Not Enemy"],
											desc = L["If checked, target must not be an enemy."],
											disabled = function(info) return IsTestFieldOff("Target Status", "isEnemy") end,
											get = function(info) return GetTestField("Target Status", "isEnemy") == false end,
											set = function(info, value) SetTestField("Target Status", "isEnemy", false) end,
										},
									},
								},
								FriendGroup = {
									type = "group", order = 20, name = L["Friendly"], inline = true,
									args = {
										CheckFriend = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, test if the target is friendly."],
											get = function(info) return IsTestFieldOn("Target Status", "isFriend") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Target Status", "isFriend", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = L["Is Friendly"],
											desc = L["If checked, target must be friendly."],
											disabled = function(info) return IsTestFieldOff("Target Status", "isFriend") end,
											get = function(info) return GetTestField("Target Status", "isFriend") == true end,
											set = function(info, value) SetTestField("Target Status", "isFriend", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["Not Friendly"],
											desc = L["If checked, target must not be friendly."],
											disabled = function(info) return IsTestFieldOff("Target Status", "isFriend") end,
											get = function(info) return GetTestField("Target Status", "isFriend") == false end,
											set = function(info, value) SetTestField("Target Status", "isFriend", false) end,
										},
									},
								},
								DeadGroup = {
									type = "group", order = 22, name = L["Dead"], inline = true,
									args = {
										CheckDead = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, test if the target is dead."],
											get = function(info) return IsTestFieldOn("Target Status", "isDead") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Target Status", "isDead", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = L["Is Dead"],
											desc = L["If checked, target must be dead."],
											disabled = function(info) return IsTestFieldOff("Target Status", "isDead") end,
											get = function(info) return GetTestField("Target Status", "isDead") == true end,
											set = function(info, value) SetTestField("Target Status", "isDead", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["Not Dead"],
											desc = L["If checked, target must not be dead."],
											disabled = function(info) return IsTestFieldOff("Target Status", "isDead") end,
											get = function(info) return GetTestField("Target Status", "isDead") == false end,
											set = function(info, value) SetTestField("Target Status", "isDead", false) end,
										},
									},
								},
								Classification = {
									type = "group", order = 25, name = L["Classification"], inline = true,
									args = {
										Enable = {
											type = "toggle", order = 1, name = L["Enable"], width = "half",
											desc = L["If checked, test the target's classification."],
											get = function(info) return GetTestField("Target Status", "classify") end,
											set = function(info, value) SetTestField("Target Status", "classify", value) end,
										},
										spacer1 = { type = "description", name = "", order = 10 },
										Normal = {
											type = "toggle", order = 20, name = L["Normal"], width = "half",
											desc = L["If checked, target must be normal."],
											disabled = function(info) return not GetTestField("Target Status", "classify") end,
											get = function(info) return GetTestField("Target Status", "classification") == "normal" end,
											set = function(info, value) SetTestField("Target Status", "classification", "normal") end,
										},
										Boss = {
											type = "toggle", order = 21, name = L["Boss"], width = "half",
											desc = L["If checked, target must be a boss."],
											disabled = function(info) return not GetTestField("Target Status", "classify") end,
											get = function(info) return GetTestField("Target Status", "classification") == "worldboss" end,
											set = function(info, value) SetTestField("Target Status", "classification", "worldboss") end,
										},
										Rare = {
											type = "toggle", order = 22, name = L["Rare"], width = "half",
											desc = L["If checked, target must be rare."],
											disabled = function(info) return not GetTestField("Target Status", "classify") end,
											get = function(info) return GetTestField("Target Status", "classification") == "rare" end,
											set = function(info, value) SetTestField("Target Status", "classification", "rare") end,
										},
										Elite = {
											type = "toggle", order = 23, name = L["Elite"], width = "half",
											desc = L["If checked, target must be elite."],
											disabled = function(info) return not GetTestField("Target Status", "classify") end,
											get = function(info) return GetTestField("Target Status", "classification") == "elite" end,
											set = function(info, value) SetTestField("Target Status", "classification", "elite") end,
										},
										RareElite = {
											type = "toggle", order = 25, name = L["Rare Elite"],
											desc = L["If checked, target must be rare elite."],
											disabled = function(info) return not GetTestField("Target Status", "classify") end,
											get = function(info) return GetTestField("Target Status", "classification") == "rareelite" end,
											set = function(info, value) SetTestField("Target Status", "classification", "rareelite") end,
										},
									},
								},
								RangeGroup = {
									type = "group", order = 30, name = L["Range"], inline = true,
									args = {
										CheckRange = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, test if the target is in range."],
											get = function(info) return IsTestFieldOn("Target Status", "inRange") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Target Status", "inRange", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = L["In Range"],
											desc = L["If checked, target must be in range."],
											disabled = function(info) return IsTestFieldOff("Target Status", "inRange") end,
											get = function(info) return GetTestField("Target Status", "inRange") == true end,
											set = function(info, value) SetTestField("Target Status", "inRange", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["Out Of Range"],
											desc = L["If checked, target must not be in range."],
											disabled = function(info) return IsTestFieldOff("Target Status", "inRange") end,
											get = function(info) return GetTestField("Target Status", "inRange") == false end,
											set = function(info, value) SetTestField("Target Status", "inRange", false) end,
										},
									},
								},
								StealableGroup = {
									type = "group", order = 35, name = L["Spellsteal"], inline = true,
									args = {
										CheckSteal = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, test for a buff that can be transferred to the player with Spellsteal."],
											get = function(info) return IsTestFieldOn("Target Status", "isSteal") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Target Status", "isSteal", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = L["Spellsteal"],
											desc = L["If checked, there must be a spellstealable buff."],
											disabled = function(info) return IsTestFieldOff("Target Status", "isSteal") end,
											get = function(info) return GetTestField("Target Status", "isSteal") == true end,
											set = function(info, value) SetTestField("Target Status", "isSteal", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["Not Spellsteal"],
											desc = L["If checked, there must not be a spellstealable buff."],
											disabled = function(info) return IsTestFieldOff("Target Status", "isSteal") end,
											get = function(info) return GetTestField("Target Status", "isSteal") == false end,
											set = function(info, value) SetTestField("Target Status", "isSteal", false) end,
										},
									},
								},
								CheckHealthGroup = {
									type = "group", order = 50, name = L["Health"], inline = true, 
									args = {
										CheckHealthEnable = {
											type = "toggle", order = 1, name = L["Enable"], width = "half",
											desc = L["If checked, test the target's health."],
											get = function(info) return IsTestFieldOn("Target Status", "checkHealth") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Target Status", "checkHealth", v) end,
										},
										CheckHealth = {
											type = "toggle", order = 2, name = L["Minimum"],
											desc = L["If checked, target's health must be at least this percentage, otherwise must be less."],
											disabled = function(info) return IsTestFieldOff("Target Status", "checkHealth") end,
											get = function(info) return GetTestField("Target Status", "checkHealth") == true end,
											set = function(info, value) SetTestField("Target Status", "checkHealth", value) end,
										},
										HealthRange = {
											type = "range", order = 3, name = "", min = 1, max = 100, step = 1,
											disabled = function(info) return IsTestFieldOff("Target Status", "checkHealth") end,
											get = function(info) return GetTestField("Target Status", "minHealth") end,
											set = function(info, value) SetTestField("Target Status", "minHealth", value) end,
										},
									},
								},
								CheckPowerGroup = {
									type = "group", order = 70, name = L["Power"], inline = true, 
									args = {
										CheckPowerEnable = {
											type = "toggle", order = 1, name = L["Enable"], width = "half",
											desc = L["If checked, test the target's power (i.e., mana, rage, energy, focus, runic power)."],
											get = function(info) return IsTestFieldOn("Target Status", "checkPower") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Target Status", "checkPower", v) end,
										},
										CheckPower = {
											type = "toggle", order = 2, name = L["Minimum"],
											desc = L["If checked, target's power must be at least this percentage, otherwise must be less."],
											disabled = function(info) return IsTestFieldOff("Target Status", "checkPower") end,
											get = function(info) return GetTestField("Target Status", "checkPower") == true end,
											set = function(info, value) SetTestField("Target Status", "checkPower", value) end,
										},
										PowerRange = {
											type = "range", order = 3, name = "", min = 1, max = 100, step = 1,
											disabled = function(info) return IsTestFieldOff("Target Status", "checkPower") end,
											get = function(info) return GetTestField("Target Status", "minPower") end,
											set = function(info, value) SetTestField("Target Status", "minPower", value) end,
										},
									},
								},
							},
						},
						FocusStatusGroup = {
							type = "group", order = 40, name = L["Focus Status"],
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = L["Enable Test"], inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, enable this test."],
											get = function(info) return GetTestField("Focus Status", "enable") end,
											set = function(info, value) SetTestField("Focus Status", "enable", value) end,
										},
									},
								},
								EnemyGroup = {
									type = "group", order = 10, name = L["Enemy"], inline = true,
									args = {
										CheckEnemy = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, test if the focus is an enemy."],
											get = function(info) return IsTestFieldOn("Focus Status", "isEnemy") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Focus Status", "isEnemy", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = L["Is Enemy"],
											desc = L["If checked, focus must be an enemy."],
											disabled = function(info) return IsTestFieldOff("Focus Status", "isEnemy") end,
											get = function(info) return GetTestField("Focus Status", "isEnemy") == true end,
											set = function(info, value) SetTestField("Focus Status", "isEnemy", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["Not Enemy"],
											desc = L["If checked, focus must not be an enemy."],
											disabled = function(info) return IsTestFieldOff("Focus Status", "isEnemy") end,
											get = function(info) return GetTestField("Focus Status", "isEnemy") == false end,
											set = function(info, value) SetTestField("Focus Status", "isEnemy", false) end,
										},
									},
								},
								FriendGroup = {
									type = "group", order = 20, name = L["Friendly"], inline = true,
									args = {
										CheckFriend = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, test if the focus is friendly."],
											get = function(info) return IsTestFieldOn("Focus Status", "isFriend") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Focus Status", "isFriend", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = L["Is Friendly"],
											desc = L["If checked, focus must be friendly."],
											disabled = function(info) return IsTestFieldOff("Focus Status", "isFriend") end,
											get = function(info) return GetTestField("Focus Status", "isFriend") == true end,
											set = function(info, value) SetTestField("Focus Status", "isFriend", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["Not Friendly"],
											desc = L["If checked, focus must not be friendly."],
											disabled = function(info) return IsTestFieldOff("Focus Status", "isFriend") end,
											get = function(info) return GetTestField("Focus Status", "isFriend") == false end,
											set = function(info, value) SetTestField("Focus Status", "isFriend", false) end,
										},
									},
								},
								DeadGroup = {
									type = "group", order = 22, name = L["Dead"], inline = true,
									args = {
										CheckFriend = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, test if the focus is dead."],
											get = function(info) return IsTestFieldOn("Focus Status", "isDead") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Focus Status", "isDead", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = L["Is Dead"],
											desc = L["If checked, focus must be dead."],
											disabled = function(info) return IsTestFieldOff("Focus Status", "isDead") end,
											get = function(info) return GetTestField("Focus Status", "isDead") == true end,
											set = function(info, value) SetTestField("Focus Status", "isDead", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["Not Dead"],
											desc = L["If checked, focus must not be dead."],
											disabled = function(info) return IsTestFieldOff("Focus Status", "isDead") end,
											get = function(info) return GetTestField("Focus Status", "isDead") == false end,
											set = function(info, value) SetTestField("Focus Status", "isDead", false) end,
										},
									},
								},
								Classification = {
									type = "group", order = 25, name = L["Classification"], inline = true,
									args = {
										Enable = {
											type = "toggle", order = 1, name = L["Enable"], width = "half",
											desc = L["If checked, test the focus's classification."],
											get = function(info) return GetTestField("Focus Status", "classify") end,
											set = function(info, value) SetTestField("Focus Status", "classify", value) end,
										},
										spacer1 = { type = "description", name = "", order = 10 },
										Normal = {
											type = "toggle", order = 20, name = L["Normal"], width = "half",
											desc = L["If checked, focus must be normal."],
											disabled = function(info) return not GetTestField("Focus Status", "classify") end,
											get = function(info) return GetTestField("Focus Status", "classification") == "normal" end,
											set = function(info, value) SetTestField("Focus Status", "classification", "normal") end,
										},
										Boss = {
											type = "toggle", order = 21, name = L["Boss"], width = "half",
											desc = L["If checked, focus must be a boss."],
											disabled = function(info) return not GetTestField("Focus Status", "classify") end,
											get = function(info) return GetTestField("Focus Status", "classification") == "worldboss" end,
											set = function(info, value) SetTestField("Focus Status", "classification", "worldboss") end,
										},
										Rare = {
											type = "toggle", order = 22, name = L["Rare"], width = "half",
											desc = L["If checked, focus must be rare."],
											disabled = function(info) return not GetTestField("Focus Status", "classify") end,
											get = function(info) return GetTestField("Focus Status", "classification") == "rare" end,
											set = function(info, value) SetTestField("Focus Status", "classification", "rare") end,
										},
										Elite = {
											type = "toggle", order = 23, name = L["Elite"], width = "half",
											desc = L["If checked, focus must be elite."],
											disabled = function(info) return not GetTestField("Focus Status", "classify") end,
											get = function(info) return GetTestField("Focus Status", "classification") == "elite" end,
											set = function(info, value) SetTestField("Focus Status", "classification", "elite") end,
										},
										RareElite = {
											type = "toggle", order = 25, name = L["Rare Elite"],
											desc = L["If checked, focus must be rare elite."],
											disabled = function(info) return not GetTestField("Focus Status", "classify") end,
											get = function(info) return GetTestField("Focus Status", "classification") == "rareelite" end,
											set = function(info, value) SetTestField("Focus Status", "classification", "rareelite") end,
										},
									},
								},
								RangeGroup = {
									type = "group", order = 30, name = L["Range"], inline = true,
									args = {
										CheckRange = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, test if the focus is in range."],
											get = function(info) return IsTestFieldOn("Focus Status", "inRange") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Focus Status", "inRange", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = L["In Range"],
											desc = L["If checked, focus must be in range."],
											disabled = function(info) return IsTestFieldOff("Focus Status", "inRange") end,
											get = function(info) return GetTestField("Focus Status", "inRange") == true end,
											set = function(info, value) SetTestField("Focus Status", "inRange", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["Out Of Range"],
											desc = L["If checked, focus must not be in range."],
											disabled = function(info) return IsTestFieldOff("Focus Status", "inRange") end,
											get = function(info) return GetTestField("Focus Status", "inRange") == false end,
											set = function(info, value) SetTestField("Focus Status", "inRange", false) end,
										},
									},
								},
								StealableGroup = {
									type = "group", order = 35, name = L["Spellsteal"], inline = true,
									args = {
										CheckSteal = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, test for a buff that can be transferred to the player with Spellsteal."],
											get = function(info) return IsTestFieldOn("Focus Status", "isSteal") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Focus Status", "isSteal", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = L["Spellsteal"],
											desc = L["If checked, there must be a spellstealable buff."],
											disabled = function(info) return IsTestFieldOff("Focus Status", "isSteal") end,
											get = function(info) return GetTestField("Focus Status", "isSteal") == true end,
											set = function(info, value) SetTestField("Focus Status", "isSteal", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["Not Spellsteal"],
											desc = L["If checked, there must not be a spellstealable buff."],
											disabled = function(info) return IsTestFieldOff("Focus Status", "isSteal") end,
											get = function(info) return GetTestField("Focus Status", "isSteal") == false end,
											set = function(info, value) SetTestField("Focus Status", "isSteal", false) end,
										},
									},
								},
								CheckHealthGroup = {
									type = "group", order = 50, name = L["Health"], inline = true, 
									args = {
										CheckHealthEnable = {
											type = "toggle", order = 1, name = L["Enable"], width = "half",
											desc = L["If checked, test the focus's health."],
											get = function(info) return IsTestFieldOn("Focus Status", "checkHealth") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Focus Status", "checkHealth", v) end,
										},
										CheckHealth = {
											type = "toggle", order = 2, name = L["Minimum"],
											desc = L["If checked, focus's health must be at least this percentage, otherwise must be less."],
											disabled = function(info) return IsTestFieldOff("Focus Status", "checkHealth") end,
											get = function(info) return GetTestField("Focus Status", "checkHealth") == true end,
											set = function(info, value) SetTestField("Focus Status", "checkHealth", value) end,
										},
										HealthRange = {
											type = "range", order = 3, name = "", min = 1, max = 100, step = 1,
											disabled = function(info) return IsTestFieldOff("Focus Status", "checkHealth") end,
											get = function(info) return GetTestField("Focus Status", "minHealth") end,
											set = function(info, value) SetTestField("Focus Status", "minHealth", value) end,
										},
									},
								},
								CheckPowerGroup = {
									type = "group", order = 70, name = L["Power"], inline = true, 
									args = {
										CheckPowerEnable = {
											type = "toggle", order = 1, name = L["Enable"], width = "half",
											desc = L["If checked, test the focus's power (i.e., mana, rage, energy, focus, runic power)."],
											get = function(info) return IsTestFieldOn("Focus Status", "checkPower") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Focus Status", "checkPower", v) end,
										},
										CheckPower = {
											type = "toggle", order = 2, name = L["Minimum"],
											desc = L["If checked, focus's power must be at least this percentage, otherwise must be less."],
											disabled = function(info) return IsTestFieldOff("Focus Status", "checkPower") end,
											get = function(info) return GetTestField("Focus Status", "checkPower") == true end,
											set = function(info, value) SetTestField("Focus Status", "checkPower", value) end,
										},
										PowerRange = {
											type = "range", order = 3, name = "", min = 1, max = 100, step = 1,
											disabled = function(info) return IsTestFieldOff("Focus Status", "checkPower") end,
											get = function(info) return GetTestField("Focus Status", "minPower") end,
											set = function(info, value) SetTestField("Focus Status", "minPower", value) end,
										},
									},
								},
							},
						},
						AllBuffsGroup = {
							type = "group", order = 45, name = L["All Buffs"],
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = L["Enable Test"], inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, enable this test."],
											get = function(info) return GetTestField("All Buffs", "enable") end,
											set = function(info, value) SetTestField("All Buffs", "enable", value) end,
										},
									},
								},
								AurasGroup = {
									type = "group", order = 2, name = L["Buff List Entry"], inline = true,
									args = {
										AuraList = {
											type = "input", order = 1, name = L["Buffs"], width = "full",
											desc = L["Enter comma-separated list of buffs."],
											get = function(info) return GetTestFieldSpellList("All Buffs", "auras") end,
											set = function(info, value) SetTestFieldSpellList("All Buffs", "auras", value) end,
										},
									},
								},
								ToggleGroup = {
									type = "group", order = 5, name = L["Test Buffs"], inline = true,
									args = {
										AllActive = {
											type = "toggle", order = 10, name = L["All Active"],
											desc = L["If checked, test if all the buffs are active."],
											get = function(info) return GetTestField("All Buffs", "toggle") ~= true end,
											set = function(info, value) SetTestField("All Buffs", "toggle", Off)  end,
										},
										NotAllActive = {
											type = "toggle", order = 20, name = L["Not All Active"],
											desc = L["If checked, test if any of the buffs are not active."],
											get = function(info) return GetTestField("All Buffs", "toggle") == true end,
											set = function(info, value) SetTestField("All Buffs", "toggle", true)  end,
										},
									},
								},
								UnitGroup = {
									type = "group", order = 10, name = L["Who Has Buff"], inline = true,
									args = {
										PlayerBuff = {
											type = "toggle", order = 10, name = L["Player"], width = "half",
											desc = L["If checked, test if buff is on player."],
											get = function(info) return GetTestField("All Buffs", "unit") == "player" end,
											set = function(info, value) SetTestField("All Buffs", "unit", "player")  end,
										},
										PetBuff = {
											type = "toggle", order = 20, name = L["Pet"], width = "half",
											desc = L["If checked, test if buff is on pet."],
											get = function(info) return GetTestField("All Buffs", "unit") == "pet" end,
											set = function(info, value) SetTestField("All Buffs", "unit", "pet")  end,
										},
										TargetBuff = {
											type = "toggle", order = 30, name = L["Target"], width = "half",
											desc = L["If checked, test if buff is on target."],
											get = function(info) return GetTestField("All Buffs", "unit") == "target" end,
											set = function(info, value) SetTestField("All Buffs", "unit", "target")  end,
										},
										FocusBuff = {
											type = "toggle", order = 40, name = L["Focus"], width = "half",
											desc = L["If checked, test if buff is on focus."],
											get = function(info) return GetTestField("All Buffs", "unit") == "focus" end,
											set = function(info, value) SetTestField("All Buffs", "unit", "focus")  end,
										},
									},
								},
								IsMineGroup = {
									type = "group", order = 15, name = L["Who Cast Buff"], inline = true,
									args = {
										DoPlayer = {
											type = "toggle", order = 1, name = L["Cast By Player"],
											desc = L["If checked, test if the buffs were cast by the player."],
											get = function(info) return GetTestField("All Buffs", "isMine") == true end,
											set = function(info, value) SetTestField("All Buffs", "isMine", true) end,
										},
										DoOther = {
											type = "toggle", order = 2, name = L["Cast By Other"],
											desc = L["If checked, test if the buffs were cast by anyone other than the player."],
											get = function(info) return GetTestField("All Buffs", "isMine") == false end,
											set = function(info, value) SetTestField("All Buffs", "isMine", false) end,
										},
										DoAnyone = {
											type = "toggle", order = 3, name = L["Cast By Anyone"],
											desc = L["If checked, buffs may be cast by anyone, including the player."],
											get = function(info) return IsOff(GetTestField("All Buffs", "isMine")) end,
											set = function(info, value) SetTestField("All Buffs", "isMine", Off) end,
										},
									},
								},
							},
						},
						AnyBuffsGroup = {
							type = "group", order = 46, name = L["Any Buffs"],
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = L["Enable Test"], inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, enable this test."],
											get = function(info) return GetTestField("Any Buffs", "enable") end,
											set = function(info, value) SetTestField("Any Buffs", "enable", value) end,
										},
									},
								},
								AurasGroup = {
									type = "group", order = 5, name = L["Buff List Entry"], inline = true,
									args = {
										AuraList = {
											type = "input", order = 1, name = L["Buffs"], width = "full",
											desc = L["Enter comma-separated list of buffs."],
											get = function(info) return GetTestFieldSpellList("Any Buffs", "auras") end,
											set = function(info, value) SetTestFieldSpellList("Any Buffs", "auras", value) end,
										},
									},
								},
								ToggleGroup = {
									type = "group", order = 5, name = L["Test Buffs"], inline = true,
									args = {
										AllActive = {
											type = "toggle", order = 10, name = L["Any Active"],
											desc = L["If checked, test if any of the buffs are active."],
											get = function(info) return GetTestField("Any Buffs", "toggle") ~= true end,
											set = function(info, value) SetTestField("Any Buffs", "toggle", Off)  end,
										},
										NotAllActive = {
											type = "toggle", order = 20, name = L["None Active"],
											desc = L["If checked, test if none of the buffs are active."],
											get = function(info) return GetTestField("Any Buffs", "toggle") == true end,
											set = function(info, value) SetTestField("Any Buffs", "toggle", true)  end,
										},
									},
								},
								UnitGroup = {
									type = "group", order = 10, name = L["Who Has Buff"], inline = true,
									args = {
										PlayerBuff = {
											type = "toggle", order = 10, name = L["Player"], width = "half",
											desc = L["If checked, test if buff is on player."],
											get = function(info) return GetTestField("Any Buffs", "unit") == "player" end,
											set = function(info, value) SetTestField("Any Buffs", "unit", "player")  end,
										},
										PetBuff = {
											type = "toggle", order = 20, name = L["Pet"], width = "half",
											desc = L["If checked, test if buff is on pet."],
											get = function(info) return GetTestField("Any Buffs", "unit") == "pet" end,
											set = function(info, value) SetTestField("Any Buffs", "unit", "pet")  end,
										},
										TargetBuff = {
											type = "toggle", order = 30, name = L["Target"], width = "half",
											desc = L["If checked, test if buff is on target."],
											get = function(info) return GetTestField("Any Buffs", "unit") == "target" end,
											set = function(info, value) SetTestField("Any Buffs", "unit", "target")  end,
										},
										FocusBuff = {
											type = "toggle", order = 40, name = L["Focus"], width = "half",
											desc = L["If checked, test if buff is on focus."],
											get = function(info) return GetTestField("Any Buffs", "unit") == "focus" end,
											set = function(info, value) SetTestField("Any Buffs", "unit", "focus")  end,
										},
									},
								},
								IsMineGroup = {
									type = "group", order = 15, name = L["Who Cast Buff"], inline = true,
									args = {
										DoPlayer = {
											type = "toggle", order = 1, name = L["Cast By Player"],
											desc = L["If checked, test if the buffs were cast by the player."],
											get = function(info) return GetTestField("Any Buffs", "isMine") == true end,
											set = function(info, value) SetTestField("Any Buffs", "isMine", true) end,
										},
										DoOther = {
											type = "toggle", order = 2, name = L["Cast By Other"],
											desc = L["If checked, test if the buffs were cast by anyone other than the player."],
											get = function(info) return GetTestField("Any Buffs", "isMine") == false end,
											set = function(info, value) SetTestField("Any Buffs", "isMine", false) end,
										},
										DoAnyone = {
											type = "toggle", order = 3, name = L["Cast By Anyone"],
											desc = L["If checked, buffs may be cast by anyone, including the player."],
											get = function(info) return IsOff(GetTestField("Any Buffs", "isMine")) end,
											set = function(info, value) SetTestField("Any Buffs", "isMine", Off) end,
										},
									},
								},
							},
						},
						BuffTimeLeftGroup = {
							type = "group", order = 47, name = L["Buff Time Left"],
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = L["Enable Test"], inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, enable this test."],
											get = function(info) return GetTestField("Buff Time Left", "enable") end,
											set = function(info, value) SetTestField("Buff Time Left", "enable", value) end,
										},
									},
								},
								AuraGroup = {
									type = "group", order = 5, name = L["Buff Entry"], inline = true,
									args = {
										AuraList = {
											type = "input", order = 1, name = L["Buff"], width = "full",
											desc = L["Enter the buff to be tested."],
											get = function(info) return GetTestField("Buff Time Left", "aura") end,
											set = function(info, value) SetTestFieldString("Buff Time Left", "aura", value) end,
										},
									},
								},
								ToggleGroup = {
									type = "group", order = 10, name = L["Time Left"], inline = true,
									args = {
										TimeLeftMinutes= {
											type = "range", order = 1, name = L["Minutes"], min = 0, max = 120, step = 1,
											desc = L["Enter minutes for time left check."],
											get = function(info) local d = GetTestField("Buff Time Left", "timeLeft"); if d then return math.floor(d / 60) else return 0 end end,
											set = function(info, value) local d = GetTestField("Buff Time Left", "timeLeft"); if not d then d = 0 end; SetTestField("Buff Time Left", "timeLeft", (value * 60) + (d % 60)) end,
										},
										TimeLeftSeconds = {
											type = "range", order = 5, name = L["Seconds"], min = 0, max = 59, step = 1,
											desc = L["Enter seconds for time left check."],
											get = function(info) local d = GetTestField("Buff Time Left", "timeLeft"); if d then return d % 60 else return 0 end end,
											set = function(info, value) local d = GetTestField("Buff Time Left", "timeLeft"); if not d then d = 0 end; SetTestField("Buff Time Left", "timeLeft", value + (60 * math.floor(d / 60))) end,
										},
										TimeLeftMinMax = {
											type = "select", order = 6, name = L["Time Left"],
											get = function(info) if GetTestField("Buff Time Left", "toggle") == true then return 1 else return 2 end end,
											set = function(info, value) if value == 1 then SetTestField("Buff Time Left", "toggle", true) else SetTestField("Buff Time Left", "toggle", Off) end end,
											values = function(info)
												local d = GetTestField("Buff Time Left", "timeLeft")
												if not d then d = 0 end
												local ds = string.format("%0d:%02d", math.floor(d / 60), d % 60)
												return { L["Less Than"] .. " " .. ds, ds .. " " .. L["Or More"] }
											end,
											style = "dropdown",
										},
									},
								},
								UnitGroup = {
									type = "group", order = 20, name = L["Who Has Buff"], inline = true,
									args = {
										PlayerBuff = {
											type = "toggle", order = 10, name = L["Player"], width = "half",
											desc = L["If checked, test if buff is on player."],
											get = function(info) return GetTestField("Buff Time Left", "unit") == "player" end,
											set = function(info, value) SetTestField("Buff Time Left", "unit", "player")  end,
										},
										PetBuff = {
											type = "toggle", order = 20, name = L["Pet"], width = "half",
											desc = L["If checked, test if buff is on pet."],
											get = function(info) return GetTestField("Buff Time Left", "unit") == "pet" end,
											set = function(info, value) SetTestField("Buff Time Left", "unit", "pet")  end,
										},
										TargetBuff = {
											type = "toggle", order = 30, name = L["Target"], width = "half",
											desc = L["If checked, test if buff is on target."],
											get = function(info) return GetTestField("Buff Time Left", "unit") == "target" end,
											set = function(info, value) SetTestField("Buff Time Left", "unit", "target")  end,
										},
										FocusBuff = {
											type = "toggle", order = 40, name = L["Focus"], width = "half",
											desc = L["If checked, test if buff is on focus."],
											get = function(info) return GetTestField("Buff Time Left", "unit") == "focus" end,
											set = function(info, value) SetTestField("Buff Time Left", "unit", "focus")  end,
										},
									},
								},
								IsMineGroup = {
									type = "group", order = 25, name = L["Who Cast Buff"], inline = true,
									args = {
										DoPlayer = {
											type = "toggle", order = 1, name = L["Cast By Player"],
											desc = L["If checked, test if the buff was cast by the player."],
											get = function(info) return GetTestField("Buff Time Left", "isMine") == true end,
											set = function(info, value) SetTestField("Buff Time Left", "isMine", true) end,
										},
										DoOther = {
											type = "toggle", order = 2, name = L["Cast By Other"],
											desc = L["If checked, test if the buff was cast by anyone other than the player."],
											get = function(info) return GetTestField("Buff Time Left", "isMine") == false end,
											set = function(info, value) SetTestField("Buff Time Left", "isMine", false) end,
										},
										DoAnyone = {
											type = "toggle", order = 3, name = L["Cast By Anyone"],
											desc = L["If checked, buff may be cast by anyone, including the player."],
											get = function(info) return IsOff(GetTestField("Buff Time Left", "isMine")) end,
											set = function(info, value) SetTestField("Buff Time Left", "isMine", Off) end,
										},
									},
								},
							},
						},
						BuffCountGroup = {
							type = "group", order = 48, name = L["Buff Count"],
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = L["Enable Test"], inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, enable this test."],
											get = function(info) return GetTestField("Buff Count", "enable") end,
											set = function(info, value) SetTestField("Buff Count", "enable", value) end,
										},
									},
								},
								AuraGroup = {
									type = "group", order = 5, name = L["Buff Entry"], inline = true,
									args = {
										AuraList = {
											type = "input", order = 1, name = L["Buff"], width = "full",
											desc = L["Enter buff to test for stack count (if buff doesn't stack then its count is 0)."],
											get = function(info) return GetTestField("Buff Count", "aura") end,
											set = function(info, value) SetTestFieldString("Buff Count", "aura", value) end,
										},
									},
								},
								ToggleGroup = {
									type = "group", order = 10, name = L["Stack Count"], inline = true,
									args = {
										CountValue = {
											type = "range", order = 1, name = L["Count"], min = 1, max = 10, step = 1,
											desc = L["Enter value to compare with the buff stack count."],
											get = function(info) local d = GetTestField("Buff Count", "count"); if d then return d else return 1 end end,
											set = function(info, value) local d = GetTestField("Buff Count", "count"); if not d then d = 1 end; SetTestField("Buff Count", "count", value) end,
										},
										CountMinMax = {
											type = "select", order = 6, name = L["Comparison"],
											get = function(info) if GetTestField("Buff Count", "toggle") == true then return 1 else return 2 end end,
											set = function(info, value) if value == 1 then SetTestField("Buff Count", "toggle", true) else SetTestField("Buff Count", "toggle", Off) end end,
											values = function(info)
												local d = GetTestField("Buff Count", "count")
												if not d then d = 1 end
												return { "Less Than " .. d, d .. " Or More" }
											end,
											style = "dropdown",
										},
									},
								},
								UnitGroup = {
									type = "group", order = 15, name = L["Who Has Buff"], inline = true,
									args = {
										PlayerBuff = {
											type = "toggle", order = 10, name = L["Player"], width = "half",
											desc = L["If checked, test if buff is on player."],
											get = function(info) return GetTestField("Buff Count", "unit") == "player" end,
											set = function(info, value) SetTestField("Buff Count", "unit", "player")  end,
										},
										PetBuff = {
											type = "toggle", order = 20, name = L["Pet"], width = "half",
											desc = L["If checked, test if buff is on pet."],
											get = function(info) return GetTestField("Buff Count", "unit") == "pet" end,
											set = function(info, value) SetTestField("Buff Count", "unit", "pet")  end,
										},
										TargetBuff = {
											type = "toggle", order = 30, name = L["Target"], width = "half",
											desc = L["If checked, test if buff is on target."],
											get = function(info) return GetTestField("Buff Count", "unit") == "target" end,
											set = function(info, value) SetTestField("Buff Count", "unit", "target")  end,
										},
										FocusBuff = {
											type = "toggle", order = 40, name = L["Focus"], width = "half",
											desc = L["If checked, test if buff is on focus."],
											get = function(info) return GetTestField("Buff Count", "unit") == "focus" end,
											set = function(info, value) SetTestField("Buff Count", "unit", "focus")  end,
										},
									},
								},
								IsMineGroup = {
									type = "group", order = 20, name = L["Who Cast Buff"], inline = true,
									args = {
										DoPlayer = {
											type = "toggle", order = 1, name = L["Cast By Player"],
											desc = L["If checked, test if the buff was cast by the player."],
											get = function(info) return GetTestField("Buff Count", "isMine") == true end,
											set = function(info, value) SetTestField("Buff Count", "isMine", true) end,
										},
										DoOther = {
											type = "toggle", order = 2, name = L["Cast By Other"],
											desc = L["If checked, test if the buff was cast by anyone other than the player."],
											get = function(info) return GetTestField("Buff Count", "isMine") == false end,
											set = function(info, value) SetTestField("Buff Count", "isMine", false) end,
										},
										DoAnyone = {
											type = "toggle", order = 3, name = L["Cast By Anyone"],
											desc = L["If checked, buff may be cast by anyone, including the player."],
											get = function(info) return IsOff(GetTestField("Buff Count", "isMine")) end,
											set = function(info, value) SetTestField("Buff Count", "isMine", Off) end,
										},
									},
								},
							},
						},
						BuffTypeGroup = {
							type = "group", order = 49, name = L["Buff Type"],
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = L["Enable Test"], inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, enable this test."],
											get = function(info) return GetTestField("Buff Type", "enable") end,
											set = function(info, value) SetTestField("Buff Type", "enable", value) end,
										},
									},
								},
								CheckBuffGroup = {
									type = "group", order = 10, name = L["Player Has Buff"], inline = true, args = {
										CheckMainhand = {
											type = "toggle", order = 2, name = L["Mainhand"],
											desc = L["If checked, player must have a mainhand buff."],
											disabled = function(info) return IsTestFieldOff("Buff Type", "hasBuff") end,
											get = function(info)
												if not GetTestField("Buff Type", "hasBuff") then SetTestField("Buff Type", "hasBuff", "Mainhand") end
												return GetTestField("Buff Type", "hasBuff") == "Mainhand"
											end,
											set = function(info, value) if value then SetTestField("Buff Type", "hasBuff", "Mainhand") end end,
										},
										CheckOffhand = {
											type = "toggle", order = 3, name = L["Offhand"],
											desc = L["If checked, player must have an offhand buff."],
											disabled = function(info) return IsTestFieldOff("Buff Type", "hasBuff") end,
											get = function(info) return GetTestField("Buff Type", "hasBuff") == "Offhand" end,
											set = function(info, value) if value then SetTestField("Buff Type", "hasBuff", "Offhand") end end,
										},
										CheckRanged = {
											type = "toggle", order = 4, name = L["Ranged"],
											desc = L["If checked, player must have a ranged buff."],
											disabled = function(info) return IsTestFieldOff("Buff Type", "hasBuff") end,
											get = function(info) return GetTestField("Buff Type", "hasBuff") == "Ranged" end,
											set = function(info, value) if value then SetTestField("Buff Type", "hasBuff", "Ranged") end end,
										},
									},
								},
								ToggleGroup = {
									type = "group", order = 20, name = L["Result"], inline = true,
									args = {
										CheckPresent = {
											type = "toggle", order = 2, name = L["Present"],
											desc = L["If checked, true if player has a buff of the specified type."],
											get = function(info) return GetTestField("Buff Type", "toggle") ~= true end,
											set = function(info, value) SetTestField("Buff Type", "toggle", false) end,
										},
										CheckMissing = {
											type = "toggle", order = 3, name = L["Missing"],
											desc = L["If checked, true if player does not have a buff of the specified type."],
											get = function(info) return GetTestField("Buff Type", "toggle") == true end,
											set = function(info, value) SetTestField("Buff Type", "toggle", true) end,
										},
									},
								},
							},
						},
						AllDebuffsGroup = {
							type = "group", order = 50, name = L["All Debuffs"],
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = L["Enable Test"], inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, enable this test."],
											get = function(info) return GetTestField("All Debuffs", "enable") end,
											set = function(info, value) SetTestField("All Debuffs", "enable", value) end,
										},
									},
								},
								AurasGroup = {
									type = "group", order = 2, name = L["Debuff List Entry"], inline = true,
									args = {
										AuraList = {
											type = "input", order = 1, name = L["Debuffs"], width = "full",
											desc = L["Enter comma-separated list of debuffs."],
											get = function(info) return GetTestFieldSpellList("All Debuffs", "auras") end,
											set = function(info, value) SetTestFieldSpellList("All Debuffs", "auras", value) end,
										},
									},
								},
								ToggleGroup = {
									type = "group", order = 5, name = L["Test Debuffs"], inline = true,
									args = {
										AllActive = {
											type = "toggle", order = 10, name = L["All Active"],
											desc = L["If checked, test if all the debuffs are active."],
											get = function(info) return GetTestField("All Debuffs", "toggle") ~= true end,
											set = function(info, value) SetTestField("All Debuffs", "toggle", Off)  end,
										},
										NotAllActive = {
											type = "toggle", order = 20, name = L["Not All Active"],
											desc = L["If checked, test if any of the debuffs are not active."],
											get = function(info) return GetTestField("All Debuffs", "toggle") == true end,
											set = function(info, value) SetTestField("All Debuffs", "toggle", true)  end,
										},
									},
								},
								UnitGroup = {
									type = "group", order = 10, name = L["Who Has Debuff"], inline = true,
									args = {
										PlayerBuff = {
											type = "toggle", order = 10, name = L["Player"], width = "half",
											desc = L["If checked, test if debuff is on player."],
											get = function(info) return GetTestField("All Debuffs", "unit") == "player" end,
											set = function(info, value) SetTestField("All Debuffs", "unit", "player")  end,
										},
										PetBuff = {
											type = "toggle", order = 20, name = L["Pet"], width = "half",
											desc = L["If checked, test if debuff is on pet."],
											get = function(info) return GetTestField("All Debuffs", "unit") == "pet" end,
											set = function(info, value) SetTestField("All Debuffs", "unit", "pet")  end,
										},
										TargetBuff = {
											type = "toggle", order = 30, name = L["Target"], width = "half",
											desc = L["If checked, test if debuff is on target."],
											get = function(info) return GetTestField("All Debuffs", "unit") == "target" end,
											set = function(info, value) SetTestField("All Debuffs", "unit", "target")  end,
										},
										FocusBuff = {
											type = "toggle", order = 40, name = L["Focus"], width = "half",
											desc = L["If checked, test if debuff is on focus."],
											get = function(info) return GetTestField("All Debuffs", "unit") == "focus" end,
											set = function(info, value) SetTestField("All Debuffs", "unit", "focus")  end,
										},
									},
								},
								IsMineGroup = {
									type = "group", order = 15, name = L["Who Cast Debuff"], inline = true,
									args = {
										DoPlayer = {
											type = "toggle", order = 1, name = L["Cast By Player"],
											desc = L["If checked, test if the debuffs were cast by the player."],
											get = function(info) return GetTestField("All Debuffs", "isMine") == true end,
											set = function(info, value) SetTestField("All Debuffs", "isMine", true) end,
										},
										DoOther = {
											type = "toggle", order = 2, name = L["Cast By Other"],
											desc = L["If checked, test if the debuffs were cast by anyone other than the player."],
											get = function(info) return GetTestField("All Debuffs", "isMine") == false end,
											set = function(info, value) SetTestField("All Debuffs", "isMine", false) end,
										},
										DoAnyone = {
											type = "toggle", order = 3, name = L["Cast By Anyone"],
											desc = L["If checked, debuffs may be cast by anyone, including the player."],
											get = function(info) return IsOff(GetTestField("All Debuffs", "isMine")) end,
											set = function(info, value) SetTestField("All Debuffs", "isMine", Off) end,
										},
									},
								},
							},
						},
						AnyDebuffsGroup = {
							type = "group", order = 51, name = L["Any Debuffs"],
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = L["Enable Test"], inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, enable this test."],
											get = function(info) return GetTestField("Any Debuffs", "enable") end,
											set = function(info, value) SetTestField("Any Debuffs", "enable", value) end,
										},
									},
								},
								AurasGroup = {
									type = "group", order = 2, name = L["Debuff List Entry"], inline = true,
									args = {
										AuraList = {
											type = "input", order = 1, name = L["Debuffs"], width = "full",
											desc = L["Enter comma-separated list of debuffs."],
											get = function(info) return GetTestFieldSpellList("Any Debuffs", "auras") end,
											set = function(info, value) SetTestFieldSpellList("Any Debuffs", "auras", value) end,
										},
									},
								},
								ToggleGroup = {
									type = "group", order = 5, name = L["Test Debuffs"], inline = true,
									args = {
										AllActive = {
											type = "toggle", order = 10, name = L["Any Active"],
											desc = L["If checked, test if any of the debuffs are active."],
											get = function(info) return GetTestField("Any Debuffs", "toggle") ~= true end,
											set = function(info, value) SetTestField("Any Debuffs", "toggle", Off)  end,
										},
										NotAllActive = {
											type = "toggle", order = 20, name = L["None Active"],
											desc = L["If checked, test if none of the debuffs are active."],
											get = function(info) return GetTestField("Any Debuffs", "toggle") == true end,
											set = function(info, value) SetTestField("Any Debuffs", "toggle", true)  end,
										},
									},
								},
								UnitGroup = {
									type = "group", order = 10, name = L["Who Has Debuff"], inline = true,
									args = {
										PlayerBuff = {
											type = "toggle", order = 10, name = L["Player"], width = "half",
											desc = L["If checked, test if debuff is on player."],
											get = function(info) return GetTestField("Any Debuffs", "unit") == "player" end,
											set = function(info, value) SetTestField("Any Debuffs", "unit", "player")  end,
										},
										PetBuff = {
											type = "toggle", order = 20, name = L["Pet"], width = "half",
											desc = L["If checked, test if debuff is on pet."],
											get = function(info) return GetTestField("Any Debuffs", "unit") == "pet" end,
											set = function(info, value) SetTestField("Any Debuffs", "unit", "pet")  end,
										},
										TargetBuff = {
											type = "toggle", order = 30, name = L["Target"], width = "half",
											desc = L["If checked, test if debuff is on target."],
											get = function(info) return GetTestField("Any Debuffs", "unit") == "target" end,
											set = function(info, value) SetTestField("Any Debuffs", "unit", "target")  end,
										},
										FocusBuff = {
											type = "toggle", order = 40, name = L["Focus"], width = "half",
											desc = L["If checked, test if debuff is on focus."],
											get = function(info) return GetTestField("Any Debuffs", "unit") == "focus" end,
											set = function(info, value) SetTestField("Any Debuffs", "unit", "focus")  end,
										},
									},
								},
								IsMineGroup = {
									type = "group", order = 15, name = L["Who Cast Debuff"], inline = true,
									args = {
										DoPlayer = {
											type = "toggle", order = 1, name = L["Cast By Player"],
											desc = L["If checked, test if the debuffs were cast by the player."],
											get = function(info) return GetTestField("Any Debuffs", "isMine") == true end,
											set = function(info, value) SetTestField("Any Debuffs", "isMine", true) end,
										},
										DoOther = {
											type = "toggle", order = 2, name = L["Cast By Other"],
											desc = L["If checked, test if the debuffs were cast by anyone other than the player."],
											get = function(info) return GetTestField("Any Debuffs", "isMine") == false end,
											set = function(info, value) SetTestField("Any Debuffs", "isMine", false) end,
										},
										DoAnyone = {
											type = "toggle", order = 3, name = L["Cast By Anyone"],
											desc = L["If checked, debuffs may be cast by anyone, including the player."],
											get = function(info) return IsOff(GetTestField("Any Debuffs", "isMine")) end,
											set = function(info, value) SetTestField("Any Debuffs", "isMine", Off) end,
										},
									},
								},
							},
						},
						DebuffTimeLeftGroup = {
							type = "group", order = 52, name = L["Debuff Time Left"],
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = L["Enable Test"], inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, enable this test."],
											get = function(info) return GetTestField("Debuff Time Left", "enable") end,
											set = function(info, value) SetTestField("Debuff Time Left", "enable", value) end,
										},
									},
								},
								AuraGroup = {
									type = "group", order = 2, name = L["Debuff Entry"], inline = true,
									args = {
										AuraList = {
											type = "input", order = 1, name = L["Debuff"], width = "full",
											desc = L["Enter the debuff to be tested."],
											get = function(info) return GetTestField("Debuff Time Left", "aura") end,
											set = function(info, value) SetTestFieldString("Debuff Time Left", "aura", value) end,
										},
									},
								},
								ToggleGroup = {
									type = "group", order = 5, name = L["Time Left"], inline = true,
									args = {
										TimeLeftMinutes= {
											type = "range", order = 1, name = L["Minutes"], min = 0, max = 120, step = 1,
											desc = L["Enter minutes for time left check."],
											get = function(info) local d = GetTestField("Debuff Time Left", "timeLeft"); if d then return math.floor(d / 60) else return 0 end end,
											set = function(info, value) local d = GetTestField("Debuff Time Left", "timeLeft"); if not d then d = 0 end; SetTestField("Debuff Time Left", "timeLeft", (value * 60) + (d % 60)) end,
										},
										TimeLeftSeconds = {
											type = "range", order = 5, name = L["Seconds"], min = 0, max = 59, step = 1,
											desc = L["Enter seconds for time left check."],
											get = function(info) local d = GetTestField("Debuff Time Left", "timeLeft"); if d then return d % 60 else return 0 end end,
											set = function(info, value) local d = GetTestField("Debuff Time Left", "timeLeft"); if not d then d = 0 end; SetTestField("Debuff Time Left", "timeLeft", value + (60 * math.floor(d / 60))) end,
										},
										TimeLeftMinMax = {
											type = "select", order = 6, name = L["Time Left"],
											get = function(info) if GetTestField("Debuff Time Left", "toggle") == true then return 1 else return 2 end end,
											set = function(info, value) if value == 1 then SetTestField("Debuff Time Left", "toggle", true) else SetTestField("Debuff Time Left", "toggle", Off) end end,
											values = function(info)
												local d = GetTestField("Debuff Time Left", "timeLeft")
												if not d then d = 0 end
												local ds = string.format("%0d:%02d", math.floor(d / 60), d % 60)
												return { "Less Than " .. ds, ds .. " Or More" }
											end,
											style = "dropdown",
										},
									},
								},
								UnitGroup = {
									type = "group", order = 10, name = L["Who Has Debuff"], inline = true,
									args = {
										PlayerBuff = {
											type = "toggle", order = 10, name = L["Player"], width = "half",
											desc = L["If checked, test if debuff is on player."],
											get = function(info) return GetTestField("Debuff Time Left", "unit") == "player" end,
											set = function(info, value) SetTestField("Debuff Time Left", "unit", "player")  end,
										},
										PetBuff = {
											type = "toggle", order = 20, name = L["Pet"], width = "half",
											desc = L["If checked, test if debuff is on pet."],
											get = function(info) return GetTestField("Debuff Time Left", "unit") == "pet" end,
											set = function(info, value) SetTestField("Debuff Time Left", "unit", "pet")  end,
										},
										TargetBuff = {
											type = "toggle", order = 30, name = L["Target"], width = "half",
											desc = L["If checked, test if debuff is on target."],
											get = function(info) return GetTestField("Debuff Time Left", "unit") == "target" end,
											set = function(info, value) SetTestField("Debuff Time Left", "unit", "target")  end,
										},
										FocusBuff = {
											type = "toggle", order = 40, name = L["Focus"], width = "half",
											desc = L["If checked, test if debuff is on focus."],
											get = function(info) return GetTestField("Debuff Time Left", "unit") == "focus" end,
											set = function(info, value) SetTestField("Debuff Time Left", "unit", "focus")  end,
										},
									},
								},
								IsMineGroup = {
									type = "group", order = 15, name = L["Who Cast Debuff"], inline = true,
									args = {
										DoPlayer = {
											type = "toggle", order = 1, name = L["Cast By Player"],
											desc = L["If checked, test if the debuff was cast by the player."],
											get = function(info) return GetTestField("Debuff Time Left", "isMine") == true end,
											set = function(info, value) SetTestField("Debuff Time Left", "isMine", true) end,
										},
										DoOther = {
											type = "toggle", order = 2, name = L["Cast By Other"],
											desc = L["If checked, test if the debuff was cast by anyone other than the player."],
											get = function(info) return GetTestField("Debuff Time Left", "isMine") == false end,
											set = function(info, value) SetTestField("Debuff Time Left", "isMine", false) end,
										},
										DoAnyone = {
											type = "toggle", order = 3, name = L["Cast By Anyone"],
											desc = L["If checked, debuff may be cast by anyone, including the player."],
											get = function(info) return IsOff(GetTestField("Debuff Time Left", "isMine")) end,
											set = function(info, value) SetTestField("Debuff Time Left", "isMine", Off) end,
										},
									},
								},
							},
						},
						DebuffCountGroup = {
							type = "group", order = 53, name = L["Debuff Count"],
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = L["Enable Test"], inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, enable this test."],
											get = function(info) return GetTestField("Debuff Count", "enable") end,
											set = function(info, value) SetTestField("Debuff Count", "enable", value) end,
										},
									},
								},
								AuraGroup = {
									type = "group", order = 2, name = L["Debuff Entry"], inline = true,
									args = {
										AuraList = {
											type = "input", order = 1, name = L["Debuff"], width = "full",
											desc = L["Enter debuff to test for stack count (if debuff doesn't stack then its count is 0)."],
											get = function(info) return GetTestField("Debuff Count", "aura") end,
											set = function(info, value) SetTestFieldString("Debuff Count", "aura", value) end,
										},
									},
								},
								ToggleGroup = {
									type = "group", order = 5, name = L["Stack Count"], inline = true,
									args = {
										CountValue = {
											type = "range", order = 1, name = L["Count"], min = 1, max = 10, step = 1,
											desc = L["Enter value to compare with the debuff stack count."],
											get = function(info) local d = GetTestField("Debuff Count", "count"); if d then return d else return 1 end end,
											set = function(info, value) local d = GetTestField("Debuff Count", "count"); if not d then d = 1 end; SetTestField("Debuff Count", "count", value) end,
										},
										CountMinMax = {
											type = "select", order = 6, name = L["Comparison"],
											get = function(info) if GetTestField("Debuff Count", "toggle") == true then return 1 else return 2 end end,
											set = function(info, value) if value == 1 then SetTestField("Debuff Count", "toggle", true) else SetTestField("Debuff Count", "toggle", Off) end end,
											values = function(info)
												local d = GetTestField("Debuff Count", "count")
												if not d then d = 1 end
												return { "Less Than " .. d, d .. " Or More" }
											end,
											style = "dropdown",
										},
									},
								},
								UnitGroup = {
									type = "group", order = 10, name = L["Who Has Debuff"], inline = true,
									args = {
										PlayerBuff = {
											type = "toggle", order = 10, name = L["Player"], width = "half",
											desc = L["If checked, test if debuff is on player."],
											get = function(info) return GetTestField("Debuff Count", "unit") == "player" end,
											set = function(info, value) SetTestField("Debuff Count", "unit", "player")  end,
										},
										PetBuff = {
											type = "toggle", order = 20, name = L["Pet"], width = "half",
											desc = L["If checked, test if debuff is on pet."],
											get = function(info) return GetTestField("Debuff Count", "unit") == "pet" end,
											set = function(info, value) SetTestField("Debuff Count", "unit", "pet")  end,
										},
										TargetBuff = {
											type = "toggle", order = 30, name = L["Target"], width = "half",
											desc = L["If checked, test if debuff is on target."],
											get = function(info) return GetTestField("Debuff Count", "unit") == "target" end,
											set = function(info, value) SetTestField("Debuff Count", "unit", "target")  end,
										},
										FocusBuff = {
											type = "toggle", order = 40, name = L["Focus"], width = "half",
											desc = L["If checked, test if debuff is on focus."],
											get = function(info) return GetTestField("Debuff Count", "unit") == "focus" end,
											set = function(info, value) SetTestField("Debuff Count", "unit", "focus")  end,
										},
									},
								},
								IsMineGroup = {
									type = "group", order = 15, name = L["Who Cast Debuff"], inline = true,
									args = {
										DoPlayer = {
											type = "toggle", order = 1, name = L["Cast By Player"],
											desc = L["If checked, test if the debuff was cast by the player."],
											get = function(info) return GetTestField("Debuff Count", "isMine") == true end,
											set = function(info, value) SetTestField("Debuff Count", "isMine", true) end,
										},
										DoOther = {
											type = "toggle", order = 2, name = L["Cast By Other"],
											desc = L["If checked, test if the debuff was cast by anyone other than the player."],
											get = function(info) return GetTestField("Debuff Count", "isMine") == false end,
											set = function(info, value) SetTestField("Debuff Count", "isMine", false) end,
										},
										DoAnyone = {
											type = "toggle", order = 3, name = L["Cast By Anyone"],
											desc = L["If checked, debuff may be cast by anyone, including the player."],
											get = function(info) return IsOff(GetTestField("Debuff Count", "isMine")) end,
											set = function(info, value) SetTestField("Debuff Count", "isMine", Off) end,
										},
									},
								},
							},
						},
						DebuffTypeGroup = {
							type = "group", order = 54, name = L["Debuff Type"],
							args = {
								EnableTestGroup = {
									type = "group", order = 10, name = L["Enable Test"], inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, enable this test."],
											get = function(info) return GetTestField("Debuff Type", "enable") end,
											set = function(info, value) SetTestField("Debuff Type", "enable", value) end,
										},
									},
								},
								CheckDebuffGroup = {
									type = "group", order = 20, name = L["Player Has Debuff"], inline = true, args = {
										CheckPoisonDebuff = {
											type = "toggle", order = 2, name = L["Poison"], width = "half",
											desc = L["If checked, player must have a poison debuff."],
											get = function(info)
												if not GetTestField("Debuff Type", "hasDebuff") then SetTestField("Debuff Type", "hasDebuff", "Poison") end
												return GetTestField("Debuff Type", "hasDebuff") == "Poison"
												end,
											set = function(info, value) if value then SetTestField("Debuff Type", "hasDebuff", "Poison") end end,
										},
										CheckDiseaseDebuff = {
											type = "toggle", order = 3, name = L["Disease"], width = "half",
											desc = L["If checked, player must have a disease debuff."],
											get = function(info) return GetTestField("Debuff Type", "hasDebuff") == "Disease" end,
											set = function(info, value) if value then SetTestField("Debuff Type", "hasDebuff", "Disease") end end,
										},
										CheckCurseDebuff = {
											type = "toggle", order = 4, name = L["Curse"], width = "half",
											desc = L["If checked, player must have a curse debuff."],
											get = function(info) return GetTestField("Debuff Type", "hasDebuff") == "Curse" end,
											set = function(info, value) if value then SetTestField("Debuff Type", "hasDebuff", "Curse") end end,
										},
										CheckMagicDebuff = {
											type = "toggle", order = 5, name = L["Magic"], width = "half",
											desc = L["If checked, player must have a magic debuff."],
											get = function(info) return GetTestField("Debuff Type", "hasDebuff") == "Magic" end,
											set = function(info, value) if value then SetTestField("Debuff Type", "hasDebuff", "Magic") end end,
										},
									},
								},
								ToggleGroup = {
									type = "group", order = 30, name = L["Result"], inline = true,
									args = {
										CheckPresent = {
											type = "toggle", order = 2, name = L["Present"], 
											desc = L["If checked, true if player has a debuff of the specified type."],
											get = function(info) return GetTestField("Debuff Type", "toggle") ~= true end,
											set = function(info, value) SetTestField("Debuff Type", "toggle", false) end,
										},
										CheckMissing = {
											type = "toggle", order = 3, name = L["Missing"], 
											desc = L["If checked, true if player does not have a debuff of the specified type."],
											get = function(info) return GetTestField("Debuff Type", "toggle") == true end,
											set = function(info, value) SetTestField("Debuff Type", "toggle", true) end,
										},
									},
								},
							},
						},
						AllCooldownsGroup = {
							type = "group", order = 56, name = L["All Cooldowns"],
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = L["Enable Test"], inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, enable this test."],
											get = function(info) return GetTestField("All Cooldowns", "enable") end,
											set = function(info, value) SetTestField("All Cooldowns", "enable", value) end,
										},
									},
								},
								SpellsGroup = {
									type = "group", order = 2, name = L["Spell List Entry"], inline = true,
									args = {
										SpellList = {
											type = "input", order = 1, name = L["Spells"], width = "full",
											desc = L["Enter comma-separated list of spells."],
											get = function(info) return GetTestFieldSpellList("All Cooldowns", "spells") end,
											set = function(info, value) SetTestFieldSpellList("All Cooldowns", "spells", value) end,
										},
									},
								},
								ToggleGroup = {
									type = "group", order = 10, name = L["Time Left"], inline = true,
									args = {
										TimeLeftSeconds = {
											type = "range", order = 5, name = L["Seconds"], min = 0, max = 60, step = 0.1,
											desc = L["Enter seconds for cooldown time left check."],
											get = function(info) local d = GetTestField("All Cooldowns", "timeLeft"); if d then return d else return 0 end end,
											set = function(info, value) local d = GetTestField("All Cooldowns", "timeLeft"); if not d then d = 0 end; SetTestField("All Cooldowns", "timeLeft", value) end,
										},
										TimeLeftMinMax = {
											type = "select", order = 6, name = L["Time Left"],
											get = function(info) if GetTestField("All Cooldowns", "toggle") == true then return 1 else return 2 end end,
											set = function(info, value) if value == 1 then SetTestField("All Cooldowns", "toggle", true) else SetTestField("All Cooldowns", "toggle", Off) end end,
											values = function(info)
												local d = GetTestField("All Cooldowns", "timeLeft")
												if not d then d = 0 end
												local ds = string.format("%0.1f", d)
												return { "Less Than " .. ds .. " Seconds", ds .. " Seconds Or More" }
											end,
											style = "dropdown",
										},
									},
								},
							},
						},
						SpellReadyGroup = {
							type = "group", order = 60, name = L["Spell Ready"],
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = L["Enable Test"], inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, enable this test."],
											get = function(info) return GetTestField("Spell Ready", "enable") end,
											set = function(info, value) SetTestField("Spell Ready", "enable", value) end,
										},
										EnableGCD = {
											type = "toggle", order = 2, name = L["Global Cooldown"],
											desc = L["If checked, test if spell is on global cooldown, otherwise ignore global cooldown."],
											get = function(info) return GetTestField("Spell Ready", "gcd") end,
											set = function(info, value) SetTestField("Spell Ready", "gcd", value) end,
										},
										IsUsable = {
											type = "toggle", order = 2, name = L["Is Usable"],
											desc = L["If checked, test if spell is usable (i.e., enough mana, reagents, etc.)."],
											get = function(info) return not GetTestField("Spell Ready", "notUsable") end,
											set = function(info, value) SetTestField("Spell Ready", "notUsable", not value) end,
										},
									},
								},
								SpellGroup = {
									type = "group", order = 5, name = L["Spell Entry"], inline = true,
									args = {
										SpellName = {
											type = "input", order = 1, name = L["Spell"], width = "full",
											desc = L["Enter spell to test if ready to be cast."],
											get = function(info) return GetTestField("Spell Ready", "spell") end,
											set = function(info, value) SetTestFieldString("Spell Ready", "spell", value) end,
										},
									},
								},
								RangeGroup = {
									type = "group", order = 10, name = L["Range"], inline = true,
									args = {
										CheckRange = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, also test if the target is in range of the spell."],
											get = function(info) return IsTestFieldOn("Spell Ready", "inRange") end,
											set = function(info, value) local v = Off if value then v = true end SetTestField("Spell Ready", "inRange", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = L["Target In Range"],
											desc = L["If checked, target must be in range."],
											disabled = function(info) return IsTestFieldOff("Spell Ready", "inRange") end,
											get = function(info) return GetTestField("Spell Ready", "inRange") == true end,
											set = function(info, value) SetTestField("Spell Ready", "inRange", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["Target Out Of Range"],
											desc = L["If checked, target must be out of range."],
											disabled = function(info) return IsTestFieldOff("Spell Ready", "inRange") end,
											get = function(info) return GetTestField("Spell Ready", "inRange") == false end,
											set = function(info, value) SetTestField("Spell Ready", "inRange", false) end,
										},
									},
								},
							},
						},
						SpellCastingGroup = {
							type = "group", order = 65, name = L["Spell Casting"],
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = L["Enable Test"], inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, enable this test."],
											get = function(info) return GetTestField("Spell Casting", "enable") end,
											set = function(info, value) SetTestField("Spell Casting", "enable", value) end,
										},
									},
								},
								SpellGroup = {
									type = "group", order = 5, name = L["Spell Entry"], inline = true,
									args = {
										SpellName = {
											type = "input", order = 1, name = L["Spell"], width = "full",
											desc = L["Enter spell to test if being cast or channeled."],
											get = function(info) return GetTestField("Spell Casting", "spell") end,
											set = function(info, value) SetTestFieldString("Spell Casting", "spell", value) end,
										},
									},
								},
								UnitGroup = {
									type = "group", order = 20, name = L["Who Is Casting"], inline = true,
									args = {
										PlayerBuff = {
											type = "toggle", order = 10, name = L["Cast By Player"],
											desc = L["If checked, test player casting."],
											get = function(info) return GetTestField("Spell Casting", "unit") == "player" end,
											set = function(info, value) SetTestField("Spell Casting", "unit", "player")  end,
										},
										TargetBuff = {
											type = "toggle", order = 20, name = L["Cast By Target"],
											desc = L["If checked, test target casting."],
											get = function(info) return GetTestField("Spell Casting", "unit") == "target" end,
											set = function(info, value) SetTestField("Spell Casting", "unit", "target")  end,
										},
										FocusBuff = {
											type = "toggle", order = 30, name = L["Cast By Focus"],
											desc = L["If checked, test focus casting."],
											get = function(info) return GetTestField("Spell Casting", "unit") == "focus" end,
											set = function(info, value) SetTestField("Spell Casting", "unit", "focus")  end,
										},
									},
								},
							},
						},
						ItemReadyGroup = {
							type = "group", order = 70, name = L["Item Ready"],
							args = {
								EnableTestGroup = {
									type = "group", order = 1, name = L["Enable Test"], inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, enable this test."],
											get = function(info) return GetTestField("Item Ready", "enable") end,
											set = function(info, value) SetTestField("Item Ready", "enable", value) end,
										},
									},
								},
								ItemGroup = {
									type = "group", order = 2, name = L["Item Entry"], inline = true,
									args = {
										ItemName = {
											type = "input", order = 1, name = L["Item"], width = "full",
											desc = L["Enter name or itemID of item to test. Item must be in the player's bags. Use generic cooldown names for Potions and Elixirs."],
											get = function(info) return GetTestField("Item Ready", "item") end,
											set = function(info, value) SetTestFieldString("Item Ready", "item", value) end,
										},
									},
								},
								ReadyOrNot = {
									type = "group", order = 10, name = L["Ready Check"], inline = true,
									args = {
										DoTrue = {
											type = "toggle", order = 2, name = L["Item Is Ready"],
											desc = L["If checked, item must be ready."],
											get = function(info) return GetTestField("Item Ready", "toggle") ~= true end,
											set = function(info, value) SetTestField("Item Ready", "toggle", Off) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["Item Is Not Ready"],
											desc = L["If checked, item must not be ready."],
											get = function(info) return GetTestField("Item Ready", "toggle") == true end,
											set = function(info, value) SetTestField("Item Ready", "toggle", true) end,
										},
									},
								},
							},
						},
						DependenciesGroup = {
							type = "group", order = 95, name = L["Result"],
							args = {
								SelectDependencies = {
									type = "group", order = 10, name = L["Dependencies"], inline = true,
									args = {
										intro = {
											type = "description", order = 1,
											name = L["Dependencies string"],
										},
										SelectCondition = {
											type = "select", order = 10, name = L["Condition"],
											get = function(info) return GetSelectedDependency() end,
											set = function(info, value) selectedDependency = value end,
											values = function(info) return GetDependenciesList() end,
											style = "dropdown",
										},
										TrueConditionButton = {
											type = "execute", order = 20, name = L["True"], width = "half",
											desc = L["Condition is true only if the selected condition evaluates to true."],
											func = function(info) SetDependency(GetDependenciesList()[selectedDependency], true) end,
										},
										FalseConditionButton = {
											type = "execute", order = 30, name = L["False"], width = "half",
											desc = L["Condition is true only if the selected condition evaluates to false."],
											func = function(info) SetDependency(GetDependenciesList()[selectedDependency], false) end,
										},
										RequiredConditionButton = {
											type = "execute", order = 35, name = L["And/Or"], width = "half",
											hidden = function(info) return not CheckDependency(GetDependenciesList()[selectedDependency]) end,
											desc = L["Toggle between logical \"And\" and \"Or\" for this dependecy."],
											func = function(info)
												local dep = GetDependenciesList()[selectedDependency]
												local ctype = GetDependencyType(dep)
												SetDependencyType(dep, not ctype)
											end,
										},
										NoConditionButton = {
											type = "execute", order = 40, name = L["Delete"], width = "half",
											hidden = function(info) return not CheckDependency(GetDependenciesList()[selectedDependency]) end,
											desc = L["Delete this condition from the dependencies list."],
											func = function(info) SetDependency(GetDependenciesList()[selectedDependency], nil) end,
										},
									},
								},
								ToggleResult = {
									type = "group", order = 20, name = L["Toggle Result"], inline = true,
									args = {
										EnableTest = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, the result for this condition, after evaluating all tests and checking dependencies, is toggled."],
											get = function(info) return GetConditionField("toggleResult") end,
											set = function(info, value) SetConditionField("toggleResult", value) end,
										},
									},
								},
								SetResult = {
									type = "group", order = 30, name = L["Set Result"], inline = true,
									args = {
										EnableSet = {
											type = "toggle", order = 1, name = L["Enable"],
											desc = L["If checked, the result for this condition is set to true or false (this feature is provided to help debug conditions)."],
											get = function(info) return IsOn(GetConditionField("setResult")) end,
											set = function(info, value) local v = Off if value then v = true end SetConditionField("setResult", v) end,
										},
										DoTrue = {
											type = "toggle", order = 2, name = L["True"],
											desc = L["If checked, result is set to true."],
											disabled = function(info) return IsOff(GetConditionField("setResult")) end,
											get = function(info) return GetConditionField("setResult") == true end,
											set = function(info, value) SetConditionField("setResult", true) end,
										},
										DoFalse = {
											type = "toggle", order = 3, name = L["False"],
											desc = L["If checked, result is set to false."],
											disabled = function(info) return IsOff(GetConditionField("setResult")) end,
											get = function(info) return GetConditionField("setResult") == false end,
											set = function(info, value) SetConditionField("setResult", false) end,
										},
									},
								},
							},
						},
					},
				},
			},
		},
		InCombat = {
			type = "group", order = 40, name = L["In-Combat Bar"], 
			disabled = function(info) return InMode() end,
			args = {
				intro = {
					type = "description", order = 1,
					name = L["In-combat string"],
				},
				EnableGroup = {
					type = "group", order = 10, name = L["Enable"], inline = true,
					args = {
						EnableOverlay = {
							type = "toggle", order = 10, name = L["Enable In-Combat Bar"], 
							desc = L["Enable in-combat buffs string"],
							get = function(info) return MOD.db.profile.InCombatBar.enable end,
							set = function(info, value) MOD.db.profile.InCombatBar.enable = value end,
						},
						LockOverlay = {
							type = "toggle", order = 20, name = L["Lock Bar Layout"],
							desc = L["Lock in-combat string"],
							disabled = function(info) return not MOD.db.profile.InCombatBar.enable end,
							get = function(info) return MOD.db.profile.InCombatBar.lock end,
							set = function(info, value) MOD.db.profile.InCombatBar.lock = value end,
						},
						OOCOverlay = {
							type = "toggle", order = 30, name = L["Out Of Combat"],
							desc = L["If checked, also display buffs on the bar when out of combat."],
							disabled = function(info) return not MOD.db.profile.InCombatBar.enable end,
							get = function(info) return MOD.db.profile.InCombatBar.ooc end,
							set = function(info, value) MOD.db.profile.InCombatBar.ooc = value end,
						},
						LinkOverlay = {
							type = "toggle", order = 40, name = L["Link Bar Layout"],
							desc = L["Link in-combat settings string"],
							disabled = function(info) return not MOD.db.profile.InCombatBar.enable end,
							get = function(info) return MOD.db.profile.InCombatBar.link end,
							set = function(info, value) MOD.db.profile.InCombatBar.link = value end,
						},
					},
				},
				BuffsGroup = {
					type = "group", order = 20, name = L["Buffs"], inline = true,
					args = {
						AddBuff = {
							type = "input", order = 10, name = L["Enter Buff"],
							desc = L["Enter in-combat buff string"],
							disabled = function(info) return not MOD.db.profile.InCombatBar.enable end,
							get = function(info) return nil end,
							set = function(info, value)
								if weaponBuffs[value] or ValidateSpellName(value) then
									local found = false
									for k, v in pairs(MOD.db.profile.InCombatBuffs) do if v == value then found = true end end
									if not found then
										table.insert(MOD.db.profile.InCombatBuffs, value)
										table.sort(MOD.db.profile.InCombatBuffs)
									end
									for k, v in pairs(MOD.db.profile.InCombatBuffs) do if v == value then selectedBuffName = k end end
								end
							end,
						},
						SelectBuff = {
							type = "select", order = 20, name = L["Buff List"],
							disabled = function(info) return not MOD.db.profile.InCombatBar.enable end,
							get = function(info)
								if not selectedBuffName and (#MOD.db.profile.InCombatBuffs > 0) then selectedBuffName = 1 end
								return selectedBuffName
							end,
							set = function(info, value) selectedBuffName = value end,
							values = function(info) return MOD.db.profile.InCombatBuffs end,
							style = "dropdown",
						},
						DeleteBuff = {
							type = "execute", order = 30, name = L["Delete"], width = "half",
							disabled = function(info) return not MOD.db.profile.InCombatBar.enable end,
							desc = L["Delete the selected buff from the in-combat list."],
							func = function(info)
								if selectedBuffName then
									table.remove(MOD.db.profile.InCombatBuffs, selectedBuffName)
									table.sort(MOD.db.profile.InCombatBuffs)
									selectedBuffName = next(MOD.db.profile.InCombatBuffs)
								end
							end,
						},
						ResetBuff = {
							type = "execute", order = 35, name = L["Reset"], width = "half",
							disabled = function(info) return not MOD.db.profile.InCombatBar.enable end,
							desc = L["Reset the in-combat buff list."],
							func = function(info) MOD.db.profile.InCombatBuffs = {}; selectedBuffName = nil end,
						},
					},
				},
				LayoutGroup = {
					type = "group", order = 30, name = L["Layout"], inline = true,
					args = {
						Row = {
							type = "toggle", order = 10, name = L["Horizontal Bar"],
							desc = L["Configure as a horizontal bar of buff icons."],
							disabled = function(info) return not MOD.db.profile.InCombatBar.enable end,
							get = function(info) return MOD.db.profile.InCombatBar.layout end,
							set = function(info, value) MOD.db.profile.InCombatBar.layout = true end,
						},
						Column = {
							type = "toggle", order = 20, name = L["Vertical Bar"],
							desc = L["Configure as a vertical bar of buff icons."],
							disabled = function(info) return not MOD.db.profile.InCombatBar.enable end,
							get = function(info) return not MOD.db.profile.InCombatBar.layout end,
							set = function(info, value) MOD.db.profile.InCombatBar.layout = false end,
						},
						Direction = {
							type = "toggle", order = 30, name = L["Direction"],
							desc = L["If checked, grow up or to the right, otherwise grow down or to the left."],
							disabled = function(info) return not MOD.db.profile.InCombatBar.enable end,
							get = function(info) return MOD.db.profile.InCombatBar.direction end,
							set = function(info, value) MOD.db.profile.InCombatBar.direction = value end,
						},
						spacer = { type = "description", name = "", order = 50 },
						GridSize = {
							type = "range", order = 60, name = L["Icon Size"], min = 5, max = 25, step = 1,
							desc = L["Set size for the buff icons."],
							disabled = function(info) return not MOD.db.profile.InCombatBar.enable end,
							get = function(info) return MOD.db.profile.InCombatBar.size end,
							set = function(info, value) MOD.db.profile.InCombatBar.size = value end,
						},
						GridSpacing = {
							type = "range", order = 70, name = L["Spacing"], min = 1, max = 10, step = 1,
							desc = L["Set spacing between the buff icons."],
							disabled = function(info) return not MOD.db.profile.InCombatBar.enable end,
							get = function(info) return MOD.db.profile.InCombatBar.spacing end,
							set = function(info, value) MOD.db.profile.InCombatBar.spacing = value end,
						},
						GridScale = {
							type = "range", order = 80, name = L["Scale"], min = 0.1, max = 2, step = 0.05,
							desc = L["Set scale factor for the bar."],
							disabled = function(info) return not MOD.db.profile.InCombatBar.enable end,
							get = function(info) return MOD.db.profile.InCombatBar.scale end,
							set = function(info, value) MOD.db.profile.InCombatBar.scale = value end,
						},
						GridAlpha = {
							type = "range", order = 90, name = L["Opacity"], min = 0, max = 1, step = 0.05,
							desc = L["Set opacity for the bar."],
							disabled = function(info) return not MOD.db.profile.InCombatBar.enable end,
							get = function(info) return MOD.db.profile.InCombatBar.alpha end,
							set = function(info, value) MOD.db.profile.InCombatBar.alpha = value end,
						},
					},
				},
				PositionGroup = {
					type = "group", order = 40, name = L["Display Position"], inline = true,
					args = {
						GridOffsetX = {
							type = "range", order = 10, name = L["Horizontal"], min = 0, max = 100, step = 0.01,
							desc = L["Set horizontal position as percentage of overall width (cannot move beyond edge of display)."],
							disabled = function(info) return not MOD.db.profile.InCombatBar.enable end,
							get = function(info) return MOD.db.profile.InCombatBar.offsetX * 100 end,
							set = function(info, value) MOD.db.profile.InCombatBar.offsetX = value / 100 end,
						},
						GridOffsetY = {
							type = "range", order = 20, name = L["Vertical"], min = 0, max = 100, step = 0.01,
							desc = L["Set vertical position as percentage of overall height (cannot move beyond edge of display)."],
							disabled = function(info) return not MOD.db.profile.InCombatBar.enable end,
							get = function(info) return MOD.db.profile.InCombatBar.offsetY * 100 end,
							set = function(info, value) MOD.db.profile.InCombatBar.offsetY = value / 100 end,
						},
						Space1 = { type = "description", name = "", order = 30, width = "half" },
						TooltipAnchor = {
							type = "select", order = 40, name = L["Tooltip Anchor"],
							desc = L["Tooltip anchor string"],
							disabled = function(info) return not MOD.db.profile.InCombatBar.enable end,
							get = function(info) return MOD.db.profile.InCombatBar.anchorTips end,
							set = function(info, value) MOD.db.profile.InCombatBar.anchorTips = value end,
							values = function(info) return anchorPoints end,
							style = "dropdown",
						},
					},
				},
				EffectsGroup = {
					type = "group", order = 60, name = L["Special Effects"],  inline = true,
					args = {
						PulseStart = {
							type = "toggle", order = 10, name = L["Pulse At Start"],
							desc = L["Enable icon pulse when bar is started."],
							get = function(info) return MOD.db.profile.InCombatBar.pulseStart end,
							set = function(info, value) MOD.db.profile.InCombatBar.pulseStart = value end,
						},
						PulseEnd = {
							type = "toggle", order = 15, name = L["Pulse When Expiring"],
							desc = L["Enable icon pulse when bar is expiring."],
							get = function(info) return MOD.db.profile.InCombatBar.pulseEnd end,
							set = function(info, value) MOD.db.profile.InCombatBar.pulseEnd = value end,
						},
						FlashExpiring = {
							type = "toggle", order = 20, name = L["Flash When Expiring"],
							desc = L["Enable flashing of expiring bars."],
							get = function(info) return MOD.db.profile.InCombatBar.flashExpiring end,
							set = function(info, value) MOD.db.profile.InCombatBar.flashExpiring = value end,
						},
						FlashTime = {
							type = "range", order = 25, name = L["Flash Time"], min = 1, max = 300, step = 1,
							desc = L["Set number of seconds before expiration that bar should start flashing."],
							disabled = function(info) return not MOD.db.profile.InCombatBar.flashExpiring end,
							get = function(info) return MOD.db.profile.InCombatBar.flashTime end,
							set = function(info, value) MOD.db.profile.InCombatBar.flashTime = value end,
						},
					},
				},
			},
		},
		Highlights = {
			type = "group", order = 50, name = L["Highlights"], 
			hidden = function(info) return not MOD.highlights end, -- must have compatible action bar mod
			disabled = function(info) return InMode() end, -- must have compatible action bar mod
			args = {
				intro = {
					type = "description", order = 1,
					name = L["Highlights string"],
				},
				EnableGroup = {
					type = "group", order = 10, name = L["Enable"], inline = true,
					args = {
						EnableGroup = {
							type = "toggle", order = 10, name = L["Enable Highlights"],
							desc = L["If checked, enable Raven's support for button highlights and cooldown counts."],
							get = function(info) return MOD.db.global.HighlightsEnabled end,
							set = function(info, value) MOD.db.global.HighlightsEnabled = value end,
						},
						ButtonFacade = {
							type = "toggle", order = 20, name = L["Use ButtonFacade"],
							desc = L["Enable setting highlights through ButtonFacade if it is loaded. This is not ideal for every BF skin which is why it can be disabled. Try both enabled and disabled settings and see which works best."],
							disabled = function(info) return not MOD.db.global.HighlightsEnabled or not LBF end,
							get = function(info) return MOD.db.global.ButtonFacade end,
							set = function(info, value) MOD.db.global.ButtonFacade = value end,
						},
					},
				},
				PlayerGroup = {
					type = "group", order = 20, name = L["Player"], inline = true, 
					disabled = function(info) return not MOD.db.global.HighlightsEnabled end,
					args = {
						PlayerBuffHighlightsEnable = {
							type = "toggle", order = 4, name = L["Highlight Buffs"],
							desc = L["Enable player buff highlights on buttons."],
							get = function(info) return MOD.db.global.PlayerBuffHighlights end,
							set = function(info, value) MOD.db.global.PlayerBuffHighlights = value end,
						},
						PlayerBuffColor = {
							type = "color", order = 5, name = L["Buff Color"], hasAlpha = false,
							desc = L["Set player buff highlight color."],
							get = function(info)
								local t = MOD.db.global.PlayerBuffColor
								return t.r, t.g, t.b, t.a
							end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.PlayerBuffColor
								t.r, t.g, t.b, t.a = r, g, b, a
							end,
						},
						PlayerDebuffHighlightsEnable = {
							type = "toggle", order = 23, name = L["Highlight Debuffs"],
							desc = L["Enable player debuff highlights on buttons."],
							get = function(info) return MOD.db.global.PlayerDebuffHighlights end,
							set = function(info, value) MOD.db.global.PlayerDebuffHighlights = value end,
						},
						PlayerDebuffColor = {
							type = "color", order = 24, name = L["Debuff Color"], hasAlpha = false,
							desc = L["Set player debuff highlight color."],
							get = function(info)
								local t = MOD.db.global.PlayerDebuffColor
								return t.r, t.g, t.b, t.a
							end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.PlayerDebuffColor
								t.r, t.g, t.b, t.a = r, g, b, a
							end,
						},
					},
				},
				TargetGroup = {
					type = "group", order = 30, name = L["Target"], inline = true,
					disabled = function(info) return not MOD.db.global.HighlightsEnabled end,
					args = {
						TargetBuffHighlightsEnable = {
							type = "toggle", order = 3, name = L["Highlight Buffs"],
							desc = L["Enable target buff highlights on buttons."],
							get = function(info) return MOD.db.global.TargetBuffHighlights end,
							set = function(info, value) MOD.db.global.TargetBuffHighlights = value end,
						},
						TargetBuffColor = {
							type = "color", order = 4, name = L["Buff Color"], hasAlpha = false,
							desc = L["Set target buff highlight color."],
							get = function(info)
								local t = MOD.db.global.TargetBuffColor
								return t.r, t.g, t.b, t.a
							end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.TargetBuffColor
								t.r, t.g, t.b, t.a = r, g, b, a
							end,
						},
						TargetDebuffHighlightsEnable = {
							type = "toggle", order = 23, name = L["Highlight Debuffs"],
							desc = L["Enable target debuff highlights on buttons."],
							get = function(info) return MOD.db.global.TargetDebuffHighlights end,
							set = function(info, value) MOD.db.global.TargetDebuffHighlights = value end,
						},
						TargetDebuffColor = {
							type = "color", order = 24, name = L["Debuff Color"], hasAlpha = false,
							desc = L["Set target debuff highlight color."],
							get = function(info)
								local t = MOD.db.global.TargetDebuffColor
								return t.r, t.g, t.b, t.a
							end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.TargetDebuffColor
								t.r, t.g, t.b, t.a = r, g, b, a
							end,
						},
					},
				},
				FocusGroup = {
					type = "group", order = 40, name = L["Focus"], inline = true,
					disabled = function(info) return not MOD.db.global.HighlightsEnabled end,
					args = {
						FocusBuffHighlightsEnable = {
							type = "toggle", order = 3, name = L["Highlight Buffs"],
							desc = L["Enable focus buff highlights on buttons."],
							get = function(info) return MOD.db.global.FocusBuffHighlights end,
							set = function(info, value) MOD.db.global.FocusBuffHighlights = value end,
						},
						FocusBuffColor = {
							type = "color", order = 4, name = L["Buff Color"], hasAlpha = false,
							desc = L["Set focus buff highlight color."],
							get = function(info)
								local t = MOD.db.global.FocusBuffColor
								return t.r, t.g, t.b, t.a
							end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.FocusBuffColor
								t.r, t.g, t.b, t.a = r, g, b, a
							end,
						},
						FocusDebuffHighlightsEnable = {
							type = "toggle", order = 23, name = L["Highlight Debuffs"],
							desc = L["Enable focus debuff highlights on buttons."],
							get = function(info) return MOD.db.global.FocusDebuffHighlights end,
							set = function(info, value) MOD.db.global.FocusDebuffHighlights = value end,
						},
						FocusDebuffColor = {
							type = "color", order = 24, name = L["Debuff Color"], hasAlpha = false,
							desc = L["Set focus debuff highlight color."],
							get = function(info)
								local t = MOD.db.global.FocusDebuffColor
								return t.r, t.g, t.b, t.a
							end,
							set = function(info, r, g, b, a)
								local t = MOD.db.global.FocusDebuffColor
								t.r, t.g, t.b, t.a = r, g, b, a
							end,
						},
					},
				},
				Flashing = {
					type = "group", order = 80, name = L["Flashing"],  inline = true,
					disabled = function(info) return not MOD.db.global.HighlightsEnabled end,
					args = {
						FlashExpiring = {
							type = "toggle", order = 2, name = L["Flash When Expiring"],
							desc = L["Enable highlight flashing for expiring buffs and debuffs."],
							get = function(info) return MOD.db.global.FlashExpiring end,
							set = function(info, value) MOD.db.global.FlashExpiring = value end,
						},
						FlashTime = {
							type = "range", order = 3, name = L["Seconds"], min = 1, max = 30, step = 1,
							desc = L["Set number of seconds to flash before buffs and debuffs expire."],
							disabled = function(info) return not MOD.db.global.FlashExpiring end,
							get = function(info) return MOD.db.global.FlashTime end,
							set = function(info, value) MOD.db.global.FlashTime = value end,
						},
					},
				},
				Cooldowns = {
					type = "group", order = 90, name = L["Cooldowns"],  inline = true,
					disabled = function(info) return not MOD.db.global.HighlightsEnabled end,
					args = {
						CooldownText = {
							type = "toggle", order = 10, name = L["Cooldown Time"],
							desc = L["Show cooldown time remaining on buttons."],
							get = function(info) return MOD.db.global.CooldownText end,
							set = function(info, value) MOD.db.global.CooldownText = value end,
						},
						CooldownFontSize = {
							type = "range", order = 20, name = L["Font Size"], min = 5, max = 30, step = 1,
							desc = L["Set font size for cooldown time text overlay."],
							disabled = function(info) return not MOD.db.global.CooldownText end,
							get = function(info) return MOD.db.global.CooldownFsize end,
							set = function(info, value) MOD.db.global.CooldownFsize = value end,
						},
						CooldownFont = {
							type = "select", order = 30, name = L["Font"], 
							desc = L["Select font for cooldown time text overlay."],
							disabled = function(info) return not MOD.db.global.CooldownText end,
							dialogControl = 'LSM30_Font',
							values = AceGUIWidgetLSMlists.font,
							get = function(info) return MOD.db.global.CooldownFont end,
							set = function(info, value) MOD.db.global.CooldownFont = value end,
						},
						TimeFormatGroup = {
							type = "group", order = 35, name = L["Time Format"],  inline = true,
							args = {
								TimeFormat = {
									type = "select", order = 10, name = L["Options"], width = "double",
									desc = L["Cooldown format string"],
									get = function(info) return MOD.db.global.CooldownTimeFormat end,
									set = function(info, value) MOD.db.global.CooldownTimeFormat = value end,
									values = function(info)
										local s = MOD.db.global.CooldownTimeSpaces
										local c = MOD.db.global.CooldownTimeCase
										return GetTimeFormatList(s, c)
									end,
									style = "dropdown",
								},
								Space1 = { type = "description", name = "", order = 15, width = "half" },
								Spaces = {
									type = "toggle", order = 20, name = L["Spaces"], width = "half",
									desc = L["Include spaces between values in time format."],
									get = function(info) return MOD.db.global.CooldownTimeSpaces end,
									set = function(info, value) MOD.db.global.CooldownTimeSpaces = value end,
								},
								Capitals = {
									type = "toggle", order = 30, name = L["Uppercase"],
									desc = L["If checked, use uppercase H, M and S in time format, otherwise use lowercase."],
									get = function(info) return MOD.db.global.CooldownTimeCase end,
									set = function(info, value) MOD.db.global.CooldownTimeCase = value end,
								},
							},
						},
						CooldownAlignment = {
							type = "group", order = 40, name = L["Alignment"], inline = true,
							args = {
								VText = { type = "description", name = L["Vertical:"], order = 1, width = "half" },
								VLeftAlign = {
									type = "toggle", order = 10, name = L["Top"],
									desc = L["If checked, vertically align cooldown text to top."],
									get = function(info) return MOD.db.global.CooldownVertical == "Top" end,
									set = function(info, value) MOD.db.global.CooldownVertical = "Top" end,
								},
								VCenterAlign = {
									type = "toggle", order = 20, name = L["Middle"],
									desc = L["If checked, vertically align cooldown text in middle."],
									get = function(info) return not MOD.db.global.CooldownVertical end,
									set = function(info, value) MOD.db.global.CooldownVertical = false end,
								},
								VRightAlign = {
									type = "toggle", order = 30, name = L["Bottom"],
									desc = L["If checked, vertically align cooldown text to bottom."],
									get = function(info) return MOD.db.global.CooldownVertical == "Bottom" end,
									set = function(info, value) MOD.db.global.CooldownVertical = "Bottom" end,
								},
								spacer = { type = "description", name = "", order = 40 },
								HText = { type = "description", name = L["Horizontal:"], order = 41, width = "half" },
								HLeftAlign = {
									type = "toggle", order = 50, name = L["Left"],
									desc = L["If checked, horizontally align cooldown text to left."],
									get = function(info) return MOD.db.global.CooldownHorizontal == "Left" end,
									set = function(info, value) MOD.db.global.CooldownHorizontal = "Left" end,
								},
								HCenterAlign = {
									type = "toggle", order = 60, name = L["Center"],
									desc = L["If checked, horizontally align cooldown text in center."],
									get = function(info) return not MOD.db.global.CooldownHorizontal end,
									set = function(info, value) MOD.db.global.CooldownHorizontal = false end,
								},
								HRightAlign = {
									type = "toggle", order = 70, name = L["Right"],
									desc = L["If checked, horizontally align cooldown text to right."],
									get = function(info) return MOD.db.global.CooldownHorizontal == "Right" end,
									set = function(info, value) MOD.db.global.CooldownHorizontal = "Right" end,
								},
							},
						},
					},
				},
			},
		},
	},
}

-- This table gets inserted into the bar settings tab for each active bar as the args field for a group.
-- Bar fields: name, enableBar, barLabel, barType, barData, monitor, isMine, color
MOD.barOptions = {
	SummaryGroup = {
		type = "group", order = 1, name = L["Bar Information"], inline = true,
		hidden = function(info) return NoBar() end,
		args = {
			BarDescription = {
				type = "description", order = 1, name = function(info) return GetBarDescription(info) end,
			},
		},
	},
	SortingGroup = {
		type = "group", order = 5, name = L["Custom Sort Order"], inline = true,
		hidden = function(info) return NoBar() or GetBarGroupField("sor") ~= "X" end,
		args = {
			PromoteBar = {
				type = "execute", order = 60, name = L["Move Up"],
				desc = L["Move the bar up, overriding sort order."],
				func = function(info) MoveBarInList(info, "up") end,
			},
			DemoteBar = {
				type = "execute", order = 70, name = L["Move Down"],
				desc = L["Move the bar down, overriding sort order."],
				func = function(info) MoveBarInList(info, "down") end,
			},
		},
	},
	OptionsGroup = {
		type = "group", order = 10, name = L["Settings"], inline = true,
		hidden = function(info) return NoBar() end,
		args = {
			EnableBar = {
				type = "toggle", order = 10, name = L["Enable"], width = "half",
				desc = L["If checked, enable showing the bar."],
				get = function(info) return GetBarField(info, "enableBar") end,
				set = function(info, value) SetBarField(info, "enableBar", value); MOD:UpdateAllBarGroups() end,
			},
			BarLabel = {
				type = "input", order = 20, name = L["Label"],
				desc = L["Enter label text for the bar."],
				get = function(info) return GetBarField(info, "barLabel") end,
				set = function(info, value) SetBarLabel(info, value, not GetBarField(info, "labelLink")); MOD:UpdateAllBarGroups() end,
			},
			LabelLink = {
				type = "toggle", order = 30, name = L["Label Link"],
				desc = L["If checked, label is linked to the associated spell and changing it here will change it for all bars linked to the same spell."],
				hidden = function(info) return GetBarField(info, "barType") == "Notification" end,
				get = function(info) return not GetBarField(info, "labelLink") end,
				set = function(info, value)
					if value then
						local bar = GetBarEntry(info)
						local label = MOD:GetLabel(bar.action)
						value = SetBarLabel(info, label, false) -- no need to update the cached label, also don't link if fails validation test
					end
					SetBarField(info, "labelLink", not value)
					MOD:UpdateAllBarGroups()
				end,
			},
			spacer = { type = "description", name = "", order = 50, hidden = function(info) return GetBarField(info, "barType") ~= "Cooldown" end, },
			EnableCooldownReadyBar = {
				type = "toggle", order = 60, name = L["Show When Ready"],
				desc = L["If checked, show ready bar when action is not on cooldown."],
				hidden = function(info) return GetBarField(info, "barType") ~= "Cooldown" end,
				get = function(info) return GetBarField(info, "enableReady") end,
				set = function(info, value) SetBarField(info, "enableReady", value); MOD:UpdateAllBarGroups() end,
			},
			EnableAuraReadyBar = {
				type = "toggle", order = 65, name = L["Show Not Active"],
				desc = L["If checked, show ready bar when action is not active."],
				hidden = function(info) local bt = GetBarField(info, "barType"); return (bt ~= "Buff") and (bt ~= "Debuff") end,
				get = function(info) return GetBarField(info, "enableReady") end,
				set = function(info, value) SetBarField(info, "enableReady", value); MOD:UpdateAllBarGroups() end,
			},
			EnableUsableTest = {
				type = "toggle", order = 70, name = L["Usable"], width = "half",
				desc = L["If checked, show ready bar only if spell is usable (i.e., enough mana, reagents, etc.)."],
				hidden = function(info) return GetBarField(info, "barType") ~= "Cooldown" end,
				disabled = function(info) return not GetBarField(info, "enableReady") end,
				get = function(info) return not GetBarField(info, "readyNotUsable") end,
				set = function(info, value) SetBarField(info, "readyNotUsable", not value); MOD:UpdateAllBarGroups() end,
			},
			ShowTime = {
				type = "range", order = 75, name = L["Show Time"], min = 0, max = 60, step = 1,
				desc = L["Set number of seconds to show the ready bar (0 for unlimited time)."],
				hidden = function(info) return GetBarField(info, "barType") == "Notification" end,
				disabled = function(info) return not GetBarField(info, "enableReady") end,
				get = function(info) return GetBarField(info, "readyTime") or 0 end,
				set = function(info, value) SetBarField(info, "readyTime", value) end,
			},
		},
	},
	ColorGroup = {
		type = "group", order = 20, name = L["Color and Icon"], inline = true,
		hidden = function(info) return NoBar() end,
		args = {
			UseColor = {
				type = "toggle", order = 10, name = L["Use Spell Color"], 
				desc = L["If checked, use color from associated spell."],
				hidden = function(info) return GetBarField(info, "barType") ~= "Notification" end,
				disabled = function(info)
					local bar = GetBarEntry(info)
					return bar.barType == "Notification" and not MOD:GetAssociatedSpellForBar(bar)
				end,
				get = function(info) return not GetBarField(info, "notColor") end,
				set = function(info, value)
					SetBarField(info, "notColor", not value)
					if value then SetBarField(info, "color", nil) end
					MOD:UpdateAllBarGroups()
				end,
			},
			UseIcon = {
				type = "toggle", order = 10, name = L["Use Spell Icon"], 
				desc = L["If checked, use icon from associated spell."],
				hidden = function(info) return GetBarField(info, "barType") ~= "Notification" end,
				disabled = function(info)
					local bar = GetBarEntry(info)
					return bar.barType == "Notification" and not MOD:GetAssociatedSpellForBar(bar)
				end,
				get = function(info) return not GetBarField(info, "notIcon") end,
				set = function(info, value) SetBarField(info, "notIcon", not value); MOD:UpdateAllBarGroups() end,
			},
			spacer = { type = "description", name = "", order = 20, hidden = function(info) return GetBarField(info, "barType") ~= "Notification" end, },
			StandardColors = {
				type = "select", order = 30, name = L["Color"], width = "half",
				desc = L["Select a standard color or click to set a custom color."],
				get = function(info) return nil end,
				set = function(info, value)
					local r, g, b, a = GetStandardColor(value)
					local bar = GetBarEntry(info)
					MOD:SetSpellColorForBar(bar, r, g, b, a)
					MOD:UpdateAllBarGroups()
				end,
				values = function(info) return GetStandardColorList() end,
				style = "dropdown",
			},
			BarColor = {
				type = "color", order = 40, name = "", hasAlpha = false, width = "half",
				get = function(info)
					local bar = GetBarEntry(info)
					local c = MOD:GetSpellColorForBar(bar)
					if not c then return 1, 1, 1, 1 end -- better than nothing
					return c.r, c.g, c.b, c.a
				end,
				set = function(info, r, g, b, a)
					local bar = GetBarEntry(info)
					MOD:SetSpellColorForBar(bar, r, g, b, a)
					MOD:UpdateAllBarGroups()
				end,
			},
			ColorLink = {
				type = "toggle", order = 50, name = L["Color Link"],
				desc = L["If checked, the color is linked to the associated spell and changing it here will change it for all bars linked to the same spell."],
				disabled = function(info)
					local bar = GetBarEntry(info)
					return bar.barType == "Notification" and not MOD:GetAssociatedSpellForBar(bar)
				end,
				get = function(info) return not GetBarField(info, "colorLink") end,
				set = function(info, value)
					local bar = GetBarEntry(info)
					bar.colorLink = not value
					MOD:LinkSpellColorForBar(bar)
					MOD:UpdateAllBarGroups()
				end,
			},
			BarIcon = {
				type = "description", order = 60, name = "", width = "half", 
				hidden = function(info) return not GetBarIcon(info) end,
				image = function(info) return GetBarIcon(info) end,
				imageWidth = 24, imageHeight = 24,
			},
		},
	},
	EffectsGroup = {
		type = "group", order = 30, name = L["Special Effects"],  inline = true,
		hidden = function(info) return NoBar() end,
		args = {
			PulseStart = {
				type = "toggle", order = 10, name = L["Pulse At Start"],
				desc = L["Enable icon pulse when bar is started."],
				get = function(info) return GetBarField(info, "pulseStart") end,
				set = function(info, value) SetBarField(info, "pulseStart", value) end,
			},
			PulseEnd = {
				type = "toggle", order = 15, name = L["Pulse When Expiring"],
				desc = L["Enable icon pulse when bar is expiring."],
				get = function(info) return GetBarField(info, "pulseEnd") end,
				set = function(info, value) SetBarField(info, "pulseEnd", value) end,
			},
			space1 = { type = "description", name = "", order = 20 },
			FlashExpiring = {
				type = "toggle", order = 25, name = L["Flash When Expiring"],
				desc = L["Enable flashing of expiring bars."],
				get = function(info) return GetBarField(info, "flashExpiring") end,
				set = function(info, value) SetBarField(info, "flashExpiring", value) end,
			},
			FlashTime = {
				type = "range", order = 30, name = L["Flash Time"], min = 1, max = 300, step = 1,
				desc = L["Set number of seconds before expiration that bar should start flashing."],
				disabled = function(info) return not GetBarField(info, "flashExpiring") end,
				get = function(info) return GetBarField(info, "flashTime") end,
				set = function(info, value) SetBarField(info, "flashTime", value) end,
			},
			space2 = { type = "description", name = "", order = 35 },
			HideEnable = {
				type = "toggle", order = 36, name = L["Hide After Start"],
				desc = L["Enable hide after start for timer bars (does not hide bars with unlimited duration)."],
				get = function(info) return GetBarField(info, "hide") end,
				set = function(info, value) SetBarField(info, "hide", value) end,
			},
			FadeEnable = {
				type = "toggle", order = 40, name = L["Fade After Start"],
				desc = L["Enable fade after start (i.e., switch to fade opacity after a delay)."],
				disabled = function(info) return GetBarField(info, "hide") end, -- preference to hide
				get = function(info) return GetBarField(info, "fade") end,
				set = function(info, value) SetBarField(info, "fade", value) end,
			},
			FadeDelay = {
				type = "range", order = 45, name = L["Delay"], min = 1, max = 300, step = 1,
				desc = L["Set number of seconds before bar will hide or fade."],
				disabled = function(info) return not GetBarField(info, "fade") and not GetBarField(info, "hide") end,
				get = function(info) return GetBarField(info, "delayTime") or 5 end,
				set = function(info, value) SetBarField(info, "delayTime", value) end,
			},
			spacer6 = { type = "description", name = "", order = 50 },
			NormalAlpha = {
				type = "range", order = 55, name = L["Normal Opacity"], min = 0, max = 1, step = 0.05,
				desc = L["Set normal opacity for this bar."],
				get = function(info) return GetBarField(info, "normalAlpha") or 1 end,
				set = function(info, value) SetBarField(info, "normalAlpha", value) end,
			},
			FadeAlpha = {
				type = "range", order = 58, name = L["Fade Opacity"], min = 0, max = 1, step = 0.05,
				desc = L["Set fade opacity for this bar."],
				disabled = function(info) return (GetBarField(info, "hide") or not GetBarField(info, "fade")) and IsOff(GetBarField(info, "fadeBar")) end,
				get = function(info) return GetBarField(info, "fadeAlpha") or 1 end,
				set = function(info, value) SetBarField(info, "fadeAlpha", value) end,
			},
			ReadyAlpha = {
				type = "range", order = 59, name = L["Ready Opacity"], min = 0, max = 1, step = 0.05,
				desc = L["Set opacity for ready bar."],
				disabled = function(info) return not GetBarField(info, "enableReady") end,
				get = function(info) return GetBarField(info, "readyAlpha") or 1 end,
				set = function(info, value) SetBarField(info, "readyAlpha", value) end,
			},
			spacer3 = { type = "description", name = "", order = 60 },
			HideBar = {
				type = "toggle", order = 65, name = L["Hide"], width = "half",
				desc = L["If checked, bar is conditionally hidden."],
				get = function(info) return IsOn(GetBarField(info, "hideBar")) end,
				set = function(info, value)
					local v = GetBarField(info, "hideBar")
					SetBarField(info, "hideBar", IsOn(v) and Off or true)
					MOD:UpdateAllBarGroups()
				end,
			},
			HideBarTrue = {
				type = "toggle", order = 70, name = L["True"], width = "half",
				desc = L["If checked, bar is hidden if the condition is true."],
				disabled = function(info) return IsOff(GetBarField(info, "hideBar")) end,
				get = function(info) return GetBarField(info, "hideBar") == true end,
				set = function(info, v) SetBarField(info, "hideBar", true); MOD:UpdateAllBarGroups() end,
			},
			HideBarFalse = {
				type = "toggle", order = 75, name = L["False"], width = "half",
				desc = L["If checked, bar is hidden if the condition is false."],
				disabled = function(info) return IsOff(GetBarField(info, "hideBar")) end,
				get = function(info) return GetBarField(info, "hideBar") == false end,
				set = function(info, v) SetBarField(info, "hideBar", false); MOD:UpdateAllBarGroups() end,
			},
			SelectCondition1 = {
				type = "select", order = 76, name = L["Hide Condition"],
				disabled = function(info) return IsOff(GetBarField(info, "hideBar")) end,
				get = function(info) return GetBarSelectedCondition(GetSelectConditionList(), GetBarField(info, "hideCondition")) end,
				set = function(info, value) SetBarField(info, "hideCondition", GetSelectConditionList()[value]) end,
				values = function(info) return GetSelectConditionList() end,
				style = "dropdown",
			},
			spacer4 = { type = "description", name = "", order = 79 },
			FlashBar = {
				type = "toggle", order = 80, name = L["Flash"], width = "half",
				desc = L["If checked, bar will conditionally flash."],
				get = function(info) return IsOn(GetBarField(info, "flashBar")) end,
				set = function(info, value)
					local v = GetBarField(info, "flashBar")
					SetBarField(info, "flashBar", IsOn(v) and Off or true)
					MOD:UpdateAllBarGroups()
				end,
			},
			FlashBarTrue = {
				type = "toggle", order = 85, name = L["True"], width = "half",
				desc = L["If checked, bar will flash if the condition is true."],
				disabled = function(info) return IsOff(GetBarField(info, "flashBar")) end,
				get = function(info) return GetBarField(info, "flashBar") == true end,
				set = function(info, v) SetBarField(info, "flashBar", true); MOD:UpdateAllBarGroups() end,
			},
			FlashBarFalse = {
				type = "toggle", order = 86, name = L["False"], width = "half",
				desc = L["If checked, bar will flash if the condition is false."],
				disabled = function(info) return IsOff(GetBarField(info, "flashBar")) end,
				get = function(info) return GetBarField(info, "flashBar") == false end,
				set = function(info, v) SetBarField(info, "flashBar", false); MOD:UpdateAllBarGroups() end,
			},
			SelectCondition2 = {
				type = "select", order = 87, name = L["Flash Condition"],
				disabled = function(info) return IsOff(GetBarField(info, "flashBar")) end,
				get = function(info) return GetBarSelectedCondition(GetSelectConditionList(), GetBarField(info, "flashCondition")) end,
				set = function(info, value) SetBarField(info, "flashCondition", GetSelectConditionList()[value]) end,
				values = function(info) return GetSelectConditionList() end,
				style = "dropdown",
			},
			spacer5 = { type = "description", name = "", order = 89 },
			FadeBar = {
				type = "toggle", order = 90, name = L["Fade"], width = "half",
				desc = L["If checked, bar will conditionally change from normal opacity to fade opacity."],
				get = function(info) return IsOn(GetBarField(info, "fadeBar")) end,
				set = function(info, value)
					local v = GetBarField(info, "fadeBar")
					SetBarField(info, "fadeBar", IsOn(v) and Off or true)
					MOD:UpdateAllBarGroups()
				end,
			},
			FadeBarTrue = {
				type = "toggle", order = 95, name = L["True"], width = "half",
				desc = L["If checked, bar will fade if the condition is true."],
				disabled = function(info) return IsOff(GetBarField(info, "fadeBar")) end,
				get = function(info) return GetBarField(info, "fadeBar") == true end,
				set = function(info, v) SetBarField(info, "fadeBar", true); MOD:UpdateAllBarGroups() end,
			},
			FadeBarFalse = {
				type = "toggle", order = 96, name = L["False"], width = "half",
				desc = L["If checked, bar will fade if the condition is false."],
				disabled = function(info) return IsOff(GetBarField(info, "fadeBar")) end,
				get = function(info) return GetBarField(info, "fadeBar") == false end,
				set = function(info, v) SetBarField(info, "fadeBar", false); MOD:UpdateAllBarGroups() end,
			},
			SelectCondition3 = {
				type = "select", order = 97, name = L["Fade Condition"],
				disabled = function(info) return IsOff(GetBarField(info, "fadeBar")) end,
				get = function(info) return GetBarSelectedCondition(GetSelectConditionList(), GetBarField(info, "fadeCondition")) end,
				set = function(info, value) SetBarField(info, "fadeCondition", GetSelectConditionList()[value]) end,
				values = function(info) return GetSelectConditionList() end,
				style = "dropdown",
			},
		},
	},
	BarTypeGroup = {
		type = "group", order = 40, name = "", inline = true,
		hidden = function(info) return not InMode("Bar") end,
		args = {
			SelectBarType = {
				type = "group", order = 1, name = L["Type"], inline = true,
				args = {
					NotificationBar = {
						type = "toggle", order = 10, name = L["Notify"], width = "half",
						desc = L["If checked, this is a notify bar."],
						get = function(info) return newbar.barType == "Notification" end,
						set = function(info, value) SetSelectedBarType("Notification") end,
					},
					BuffBar = {
						type = "toggle", order = 20, name = L["Buff"], width = "half",
						desc = L["If checked, this is a buff bar."],
						get = function(info) return newbar.barType == "Buff" end,
						set = function(info, value) SetSelectedBarType("Buff") end,
					},
					DebuffBar = {
						type = "toggle", order = 30, name = L["Debuff"], width = "half",
						desc = L["If checked, this is a debuff bar."],
						get = function(info) return newbar.barType == "Debuff" end,
						set = function(info, value) SetSelectedBarType("Debuff") end,
					},
					CooldownBar = {
						type = "toggle", order = 40, name = L["Cooldown"],
						desc = L["If checked, this is a cooldown bar."],
						get = function(info) return newbar.barType == "Cooldown" end,
						set = function(info, value) SetSelectedBarType("Cooldown") end,
					},
				},
			},
			SelectBarSource = {
				type = "group", order = 10, name = L["Source"], inline = true,
				hidden = function(info) return (newbar.barType == "Notification") end,
				args = {
					Class = {
						type = "toggle", order = 10, name = L["Class"], width = "half",
						desc = L["If checked, select actions associated with a class."],
						get = function(info) return newbar.barSource == "Class" end,
						set = function(info, value) SetSelectedBarSource("Class") end,
					},
					Racial = {
						type = "toggle", order = 20, name = L["Race"], width = "half",
						desc = L["If checked, select actions associated with a race."],
						get = function(info) return newbar.barSource == "Racial" end,
						set = function(info, value) SetSelectedBarSource("Racial") end,
					},
					Pet = {
						type = "toggle", order = 30, name = L["Pet"], width = "half",
						desc = L["If checked, select actions associated with a pet."],
						get = function(info) return newbar.barSource == "Pet" end,
						set = function(info, value) SetSelectedBarSource("Pet") end,
					},
					Spell = {
						type = "toggle", order = 40, name = L["Other Spell"],
						desc = L["If checked, enter spell name for an associated action."],
						get = function(info) return newbar.barSource == "Other" end,
						set = function(info, value) SetSelectedBarSource("Other") end,
					},
				},
			},
			SelectClass = {
				type = "group", order = 15, name = L["Class"], inline = true,
				hidden = function(info)
					return (newbar.barType == "Notification") or ((newbar.barSource ~= "Class") and (newbar.barSource ~= "Pet"))
				end,
				args = {
					Druid = {
						type = "toggle", order = 10, name = L["Druid"], width = "half",
						desc = L["If checked, select druid actions."],
						hidden = function(info) return newbar.barSource == "Pet" end,
						get = function(info) return newbar.barClass == 2 end,
						set = function(info, value) SetSelectedBarClass(2) end,
					},
					Hunter = {
						type = "toggle", order = 15, name = L["Hunter"], width = "half",
						desc = L["If checked, select hunter actions."],
						get = function(info) return newbar.barClass == 3 end,
						set = function(info, value) SetSelectedBarClass(3) end,
					},
					Mage = {
						type = "toggle", order = 20, name = L["Mage"], width = "half",
						desc = L["If checked, select mage actions."],
						hidden = function(info) return newbar.barSource == "Pet" end,
						get = function(info) return newbar.barClass == 4 end,
						set = function(info, value) SetSelectedBarClass(4) end,
					},
					Paladin = {
						type = "toggle", order = 25, name = L["Paladin"], width = "half",
						desc = L["If checked, select paladin actions."],
						hidden = function(info) return newbar.barSource == "Pet" end,
						get = function(info) return newbar.barClass == 5 end,
						set = function(info, value) SetSelectedBarClass(5) end,
					},
					Priest = {
						type = "toggle", order = 30, name = L["Priest"], width = "half",
						desc = L["If checked, select priest actions."],
						hidden = function(info) return newbar.barSource == "Pet" end,
						get = function(info) return newbar.barClass == 6 end,
						set = function(info, value) SetSelectedBarClass(6) end,
					},
					Rogue = {
						type = "toggle", order = 35, name = L["Rogue"], width = "half",
						desc = L["If checked, select rogue actions."],
						hidden = function(info) return newbar.barSource == "Pet" end,
						get = function(info) return newbar.barClass == 7 end,
						set = function(info, value) SetSelectedBarClass(7) end,
					},
					Shaman = {
						type = "toggle", order = 40, name = L["Shaman"], width = "half",
						desc = L["If checked, select shaman actions."],
						hidden = function(info) return newbar.barSource == "Pet" end,
						get = function(info) return newbar.barClass == 8 end,
						set = function(info, value) SetSelectedBarClass(8) end,
					},
					Warlock = {
						type = "toggle", order = 45, name = L["Warlock"], width = "half",
						desc = L["If checked, select warlock actions."],
						get = function(info) return newbar.barClass == 9 end,
						set = function(info, value) SetSelectedBarClass(9) end,
					},
					Warrior = {
						type = "toggle", order = 50, name = L["Warrior"], width = "half",
						desc = L["If checked, select warrior actions."],
						hidden = function(info) return newbar.barSource == "Pet" end,
						get = function(info) return newbar.barClass == 10 end,
						set = function(info, value) SetSelectedBarClass(10) end,
					},
					DeathKnight = {
						type = "toggle", order = 55, name = L["Death Knight"],
						desc = L["If checked, select death knight actions."],
						get = function(info) return newbar.barClass == 1 end,
						set = function(info, value) SetSelectedBarClass(1) end,
					},
				},
			},
			SelectConditionGroup = {
				type = "group", order = 20, name = L["Conditions To Monitor"], inline = true,
				hidden = function(info) return (newbar.barType ~= "Notification") or (GetBarConditionListCount() == 0) end,
				args = {
					AllOnBar = {
						type = "execute", order = 10, name = L["All On"], width = "half",
						desc = L["Select all the conditions."],
						func = function(info) SetAllBarConditions(true) end,
					},
					AllOffBar = {
						type = "execute", order = 20, name = L["All Off"], width = "half",
						desc = L["Deselect all the conditions."],
						func = function(info) SetAllBarConditions(false) end,
					},
					SelectCondition = {
						type = "multiselect", order = 30, name = "",
						get = function(info, key) return GetSelectedBarCondition(key) end,
						set = function(info, key, value) SetSelectedBarCondition(key, value) end,
						values = function(info) return GetBarConditionList() end,
					},
				},
			},
			EnterOtherSpellGroup = {
				type = "group", order = 40, name = L["Other Spell"], inline = true,
				hidden = function(info) return (newbar.barType == "Notification") or (newbar.barSource ~= "Other") end,
				args = {
					SpellName = {
						type = "input", order = 10, name = L["Spell Name"],
						desc = L["Enter spell name (or numeric identifier) to associate with a bar."],
						get = function(info) return selectedSpellName end,
						set = function(info, value)
							local n, t = value, tonumber(value)
							if t then n = GetSpellInfo(t) end
							ValidateSpellName(n)
							selectedSpellName = n
						end,
					},
				},
			},
			BarActionGroup = {
				type = "group", order = 50, name = L["Actions To Monitor"], inline = true,
				hidden = function(info) return (newbar.barSource == "Other") or (GetBarActionListCount() == 0) end,
				args = {
					AllOnBar = {
						type = "execute", order = 10, name = L["All On"], width = "half",
						desc = L["Select all the actions."],
						func = function(info) SetAllBarActions(true) end,
					},
					AllOffBar = {
						type = "execute", order = 20, name = L["All Off"], width = "half",
						desc = L["Deselect all the actions."],
						func = function(info) SetAllBarActions(false) end,
					},
					SelectBarAction = {
						type = "multiselect", order = 30, name = "",
						get = function(info, key) return GetSelectedBarAction(key) end,
						set = function(info, key, value) SetSelectedBarAction(key, value) end,
						values = function(info) return GetBarActionList() end,
					},
				},
			},
		},
	},
	MonitorUnitGroup = {
		type = "group", order = 50, name = L["Action On"], inline = true,
		hidden = function(info) return not InMode("Bar") or ((newbar.barType ~= "Buff") and (newbar.barType ~= "Debuff")) end,
		args = {
			PlayerBuff = {
				type = "toggle", order = 10, name = L["Player"],
				desc = L["If checked, must be on the player."],
				get = function(info) return newbar.monitor == "player" end,
				set = function(info, value) newbar.monitor = "player" end,
			},
			PetBuff = {
				type = "toggle", order = 15, name = L["Pet"],
				desc = L["If checked, must be on the player's pet."],
				get = function(info) return newbar.monitor == "pet" end,
				set = function(info, value) newbar.monitor = "pet" end,
			},
			TargetBuff = {
				type = "toggle", order = 20, name = L["Target"],
				desc = L["If checked, must be on the target."],
				get = function(info) return newbar.monitor == "target" end,
				set = function(info, value) newbar.monitor = "target" end,
			},
			FocusBuff = {
				type = "toggle", order = 25, name = L["Focus"],
				desc = L["If checked, must be on the focus."],
				get = function(info) return newbar.monitor == "focus" end,
				set = function(info, value) newbar.monitor = "focus" end,
			},
			MouseoverBuff = {
				type = "toggle", order = 30, name = L["Mouseover"],
				desc = L["If checked, must be on the mouseover unit."],
				get = function(info) return newbar.monitor == "mouseover" end,
				set = function(info, value) newbar.monitor = "mouseover" end,
			},
			PetTargetBuff = {
				type = "toggle", order = 35, name = L["Pet's Target"],
				desc = L["If checked, must be on the pet's target."],
				get = function(info) return newbar.monitor == "pettarget" end,
				set = function(info, value) newbar.monitor = "pettarget" end,
			},
			TargetTargetBuff = {
				type = "toggle", order = 40, name = L["Target's Target"],
				desc = L["If checked, must be on the target's target."],
				get = function(info) return newbar.monitor == "targettarget" end,
				set = function(info, value) newbar.monitor = "targettarget" end,
			},
			FocusTargetBuff = {
				type = "toggle", order = 45, name = L["Focus's Target"],
				desc = L["If checked, must be on the focus's target."],
				get = function(info) return newbar.monitor == "focustarget" end,
				set = function(info, value) newbar.monitor = "focustarget" end,
			},
		},
	},
	CastUnitGroup = {
		type = "group", order = 60, name = L["Cast By"], inline = true,
		hidden = function(info) return not InMode("Bar") or ((newbar.barType ~= "Buff") and (newbar.barType ~= "Debuff")) end,
		args = {
			PlayerBuff = {
				type = "toggle", order = 10, name = L["Player"], width = "half",
				desc = L["If checked, only track if cast by the player."],
				get = function(info) return newbar.castBy == "player" end,
				set = function(info, value) newbar.castBy = "player" end,
			},
			PetBuff = {
				type = "toggle", order = 20, name = L["Pet"], width = "half",
				desc = L["If checked, only track if cast by the player's pet."],
				get = function(info) return newbar.castBy == "pet" end,
				set = function(info, value) newbar.castBy = "pet" end,
			},
			OtherBuff = {
				type = "toggle", order = 30, name = L["Other"], width = "half",
				desc = L["If checked, only track if cast by anyone other than the player."],
				get = function(info) return newbar.castBy == "other" end,
				set = function(info, value) newbar.castBy = "other" end,
			},
			AnyoneBuff = {
				type = "toggle", order = 40, name = L["Anyone"], width = "half",
				desc = L["If checked, track if cast by anyone, including player and pet."],
				get = function(info) return newbar.castBy == "anyone" end,
				set = function(info, value) newbar.castBy = "anyone" end,
			},
		},
	},
	OKNewBar = {
		type = "execute", order = 93, name = L["OK"], width = "half",
		desc = L["Confirm creating new bars for the selected actions/conditions."],
		hidden = function(info) return not InMode("Bar") end,
		func = function(info) EnterNewBar("ok") end,
	},
	CancelNewBar = {
		type = "execute", order = 94, name = L["Cancel"], width = "half",
		desc = L["Cancel creating new bars."],
		hidden = function(info) return not InMode("Bar") end,
		func = function(info) EnterNewBar("cancel") end,
	},
}
